#include <errno.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "cfg.h"

#include "guiutils.h"
#include "cfg_gtk.h"
#include "tool_bar.h"
#include "menubutton.h"
#include "cdialog.h"
#include "fb.h"

#include "editor.h"
#include "editor_find_in_pages.h"
#include "editor_undo.h"
#include "editor_cb.h"
#include "editor_values.h"
#include "editor_dnd.h"
#include "editor_fio.h"
#include "viewer_dnd.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_op.h"
#include "messages.h"

#include "me_cfg_list.h"
#include "config.h"

#include "mansuffix.h"

#include "images/icon_manedit_editor_48x48.xpm"
#include "images/icon_manedit_editor_16x16.xpm"
#include "images/icon_manedit_editor_20x20.xpm"
#include "images/icon_manedit_viewer_16x16.xpm"
#include "images/icon_manedit_preview_20x20.xpm"

#include "images/icon_new_20x20.xpm"
#include "images/icon_open_20x20.xpm"
#include "images/icon_save_20x20.xpm"
#include "images/icon_save_as_20x20.xpm"
#include "images/icon_revert_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_exit_20x20.xpm"

#include "images/icon_undo_20x20.xpm"
#include "images/icon_redo_20x20.xpm"
#include "images/icon_cut_20x20.xpm"
#include "images/icon_copy_20x20.xpm"
#include "images/icon_paste_20x20.xpm"
#include "images/icon_options_20x20.xpm"

#include "images/icon_add_20x20.xpm"
#include "images/icon_remove_20x20.xpm"
#include "images/icon_properties_20x20.xpm"

#include "images/icon_manual_opened_20x20.xpm"
#include "images/icon_manpage_heading_20x20.xpm"
#include "images/icon_manpage_section_20x20.xpm"

#include "images/icon_paragraphi_20x20.xpm"
#include "images/icon_paragraphl_20x20.xpm"
#include "images/icon_paragraphr_20x20.xpm"
#include "images/icon_listitems_20x20.xpm"

#include "images/icon_bold_20x20.xpm"
#include "images/icon_underline_20x20.xpm"

#include "images/icon_break_20x20.xpm"

#include "images/icon_amp_20x20.xpm"
#include "images/icon_lessthan_20x20.xpm"
#include "images/icon_greaterthan_20x20.xpm"

#include "images/icon_strip_tags_20x20.xpm"

#include "images/icon_search_20x20.xpm"
#include "images/icon_replace_20x20.xpm"
#include "images/icon_replace_all_20x20.xpm"


/* EditorItem */
editor_item_struct *EditorItemNew(void);
editor_item_struct *EditorItemNewWithValues(
	editor_item_type type,
	editor_struct *editor,
	GtkCTreeNode *node
);
editor_item_struct *EditorItemCopy(editor_item_struct *item);
void EditorItemSetLines(
	editor_item_struct *item,
	gchar **lines_list, const gint nlines
);
void EditorItemDelete(editor_item_struct *item);

/* Select */
GtkCTreeNode *EditorCurrentSelectedNode(editor_struct *editor);
editor_item_struct *EditorCurrentSelectedItem(editor_struct *editor);
void EditorSelect(
	editor_struct *editor,
	GtkCTreeNode *node
);

/* Set/Get EditorItem */
void EditorBranchSetData(
	GtkCTree *ctree,
	GtkCTreeNode *node,
	editor_item_struct *item, GtkDestroyNotify destroy
);
editor_item_struct *EditorBranchGetData(
	GtkCTree *ctree,
	GtkCTreeNode *node
);
void EditorSetNodeHasChanges(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean has_changes
);

GtkCTreeNode *EditorGetFirstToplevelNode(editor_struct *editor);
GtkCTreeNode *EditorGetToplevelNode(GtkCTreeNode *node);

/* Open & Save As Query */
gint EditorNewTemplateQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
);
gint EditorOpenQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
);
gint EditorSaveAsQuery(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean verbose
);

/* Editor */
static void EditorCreateMenuBar(
	editor_struct *editor,
	GtkWidget *parent
);
static void EditorCreateToolBar(
	editor_struct *editor,
	GtkWidget *parent
);
static void EditorCreateStatusBar(
	editor_struct *editor,
	GtkWidget *parent
);
editor_struct *EditorNew(me_core_struct *core);
void EditorRecordConfiguration(editor_struct *editor);
void EditorUpdateMenus(editor_struct *editor);
void EditorSetStatusPosition(
	editor_struct *editor,
	const gint row, const gint column
);
void EditorSetStatusMessage(
	editor_struct *editor,
	const gchar *msg
);
void EditorSetStatusProgress(
	editor_struct *editor,
	const gfloat v
);
void EditorClearUndosList(editor_struct *editor);
void EditorSetBusy(
	editor_struct *editor,
	const gboolean busy
);
gboolean EditorIsMapped(editor_struct *editor);
void EditorMap(editor_struct *editor);
void EditorUnmap(editor_struct *editor);
void EditorDelete(editor_struct *editor);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Creates a new EditorItem.
 */
editor_item_struct *EditorItemNew(void)
{
	return(EDITOR_ITEM(g_malloc0(sizeof(editor_item_struct))));
}

/*
 *	Creates a new EditorItem with values.
 */
editor_item_struct *EditorItemNewWithValues(
	editor_item_type type,
	editor_struct *editor,
	GtkCTreeNode *node
)
{
	editor_item_struct *item = EditorItemNew();
	if(item == NULL)
	    return(NULL);

	item->type = type;
	item->editor = editor;
	item->node = node;

	return(item);
}

/*
 *	Coppies the EditorItem.
 */
editor_item_struct *EditorItemCopy(editor_item_struct *item)
{
	editor_item_struct	*tar_item,
				*src_item = item;

	if(src_item == NULL)
	    return(NULL);

	tar_item = EditorItemNew();
	if(tar_item == NULL)
	    return(NULL);

	tar_item->type = src_item->type;
	tar_item->editor = src_item->editor;
	tar_item->node = src_item->node;

	tar_item->has_changes = src_item->has_changes;

	tar_item->name = STRDUP(src_item->name);
	tar_item->path = STRDUP(src_item->path);

	tar_item->header_name = STRDUP(src_item->header_name);
	tar_item->header_section_number = STRDUP(src_item->header_section_number);
	tar_item->header_version = STRDUP(src_item->header_version);
	tar_item->header_author = STRDUP(src_item->header_author);
	tar_item->header_catagory = STRDUP(src_item->header_catagory);

	tar_item->section_name = STRDUP(src_item->section_name);

	tar_item->nlines = src_item->nlines;
	if(tar_item->nlines > 0)
	{
	    const gint n = tar_item->nlines;
	    gint i;

	    tar_item->lines_list = (gchar **)g_malloc(
		(n + 1) * sizeof(gchar *)
	    );
	    if(tar_item->lines_list != NULL)
	    {
		for(i = 0; i < n; i++)
		    tar_item->lines_list[i] = STRDUP(src_item->lines_list[i]);

		tar_item->lines_list[n] = NULL;
	    }
	}
	else
	{
	    tar_item->lines_list = NULL;
	}

	return(tar_item);
}

/*
 *	Transfers the lines list to the EditorItem.
 *
 *	The lines_list and nlines specifies the lines list which will
 *	be transfered to the EditorItem and should not be referenced
 *	again after this call.
 */
void EditorItemSetLines(
	editor_item_struct *item,
	gchar **lines_list, const gint nlines
)
{
	if(item == NULL)
	{
	    gint i;

	    for(i = 0; i < nlines; i++)
		g_free(lines_list[i]);
	    g_free(lines_list);

	    return;
	}

	/* Delete the existing lines list on the EditorItem */
	g_strfreev(item->lines_list);

	/* Transfer the new lines list to the EditorItem */
	item->nlines = nlines;
	if(nlines > 0)
	{
	    gint i;

	    item->lines_list = (gchar **)g_malloc(
		(nlines + 1) * sizeof(gchar *)
	    );
	    if(item->lines_list != NULL)
	    {
		for(i = 0; i < nlines; i++)
		    item->lines_list[i] = lines_list[i];

		item->lines_list[nlines] = NULL;
	    }
	}
	else
	{
	    item->lines_list = NULL;
	}

	/* Delete the lines list pointer array which was not transfered */
	g_free(lines_list);
}

/*
 *	Deletes the EditorItem.
 */
void EditorItemDelete(editor_item_struct *item)
{
	if(item == NULL)
	    return;

	g_free(item->name);
	g_free(item->path);

	g_free(item->header_name);
	g_free(item->header_section_number);
	g_free(item->header_version);
	g_free(item->header_author);
	g_free(item->header_catagory);

	g_free(item->section_name);

	g_strfreev(item->lines_list);

	g_free(item);
}


/*
 *	Gets the current selected node.
 */
GtkCTreeNode *EditorCurrentSelectedNode(editor_struct *editor)
{
	GList *glist;
	GtkCList *clist;

	if(editor == NULL)
	    return(NULL);

	clist = GTK_CLIST(editor->layout_ctree);
	glist = clist->selection_end;
	if(glist == NULL)
	    return(NULL);

	return((GtkCTreeNode *)glist->data);
}

/*
 *	Gets the current selected EditorItem.
 */
editor_item_struct *EditorCurrentSelectedItem(editor_struct *editor)
{
	GList *glist;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *node;

	if(editor == NULL)
	    return(NULL);

	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	glist = clist->selection_end;
	if(glist == NULL)
	    return(NULL);

	node = (GtkCTreeNode *)glist->data;
	if(node == NULL)
	    return(NULL);

	return(EDITOR_ITEM(gtk_ctree_node_get_row_data(
	    ctree,
	    node
	)));
}

/*
 *	Selects the node.
 *
 *	Expands the parent branches if any and as needed.
 *
 *	Will call the expand, select, and unselect callbacks.
 */
void EditorSelect(
	editor_struct *editor,
	GtkCTreeNode *node
)
{
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *parent_node;

	if((editor == NULL) || (node == NULL))
	    return;

	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);

	/* Get the parent node of the specified node */
	parent_node = GTK_CTREE_NODE_PARENT(node);

	gtk_clist_freeze(clist);

	/* Is parent node node expanded? */
	if(!GTK_CTREE_NODE_EXPANDED(parent_node))
	{
	    /* Expand all the parent nodes */
	    while(parent_node != NULL)
	    {
		gtk_ctree_expand(
		    ctree,
		    parent_node
		);

		/* Get parent node of this parent node */
		parent_node = GTK_CTREE_NODE_PARENT(parent_node);
	    }
	}

	/* Select the node, this will call the unselect and then
	 * select callbacks
	 */
	gtk_ctree_select(ctree, node);

	gtk_clist_thaw(clist);
}


/*
 *	Sets the EditorItem as the node's data.
 */
void EditorBranchSetData(
	GtkCTree *ctree,
	GtkCTreeNode *node,
	editor_item_struct *item, GtkDestroyNotify destroy
)
{
	if((ctree == NULL) || (node == NULL))
	    return;

	gtk_ctree_node_set_row_data_full(
	    ctree,
	    node,
	    item, destroy
	);
}

/*
 *	Gets the EditorItem from the node.
 */
editor_item_struct *EditorBranchGetData(
	GtkCTree *ctree,
	GtkCTreeNode *node
)
{
	if((ctree == NULL) || (node == NULL))
	    return(NULL);

	return(EDITOR_ITEM(gtk_ctree_node_get_row_data(
	    ctree,
	    node
	)));
}


/*
 *	Marks the manual page as having or not having changes.
 *
 *	The node specifies the GtkCTreeNode of the manual page.
 */
void EditorSetNodeHasChanges(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean has_changes
)
{
	GtkCTree *ctree;
	editor_item_struct *item;

	if(editor == NULL)
	    return;

	ctree = GTK_CTREE(editor->layout_ctree);

	while(node != NULL)
	{
	    item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
		ctree,
		node
	    ));
	    if(item != NULL)
		item->has_changes = has_changes;

	    node = GTK_CTREE_NODE_PARENT(node);
	}
}


/*
 *	Gets the first toplevel GtkCTreeNode.
 */
GtkCTreeNode *EditorGetFirstToplevelNode(editor_struct *editor)
{
	GtkCTree *ctree;

	if(editor == NULL)
	    return(NULL);

	ctree = GTK_CTREE(editor->layout_ctree);

	return(gtk_ctree_node_nth(ctree, 0));
}

/*
 *	Gets the toplevel GtkCTreeNode.
 */
GtkCTreeNode *EditorGetToplevelNode(GtkCTreeNode *node)
{
	GtkCTreeNode *parent;

	while(node != NULL)
	{
	    parent = GTK_CTREE_NODE_PARENT(node);
	    if(parent == NULL)
		return(node);

	    node = parent;
	}

	return(NULL);
}


/*
 *	Queries the user to create a new manual page from a template
 *	and creates the new manual page.
 *
 *	The insert_node specifies the GtkCTreeNode to insert the
 *	new manual page GtkCTreeNode after.
 *
 *	Returns 0 on success, -4 if the user canceled, or non-zero on
 *	error.
 */
gint EditorNewTemplateQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
)
{
	gboolean response;
	gint            status,
			error_code,
			npaths = 0,
			nftypes = 0;
	gchar           **paths_list,
			*templates_path;
	const gchar *s;
	GtkWidget *toplevel;
	CfgList *cfg_list;
	fb_type_struct  **ftypes_list = NULL,
			*ftype_rtn;
	me_core_struct *core;

	if(editor == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	EditorSetBusy(editor, TRUE);

	toplevel = editor->toplevel;
	core = editor->core;
	cfg_list = core->cfg_list;

	/* Get the templates directory */
	s = ME_GET_S(ME_CFG_PARM_PATH_GLOBAL_DATA);
	if(s == NULL)
	    s = ME_PATH_GLOBAL_DATA;
	templates_path = STRDUP(PrefixPaths(
	    s,
	    ME_PATH_TEMPLATES
	));

	/* Create the file types list */
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    ".mpt",
	    "Manual page template"
	);
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*",
	    "All files"
	);

	status = -4;
	error_code = 0;

	/* Query the user to open a manual page as a template */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
	    "Open Manual Page Template",
	    "Open",
	    "Cancel",
	    templates_path,
	    ftypes_list, nftypes,
	    &paths_list, &npaths,
	    &ftype_rtn
	);
	g_free(templates_path);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gint i;
	    gchar *path;

	    /* Open the manual page templates from last to first */
	    for(i = npaths - 1; i >= 0; i--)
	    {
		path = STRDUP(paths_list[i]);
		if(path == NULL)
		    continue;

		/* Open this manual page as a template */
		status = EditorFileOpen(
		    editor,
		    path,
		    insert_node,		/* Insert at this manual page */
		    TRUE                        /* Open as template */
		);
		error_code = (gint)errno;
		switch(status)
		{
		  case -5:			/* User responded with no */
		  case -4:			/* User aborted */
		    break;

		  case 0:			/* Successfully opened */
		    break;

		  default:
		    if(verbose)
		    {
			gchar *msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
			    g_strerror(error_code),
			    path
			);
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Open Failed",
			    msg,
			    NULL,
			    CDIALOG_ICON_ERROR,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			g_free(msg);
			CDialogSetTransientFor(NULL);
		    }
		    break;
		}

		g_free(path);

		if(status == -4)
		    break;
	    }
	}

	/* Delete the file types list */
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	EditorSetBusy(editor, FALSE);

	if((status == -4) || (status == -5))
	    errno = EINTR;
	else
	    errno = (int)error_code;

	return(status);
}

/*
 *	Queries the user to open the manual page and opens the manual
 *	page.
 *
 *	The insert_node specifies the GtkCTreeNode to insert the
 *	opened manual page GtkCTreeNode after.
 *
 *	Returns 0 on success, -4 if the user canceled, or non-zero on
 *	error.
 */
gint EditorOpenQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
)
{
	gboolean response;
	gint            status,
			error_code,
			npaths = 0,
			nftypes = 0;
	gchar **paths_list;
	GtkWidget *toplevel;
	fb_type_struct  **ftypes_list = NULL,
			*ftype_rtn = NULL;

	if(editor == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	EditorSetBusy(editor, TRUE);

	toplevel = editor->toplevel;

	/* Create the file types list */
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    MANPAGE_SUFFIX,
	    "Manual page"
	);
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    ".mpt",
	    "Manual page template"
	);
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*", "All files"
	);

	status = -4;
	error_code = 0;

	/* Query the user to open a manual page */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
	    "Open Manual Page",
	    "Open",
	    "Cancel",
	    editor->last_open_path,
	    ftypes_list, nftypes,
	    &paths_list, &npaths,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gint i;
	    gchar *path;

	    /* Open the manual pages from last to first */
	    for(i = npaths - 1; i >= 0; i--)
	    {
		path = STRDUP(paths_list[i]);
		if(path == NULL)
		    continue;

		/* Open this manual page file */
		status = EditorFileOpen(
		    editor,
		    path,
		    insert_node,
		    FALSE			/* Not as a template */
		);
		error_code = (gint)errno;
		switch(status)
		{
		  case -5:			/* User responded with no */
		  case -4:			/* User aborted */
		    break;

		  case 0:			/* Successfully opened */
		    /* Check if the manual page is not writable */
		    if(access((const char *)path, W_OK) && verbose)
		    {
			/* Warn that the manual page file is read-only */
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
"Manual Page Read-Only",
"This manual page is set read-only,\n\
you will not be able to save it.",
			    NULL,
			    CDIALOG_ICON_INFO,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
		    }
		    break;

		  default:
		    if(verbose)
		    {
			gchar *msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
			    g_strerror(error_code),
			    path
			);
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Open Failed",
			    msg,
			    NULL,
			    CDIALOG_ICON_ERROR,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			g_free(msg);
			CDialogSetTransientFor(NULL);
		    }
		    break;
		}

		g_free(path);

		if(status == -4)
		    break;
	    }
	}

	/* Delete the file types list */
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	EditorSetBusy(editor, FALSE);

	if((status == -4) || (status == -5))
	    errno = EINTR;
	else
	    errno = (int)error_code;

	return(status);
}

/*
 *	Queries the user to save the manual page as and saves the
 *	manual page as. If the manual page was saved successfully then
 *	the manual page will be marked as no longer haing changes.
 *
 *	The node specifies the toplevel GtkCTreeNode of the manual
 *	page to save as.
 *
 *	Returns 0 on success, -4 if the user canceled, or non-zero on
 *	error.
 */
gint EditorSaveAsQuery(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean verbose
)
{
	gboolean response;
	gint		status,
			error_code,
			npaths = 0,
			nftypes = 0;
	gchar **paths_list;
	GtkWidget *toplevel;
	fb_type_struct	**ftypes_list = NULL,
			*ftype_rtn = NULL;

	if(editor == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	EditorSetBusy(editor, TRUE);

	toplevel = editor->toplevel;

	/* Create the file types list */
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    MANPAGE_SUFFIX,
	    "Manual page"
	);
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    ".mpt",
	    "Manual page template"
	);
	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*", "All files"
	);

	status = -4;
	error_code = 0;

	/* Query the user to save the manual page as */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
	    "Save Manual Page As",
	    "Save",
	    "Cancel",
	    editor->last_save_as_path,
	    ftypes_list, nftypes,
	    &paths_list, &npaths,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gchar *path = STRDUP(paths_list[npaths - 1]);
	    if(path != NULL)
	    {
		gchar *s = FileBrowserTypeCompleteExtension(
		    path,
		    ftype_rtn,
		    FALSE			/* Do not replace if specified */
		);
		if(s != NULL)
		{
		    g_free(path);
		    path = s;
		}

		/* Save the manual page as */
		status = EditorFileSaveAs(
		    editor,
		    path,
		    node
		);
		error_code = (gint)errno;
		switch(status)
		{
	          case -5:			/* User responded with "no" */
	          case -4:			/* User aborted */
		    break;

		  case 0:			/* Save successful */
		    /* Mark this manual page as no longer having changes */
		    EditorSetNodeHasChanges(
			editor,
			node,
			FALSE			/* No changes */
		    );
		    /* Reget the values for the selected node */
		    EditorGetValues(
			editor,
			EditorCurrentSelectedNode(editor)
		    );
		    break;

		  default:			/* Other error */
		    if(verbose)
		    {
			gchar *msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
			    g_strerror(error_code),
			    path
			);
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Save Failed",
			    msg,
			    NULL,
			    CDIALOG_ICON_ERROR,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			g_free(msg);
			CDialogSetTransientFor(NULL);
		    }
		    break;
		}

		g_free(path);
	    }
	}

	/* Delete the file types list */
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	/* Reset the file selector so the new file shows up in its list */
	FileBrowserReset();

	EditorSetBusy(editor, FALSE);

	if((status == -4) || (status == -5))
	    errno = EINTR;
	else
	    errno = (int)error_code;

	return(status);
}


/*
 *	Creates the menu bar for the editor, this function should be
 *	called by EditorNew().
 *
 *	Given parent is assumed to be a vbox.
 */
static void EditorCreateMenuBar(editor_struct *editor, GtkWidget *parent)
{
	guint		accel_key,
			accel_mods;
	gchar *label = NULL;
	guint8 **icon;
	GtkWidget	*w,
			*parent2,
			*menu_bar,
			*menu;
	GtkAccelGroup *accelgrp = editor->accelgrp;
	gpointer client_data = editor;
	gpointer old_client_data;
	void (*func_cb)(GtkWidget *w, gpointer);

	/* GtkMenuBar */
	editor->menu_bar = menu_bar = GUIMenuBarCreate(NULL);
	gtk_widget_show(menu_bar);

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, client_data				\
 );						\
}

#define ADD_MENU_ITEM_SUBMENU	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SUBMENU, accelgrp,	\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, client_data				\
 );						\
 if(w != NULL)					\
  GUIMenuItemSetSubMenu(w, submenu);		\
}

#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, client_data				\
 );						\
}

#define ADD_MENU_ITEM_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,	\
  NULL, NULL, 0, 0,				\
  NULL, NULL					\
 );						\
}

	/* File */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New";
	    accel_key = GDK_n;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualNewCB;
	    ADD_MENU_ITEM_LABEL
	    editor->new_mi = w;

	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New From Template...";
	    accel_key = GDK_t;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualNewFromTemplateCB;
	    ADD_MENU_ITEM_LABEL
	    editor->new_from_template_mi = w;

	    icon = (guint8 **)icon_open_20x20_xpm;
	    label = "Open...";
	    accel_key = GDK_o;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualOpenCB;
	    ADD_MENU_ITEM_LABEL
	    editor->open_mi = w;

	    icon = (guint8 **)icon_save_20x20_xpm;
	    label = "Save";
	    accel_key = 's';
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualSaveCB;
	    ADD_MENU_ITEM_LABEL
	    editor->save_mi = w;

	    icon = (guint8 **)icon_save_as_20x20_xpm;
	    label = "Save As...";
	    accel_key = GDK_a;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualSaveAsCB;
	    ADD_MENU_ITEM_LABEL
	    editor->save_as_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_close_20x20_xpm;
	    label = "Close Manual";
	    accel_key = GDK_e;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualCloseCB;
	    ADD_MENU_ITEM_LABEL
	    editor->close_manual_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_revert_20x20_xpm;
	    label = "Revert";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualRevertCB;
	    ADD_MENU_ITEM_LABEL
	    editor->revert_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    /* This closes the editor, not the selected manual page */
	    icon = (guint8 **)icon_close_20x20_xpm;
	    label = "Close";
	    accel_key = GDK_w;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorCloseCB;
	    ADD_MENU_ITEM_LABEL
	    editor->close_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_exit_20x20_xpm;
	    label = "Exit";
	    accel_key = GDK_x;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorExitCB;
	    ADD_MENU_ITEM_LABEL
	    editor->exit_mi = w;
	}
	editor->file_mh = GUIMenuAddToMenuBar(
	    menu_bar, menu,
	    "File",
	    GTK_JUSTIFY_LEFT
	);

	/* Edit */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_undo_20x20_xpm;
	    label = "Undo";
	    accel_key = GDK_z;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorUndoCB;
	    ADD_MENU_ITEM_LABEL
	    editor->undo_mi = w;
	    editor->undo_milabel = GUIMenuItemGetLabelWidget(w);

	    icon = (guint8 **)icon_redo_20x20_xpm;
	    label = "Redo";
	    accel_key = GDK_r;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorRedoCB;
	    ADD_MENU_ITEM_LABEL
	    editor->redo_mi = w;
	    editor->redo_milabel = GUIMenuItemGetLabelWidget(w);

	    ADD_MENU_ITEM_SEPARATOR 

	    icon = (guint8 **)icon_cut_20x20_xpm;
	    label = "Cut";
	    accel_key = GDK_x;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorCutCB;
	    ADD_MENU_ITEM_LABEL
	    editor->cut_mi = w;

	    icon = (guint8 **)icon_copy_20x20_xpm;
	    label = "Copy";
	    accel_key = GDK_c;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorCopyCB;
	    ADD_MENU_ITEM_LABEL
	    editor->copy_mi = w;

	    icon = (guint8 **)icon_paste_20x20_xpm;
	    label = "Paste";
	    accel_key = GDK_v;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorPasteCB;
	    ADD_MENU_ITEM_LABEL
	    editor->paste_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Select All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorSelectAllCB;
	    ADD_MENU_ITEM_LABEL
	    editor->select_all_mi = w;

	    icon = NULL;
	    label = "Unselect All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorUnselectAllCB;
	    ADD_MENU_ITEM_LABEL
	    editor->unselect_all_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_search_20x20_xpm;
	    label = "Find In Pages...";
	    accel_key = GDK_f;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorMapFIPCB;
	    ADD_MENU_ITEM_LABEL
	    editor->find_in_pages_mi = w;
	}
	editor->edit_mh = GUIMenuAddToMenuBar(
	    menu_bar, menu,
	    "Edit",
	    GTK_JUSTIFY_LEFT
	);

	/* View */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_manedit_preview_20x20_xpm;
	    label = "Preview";
	    accel_key = GDK_F9;
	    accel_mods = 0;
	    func_cb = EditorManualPreviewCB;
	    ADD_MENU_ITEM_LABEL
	    editor->preview_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Syntax Highlighting";
	    accel_key = GDK_h;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorManualSyntaxHighlightToggleCB;
	    ADD_MENU_ITEM_CHECK
	    editor->syntax_highlight_cmi = w;
	}
	editor->view_mh = GUIMenuAddToMenuBar(
	    menu_bar, menu,
	    "View",
	    GTK_JUSTIFY_LEFT
	);

	/* Layout */   
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_add_20x20_xpm;
	    label = "Add Header";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualAddHeaderCB;
	    ADD_MENU_ITEM_LABEL
	    editor->add_header_mi = w;

	    icon = (guint8 **)icon_add_20x20_xpm; 
	    label = "Add Section";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualAddSectionCB;
	    ADD_MENU_ITEM_LABEL
	    editor->add_section_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_remove_20x20_xpm;
	    label = "Remove";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualRemoveSectionCB;
	    ADD_MENU_ITEM_LABEL
	    editor->remove_mi = w;

	    ADD_MENU_ITEM_SEPARATOR
	    
	    icon = (guint8 **)icon_properties_20x20_xpm;
	    label = "Properties...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualPropertiesCB;
	    ADD_MENU_ITEM_LABEL
	    editor->properties_mi = w;
	}
	editor->layout_mh = GUIMenuAddToMenuBar(   
	    menu_bar, menu,
	    "Layout",
	    GTK_JUSTIFY_LEFT
	);

	/* Format */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_paragraphl_20x20_xpm;
	    label = "Paragraph Left";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphLeftCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_pl_mi = w;

	    icon = (guint8 **)icon_paragraphr_20x20_xpm;
	    label = "Paragraph Right";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphRightCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_pr_mi = w;

	    icon = (guint8 **)icon_paragraphi_20x20_xpm;
	    label = "Paragraph Indented";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphIndentedCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_pi_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_listitems_20x20_xpm;
	    label = "List Item";
	    accel_key = 0; 
	    accel_mods = 0;
	    func_cb = EditorFmtListItemCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_li_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_bold_20x20_xpm;
	    label = "Bold";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtBoldCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_b_mi = w;

	    icon = (guint8 **)icon_underline_20x20_xpm;
	    label = "Underline";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtUnderlineCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_u_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_break_20x20_xpm;
	    label = "Line Break";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtLineBreakCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_br_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_amp_20x20_xpm;
	    label = "Ampersand";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtAmpersandCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_amp_mi = w;

	    icon = (guint8 **)icon_lessthan_20x20_xpm;
	    label = "Less Than";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtLessThanCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_lt_mi = w;

	    icon = (guint8 **)icon_greaterthan_20x20_xpm;
	    label = "Greater Than";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtGreaterThanCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_gt_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_strip_tags_20x20_xpm;
	    label = "Strip Tags";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorStripTagsCB;
	    ADD_MENU_ITEM_LABEL
	    editor->fmt_strip_tags_mi = w;
	}
	editor->fmt_mh = GUIMenuAddToMenuBar(   
	    menu_bar,
	    menu,
	    "Tags",
	    GTK_JUSTIFY_LEFT
	);

	/* Settings */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_options_20x20_xpm;
	    label = "Options...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorOptionsCB;
	    ADD_MENU_ITEM_LABEL
	    editor->settings_options_mi = w;
	}
	editor->settings_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Settings",
	    GTK_JUSTIFY_LEFT
	);

	/* Windows */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    old_client_data = client_data;
	    client_data = editor->core;

	    icon = (guint8 **)icon_manedit_editor_20x20_xpm;
	    label = "New Editor";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = MENewEditorCB;
	    ADD_MENU_ITEM_LABEL
	    editor->windows_new_editor = w;

	    icon = (guint8 **)icon_manual_opened_20x20_xpm;
	    label = "New Viewer";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = MENewViewerCB;
	    ADD_MENU_ITEM_LABEL
	    editor->windows_new_viewer = w;

	    client_data = old_client_data;
	}
	editor->windows_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Windows",
	    GTK_JUSTIFY_LEFT
	);

	/* Help */
	menu = MEBuildHelpMenu(
	    editor->core,
	    accelgrp
	);
	GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Help",
	    GTK_JUSTIFY_RIGHT
	);

#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR


	/* Menu Bar GtkHandleBox */
	editor->menu_bar_dock = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	gtk_container_add(GTK_CONTAINER(parent2), menu_bar);
}

/*
 *	Creates the tool bar for the editor, this function should be
 *	called by EditorNew().
 *
 *	Given parent is assumed to be a vbox.
 */
static void EditorCreateToolBar(editor_struct *editor, GtkWidget *parent)
{
	const gint border_minor = 2;
	gint bw, bh;
	guint		accel_key,
			accel_mods;
	const gchar *label;
	guint8 **icon;
	gpointer combo_rtn, mclient_data;
	GList *items_list;
	GtkAccelGroup *accelgrp = editor->accelgrp;
	GtkWidget	*w,
			*menu,
			*parent2, *parent3;
	void (*func_cb)(GtkWidget *w, gpointer);
	me_core_struct *core = editor->core;
	CfgList *cfg_list = core->cfg_list;

#define ADD_MENU_ITEM_LABEL {			\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU {			\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SUBMENU, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
 GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_ITEM_CHECK {			\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
}
#define ADD_MENU_ITEM_SEPARATOR {		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,	\
  NULL, NULL, 0, 0,				\
  NULL, NULL					\
 );						\
}

	/* ToolBarItems */
	items_list = NULL;
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_MENU_BUTTON,
		"New",
		(guint8 **)icon_new_20x20_xpm,
		EDITOR_TT_NEW_MENU_BTN,
		EDITOR_OPID_NEW,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Open",
		(guint8 **)icon_open_20x20_xpm,
		EDITOR_TT_OPEN,
		EDITOR_OPID_OPEN,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Save",
		(guint8 **)icon_save_20x20_xpm,
		EDITOR_TT_SAVE,
		EDITOR_OPID_SAVE,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Save As",
		(guint8 **)icon_save_as_20x20_xpm,
		EDITOR_TT_SAVE_AS,
		EDITOR_OPID_SAVE_AS,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Close",
		(guint8 **)icon_close_20x20_xpm,
		EDITOR_TT_CLOSE_MANUAL,
		EDITOR_OPID_CLOSE_MANUAL,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Undo",
		(guint8 **)icon_undo_20x20_xpm,
		EDITOR_TT_UNDO,
		EDITOR_OPID_UNDO,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Redo",
		(guint8 **)icon_redo_20x20_xpm,
		EDITOR_TT_REDO,
		EDITOR_OPID_REDO,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Cut",
		(guint8 **)icon_cut_20x20_xpm,
		EDITOR_TT_CUT,
		EDITOR_OPID_CUT,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Copy",
		(guint8 **)icon_copy_20x20_xpm,
		EDITOR_TT_COPY,
		EDITOR_OPID_COPY,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Paste",
		(guint8 **)icon_paste_20x20_xpm,
		EDITOR_TT_PASTE,
		EDITOR_OPID_PASTE,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Header",
		(guint8 **)icon_manpage_heading_20x20_xpm,
		EDITOR_TT_NEW_HEADER,
		EDITOR_OPID_NEW_HEADER,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Section",
		(guint8 **)icon_manpage_section_20x20_xpm,
		EDITOR_TT_NEW_SECTION,
		EDITOR_OPID_NEW_SECTION,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Remove",
		(guint8 **)icon_remove_20x20_xpm,
		EDITOR_TT_REMOVE,
		EDITOR_OPID_REMOVE,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Preview",
		(guint8 **)icon_manedit_preview_20x20_xpm,
		EDITOR_TT_PREVIEW,
		EDITOR_OPID_PREVIEW,
		EditorToolBarItemActivateCB, editor,
		NULL, NULL,
		NULL, NULL
	    )
	);

	/* GtkHandleBox */
	editor->tool_bar_dock = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* ToolBar */
	editor->tool_bar = w = ToolBarNew(
	    GTK_ORIENTATION_HORIZONTAL,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS),
	    (GtkReliefStyle)ME_GET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF),
	    GTK_POLICY_AUTOMATIC,
	    items_list
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	gtk_widget_show(w);

	if(items_list != NULL)
	{
	    g_list_foreach(
		items_list,
		(GFunc)ToolBarItemDelete,
		NULL
	    );
	    g_list_free(items_list);
	}

	/* Set up the new button and menu */
	MenuButtonSetMapTrigger(
	    ToolBarGetItemWidget(w, EDITOR_OPID_NEW),
	    MENU_BUTTON_MAP_TYPE_PRESSED_DRAG
	);
	menu = ToolBarGetItemMenu(w, EDITOR_OPID_NEW);
	mclient_data = editor;
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualNewCB;
	    ADD_MENU_ITEM_LABEL

	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New From Template...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualNewFromTemplateCB;
	    ADD_MENU_ITEM_LABEL
	}


	/* ********************************************************** */
	/* Format Bar GtkHandleBox */
	w = gtk_handle_box_new();
	editor->fmt_bar_dock = w;
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Format Bar GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	editor->fmt_bar = w;
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), 2);
	gtk_widget_show(w);
	parent3 = w;

	bw = 24;
	bh = 24;

	/* Paragraph left */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_paragraphl_20x20_xpm
	);
	editor->fmt_pl_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtParagraphLeftCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_LP);
	gtk_widget_show(w);

	/* Paragraph right */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_paragraphr_20x20_xpm
	);
	editor->fmt_pr_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtParagraphRightCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_RP);
	gtk_widget_show(w);

	/* Paragraph indented */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_paragraphi_20x20_xpm
	);
	editor->fmt_pi_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtParagraphIndentedCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_IP);
	gtk_widget_show(w);

	/* Separator */
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* List item */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_listitems_20x20_xpm
	);
	editor->fmt_li_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtListItemCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_LIST);
	gtk_widget_show(w);

	/* Separator */   
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Bold */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_bold_20x20_xpm
	);
	editor->fmt_b_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtBoldCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_BOLD);
	gtk_widget_show(w);

	/* Underlined */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_underline_20x20_xpm
	);
	editor->fmt_u_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtUnderlineCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_UNDERLINE);
	gtk_widget_show(w);

	/* Separator */   
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Break */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_break_20x20_xpm
	);
	editor->fmt_br_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtLineBreakCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_BREAK);
	gtk_widget_show(w);

	/* Separator */
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Ampersand */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_amp_20x20_xpm
	);
	editor->fmt_amp_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtAmpersandCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_AMP);
	gtk_widget_show(w);

	/* Less than */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_lessthan_20x20_xpm
	);
	editor->fmt_lt_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtLessThanCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_LT);
	gtk_widget_show(w);

	/* Greater than */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_greaterthan_20x20_xpm
	);
	editor->fmt_gt_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFmtGreaterThanCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_GT);
	gtk_widget_show(w);

	/* Separator */ 
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Strip tags */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_strip_tags_20x20_xpm
	);
	editor->fmt_strip_tags_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorStripTagsCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_FMT_STRIP_TAGS);
	gtk_widget_show(w);


	/* ********************************************************** */

	/* Handle for find bar hbox */
	w = gtk_handle_box_new();
	editor->find_bar_dock = w;
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Find bar hbox inside handle */
	w = gtk_hbox_new(FALSE, 0);
	editor->find_bar = w;
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), 2);
	gtk_widget_show(w);
	parent3 = w;

	/* Find Menu Button */
	editor->find_btn = w = MenuButtonNewH(
	    "Find:",
	    (guchar **)icon_search_20x20_xpm,
	    &menu
	);
	MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED);
	gtk_widget_set_usize(w, 80, -1);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, EDITOR_TT_FIND_MB); 
	gtk_widget_show(w);

	/* Find GtkMenu from Menu Button */
	editor->find_bar_find_menu = menu; 
	mclient_data = editor;
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_search_20x20_xpm;
	    label = "in Pages...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorMapFIPCB;
	    ADD_MENU_ITEM_LABEL
	    editor->find_bar_find_in_pages_mi = w;
	}

	/* Find GtkCombo */
	w = (GtkWidget *)GUIComboCreate(
	    NULL,
	    NULL,
	    NULL,
	    20,					/* Max items */
	    &combo_rtn,
	    editor,
	    EditorFindComboActivateCB,
	    NULL
	);
	editor->find_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(GTK_COMBO(combo_rtn)->entry, EDITOR_TT_FIND);
	gtk_widget_show(w);

	/* Separator */
	w = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Replace Menu Button */
	editor->replace_btn = w = MenuButtonNewH(
	    "Replace:",
	    (guchar **)icon_replace_20x20_xpm,
	    &menu
	);
	MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED);
	gtk_widget_set_usize(w, 80, -1);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, EDITOR_TT_REPLACE_MB);
	gtk_widget_show(w);

	/* Replace GtkMenu from Menu Button */
	editor->find_bar_replace_menu = menu;
	mclient_data = editor;
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_replace_all_20x20_xpm;
	    label = "Through Entire Manual Page";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorReplaceEntirePageCB;
	    ADD_MENU_ITEM_LABEL
	    editor->find_bar_replace_entire_page_mi = w;
	}

	/* Replace GtkCombo */
	w = GUIComboCreate(
	    NULL,
	    NULL,
	    NULL,
	    20,					/* Max items */
	    &combo_rtn,
	    editor,
	    EditorReplaceComboActivateCB,
	    NULL
	);
	editor->replace_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(GTK_COMBO(combo_rtn)->entry, EDITOR_TT_REPLACE);
	gtk_widget_show(w);

	/* Replace all */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_replace_all_20x20_xpm
	);
	editor->replace_all_btn = w;
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, bw, bh);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorReplaceAllCB), editor
	);
	GUISetWidgetTip(w, EDITOR_TT_REPLACE_ALL);
	gtk_widget_show(w);


#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR
}

/*
 *	Creates the status bar for the editor, this function should be
 *	called by EditorNew().
 *
 *	Given parent is assumed to be a vbox.
 */
static void EditorCreateStatusBar(
	editor_struct *editor,
	GtkWidget *parent
)
{
	const gint border_minor = 2;
	GtkAdjustment *adj;
	GtkWidget	*w,
			*parent2, *parent3;

	/* Outer GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_widget_set_usize(
	    w,
	    -1,
	    ME_STATUS_BAR_HEIGHT
	);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Main GtkTable */
	w = gtk_table_new(
	    1, 4,				/* Rows columns */
	    FALSE				/* Not homogenious */
	);
	gtk_table_set_col_spacings(GTK_TABLE(w), 1);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkProgressBar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1.0f, 100.0f,
	    0.0f, 0.0f, 0.0f
	);
	editor->status_bar_progress = w = gtk_progress_bar_new_with_adjustment(adj);
	gtk_widget_set_usize(
	    w,
	    100,
	    ME_STATUS_BAR_HEIGHT - (2 * 3)
	);
	gtk_progress_bar_set_orientation(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_LEFT_TO_RIGHT
	);
	gtk_progress_bar_set_bar_style(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_CONTINUOUS
	);
	gtk_progress_set_activity_mode(
	    GTK_PROGRESS(w),
	    FALSE
	);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    0, 1,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);


	/* Message GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    1, 2,
	    0, 1,
	    GTK_SHRINK | GTK_EXPAND | GTK_FILL,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Message GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Message Gtklabel */
	editor->status_bar_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);


	/* Cursor Position GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    2, 3,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Cursor position GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Cursor position GtkLabel */
	editor->status_bar_cursor_label = w = gtk_label_new("");
	gtk_widget_set_usize(w, 100, -1);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);


	/* Window buttons GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    3, 4,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Editor GtkButton */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_manedit_editor_16x16_xpm
	);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, 25, -1);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MENewEditorCB), editor->core
	);
	GUISetWidgetTip(w, MEDIT_SB_TT_NEW_EDITOR);
	gtk_widget_show(w);

	/* Viewer GtkButton */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_manedit_viewer_16x16_xpm
	);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, 25, -1);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MENewViewerCB), editor->core
	);
	GUISetWidgetTip(w, MEDIT_SB_TT_NEW_VIEWER);
	gtk_widget_show(w);
}

/*
 *	Creates a new Editor.
 */
editor_struct *EditorNew(me_core_struct *core)
{
	const gint	border_major = 5,
			border_minor = 2;
	gint header_label_width;
	guint		accel_key,
			accel_mods;
	const gchar *label;
	guint8 **icon;
	gchar *title[1];
	const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",				0,	MEDIT_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",			0,	MEDIT_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",				0,	MEDIT_DND_TYPE_INFO_STRING},
{MEDIT_DND_TARGET_EDITOR_BRANCH,	GTK_TARGET_SAME_APP,	MEDIT_DND_TYPE_INFO_EDITOR_BRANCH_CMD},
	};
	const GtkTargetEntry dnd_src_types[] = {
{MEDIT_DND_TARGET_EDITOR_BRANCH,	GTK_TARGET_SAME_APP,	MEDIT_DND_TYPE_INFO_EDITOR_BRANCH_CMD}
	};
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*menu, *submenu,
			*parent, *parent2, *parent3, *parent4,
			*parent5, *parent6,
			*toplevel;
	GtkText *text;
	GtkCList *clist;
	GtkCTree *ctree;
	gpointer mclient_data;
	void (*func_cb)(GtkWidget *w, gpointer);
	CfgList *cfg_list;
	editor_struct *editor;

	if(core == NULL)
	    return(NULL);

	cfg_list = core->cfg_list;

	/* Create a new Editor */
	editor = EDITOR(g_malloc0(sizeof(editor_struct)));
	if(editor == NULL)
	    return(NULL);

	editor->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	editor->accelgrp = accelgrp = gtk_accel_group_new();
/*
	editor->freeze_count = 0;
	editor->busy_count = 0;
 */
	editor->syntax_highlighting = ME_GET_B(
	    ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING
	);

	editor->core = core;
	editor->viewer_num = -1;

/*
	editor->synhl_timeout_cb_id = 0;
 */
	editor->synhl_timeout_cb_cursor_pos = -1;
	editor->synhl_timeout_cb_start_pos = -1;
	editor->synhl_timeout_cb_end_pos = -1;

	editor->max_undos = 5;
/*
	editor->status_bar_progress_pos_last = 0;
 */
	editor->last_open_path = STRDUP(ME_GET_S(
	    ME_CFG_PARM_PATH_LAST_OPEN
	));
	editor->last_save_as_path = STRDUP(ME_GET_S(
	    ME_CFG_PARM_PATH_LAST_SAVE_AS
	));

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SUBMENU, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
 GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
  icon, label, accel_key, accel_mods,		\
  func_cb, mclient_data				\
 );						\
}
#define ADD_MENU_ITEM_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,	\
  NULL, NULL, 0, 0,				\
  NULL, NULL					\
 );						\
}

	editor->freeze_count++;

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, FALSE);
	gtk_widget_set_events(w,
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
	);
	gtk_signal_connect(  
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(EditorKeyEventCB), editor
	);
	gtk_signal_connect(   
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(EditorKeyEventCB), editor
	);
	if(ME_GET_B(ME_CFG_PARM_RECORD_WIN_POS))
	{
	    CfgGtkSetGtkWindow(
		cfg_list,
		ME_CFG_PARM_EDITOR_X,
		ME_CFG_PARM_EDITOR_Y,
		ME_CFG_PARM_EDITOR_WIDTH,
		ME_CFG_PARM_EDITOR_HEIGHT,
		w
	    );
	}
	else
	{
	    gtk_widget_set_usize(
		w,
		ME_EDITOR_DEF_WIDTH, ME_EDITOR_DEF_HEIGHT
	    );
	}
	gtk_widget_realize(w);
	GUISetWMIcon(
	    w->window,
	    (guint8 **)icon_manedit_editor_48x48_xpm
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(EditorDeleteEventCB), editor
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;

	/* Main GtkVBox */
	editor->main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Menu Bar */
	EditorCreateMenuBar(editor, editor->main_vbox);

	/* Tool Bar */
	EditorCreateToolBar(editor, editor->main_vbox);


	/* Main GtkPaned */
	editor->main_paned = w = gtk_hpaned_new();
	gtk_paned_set_handle_size(GTK_PANED(w), ME_DEF_PANED_HANDLE_SIZE);
	gtk_paned_set_gutter_size(GTK_PANED(w), ME_DEF_PANED_GUTTER_SIZE);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	CfgGtkSetGtkPaned(
	    cfg_list,
	    ME_CFG_PARM_EDITOR_PANED_POSITION,
	    w
	);
	gtk_widget_show(w);
	parent2 = w;


	/* GtkVBox for the Layout GtkCTree */
	w = gtk_vbox_new(FALSE, 0);
	gtk_paned_add1(GTK_PANED(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Layout GtkScrolledWindow */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Layout GtkCTree */
	title[0] = "Layout";
	editor->layout_ctree = w = gtk_ctree_new_with_titles(
	    1,
	    0,
	    title
	);
	clist = GTK_CLIST(w);
	ctree = GTK_CTREE(w);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(EditorMenuMapCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_select_row",
	    GTK_SIGNAL_FUNC(EditorLayoutCTreeSelectCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_unselect_row",
	    GTK_SIGNAL_FUNC(EditorLayoutCTreeUnselectCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_expand",
	    GTK_SIGNAL_FUNC(EditorLayoutCTreeExpandCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_collapse",
	    GTK_SIGNAL_FUNC(EditorLayoutCTreeExpandCB), editor
	);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_column_titles_hide(clist);
	gtk_clist_set_row_height(clist, ME_LIST_ROW_SPACING);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_ctree_set_line_style(ctree, GTK_CTREE_LINES_DOTTED);
	gtk_clist_set_column_width(
	    clist,
	    0,
	    10
	);
	gtk_clist_set_column_auto_resize(
	    clist,
	    0,
	    TRUE
	);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	GUIDNDSetSrc(
	    w,
	    dnd_src_types,
	    sizeof(dnd_src_types) / sizeof(GtkTargetEntry),
	    GDK_ACTION_COPY | GDK_ACTION_MOVE,  /* Actions */
	    GDK_BUTTON1_MASK,			/* Buttons */
	    NULL,
	    EditorDNDDataRequestCB,
	    EditorDNDDataDeleteCB,
	    NULL,
	    editor
	);
	GUIDNDSetTar(
	    w,
	    dnd_tar_types,
	    sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
	    GDK_ACTION_COPY | GDK_ACTION_MOVE,  /* Actions */
	    GDK_ACTION_MOVE,                    /* Default action if same */
	    GDK_ACTION_COPY,                    /* Default action */
	    EditorDNDDataRecievedCB,
	    editor,
	    TRUE				/* Highlight */
	);
	gtk_widget_show(w);


	/* Layout GtkCTree Right-Click GtkMenu */
	editor->layout_menu = menu = GUIMenuCreate();
	mclient_data = editor;
	if(menu != NULL)
	{
	    icon = NULL;
	    label = "Expand/Collapse";
	    accel_key = 0; 
	    accel_mods = 0;
	    func_cb = EditorExpandCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_expand_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_manedit_preview_20x20_xpm;
	    label = "Preview";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualPreviewCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_preview_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_add_20x20_xpm;
	    label = "New Header";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualAddHeaderCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_add_header_mi = w;

	    icon = (guint8 **)icon_add_20x20_xpm;
	    label = "New Section";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualAddSectionCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_add_section_mi = w;

	    icon = (guint8 **)icon_remove_20x20_xpm;
	    label = "Remove";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualRemoveSectionCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_remove_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_properties_20x20_xpm;
	    label = "Properties...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualPropertiesCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_properties_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualNewCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_new_manual_mi = w;

	    icon = (guint8 **)icon_new_20x20_xpm;
	    label = "New From Template...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualNewFromTemplateCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_new_from_template_mi = w;

	    icon = (guint8 **)icon_open_20x20_xpm;
	    label = "Open...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualOpenCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_open_mi = w;

	    icon = (guint8 **)icon_save_20x20_xpm;
	    label = "Save";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualSaveCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_save_mi = w;

	    icon = (guint8 **)icon_save_as_20x20_xpm;
	    label = "Save As...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualSaveAsCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_save_as_mi = w;

	    ADD_MENU_ITEM_SEPARATOR
	
	    icon = (guint8 **)icon_close_20x20_xpm;
	    label = "Close Manual";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualCloseCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_close_manual_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_revert_20x20_xpm;
	    label = "Revert";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorManualRevertCB;
	    ADD_MENU_ITEM_LABEL
	    editor->layout_menu_revert_mi = w;
	}


	/* GtkVBox for the contents toplevel GtkBoxes */
	w = gtk_vbox_new(FALSE, 0);
	gtk_paned_add2(GTK_PANED(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* File Toplevel GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	editor->contents_toplevels_list = g_list_append(
	    editor->contents_toplevels_list,
	    w
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	parent3 = w;


	/* Header Toplevel GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	editor->contents_toplevels_list = g_list_append(
	    editor->contents_toplevels_list,
	    w
	);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	header_label_width = 100;

#define ADD_ENTRY(_s_,_wr_)		{		\
 /* Name GtkEntry */					\
 w = gtk_hbox_new(FALSE, border_minor);			\
 gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0); \
 gtk_widget_show(w);					\
 parent5 = w;						\
 /* Label GtkAlignment */				\
 w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);		\
 gtk_widget_set_usize(w, header_label_width, -1);	\
 gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0); \
 gtk_widget_show(w);					\
 parent6 = w;						\
 /* GtkLabel */						\
 w = gtk_label_new(_s_);				\
 gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT); \
 gtk_container_add(GTK_CONTAINER(parent6), w);		\
 gtk_widget_show(w);					\
 /* GtkEntry */						\
 (_wr_) = w = gtk_entry_new();				\
 gtk_widget_modify_style(				\
  w,							\
  core->editable_rcstyle				\
 );							\
 gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0); \
 gtk_widget_show(w);					\
}

	ADD_ENTRY(
	    "Name:",
	    editor->header_name_entry
	);
	ADD_ENTRY(
	    "Section Number:",
	    editor->header_section_number_entry
	);
	ADD_ENTRY(
	    "Version:",
	    editor->header_version_entry
	);
	ADD_ENTRY(
	    "Author:",
	    editor->header_author_entry
	);
	ADD_ENTRY(
	    "Category:",
	    editor->header_catagory_entry
	);
#undef ADD_ENTRY

	/* Comments GtkFrame */
	w = gtk_frame_new("Comments");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Table for text and scroll bar widgets */
	w = gtk_table_new(2, 2, FALSE);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_table_set_row_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_table_set_col_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent5 = w;

	/* Comments GtkText */
	editor->header_comments_text = w = gtk_text_new(NULL, NULL);
	text = GTK_TEXT(w);
	gtk_widget_modify_style(
	    w,
	    core->editable_rcstyle
	);
	/* Need to use gtk_signal_connect_after() to connect the
	 * signals for the GtkEntry so that the right-click
	 * GtkMenu maps properly
	 */
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(EditorTextKeyEventCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(EditorTextButtonEventCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(EditorTextButtonEventCB), editor
	);
	gtk_text_set_editable(text, TRUE);
	gtk_text_set_word_wrap(text, TRUE);
	gtk_table_attach(
	    GTK_TABLE(parent5),
	    w,
	    0, 1, 0, 1,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(EditorTextChangeCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "insert_text",
	    GTK_SIGNAL_FUNC(EditorTextInsertCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_text",
	    GTK_SIGNAL_FUNC(EditorTextDeleteCB), editor
	);
	gtk_widget_realize(w);
	if(w->window != NULL)
	{
	    GdkColormap *colormap = GDK_COLORMAP_REF(gtk_widget_get_colormap(w));
	    GdkColor *ct;
	    CfgColor *cs;

	    ct = &editor->normal_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_FG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->deliminator_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_DELIMINATOR);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->tag_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_TAG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->symbol_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_SYMBOL);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    editor->colormap = colormap;

	    gdk_window_set_cursor(
		text->text_area,
		core->text_cursor
	    );
	}
	gtk_widget_show(w);

	w = gtk_vscrollbar_new(text->vadj);
	gtk_table_attach(
	    GTK_TABLE(parent5),
	    w,
	    1, 2, 0, 1,
	    GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_widget_show(w);


	/* Section Toplevel GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	editor->contents_toplevels_list = g_list_append(
	    editor->contents_toplevels_list,
	    w
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	parent3 = w;

	/* Section Name */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_label_new("Section:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Section Name GtkEntry */
	editor->section_name_entry = w = gtk_entry_new();
	gtk_widget_modify_style(
	    w,
	    core->editable_rcstyle
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);

	/* Section GtkTable */
	w = gtk_table_new(2, 2, FALSE);
	gtk_table_set_row_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_table_set_col_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w; 

	/* Section GtkText */
	editor->section_text = w = gtk_text_new(NULL, NULL);
	text = GTK_TEXT(w);
	gtk_widget_modify_style(
	    w,
	    core->editable_rcstyle
	);
	/* Need to use gtk_signal_connect_after() to connect the
	 * signals for the GtkEntry so that the right-click
	 * GtkMenu maps properly
	 */
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(EditorTextKeyEventCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(EditorTextButtonEventCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(EditorTextButtonEventCB), editor
	);
	gtk_text_set_editable(text, TRUE);
	gtk_text_set_word_wrap(text, TRUE);
	gtk_table_attach(
	    GTK_TABLE(parent4),
	    w,
	    0, 1, 0, 1,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(EditorTextChangeCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "insert_text",
	    GTK_SIGNAL_FUNC(EditorTextInsertCB), editor
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_text",
	    GTK_SIGNAL_FUNC(EditorTextDeleteCB), editor
	);
	gtk_widget_realize(w);
	gdk_window_set_cursor(
	    GTK_TEXT(w)->text_area,
	    core->text_cursor
	);
	gtk_widget_show(w);

	w = gtk_vscrollbar_new(text->vadj);
	gtk_table_attach(
	    GTK_TABLE(parent4),
	    w,
	    1, 2, 0, 1,
	    GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_widget_show(w);


	/* ************************************************************ */

	/* Editing panels right click tags submenu */
	editor->edit_panel_fmt_sub_menu = submenu = menu = GUIMenuCreate();
	mclient_data = editor;

	if(menu != NULL)
	{
	    icon = (guint8 **)icon_paragraphl_20x20_xpm;
	    label = "Paragraph Left";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphLeftCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_pl_mi = w;

	    icon = (guint8 **)icon_paragraphr_20x20_xpm;
	    label = "Paragraph Right";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphRightCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_pr_mi = w;
	    
	    icon = (guint8 **)icon_paragraphi_20x20_xpm;
	    label = "Paragraph Indented";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtParagraphIndentedCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_pi_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_listitems_20x20_xpm;
	    label = "List Item";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtListItemCB; 
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_li_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_bold_20x20_xpm;
	    label = "Bold";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtBoldCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_b_mi = w;

	    icon = (guint8 **)icon_underline_20x20_xpm;
	    label = "Underline";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtUnderlineCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_u_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_break_20x20_xpm;
	    label = "Line Break"; 
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtLineBreakCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_br_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_amp_20x20_xpm;
	    label = "Ampersand";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtAmpersandCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_amp_mi = w;

	    icon = (guint8 **)icon_lessthan_20x20_xpm;
	    label = "Less Than";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtLessThanCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_lt_mi = w;

	    icon = (guint8 **)icon_greaterthan_20x20_xpm;
	    label = "Greater Than";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFmtGreaterThanCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_gt_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_strip_tags_20x20_xpm;
	    label = "Strip Tags";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorStripTagsCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_fmt_strip_tags_mi = w;
	}


	/* Editing panels right click GtkMenu */
	editor->edit_panel_menu = menu = GUIMenuCreate();
	mclient_data = editor;
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_undo_20x20_xpm;
	    label = "Undo";
	    accel_key = GDK_BackSpace;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = EditorUndoCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_undo_mi = w;
	    editor->edit_panel_undo_milabel = GUIMenuItemGetLabelWidget(w);

	    icon = (guint8 **)icon_redo_20x20_xpm;
	    label = "Redo"; 
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorRedoCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_redo_mi = w;
	    editor->edit_panel_redo_milabel = GUIMenuItemGetLabelWidget(w);

	    ADD_MENU_ITEM_SEPARATOR 

	    /* Format sub menu */
	    icon = NULL;
	    label = "Tags";
	    accel_key = 0; 
	    accel_mods = 0;
	    func_cb = NULL;
	    /* Variable submenu already set */
	    ADD_MENU_ITEM_SUBMENU

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_cut_20x20_xpm;
	    label = "Cut";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorCutCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_cut_mi = w;

	    icon = (guint8 **)icon_copy_20x20_xpm;
	    label = "Copy";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorCopyCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_copy_mi = w;

	    icon = (guint8 **)icon_paste_20x20_xpm;
	    label = "Paste";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorPasteCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_paste_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Select All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorSelectAllCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_select_all_mi = w;

	    icon = NULL;
	    label = "Unselect All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorUnselectAllCB;
	    ADD_MENU_ITEM_LABEL
	    editor->edit_panel_unselect_all_mi = w;
	}


	/* Status Bar */
	EditorCreateStatusBar(editor, editor->main_vbox);

#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR

	EditorUpdateMenus(editor);

	editor->freeze_count--;

	return(editor);
}

/*
 *	Records the Editor's current values to the CfgList.
 */
void EditorRecordConfiguration(editor_struct *editor)
{
	CfgList *cfg_list;
	me_core_struct *core;

	if(editor == NULL)
	    return;

	core = editor->core;
	cfg_list = core->cfg_list;

	/* Toplevel GtkWindow position and size */
	CfgGtkGetGtkWindow(
	    cfg_list,
	    ME_CFG_PARM_EDITOR_X,
	    ME_CFG_PARM_EDITOR_Y,
	    ME_CFG_PARM_EDITOR_WIDTH,
	    ME_CFG_PARM_EDITOR_HEIGHT,
	    editor->toplevel
	);

	/* Main GtkPaned position */
	CfgGtkGetGtkPaned(
	    cfg_list,
	    ME_CFG_PARM_EDITOR_PANED_POSITION,
	    editor->main_paned
	);

	ME_SET_S(
	    ME_CFG_PARM_PATH_LAST_OPEN,
	    editor->last_open_path
	);

	ME_SET_S(
	    ME_CFG_PARM_PATH_LAST_SAVE_AS,
	    editor->last_save_as_path
	);
}

/*
 *	Updates the Editor's widgets and menus.
 */
void EditorUpdateMenus(editor_struct *editor)
{
	gboolean	sensitivity,
			state,
			has_changes;
	const gchar *path;
	GList *glist;
	GtkWidget	*w,
			*menu;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode	*node,
			*toplevel_node;
	CfgList *cfg_list;
	editor_item_struct *item;
	me_core_struct *core;

	if(editor == NULL)
	    return;

	editor->freeze_count++;

	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;
	cfg_list = core->cfg_list;

	glist = clist->selection_end;
	node = (glist != NULL) ? (GtkCTreeNode *)glist->data : NULL;
	item = (node != NULL) ? EDITOR_ITEM(gtk_ctree_node_get_row_data(
	    ctree,
	    node
	)) : NULL;

	toplevel_node = EditorGetToplevelNode(node);

	has_changes = FALSE;
	path = NULL;
	if(toplevel_node != NULL)
	{
	    editor_item_struct *item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
		ctree,
		toplevel_node
	    ));
	    if(item != NULL)
	    {
		if(item->type == EDITOR_ITEM_FILE)
		{
		    if(item->has_changes)
			has_changes = TRUE;

		    path = item->name;
		}
	    }
	}

#define SET_WIDGET_SENSITIVITY	{		\
 if(w != NULL)					\
  gtk_widget_set_sensitive(w, sensitivity);	\
}

#define SET_CHECK_MENU_ITEM_STATE	{	\
 GUIMenuItemSetCheck(w, state, TRUE);		\
}

	/* Title */
	w = editor->toplevel;
	if(STRISEMPTY(path))
	{
	    gtk_window_set_title(
		GTK_WINDOW(w),
		PROG_NAME
	    );
	}
	else
	{
	    gchar *title = g_strdup_printf(
		"%s: %s%s",
		PROG_NAME, path,
		(has_changes) ? " (*)" : ""
	    );
	    gtk_window_set_title(GTK_WINDOW(w), title);
	    g_free(title);
	}

	/* File Menu */
	menu = editor->file_mh;
	if(menu != NULL)
	{
	    w = editor->open_mi;
	    sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    w = editor->save_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity)
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    w = editor->save_as_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity)
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    /* Revert enable only if entire manual has changes */
	    w = editor->revert_mi;
	    sensitivity = has_changes;
	    SET_WIDGET_SENSITIVITY

	    w = editor->close_manual_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity) 
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY
	}

	/* Edit Menu */
	menu = editor->edit_mh;
	if(menu != NULL)
	{
	    editor_undo_common_struct *u_common;

	    w = editor->undo_mi;
	    sensitivity = (editor->total_undos > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->undo_milabel;
	    if(w != NULL)
	    {
		if(editor->total_undos > 0)
		{
		    u_common = (editor_undo_common_struct *)editor->undo[0];
		    if((u_common != NULL) ?
			(u_common->comment != NULL) : FALSE
		    )       
		    {
			gchar *s = g_strdup_printf(
			    "Undo %s (%i)",
			    u_common->comment, editor->total_undos
			);
			gtk_label_set_text(GTK_LABEL(w), s);
			g_free(s);
		    }
		}
		else
		{
		    gtk_label_set_text(GTK_LABEL(w), "Undo");
		}
	    }


	    w = editor->redo_mi;
	    sensitivity = ((editor->total_redos > 0) ?
		TRUE : FALSE
	    );
	    SET_WIDGET_SENSITIVITY

	    w = editor->redo_milabel;
	    if(w != NULL)
	    {
		if(editor->total_redos > 0)
		{
		    u_common = (editor_undo_common_struct *)editor->redo[0];
		    if((u_common != NULL) ?
			(u_common->comment != NULL) : FALSE
		    )
		    {
			gchar *s = g_strdup_printf(
			    "Redo %s (%i)",
			    u_common->comment, editor->total_redos
			);
			gtk_label_set_text(GTK_LABEL(w), s);
			g_free(s);
		    }
		}
		else
		{
		    gtk_label_set_text(GTK_LABEL(w), "Redo");
		}
	    }

	    w = editor->cut_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->copy_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->paste_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->select_all_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->unselect_all_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->find_in_pages_mi;
	    sensitivity = (clist->rows > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY
	}

	/* View menu */
	menu = editor->view_mh;
	if(menu != NULL)
	{
	    w = editor->preview_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->syntax_highlight_cmi;
	    state = editor->syntax_highlighting;
	    SET_CHECK_MENU_ITEM_STATE
	}


	/* Layout Menu */
	menu = editor->layout_mh;
	if(menu != NULL)
	{
	    /* Show or hide menu depending on if a branch is selected */
	    if(node != NULL)
		gtk_widget_show(menu);
	    else
		gtk_widget_hide(menu);
	}

	/* Format Menu */
	menu = editor->fmt_mh;
	if(menu != NULL)
	{
	    sensitivity = (item != NULL) ?
		(item->type == EDITOR_ITEM_SECTION) : FALSE;

	    /* Show or hide menu depending on if a branch is selected */
	    if((item != NULL) ?
		(item->type == EDITOR_ITEM_SECTION) : FALSE
	    )
		gtk_widget_show(menu);
	    else
		gtk_widget_hide(menu);
	}


	/* Layout GtkCTree Right Click GtkMenu */
	menu = editor->layout_menu;
	if(menu != NULL)
	{
	    GtkCTreeRow *row_ptr;

	    if(node != NULL)
		row_ptr = GTK_CTREE_ROW(node);
	    else
		row_ptr = NULL;
	    sensitivity = (row_ptr != NULL) ? !row_ptr->is_leaf : FALSE;
	    w = editor->layout_menu_expand_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_preview_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_add_header_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_add_section_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_remove_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_properties_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_open_mi;
	    sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_save_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity)
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_save_as_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity)
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY

	    /* Revert enable only if entire manual has changes */
	    w = editor->layout_menu_revert_mi;
	    sensitivity = has_changes;
	    SET_WIDGET_SENSITIVITY

	    w = editor->layout_menu_close_manual_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    if(sensitivity)
		sensitivity = !FileBrowserIsQuery();
	    SET_WIDGET_SENSITIVITY
	}


	/* Panel Widgets Right Click GtkMenu */
	menu = editor->edit_panel_menu;
	if(menu != NULL)
	{
	    editor_undo_common_struct *u_common;

	    w = editor->edit_panel_undo_mi;
	    sensitivity = (editor->total_undos > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_undo_milabel;
	    if(w != NULL)
	    {
		if(editor->total_undos > 0)
		{
		    u_common = (editor_undo_common_struct *)editor->undo[0];
		    if((u_common != NULL) ?
			(u_common->comment != NULL) : FALSE
		    )
		    {
			gchar *s = g_strdup_printf(
			    "Undo %s (%i)",
			    u_common->comment, editor->total_undos
			);
			gtk_label_set_text(GTK_LABEL(w), s);
			g_free(s);
		    }
		}
		else
		{
		    gtk_label_set_text(GTK_LABEL(w), "Undo");
		}
	    }


	    w = editor->edit_panel_redo_mi;
	    sensitivity = (editor->total_redos > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_redo_milabel;
	    if(w != NULL)
	    {
		if(editor->total_redos > 0)
		{
		    u_common = (editor_undo_common_struct *)editor->redo[0];
		    if((u_common != NULL) ?
			(u_common->comment != NULL) : FALSE
		    )
		    {
			gchar *s = g_strdup_printf(
			    "Redo %s (%i)",
			    u_common->comment, editor->total_redos
			);
			gtk_label_set_text(GTK_LABEL(w), s);
			g_free(s);
		    }
		}
		else
		{
		    gtk_label_set_text(GTK_LABEL(w), "Redo");
		}
	    }

	    w = editor->edit_panel_cut_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_copy_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_paste_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_select_all_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_unselect_all_mi;
	    sensitivity = (node != NULL) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY
	}
	/* Edit panel format submenu */
	menu = editor->edit_panel_fmt_sub_menu;
	if(menu != NULL)
	{
	    sensitivity = (item != NULL) ?
		(item->type == EDITOR_ITEM_SECTION) : FALSE;

	    w = editor->edit_panel_fmt_pl_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_pr_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_pi_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_li_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_b_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_u_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_br_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_amp_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_lt_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_gt_mi;
	    SET_WIDGET_SENSITIVITY

	    w = editor->edit_panel_fmt_strip_tags_mi;
	    SET_WIDGET_SENSITIVITY
	}


	/* Tool Bar */
	w = editor->tool_bar;
	ToolBarSetButtonDecals(
	    w,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS)
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_OPEN,
	    !FileBrowserIsQuery()
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_SAVE,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_SAVE_AS,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_CLOSE_MANUAL,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_UNDO,
	    (editor->total_undos > 0) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_REDO,
	    (editor->total_undos > 0) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_CUT,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_COPY,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_PASTE,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_NEW_HEADER,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_NEW_SECTION,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_REMOVE,
	    (node != NULL) ? TRUE : FALSE
	);
	ToolBarSetItemSensitive(
	    w,
	    EDITOR_OPID_PREVIEW,
	    (node != NULL) ? TRUE : FALSE
	);

	/* Format bar buttons */
	w = editor->fmt_bar;
	if(w != NULL)
	{
	    sensitivity = (item != NULL) ?
		(item->type == EDITOR_ITEM_SECTION) : FALSE;

	    w = editor->fmt_pl_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_pr_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_pi_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_li_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_b_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_u_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_br_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_amp_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_lt_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_gt_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->fmt_strip_tags_btn;
	    SET_WIDGET_SENSITIVITY
	}

	/* Find Bar */
	w = editor->find_bar;
	if(w != NULL)
	{
	    sensitivity = (node != NULL) ? TRUE : FALSE;

	    w = editor->find_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->find_combo;
	    SET_WIDGET_SENSITIVITY

	    w = editor->replace_btn;
	    SET_WIDGET_SENSITIVITY

	    w = editor->replace_combo; 
	    SET_WIDGET_SENSITIVITY

	    w = editor->replace_all_btn;
	    SET_WIDGET_SENSITIVITY
	}
	/* Find bar find button menu */
	w = editor->find_bar_find_menu;
	if(w != NULL)
	{
	    w = editor->find_bar_find_in_pages_mi;
	    sensitivity = (clist->rows > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY
	}
	/* Find bar replace button menu */
	w = editor->find_bar_replace_menu;
	if(w != NULL)
	{
	    w = editor->find_bar_replace_entire_page_mi;
	    sensitivity = (clist->rows > 0) ? TRUE : FALSE;
	    SET_WIDGET_SENSITIVITY
	}

#undef SET_WIDGET_SENSITIVITY
#undef SET_CHECK_MENU_ITEM_STATE

	editor->freeze_count--;
}

/*
 *	Updates the status bar's cursor position.
 */
void EditorSetStatusPosition(
	editor_struct *editor,
	const gint row, const gint column
)
{
	gchar *s;

	if(editor == NULL)
	    return;

	s = g_strdup_printf(
	    "Line: %i / Col: %i",
	    row + 1, column + 1
	);
	gtk_label_set_text(GTK_LABEL(editor->status_bar_cursor_label), s);
	g_free(s);

	while(gtk_events_pending() > 0)
	    gtk_main_iteration();
}

/*
 *	Sets the status bar message.
 */
void EditorSetStatusMessage(
	editor_struct *editor,
	const gchar *msg
)
{
	if(editor == NULL)
	    return;

	gtk_label_set_text(
	    GTK_LABEL(editor->status_bar_label),
	    (msg != NULL) ? msg : ""
	);

	while(gtk_events_pending() > 0)
	    gtk_main_iteration();
}

/*
 *	Sets the status bar progress.
 *
 *	The v specifies the value from 0.0 to 1.0 inclusive or -1.0 for
 *	"unknown limit" which will cause the progress bar to be displayed 
 *	in activity mode.
 *
 *	Note that gtk_main_iteration() will be called when this function
 *	is called to update the progress bar properly.
 */
void EditorSetStatusProgress(
	editor_struct *editor,
	const gfloat v
)
{
	const gboolean allow_gtk_iteration = TRUE;
	gfloat value = v;
	GtkWidget *w;
	GtkProgress *progress;
	GtkProgressBar *progress_bar;

	if(editor == NULL)
	    return;

	w = editor->status_bar_progress;
	progress = GTK_PROGRESS(w);
	progress_bar = GTK_PROGRESS_BAR(w);

	/* Do activity? */
	if(value < 0.0f)
	{
	    GtkAdjustment *adj = progress->adjustment;

	    /* Get new value based on unknown position */
	    value = gtk_progress_get_value(progress) + 1.0f;
	    if(value > adj->upper)
		value = adj->lower;

	    /* Update progress bar for `activity mode' (unknown limit)
	     * and set new value
	     */
	    gtk_progress_set_activity_mode(
		progress,
		TRUE
	    );
	    gtk_progress_set_show_text(
		progress,
		FALSE
	    );
	    gtk_progress_set_value(
		progress,
		value
	    );

	    /* Reset the last progress position to -1.0, indicating
	     * that there is no known last position since we are
	     * doing activity mode
	     */
	    editor->status_bar_progress_pos_last = -1.0f;
	}
	else
	{

	    /* Clip value to 0.0, 1.0 */
	    if(value > 1.0f)
		value = 1.0f;

	    /* Reset the last progress position if it is greater than
	     * the specified value, implying that the progress has
	     * warped back to the beginning
	     */
	    if(editor->status_bar_progress_pos_last > value)
		editor->status_bar_progress_pos_last = -1.0f;

	    /* Has the value not sufficiently changed? (less than
	     * 0.009 change in value)
	     */
	    if((editor->status_bar_progress_pos_last > 0.0f) &&
	       ((value - editor->status_bar_progress_pos_last) < 0.009f)
	    )
		return;

	    /* Update progress */
	    gtk_progress_set_activity_mode(
		progress,
		FALSE
	    );
	    gtk_progress_set_format_string(
		progress,
		"%p%%"
	    );
	    /* Display text only if value is positive */
	    gtk_progress_set_show_text(
		progress,
		(value > 0.0f) ? TRUE : FALSE
	    );
	    gtk_progress_bar_update(
		progress_bar,
		value 
	    );

	    /* Record the last progress value */
	    editor->status_bar_progress_pos_last = value;
	}

	/* Manage any GTK events so that the value changes to the
	 * GtkWidgets take affect immediately
	 */
	if(allow_gtk_iteration)
	{
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();
	}
}

/*
 *	Deletes the Editor's undo/redo list.
 */
void EditorClearUndosList(editor_struct *editor)
{
	gint i;

	if(editor == NULL)
	    return;

	/* Delete all undos and redos */
	for(i = 0; i < editor->total_undos; i++)
	    EditorUndoDelete(editor, editor->undo[i]);
	g_free(editor->undo);
	editor->undo = NULL;
	editor->total_undos = 0;
  
	for(i = 0; i < editor->total_redos; i++)
	    EditorUndoDelete(editor, editor->redo[i]);
	g_free(editor->redo);
	editor->redo = NULL;
	editor->total_redos = 0;
}

/*
 *	Sets the Editor busy or ready.
 */
void EditorSetBusy(
	editor_struct *editor,
	const gboolean busy
)
{
	GdkCursor	*cur,
			*text_cur;
	GdkWindow *window;
	GtkWidget *w;
	me_core_struct *core;

	if(editor == NULL)
	    return;

	w = editor->toplevel;
	core = editor->core;

	if(busy)
	{
	    editor->busy_count++;
	    if(editor->busy_count > 1)
		return;
	    text_cur = cur = core->busy_cursor;
	}
	else
	{
	    editor->busy_count--;
	    if(editor->busy_count < 0)
		editor->busy_count = 0;
	    if(editor->busy_count > 0)
		return;
	    cur = NULL;				/* Default cursor */
	    text_cur = core->text_cursor;
	}

	window = w->window;
	if(window != NULL)
	    gdk_window_set_cursor(window, cur);

	window = GTK_TEXT(editor->header_comments_text)->text_area;
	if(window != NULL)
	    gdk_window_set_cursor(window, text_cur);

	window = GTK_TEXT(editor->section_text)->text_area;
	if(window != NULL)
	    gdk_window_set_cursor(window, text_cur);

	gdk_flush();
}

/*
 *	Checks if the Editor is mapped.
 */
gboolean EditorIsMapped(editor_struct *editor)
{
	if(editor == NULL)
	    return(FALSE);

	return(GTK_WIDGET_MAPPED(editor->toplevel));
}

/*
 *	Maps the Editor.
 */
void EditorMap(editor_struct *editor)
{
	if(editor == NULL)
	    return;

	gtk_widget_show(editor->toplevel);
}

/*
 *	Unmaps the Editor.
 */
void EditorUnmap(editor_struct *editor)
{       
	if(editor == NULL)
	    return;

	gtk_widget_hide(editor->toplevel);
}

/*
 *	Deletes the Editor.
 */
void EditorDelete(editor_struct *editor)
{
	GtkCList *clist;

	if(editor == NULL)
	    return;

	editor->freeze_count++;

	/* Remove the syntax highlighting timeout */
	GTK_TIMEOUT_REMOVE(editor->synhl_timeout_cb_id);
	editor->synhl_timeout_cb_id = 0;

	/* Delete all the undos and redos */
	EditorClearUndosList(editor);

	/* Delete all the loaded Manual Pages */
	clist = GTK_CLIST(editor->layout_ctree);
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);

	/* Delete the Find In Pages dialog */
	EditorFIPDelete(editor->fip_dlg);

	g_list_free(editor->contents_toplevels_list);
	editor->contents_toplevels_list = NULL;

	GTK_WIDGET_DESTROY(editor->toplevel);
	GTK_ACCEL_GROUP_UNREF(editor->accelgrp);

	if(editor->colormap != NULL)
	{
	    GdkColormap *colormap = editor->colormap;
	    GDK_COLORMAP_FREE_COLOR(colormap, &editor->normal_text_color);
	    GDK_COLORMAP_FREE_COLOR(colormap, &editor->deliminator_text_color);
	    GDK_COLORMAP_FREE_COLOR(colormap, &editor->tag_text_color);
	    GDK_COLORMAP_FREE_COLOR(colormap, &editor->symbol_text_color);
	    GDK_COLORMAP_UNREF(colormap);
	}

	g_free(editor->last_open_path);
	g_free(editor->last_save_as_path);

	editor->freeze_count--;

	g_free(editor);
}
