#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <unistd.h>

#include "url.h"

#include "guiutils.h"
#include "cdialog.h"

#include "editor.h"
#include "editor_cb.h"
#include "editor_values.h"
#include "editor_dnd.h"
#include "editor_fio.h"
#include "viewer_dnd.h"

#include "manedit.h"
#include "config.h"


/* Recieved data command string parsing */
GList *EditorDNDParseCmd(
	const gchar *string,
	editor_struct **editor_rtn
);

/* Inserting to target editor routines */
static GtkCTreeNode *EditorDNDItemAdd(
	editor_struct *editor,
	editor_item_struct *item,
	GtkCTreeNode *parent,
	GtkCTreeNode *sibling
);
static void EditorDNDItemListAdd(
	editor_struct *editor,
	GList *items_list,
	const gint insert_row
);
static GtkCTreeNode *EditorDNDCopyItemIterate(
	me_pixmaps_list_struct *pixmaps_list,
	editor_struct *editor,
	editor_struct *src_editor,
	GtkCTree *ctree,
	GtkCTree *src_ctree,
	GtkCTreeNode *parent_node,
	GtkCTreeNode *sibling_node,
	GtkCTreeNode *src_node
);
static void EditorDNDCopyItems(
	editor_struct *editor,
	editor_struct *src_editor,
	GList *nodes_list,
	const gint insert_row
);


/* Callbacks */
void EditorDNDDataRequestCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);
void EditorDNDDataRecievedCB(
	GtkWidget *widget,
	GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data,
	guint info, guint t,
	gpointer data
);
void EditorDNDDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Parses the given string which should have been recieved as
 *	a DND command.
 *
 *	This function is designed to be called by
 *	EditorDNDDataRecievedCB().
 */
GList *EditorDNDParseCmd(
	const gchar *string,
	editor_struct **editor_rtn
)
{
	gint i;
	guint32 addr_val;
	gchar **strv;
	GList *glist, *nodes_list;
	GtkCList *clist;
	editor_struct *editor = NULL;

	if(editor_rtn != NULL)
	    *editor_rtn = NULL;

	if(string == NULL)
	    return(NULL);

	if(g_strcasepfx(string, "error"))
	    return(NULL);

	/* Parse command, format is as follows:
	 * <editor_ptr> <branch_ptr...>
	 */
	strv = g_strsplit(string, " ", -1);
	if(strv == NULL)
	    return(NULL);

	/* Get the editor pointer
	 *
	 * Format is in hex with no "0x" prefix
	 */
	i = 0;
	if(strv[i] != NULL)
	{
	    const gint status = sscanf(
		strv[i], "%x", (guint32 *)&addr_val
	    );
	    if(status > 0)
		editor = EDITOR(addr_val);

	    i++;
	}
	if(editor_rtn != NULL)
	    *editor_rtn = editor;

	if(editor == NULL)
	    return(NULL);

	/* Get the selected nodes on the Editor */
	clist = GTK_CLIST(editor->layout_ctree);
	nodes_list = NULL;
	for(glist = clist->selection;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	    nodes_list = g_list_append(
		nodes_list,
		glist->data
	    );

	g_strfreev(strv);

	return(nodes_list);
}


/*
 *	Adds a new EditorItem to the GtkCTree.
 */
static GtkCTreeNode *EditorDNDItemAdd(
	editor_struct *editor,
	editor_item_struct *item,
	GtkCTreeNode *parent,
	GtkCTreeNode *sibling
)
{
	gint		i,
			ncolumns;
	gchar **strv;
	GdkBitmap *closed_m = NULL, *opened_m = NULL;
	GdkPixmap *closed_p = NULL, *opened_p = NULL;
	GtkCTree *ctree = GTK_CTREE(editor->layout_ctree);
	GtkCList *clist = GTK_CLIST(ctree);
	GtkCTreeNode *node;
	GtkCTreeRow *row_ptr;
	me_core_struct *core = editor->core;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	if(strv == NULL)
	    return(NULL);

	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	/* Handle by the item data type */
	switch(item->type)
	{
	  case EDITOR_ITEM_FILE:
	    strv[0] = item->name;
	    closed_p = pixmaps_list->manual_closed_20x20;
	    closed_m = pixmaps_list->manual_closed_20x20_mask;
	    opened_p = pixmaps_list->manual_opened_20x20;
	    opened_m = pixmaps_list->manual_opened_20x20_mask;
	    break;

	  case EDITOR_ITEM_HEADER:
	    strv[0] = "Header";
	    closed_p = pixmaps_list->manpage_heading_20x20;
	    closed_m = pixmaps_list->manpage_heading_20x20_mask;
	    opened_p = closed_p;
	    opened_m = closed_m;
	    break;

	  case EDITOR_ITEM_SECTION:
	    strv[0] = item->section_name;
	    closed_p = pixmaps_list->manpage_section_20x20;
	    closed_m = pixmaps_list->manpage_section_20x20_mask;
	    opened_p = closed_p;
	    opened_m = closed_m;
	    break;
	}

	node = item->node;
	row_ptr = (node != NULL) ? GTK_CTREE_ROW(node) : NULL;

	/* Create a new node */
	node = gtk_ctree_insert_node(
	    ctree,
	    parent,
	    sibling,
	    strv,
	    ME_LIST_ICON_TEXT_SPACING,
	    closed_p, closed_m,
	    opened_p, opened_m,
	    (row_ptr != NULL) ? row_ptr->is_leaf : TRUE,
	    FALSE		/* Expanded */
	);
	if(node == NULL)
	    return(NULL);

	/* Update the specified EditorItem's values since its original
	 * Editor and GtkCTreeNode may be different
	 */
	item->editor = editor;
	item->node = node;

	/* Set the new node's data as the new item */
	EditorBranchSetData(
	    ctree,
	    node,
	    item, EditorItemDestroyCB
	);

	/* Set the new EditorItem's has changes flag */
	EditorSetNodeHasChanges(
	    editor,
	    node,
	    TRUE
	);

	g_free(strv);

	return(node);
}

/*
 *      Adds the list of EditorItems to the Editor.
 *
 *	The items_list specifies a GList of EditorItem * items to
 *	transfer to the Editor and should not be referenced after
 *	this call.
 *
 *	Only headers and sections can be added.
 */
static void EditorDNDItemListAdd(
	editor_struct *editor,
	GList *items_list,
	const gint insert_row
)
{
	gboolean append;
	GtkCTreeNode	*insert_node,
			*parent_node;
	GtkCTreeRow *row_ptr;
	GtkCTree *ctree = GTK_CTREE(editor->layout_ctree);
	GtkCList *clist = GTK_CLIST(ctree);
	editor_item_struct *item;

	/* Get the node to insert at or append */
	if((insert_row >= 0) && (insert_row < clist->rows))
	{
	    append = FALSE;
	    insert_node = gtk_ctree_node_nth(
		ctree,
		insert_row
	    );
	}
	else
	{
	    append = TRUE;
	    insert_node = gtk_ctree_node_nth(
		ctree,
		clist->rows - 1
	    );
	}
	if(insert_node == NULL)
	{
	    g_list_foreach(
		items_list,
		(GFunc)EditorItemDelete,
		NULL
	    );
	    g_list_free(items_list);
	    return;
	}

	row_ptr = GTK_CTREE_ROW(insert_node);

	gtk_clist_freeze(clist);

	/* Is the insert node a leaf? */
	if(row_ptr->is_leaf)
	{
	    GList *glist;

	    /* Insert to this node's parent node */
	    parent_node = row_ptr->parent;

	    /* Insert from last to first */
	    for(glist = g_list_last(items_list);
		glist != NULL;
		glist = g_list_previous(glist)
	    )
	    {
		item = EDITOR_ITEM(glist->data);
		if(item == NULL)
		    continue;

		insert_node = EditorDNDItemAdd(
		    editor,
		    item,
		    parent_node,
		    append ? NULL : insert_node
		);
	    }
	}
	else
	{
	    /* Inserting to a toplevel node, so insert it as its
	     * first child
	     */
	    GList *glist;

	    parent_node = insert_node;
	    insert_node = row_ptr->children;

	    /* Insert from last to first */
	    for(glist = g_list_last(items_list);
		glist != NULL;
		glist = g_list_previous(glist)
	    )
	    {
		item = EDITOR_ITEM(glist->data);
		if(item == NULL)
		    continue;

		insert_node = EditorDNDItemAdd(
		    editor,
		    item,
		    parent_node,
		    insert_node
		);
	    }
	}

	gtk_clist_thaw(clist);

	/* Delete the items list */
	g_list_free(items_list);
}


/*
 *	Called by EditorDNDCopyItems() to recursivly copy the source
 *	node from the source editor to the target editor.
 *
 *	Returns the coppied node.
 */
static GtkCTreeNode *EditorDNDCopyItemIterate(
	me_pixmaps_list_struct *pixmaps_list,
	editor_struct *editor,
	editor_struct *src_editor,
	GtkCTree *ctree,
	GtkCTree *src_ctree,
	GtkCTreeNode *parent_node,		/* Parent on target editor */
	GtkCTreeNode *sibling_node,		/* Node to insert at */
	GtkCTreeNode *src_node			/* On the source editor */
)
{
	gint		i,
			ncolumns;
	gchar **strv;
	GdkPixmap *closed_p = NULL, *opened_p = NULL;
	GdkBitmap *closed_m = NULL, *opened_m = NULL;
	GtkCList *clist = GTK_CLIST(ctree);
	GtkCTreeNode *new_node;
	GtkCTreeRow *row_ptr; 
	editor_item_struct	*item,
				*src_item;

	if(src_node == NULL)
	    return(NULL);

	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	if(strv == NULL)
	    return(NULL);

	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	/* Get the source EditorItem */
	src_item = EditorBranchGetData(
	    src_ctree,
	    src_node
	);

	/* Get the source EditorItem's values */
	if(src_item != NULL)
	{
	    switch(src_item->type)
	    {
	      case EDITOR_ITEM_FILE:
		strv[0] = src_item->name;
		closed_p = pixmaps_list->manual_closed_20x20;
		closed_m = pixmaps_list->manual_closed_20x20_mask;
		opened_p = pixmaps_list->manual_opened_20x20;
		opened_m = pixmaps_list->manual_opened_20x20_mask;
		break;

	      case EDITOR_ITEM_HEADER:
		strv[0] = "Header";
		closed_p = pixmaps_list->manpage_heading_20x20;
		closed_m = pixmaps_list->manpage_heading_20x20_mask;
		opened_p = closed_p;
		opened_m = closed_m;
		break;

	      case EDITOR_ITEM_SECTION:
		strv[0] = src_item->section_name;
		closed_p = pixmaps_list->manpage_section_20x20;
		closed_m = pixmaps_list->manpage_section_20x20_mask;
		opened_p = closed_p;
		opened_m = closed_m;
		break;
	    }
	}

	/* Insert a new node on target Editor */
	row_ptr = GTK_CTREE_ROW(src_node);
	new_node = gtk_ctree_insert_node(
	    ctree,
	    parent_node,
	    sibling_node,
	    strv,
	    ME_LIST_ICON_TEXT_SPACING,
	    closed_p, closed_m,
	    opened_p, opened_m,
	    (row_ptr != NULL) ? row_ptr->is_leaf : TRUE,
	    (row_ptr != NULL) ? row_ptr->expanded : FALSE
	);

	/* Copy the EditorItem */
	item = EditorItemCopy(src_item);
	if(item != NULL)
	{
	    /* Set the values on the coppied EditorItem to reflect
	     * the target Editor's values
	     */
	    item->editor = editor;
	    item->node = new_node;

	    EditorBranchSetData(
		ctree,
		new_node,
		item, EditorItemDestroyCB
	    );
	}

	/* Set the coppied EditorItem's has changes flag */
	EditorSetNodeHasChanges(
	    editor,
	    new_node,
	    TRUE
	);

	/* Copy any children on the source node */
	row_ptr = GTK_CTREE_ROW(src_node);
	if((row_ptr != NULL) ? (row_ptr->children != NULL) : FALSE)
	{
	    /* Set the new node as the parent node */
	    parent_node = new_node;

	    sibling_node = NULL;

	    /* Get the first child node on the source node */
	    src_node = row_ptr->children;

	    /* Copy the children */
	    while(src_node != NULL)
	    {
		(void)EditorDNDCopyItemIterate(
		    pixmaps_list,
		    editor,
		    src_editor,
		    ctree,
		    src_ctree,
		    parent_node,
		    sibling_node,		/* NULL, append */
		    src_node
		);

		src_node = GTK_CTREE_NODE_SIBLING(src_node);
	    }
	}

	g_free(strv);

	return(new_node);
}

/*
 *	Coppies the list of toplevel nodes.
 *
 *	The nodes_list specifies the GList of toplevel GtkCTreeNodes
 *	on the source editor to copy.
 */
static void EditorDNDCopyItems(
	editor_struct *editor,
	editor_struct *src_editor,
	GList *nodes_list,
	const gint insert_row
)
{
	GList *glist;
	GtkCTreeNode	*sibling_node,
			*insert_node;
	GtkCList	*clist = GTK_CLIST(editor->layout_ctree),
			*src_clist = GTK_CLIST(src_editor->layout_ctree);
	GtkCTree	*ctree = GTK_CTREE(clist),
			*src_ctree = GTK_CTREE(src_clist);
	me_core_struct *core = editor->core;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	if((insert_row >= 0) && (insert_row < clist->rows))
	    insert_node = gtk_ctree_node_nth(
		ctree,
		insert_row
	    );
	else
	    insert_node = NULL;
	if(insert_node != NULL)
	{
	    /* Get the toplevel node of the node ot insert at */
	    insert_node = EditorGetToplevelNode(insert_node);
	    if(insert_node == NULL)
		return;
	}

	gtk_clist_freeze(clist);

	/* Begin inserting/copying the nodes from the source Editor
	 * to the target Editor
	 */
	sibling_node = insert_node;
	for(glist = g_list_last(nodes_list);
	    glist != NULL;
	    glist = g_list_previous(glist)
	)
	{
	    sibling_node = EditorDNDCopyItemIterate(
		pixmaps_list,
		editor,
		src_editor,
		ctree,
		src_ctree,
		NULL,				/* Toplevel */
		sibling_node,			/* Insert before this node */
		(GtkCTreeNode *)glist->data
	    );
	}

	gtk_clist_thaw(clist);
}

/*
 *	GtkCTree "drag_data_get" signal callback.
 *
 *	Sends out a command containing the information about
 *	the given editor and selected node.
 * 
 *	The selected branches will have their values applied before
 *	the command which contains the selected branch pointers is sent
 *	out.
 */
void EditorDNDDataRequestCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data 
)
{
	gchar *buf;
	GtkCTreeNode *node;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL) || (dc == NULL))
	    return;

	/* Get the selected GtkCTreeNode */
	node = EditorCurrentSelectedNode(editor);

	/* Set the values from the widgets to the EditorItem so that
	 * we have the most up to date data
	 */
	EditorSetValues(
	    editor,
	    node
	);

	/* Allocate buffer that we will be sending out, containing
	 * the following command format:
	 *
	 * <editor_ptr> <branch_ptr...>
	 */
	if(node != NULL)
	    buf = g_strdup_printf(
		"%.8x %.8x",
		(guint32)editor,
		(guint32)node
	    );
	else
	    buf = g_strdup_printf(
		"%.8x",
		(guint32)editor
	    );

	/* Send out the requested data */
	gtk_selection_data_set(
	    selection_data,
	    GDK_SELECTION_TYPE_STRING,
	    8,					/* 8 bits per character */
	    buf, strlen(buf)
	);

	g_free(buf);
}

/*
 *	GtkCTree "drag_data_received" signal callback.
 */
void EditorDNDDataRecievedCB(
	GtkWidget *widget,
	GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data,  
	guint info, guint t,  
	gpointer data
)
{
	gint row, column;
	GtkWidget *toplevel;
	GtkCList *clist;
	GtkCTree *ctree;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (dc == NULL) || (editor == NULL))
	    return;

	if(selection_data == NULL)
	    return;
	if(selection_data->length < 0)
	    return;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(widget);
	clist = GTK_CLIST(ctree);
	core = editor->core;

	/* Get row and column that the drop occured over */
	if(!gtk_clist_get_selection_info(
	    clist,
	    x, y,
	    &row, &column
	))
	{
	    row = clist->rows;
	    column = 0;
	}
	if(row < 0)
	    row = 0;

	/* Editor Branch */
	if(info == MEDIT_DND_TYPE_INFO_EDITOR_BRANCH_CMD)
	{
	    gboolean	item_is_file = FALSE,
			item_is_header = FALSE,
			item_is_section = FALSE;
	    GList *nodes_list;
	    editor_struct *src_editor = NULL;

	    /* Parse the data to get the references to the source
	     * Editor's values
	     */
	    nodes_list = EditorDNDParseCmd(
		(const gchar *)selection_data->data,
		&src_editor
	    );
	    if(src_editor != NULL)
	    {
		GList *glist;
		editor_item_struct *item;

		/* Get the branch item's type */
		for(glist = nodes_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
			GTK_CTREE(src_editor->layout_ctree),
			(GtkCTreeNode *)glist->data
		    ));
		    if(item == NULL)
			continue;

		    switch(item->type)
		    {
		      case EDITOR_ITEM_FILE:
			item_is_file = TRUE;
			break;

		      case EDITOR_ITEM_HEADER:
			item_is_header = TRUE;
			break;

		      case EDITOR_ITEM_SECTION:
			item_is_section = TRUE;
			break;
		    }
		}
		/* Did we get a bad combination of types? */
		if(item_is_file &&
		   (item_is_header || item_is_section)
		)
		{
		    CDialogSetTransientFor(toplevel);
		    CDialogGetResponse(
			"Invalid Operation",
"Unable to drag and drop both file reference and sections\n\
data types at once.",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		    );
		    CDialogSetTransientFor(NULL);
		    g_free(nodes_list);
		    return;
		}
		/* Handle by the branch item type
		 *
		 * File
		 */
		if(item_is_file)
		{
		    /* Add each toplevel trunk branch to the target
		     * editor
		     */
		    EditorDNDCopyItems(
			editor,			/* Target */
			src_editor,		/* Source */
			nodes_list,
			row
		    );

		    gtk_clist_columns_autosize(clist);
		    EditorUpdateMenus(editor);
		}
		/* Header or section? */
		else if(item_is_header || item_is_section)
		{
		    /* Copy the items to the editor */
		    GList	*glist,
				*items_list = NULL;
		    editor_item_struct *item;

		    for(glist = nodes_list;
			glist != NULL;
			glist = g_list_next(glist)
		    )
		    {
			item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
			    GTK_CTREE(src_editor->layout_ctree),
			    (GtkCTreeNode *)glist->data
			));
			if(item == NULL)
			    continue;

			items_list = g_list_append(
			    items_list,
			    EditorItemCopy(item)
			);
		    }

		    gtk_clist_freeze(clist);

		    /* Insert the coppied items to the editor */
		    EditorDNDItemListAdd(
			editor,
			items_list,
			row
		    );

		    gtk_clist_columns_autosize(clist);
		    gtk_clist_thaw(clist);

		    EditorUpdateMenus(editor);
		}
	    }

	    /* Delete the nodes list */
	    g_list_free(nodes_list);
	}
	/* text/plain, text/uri-list, and STRING */
	else if((info == MEDIT_DND_TYPE_INFO_TEXT_PLAIN) ||
		(info == MEDIT_DND_TYPE_INFO_TEXT_URI_LIST) ||
		(info == MEDIT_DND_TYPE_INFO_STRING)
	)
	{
	    GtkCTreeNode *node = EditorCurrentSelectedNode(editor);
	    GList *url_list = URLDecode(
		selection_data->data, selection_data->length
	    );
	    if(url_list  != NULL)
	    {
		gchar *msg;
		GList *glist;
		url_struct *url;

		for(glist = url_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    url = URL(glist->data);
		    if(url == NULL)
			continue;

		    if((url->protocol == NULL) || (url->path == NULL))
			continue;

		    if(g_strcasecmp(url->protocol, "file"))
			continue;

		    /* Open the manual page */
		    switch(EditorFileOpen(
			editor,
			url->path,
			node,
			FALSE			/* Not as a template */
		    ))
		    {
		      /* User aborted */
		      case -5:
		      case -4:
			break;

		      case 0:
			/* Check if the manual page is not writable */
			if(access((const char *)url->path, W_OK))
			{
			    /* Warn that file is read only */
			    CDialogSetTransientFor(toplevel);
			    CDialogGetResponse(
"Manual Page Read-Only",
"This manual page is set read-only,\n\
you will not be able to save it.",
				NULL,
				CDIALOG_ICON_WARNING,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			    );
			    CDialogSetTransientFor(NULL);
			}
			break;

		      default:
			msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
			    g_strerror((gint)errno),
			    url->path
			);
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Open Failed",
			    msg,
			    NULL,
			    CDIALOG_ICON_WARNING,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			g_free(msg);
			CDialogSetTransientFor(NULL);
			break;
		    }

		    /* Only load one */
		    break;
		}

		g_list_foreach(
		    url_list,
		    (GFunc)URLDelete,
		    NULL
		);
		g_list_free(url_list);
	    }
	}
}

/*
 *	GtkCTree "drag_data_delete" signal callback.
 */
void EditorDNDDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
)
{
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *node;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (dc == NULL) || (editor == NULL))
	    return;

	/* Remove the selected nodes */
	ctree = GTK_CTREE(widget);
	clist = GTK_CLIST(ctree);
	gtk_clist_freeze(clist);
	while(clist->selection != NULL)
	{
	    node = (GtkCTreeNode *)clist->selection->data;
	    if(node == NULL)
		break;

	    EditorSetNodeHasChanges(
		editor,
		node,
		TRUE
	    );

	    gtk_ctree_remove_node(
		ctree,
		node
	    );
	}

	gtk_clist_columns_autosize(clist);
	gtk_clist_thaw(clist);

	EditorUpdateMenus(editor);
}
