#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "editor.h"
#include "editor_find_in_pages.h"
#include "editor_find_in_pages_cb.h"
#include "manedit.h"
#include "messages.h"
#include "config.h"

#include "images/icon_manedit_editor_48x48.xpm"

#include "images/icon_close_20x20.xpm"
#include "images/icon_clear_20x20.xpm"
#include "images/icon_goto_20x20.xpm"
#include "images/icon_search_20x20.xpm"
#include "images/icon_stop_20x20.xpm"


void EditorFIPItemDelete(editor_fip_item_struct *item);

gint EditorFIPAddItem(
	editor_fip_struct *fip,
	const gchar *manual_page,
	const gchar *section,
	const gint line_num,
	const gchar *excerpt,
	const gint text_sel_start, const gint text_sel_end,
	GtkCTreeNode *matched_node,
	editor_item_struct *editor_item
);

static void EditorFIPCreateStatusBar(
	editor_fip_struct *fip,
	GtkWidget *parent
);

editor_fip_struct *EditorFIPNew(
	me_core_struct *core,
	editor_struct *editor
);
void EditorFIPUpdateMenus(editor_fip_struct *fip);
void EditorFIPSetStatusMessage(
	editor_fip_struct *fip,
	const gchar *msg
);
void EditorFIPSetStatusProgress(
	editor_fip_struct *fip,
	const gfloat percent
);
void EditorFIPSetBusy(
	editor_fip_struct *fip,
	const gboolean busy
);
gboolean EditorFIPIsMapped(editor_fip_struct *fip);
void EditorFIPMap(editor_fip_struct *fip);
void EditorFIPUnmap(editor_fip_struct *fip);
void EditorFIPDelete(editor_fip_struct *fip);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Deletes the item.
 */
void EditorFIPItemDelete(editor_fip_item_struct *item)
{
	if(item == NULL)
	    return;

	g_free(item);
}


/*
 *	Adds a new item to the Results GtkCList.
 */
gint EditorFIPAddItem(
	editor_fip_struct *fip,
	const gchar *manual_page,
	const gchar *section,
	const gint line_num,
	const gchar *excerpt,
	const gint text_sel_start, const gint text_sel_end,
	GtkCTreeNode *matched_node,
	editor_item_struct *editor_item
)
{
	gint i, row, ncolumns;
	gchar **strv;
	GtkCList *clist;
	editor_fip_item_struct *item;

	if(fip == NULL)
	    return(-2);

	clist = GTK_CLIST(fip->results_clist);

	/* Allocate the row cell values */
	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	if(strv == NULL)
	    return(-3);

	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	/* Create a new EditorFIPItem */
	item = EDITOR_FIP_ITEM(g_malloc0(sizeof(editor_fip_item_struct)));
	if(item == NULL)
	    return(-3);

	item->fip = fip;
	item->matched_node = matched_node;
	item->editor_item = editor_item;
	item->text_sel_start = text_sel_start;
	item->text_sel_end = text_sel_end;
	item->line_num = line_num;

	gtk_clist_freeze(clist);

	if(ncolumns >= 4)
	{
	    gchar *line_num_str;

	    if(line_num < 0)
		line_num_str = g_strdup("");
	    else
		line_num_str = g_strdup_printf("%i", line_num + 1);

	    strv[0] = g_strdup((manual_page != NULL) ? manual_page : "");
	    strv[1] = g_strdup((section != NULL) ? section : "");
	    strv[2] = line_num_str;
	    strv[3] = g_strdup((excerpt != NULL) ? excerpt : "");

	    row = gtk_clist_append(
		clist,
		strv
	    );

	    g_free(strv[0]);
	    g_free(strv[1]);
	    g_free(strv[2]);
	    g_free(strv[3]);
	}
	else
	{
	    row = -1;
	}

	g_free(strv);

	if(row > -1)
	{
	    gtk_clist_set_row_data_full(
		clist,
		row,
		item, EditorFIPItemDestroyCB
	    );
	}
	else
	{
	    EditorFIPItemDelete(item);
	}

	gtk_clist_columns_autosize(clist);

	gtk_clist_thaw(clist);

	return(row);
}


/*
 *      Creates the status bar for the dialog, this function should be
 *      called by EditorFIPNew().
 *
 *      Given parent is assumed to be a vbox.
 */
static void EditorFIPCreateStatusBar(
	editor_fip_struct *fip,
	GtkWidget *parent
)
{
	const gint border_minor = 2;
	GtkAdjustment *adj;
	GtkWidget		*w,
				*parent2, *parent3;

	/* Outer GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_widget_set_usize(
	    w,
	    -1,
	    ME_STATUS_BAR_HEIGHT
	);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Main GtkTable */
	w = gtk_table_new(
	    1, 2,                               /* Rows columns */
	    FALSE                               /* Not homogenious */
	);
	gtk_table_set_col_spacings(GTK_TABLE(w), 1);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkProgressBar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1.0f, 100.0f,
	    0.0f, 0.0f, 0.0f
	);
	fip->status_bar_progress = w = gtk_progress_bar_new_with_adjustment(adj);
	gtk_widget_set_usize(
	    w,
	    100,
	    ME_STATUS_BAR_HEIGHT - (2 * 3)
	);
	gtk_progress_bar_set_orientation(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_LEFT_TO_RIGHT
	);
	gtk_progress_bar_set_bar_style(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_CONTINUOUS
	);
	gtk_progress_set_activity_mode(
	    GTK_PROGRESS(w),
	    FALSE
	);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    0, 1,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);


	/* Message GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    1, 2,
	    0, 1,
	    GTK_SHRINK | GTK_EXPAND | GTK_FILL,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Message GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Message Gtklabel */
	fip->status_bar_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);
}



/*
 *	Creates a new Find In Pages Dialog.
 */
editor_fip_struct *EditorFIPNew(
	me_core_struct *core,
	editor_struct *editor
)
{
	const gint	border_major = 5,
			border_minor = 2;
	gchar *heading[4];
	guint		accel_key,
			accel_mods;
	gpointer	combo_rtn,
			data;
	const gchar *label;
	guint8 **icon;
	GSList *gslist;
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*parent5,
			*toplevel,
			*menu;
	GtkCList *clist;
	void (*func_cb)(GtkWidget *w, gpointer);
	editor_fip_struct *fip;

	if((core == NULL) || (editor == NULL))
	    return(NULL);

	fip = EDITOR_FIP(g_malloc0(sizeof(editor_fip_struct)));
	if(fip == NULL)
	    return(NULL);

	fip->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	fip->accelgrp = accelgrp = gtk_accel_group_new();
/*
	fip->freeze_count = 0;
	fip->busy_count = 0;
	fip->stop_count = 0;

	fip->processing = FALSE;
 */
	fip->core = core;
	fip->editor = editor;

	fip->freeze_count++;

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SUBMENU, accelgrp,	\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
 if(w != NULL)					\
  GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,	\
  NULL, NULL, 0, 0,				\
  NULL, NULL					\
 );						\
}

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, FALSE);
	gtk_window_set_title(GTK_WINDOW(w), PROG_NAME ": Find In Pages");
	gtk_widget_set_usize(
	    w,
	    ME_EDITOR_FIP_DEF_WIDTH,
	    ME_EDITOR_FIP_DEF_HEIGHT
	);
	gtk_widget_realize(w);   
	GUISetWMIcon(w->window, (guint8 **)icon_manedit_editor_48x48_xpm);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(EditorFIPDeleteEventCB), fip
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;

	/* Main GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* GtkHBox for two columns */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Left side GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Match Options GtkVBox */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Match All Occurances */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	gslist = NULL;
	fip->match_all_radio = w = gtk_radio_button_new_with_label(
	    gslist,
	    "Match All Occurances"
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	gtk_widget_show(w);

	/* Match Any */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	fip->match_any_radio = w = gtk_radio_button_new_with_label(
	    gslist,
	    "Match Any Occurance"
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Phrase */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* Match phrase radio button */
	fip->match_phrase_radio = w = gtk_radio_button_new_with_label(
	    gslist,
	    "Match Phrase Occurance"
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Find GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 2);
	gtk_widget_show(w);
	parent4 = w;

	/* Find GtkCombo */
	w = GUIComboCreate(
	    "Of:",
	    NULL,
	    NULL,
	    20,
	    &combo_rtn,
	    fip,
	    EditorFIPFindActivateCB,
	    NULL
	);
	fip->find_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Additional Options GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 2);
	gtk_widget_show(w);
	parent4 = w;

	/* Case Sensitive GtkCheckButton */
	fip->case_sensitive_toggle = w = gtk_check_button_new_with_label(
	    "Case Sensitive"
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_toggle_button_set_active(
	    GTK_TOGGLE_BUTTON(w),
	    FALSE
	);
	gtk_widget_show(w);


	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* GtkScrolledWindow for the Results GtkCList */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),   
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Results GtkCList */
	heading[0] = "Manual Page";
	heading[1] = "Section";
	heading[2] = "Line";
	heading[3] = "Excerpt";
	fip->results_clist = w = gtk_clist_new_with_titles(
	    sizeof(heading) / sizeof(gchar *),
	    heading
	);
	clist = GTK_CLIST(w);
/*	gtk_widget_set_usize(w, -1, 240); */
	gtk_widget_set_events(
	    w,
	    GDK_BUTTON_PRESS_MASK
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(EditorFIPMenuMapCB), fip
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "select_row",
	    GTK_SIGNAL_FUNC(EditorFIPResultsListSelectCB), fip
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "unselect_row",
	    GTK_SIGNAL_FUNC(EditorFIPResultsListUnselectCB), fip
	);
	gtk_signal_connect( 
	    GTK_OBJECT(w), "click_column",
	    GTK_SIGNAL_FUNC(EditorFIPResultsListColumnClickCB), fip
	);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_SINGLE);
	gtk_clist_set_sort_type(clist, GTK_SORT_ASCENDING);
	gtk_clist_set_sort_column(clist, 3);	/* By manual page name */
	gtk_clist_set_row_height(clist, ME_LIST_ROW_SPACING);
	gtk_clist_set_column_justification(
	    clist,
	    0,
	    GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_width(clist, 0, 100);
	gtk_clist_set_column_auto_resize(
	    clist,
	    0,
	    TRUE
	);
	gtk_clist_set_column_justification(
	    clist,
	    1,
	    GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_width(clist, 1, 100);
	gtk_clist_set_column_auto_resize(
	    clist,
	    1,
	    TRUE
	);
	gtk_clist_set_column_justification(
	    clist,
	    2,
	    GTK_JUSTIFY_RIGHT
	);
	gtk_clist_set_column_width(clist, 2, 100); 
	gtk_clist_set_column_auto_resize(
	    clist,
	    2,
	    TRUE
	);
	gtk_clist_set_column_justification(
	    clist,
	    3,
	    GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_width(clist, 3, 100); 
	gtk_clist_set_column_auto_resize(
	    clist,
	    3,
	    TRUE
	);
	gtk_widget_show(w);

	/* Results GtkCList Right Click GtkMenu */
	fip->results_menu = menu = GUIMenuCreate();
	data = fip;
 
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_goto_20x20_xpm;
	    label = "Goto";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = EditorFIPGotoCB;
	    ADD_MENU_ITEM_LABEL
	    fip->goto_mi = w;
	}

	/* Results Buttons GtkHBox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Goto GtkButton */
	fip->goto_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_goto_20x20_xpm,
	     "GoTo",
	     NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFIPGotoCB), fip
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_g, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_g);
	gtk_widget_show(w);


	/* Right side GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Find, Stop, Clear GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Search GtkButton */
	fip->find_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_search_20x20_xpm,
	    "Search",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFIPFindCB), fip
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_s, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_s);
	gtk_widget_show(w);

	/* Stop GtkButton */
	fip->stop_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_stop_20x20_xpm,
	    "Stop",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFIPStopCB), fip
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_t, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_t);
	gtk_widget_show(w);

	/* Clear GtkButton */
	fip->clear_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_clear_20x20_xpm,
	    "Clear",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFIPClearCB), fip
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_l, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_l);
	gtk_widget_show(w);


	/* Close GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_end(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Close button */ 
	fip->close_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_close_20x20_xpm,
	    "Close",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(EditorFIPCloseMCB), fip
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_c, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);




#undef ADD_MENU_ITEM_LABEL  
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR

	/* Status Bar */
	EditorFIPCreateStatusBar(fip, parent);

	EditorFIPUpdateMenus(fip);

	fip->freeze_count--;

	return(fip);
}


/*
 *	Updates menus on dialog.
 */
void EditorFIPUpdateMenus(editor_fip_struct *fip)
{
	gboolean sensitive;
	gint row;
	GtkCList *clist;

	if(fip == NULL)
	    return;

	fip->freeze_count++;

	clist = GTK_CLIST(fip->results_clist);
	if(clist->selection_end != NULL)
	    row = (gint)clist->selection_end->data;
	else
	    row = -1;

	/* Results GtkCList Right Click GtkMenu */
	if(fip->results_menu != NULL)
	{
	    sensitive = (row > -1) ? TRUE : FALSE;
	    gtk_widget_set_sensitive(fip->goto_mi, sensitive);
	}

	sensitive = fip->processing ? FALSE : TRUE;
	gtk_widget_set_sensitive(fip->match_all_radio, sensitive);
	gtk_widget_set_sensitive(fip->match_any_radio, sensitive);
	gtk_widget_set_sensitive(fip->match_phrase_radio, sensitive);
	gtk_widget_set_sensitive(fip->case_sensitive_toggle, sensitive);

	gtk_widget_set_sensitive(fip->find_combo, sensitive);

	sensitive = fip->processing ? FALSE : TRUE;
	gtk_widget_set_sensitive(fip->find_btn, sensitive);

	sensitive = fip->processing ? TRUE : FALSE;
	gtk_widget_set_sensitive(fip->stop_btn, sensitive);

	sensitive = fip->processing ? FALSE : TRUE;
	gtk_widget_set_sensitive(fip->clear_btn, sensitive);

	sensitive = fip->processing ? FALSE : TRUE;
	gtk_widget_set_sensitive(fip->close_btn, sensitive);

	sensitive = (row > -1) ? TRUE : FALSE;
	gtk_widget_set_sensitive(fip->goto_btn, sensitive);

	fip->freeze_count--;
}

/*
 *	Updates status message.
 */
void EditorFIPSetStatusMessage(
	editor_fip_struct *fip,
	const gchar *msg
)
{
	GtkWidget *w;

	if(fip == NULL)
	    return;

	w = fip->status_bar_label;
	if(w == NULL)
	    return;
 
	gtk_label_set_text(
	    GTK_LABEL(w),
	    (msg == NULL) ? "" : msg
	);

	/* Itterate through GTK main function until all events are handled
	 * so that we ensure this label text is updated
	 */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();
}

/*
 *	Updates the progress bar on the dialog.
 */
void EditorFIPSetStatusProgress(
	editor_fip_struct *fip,
	const gfloat percent
)
{
	gfloat		_percent = percent,
			p;
	GtkWidget *w;

	if(fip == NULL)
	    return;

	w = fip->status_bar_progress;
	if(w == NULL)
	    return;  

	/* Negative? Implies just do activity */        
	if(_percent < 0.0)
	{
	    GtkAdjustment *adj = GTK_PROGRESS(w)->adjustment;
	    _percent = gtk_progress_get_value(GTK_PROGRESS(w)) + 1;

	    if(_percent > adj->upper)
		_percent = adj->lower;

	    gtk_progress_set_activity_mode(
		GTK_PROGRESS(w), TRUE
	    );
	    gtk_progress_set_value(GTK_PROGRESS(w), _percent);

	    /* Itterate through GTK main function until all events are
	     * handled so that we ensure this progress bar setting is
	     * updated.
	     */
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    return;
	}

	if(_percent > 1.0)
	    _percent = 1.0; 
	    
	gtk_progress_set_activity_mode(
	    GTK_PROGRESS(w), FALSE
	);
	gtk_progress_set_format_string(
	    GTK_PROGRESS(w), "%p%%"
	);
	gtk_progress_set_show_text(
	    GTK_PROGRESS(w), (_percent > 0) ? TRUE : FALSE
	);
	
	p = fip->status_bar_progress_pos_last;
	if(p > _percent)
	    p = _percent;

	while(1)
	{
	    if(p > _percent)
		p = _percent;

	    gtk_progress_bar_update(
		GTK_PROGRESS_BAR(w), p
	    );   

	    /* Itterate through GTK main function until all events are
	     * handled so that we ensure this progress bar setting is
	     * updated.
	     */
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();
 
	    if(p < _percent)
		p += (gfloat)0.05;
	    else
		break;
	}
	 
	fip->status_bar_progress_pos_last = _percent;
}

/*
 *	SEts the dialoga busy or ready.
 */
void EditorFIPSetBusy(
	editor_fip_struct *fip,
	const gboolean busy
)
{
	GdkCursor *cur;
	GdkWindow *window;
	GtkWidget *w;
	me_core_struct *core;

	if(fip == NULL)
	    return;

	w = fip->toplevel;
	core = fip->core;

	if(busy)
	{
	    fip->busy_count++;
	    if(fip->busy_count > 1)
		return;
	    cur = core->busy_cursor;
	}
	else
	{
	    fip->busy_count--;
	    if(fip->busy_count < 0)
		fip->busy_count = 0;
	    if(fip->busy_count > 0)
		return;
	    cur = NULL;				/* Default cursor */
	}

	window = w->window;
	if(window != NULL)
	    gdk_window_set_cursor(window, cur);
}       

/*
 *	Checks if the dialog is mapped.
 */
gboolean EditorFIPIsMapped(editor_fip_struct *fip)
{
	if(fip == NULL)
	    return(FALSE);

	return(GTK_WIDGET_MAPPED(fip->toplevel));
}

/*
 *	Maps the dialog.
 */
void EditorFIPMap(editor_fip_struct *fip)
{
	if(fip == NULL)
	    return;
	
	gtk_widget_show(fip->toplevel);
}

/*
 *	Unmaps the dialog.
 */
void EditorFIPUnmap(editor_fip_struct *fip)
{
	if(fip == NULL)
	    return;
	
	gtk_widget_hide(fip->toplevel);
}

/*
 *	Deletes the dialog.
 */
void EditorFIPDelete(editor_fip_struct *fip)
{
	GtkCList *clist;

	if(fip == NULL)
	    return;
	
	if(fip->processing)
	    return;

	fip->freeze_count++;

	clist = GTK_CLIST(fip->results_clist);
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);

	GTK_WIDGET_DESTROY(fip->results_menu);
	GTK_WIDGET_DESTROY(fip->toplevel);
	GTK_ACCEL_GROUP_UNREF(fip->accelgrp);

	fip->freeze_count--;

	g_free(fip);
}
