#define _GNU_SOURCE

#include <string.h>
#include <gtk/gtk.h>

#include "../include/string.h"

#include "editor.h"
#include "editor_cb.h"
#include "editor_find_replace.h"

#include "manedit.h"


gboolean EditorFind(
	editor_struct *editor,
	GtkText *text,
	const gchar *haystack,
	const gchar *needle,
	const gint haystack_len,
	const gint start_pos,
	const gboolean case_sensitive,
	const gboolean move_to,
	gboolean *search_wrapped
);
gboolean EditorReplace(
	editor_struct *editor,
	GtkText *text,
	const gchar *haystack,
	const gchar *needle,
	const gchar *replacement,
	const gint haystack_len,
	const gint start_pos,
	const gboolean case_sensitive
);
gint EditorReplaceAll(
	editor_struct *editor,
	GtkText *text,
	const gchar *needle,
	const gchar *replacement,
	const gboolean case_sensitive
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Find.
 *
 *	The text specifies the GtkText to find the string in.
 *
 *	The haystack specifies the string to search through.
 *
 *	The needle specifies the string to match.
 *
 *	The kaystack_len specifies the length of haystack to search.
 *
 *	The start_pos specifies the starting position in haystack.
 *
 *	If case_sensitive is TRUE then a case sensitive search will
 *	be made.
 *
 *	If move_to is TRUE then the GtkText will be scrolled to the
 *	matched string.
 *
 *	The search_wrapped specifies the return value for the search
 *	wrapped flag. If the search has wrapped then *search_wrapped
 *	will be set to TRUE.
 *
 *	Returns TRUE if a match is made.
 */
gboolean EditorFind(  
	editor_struct *editor,
	GtkText *text,
	const gchar *haystack,
	const gchar *needle,
	const gint haystack_len,
	const gint start_pos,
	const gboolean case_sensitive,
	const gboolean move_to,
	gboolean *search_wrapped
)
{
	gint		_haystack_len = haystack_len,
			_start_pos = start_pos;
	const gchar *match_ptr;
	GtkEditable *editable;

	if(search_wrapped != NULL)
	    *search_wrapped = FALSE;

	if((editor == NULL) || (text == NULL) ||
	   (haystack == NULL) || (needle == NULL)
	)
	    return(FALSE);

	editable = GTK_EDITABLE(text);

	if(_haystack_len < 0)
	    _haystack_len = strlen(haystack);

	/* Empty haystack? */
	if(_haystack_len < 1)
	    return(FALSE);

	/* Empty needle? */
	if(*needle == '\0')
	    return(FALSE);

	editor->freeze_count++;

	EditorSetStatusMessage(editor, "Searching...");

	/* Clip the starting position */
	if(_start_pos >= _haystack_len)
	    _start_pos = 0;
	else if(_start_pos < 0)
	    _start_pos = 0;

	/* Begin the search from the starting position */
	if(case_sensitive)
	    match_ptr = (const gchar *)strstr(
		(const char *)(haystack + _start_pos),
		(const char *)needle
	    );
	else
	    match_ptr = (const gchar *)strcasestr(
		(const char *)(haystack + _start_pos),
		(const char *)needle
	    );
	if(match_ptr == NULL)
	{
	    /* Did not get match, so start from the beginning of
	     * haystack and search again
	     */
	    if(case_sensitive)
		match_ptr = (const gchar *)strstr(
		    (const char *)haystack,
		    (const char *)needle
		);
	    else
		match_ptr = (const gchar *)strcasestr(
		    (const char *)haystack,
		    (const char *)needle
		);
	    if(match_ptr != NULL)
	    {
		/* Got match the second time, set search wrapped flag */
		if(search_wrapped != NULL)
		    *search_wrapped = TRUE;
	    }
	}

	EditorSetStatusMessage(editor, "Search done");

	/* Did we get a match? */
	if(match_ptr != NULL)
	{
	    /* Got match */
	    const gint needle_len = strlen(needle);
	    gint match_start_pos = (gint)(match_ptr - haystack);
	    if(match_start_pos < 0)
		match_start_pos = 0;

	    /* Scroll to the matched position? */
	    if(move_to)
	    {
		gtk_text_freeze(text);
		gtk_text_set_point(
		    text,
		    match_start_pos
		);
		gtk_editable_set_position(
		    editable,
		    match_start_pos
		);
		gtk_text_thaw(text);
	    }

	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    if(needle_len > 0)
	    {
		gtk_text_freeze(text);
		gtk_editable_select_region(
		    editable,
		    match_start_pos,
		    match_start_pos + needle_len
		);
		gtk_text_thaw(text);
	    }

	    editor->freeze_count--;

	    return(TRUE);
	}
	else
	{
	    /* No match */
	    editor->freeze_count--;
	    return(FALSE);
	}
}


/*
 *	Replace.
 *
 *	If the given GtkText widget already has a selection marked
 *	then no find will be made and the selected text will be
 *	replaced with the replacement.
 *
 *	Returns TRUE if a replace was done and FALSE if no replace
 *	was made or error.
 */
gboolean EditorReplace(
	editor_struct *editor,
	GtkText *text,
	const gchar *haystack,
	const gchar *needle,
	const gchar *replacement,
	const gint haystack_len,
	const gint start_pos,
	const gboolean case_sensitive
)
{
	gint		_start_pos = start_pos,
			delete_len,
			replacement_len;
	GtkEditable *editable;
	me_core_struct *core;

	if((editor == NULL) || (text == NULL) ||
	   (haystack == NULL) || (needle == NULL) || (replacement == NULL)
	)
	    return(FALSE);

	editor->freeze_count++;

	editable = GTK_EDITABLE(text);
	core = editor->core;

	/* Get the length of the replacement string */
	replacement_len = strlen(replacement);

	/* GtkText already has selection? */
	if(editable->has_selection)
	{
	    /* Text already selected */

	    EditorSetStatusMessage(editor, "Replacing...");

	    /* Set new starting position, overriding the given one */
	    _start_pos = editable->selection_start_pos;

	    /* Get length of text to delete, can be 0 */
	    delete_len = MAX(
		(editable->selection_end_pos - editable->selection_start_pos),
		0
	    );
	}
	else
	{
	    /* Text not selected, so we need to find first */
	    gboolean	search_wrapped,
			got_match = EditorFind(
		editor,
		text,
		haystack,
		needle,
		haystack_len,
		_start_pos,
		case_sensitive,
		TRUE,				/* Scroll to matched position */
		&search_wrapped
	    );

	    EditorSetStatusMessage(editor, "Replacing...");

	    /* Did not get a match? */
	    if(!got_match)
	    {
		/* No match, then that means we have nothing to replace */
		EditorSetStatusMessage(editor, "Replace done");
		editor->freeze_count--;
		return(FALSE);
	    }

	    /* Now hopefully the text widget has its selection and marked */
	    if(!editable->has_selection)
	    {
		/* Something went wrong, supposedly got match but GtkText
		 * widget was not marked as having a selection.
		 */
		EditorSetStatusMessage(editor, "Replace done");
		editor->freeze_count--;
		return(FALSE);
	    }

	    /* Set new starting position, overriding the given one */
	    _start_pos = editable->selection_start_pos;

	    /* Get length of text to delete, can be 0 */
	    delete_len = MAX(
		editable->selection_end_pos - editable->selection_start_pos,
		0
	    );
	}

	/* Now we do replace, start_pos and delete_len should be valid
	 * now.
	 */

	/* Any text to delete? */
	if(delete_len > 0)
	{
	    if(use_text_delete)
	    {
		EditorTextDeleteCB(
		    editable,
		    _start_pos,
		    _start_pos + delete_len,
		    editor
		);
		gtk_text_freeze(text);
		gtk_text_set_point(text, (guint)_start_pos);
		gtk_text_forward_delete(text, (guint)delete_len);
		gtk_text_thaw(text);
	    }
	    else
	    {
		gtk_text_freeze(text);
		gtk_editable_delete_text(
		    editable, _start_pos, _start_pos + delete_len
		);
		gtk_text_thaw(text);
	    }
	}

	/* Insert replacement string? */
	if(replacement_len > 0)
	{
	    gint position;
	    GdkFont *font = core->editable_font;

	    /* Set insert point and insert replacement text */
	    gtk_text_set_point(text, _start_pos);
	    gtk_text_freeze(text);
	    gtk_text_insert(
		text,
		font,
		NULL,
		NULL,
		replacement,
		replacement_len
	    );
	    gtk_text_thaw(text);

	    /* Call insert callback after inserting text */
	    position = _start_pos;
	    EditorTextInsertCB(
		editable,
		replacement,
		replacement_len,
		&position,
		editor
	    );

	    /* Move cursor position after replacement */
	    gtk_text_set_point(
		text,
		_start_pos + replacement_len
	    );
	    gtk_editable_set_position(
		editable,
		_start_pos + replacement_len
	    );

	    /* Do not select replacement */
	}


	EditorSetStatusMessage(editor, "Replace done");

	editor->freeze_count--;

	return(TRUE);
}

/*
 *	Replace all.
 *
 *      This function will freeze and thaw the given GtkText widget but
 *      not set its sensitivity.      
 *
 *      This function calls EditorFind() so the given haystack
 *      and needle may be modified. 
 *
 *	Returns the number of replacements made or 0 on error or no
 *	matches.
 */
gint EditorReplaceAll(
	editor_struct *editor,
	GtkText *text,
	const gchar *needle,
	const gchar *replacement,
	const gboolean case_sensitive
)
{
	gint		orig_pos,
			noccurances_replaced = 0,
			delete_len,
			replacement_len;
	GtkEditable *editable;
	me_core_struct *core;

	if((editor == NULL) || (text == NULL) ||
	   (needle == NULL) || (replacement == NULL)
	)
	    return(noccurances_replaced);

	editor->freeze_count++;

	editable = GTK_EDITABLE(text);
	core = editor->core;

	/* Get the length of the replacement string */
	replacement_len = strlen(replacement);

	/* Record initial position */
	orig_pos = (gint)gtk_editable_get_position(editable);

	if(TRUE)
	{
	    /* Seek back to beginning and do a through search, for each
	     * matched find, replace it with replacement.
	     */
	    gboolean	search_wrapped,
			got_match;
	    gint	start_pos,
			next_pos = 0;
	    gchar *haystack = NULL;

	    do
	    {
		/* Get latest haystack from text widget */
		g_free(haystack);
		haystack = gtk_editable_get_chars(editable, 0, -1);

		/* Find */
		got_match = EditorFind(
		    editor,
		    text,
		    haystack,
		    needle,
		    -1,
		    next_pos,
		    case_sensitive,
		    FALSE,			/* No scroll to matched position */
		    &search_wrapped
		);
		if(!got_match)
		    break;

		/* If the search wrapped then do not perform any more
		 * replacements since it may have matched an embedded
		 * part of needle in the replacement for example.
		 */
		if(search_wrapped)
		    break;

		/* Now hopefully the text widget has its selection and
		 * marked.
		 */
		if(!editable->has_selection)
		{
		    /* Something went wrong, supposedly got match but
		     * GtkText widget was not marked as having a
		     * selection.
		     */
		    break;
		}

		/* Get start of selection as start position */
		start_pos = editable->selection_start_pos;

		/* Set next starting position, being past the current
		 * selected starting position and past what would be
		 * the length of the replacement string.
		 */
		next_pos = start_pos + replacement_len;

		/* Get length of text to delete, can be 0 */
		delete_len = MAX(
		    (editable->selection_end_pos -
			editable->selection_start_pos),
		    0
		);

		/* Any text to delete? */
		if(delete_len > 0)
		{
		    if(use_text_delete)
		    {
			EditorTextDeleteCB(
			    editable,
			    start_pos,
			    start_pos + delete_len,
			    editor
			);
			gtk_text_freeze(text);
			gtk_text_set_point(text, (guint)start_pos);
			gtk_text_forward_delete(text, (guint)delete_len);
			gtk_text_thaw(text);
		    }
		    else
		    {
			gtk_text_freeze(text);
			gtk_editable_delete_text(
			    editable, start_pos, start_pos + delete_len
			);
			gtk_text_thaw(text);
		    }
		}

		/* Insert replacement string? */
		if(replacement_len > 0)
		{
		    gint position;
		    GdkFont *font = core->editable_font;

		    /* Set insert point and insert replacement text */
		    gtk_text_set_point(text, start_pos);
		    gtk_text_freeze(text);
		    gtk_text_insert(
			text,
			font,
			NULL,
			NULL,
			replacement,
			replacement_len
		    );
		    gtk_text_thaw(text);

		    /* Call insert callback after inserting text */
		    position = start_pos;
		    EditorTextInsertCB(
			editable,
			replacement,
			replacement_len,
			&position,
			editor
		    );

		    /* Move cursor position after replacement */
		    gtk_text_set_point(
			text,
			start_pos + replacement_len
		    );
#if 0
		    gtk_editable_set_position(
			editable,
			start_pos + replacement_len
		    );
#endif
		    /* Do not select the replacement */
		}

		noccurances_replaced++;

	    } while(1);

	    g_free(haystack);
	}

	/* Set back the original position */
	if(TRUE)
	{
	    const gint len = (gint)gtk_text_get_length(text);
	    if(orig_pos > len)
		orig_pos = len;
	    gtk_text_set_point(
		text,
		orig_pos
	    );
/*	    gtk_editable_set_position(
		editable,
		orig_pos
	    );
 */
	}

	editor->freeze_count--;

	return(noccurances_replaced);
}
