#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <unistd.h>

#include "../include/string.h"

#include "guiutils.h"
#include "cdialog.h"

#include "manpage_fio.h"
#include "editor.h"
#include "editor_cb.h"
#include "editor_fio.h"
#include "editor_values.h"

#include "manedit.h"
#include "config.h"


static gint EditorFileOpenProgressCB(
	gulong cur, gulong max, gpointer data
);
static gint EditorFileSaveProgressCB(
	gulong cur, gulong max, gpointer data
);

static void EditorFileOpenParseHeader(
	editor_struct *editor,
	editor_item_struct *item,
	MPHeader *head
);

static void EditorFileSaveFormatHeader(
	editor_struct *editor,
	editor_item_struct *item,
	MPHeader *head
);
static void EditorFileSaveFormatSection(
	editor_struct *editor,
	editor_item_struct *item,
	MPSection *sect
);

gint EditorFileOpen(
	editor_struct *editor,
	const gchar *path,
	GtkCTreeNode *node,
	const gboolean as_template
);
gint EditorFileSave(
	editor_struct *editor,
	GtkCTreeNode *node
);
gint EditorFileSaveAs(
	editor_struct *editor,
	const gchar *path,
	GtkCTreeNode *node
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Editor file load progress callback.
 *
 *	Updates the progress bar.
 */
static gint EditorFileOpenProgressCB(
	gulong cur, gulong max, gpointer data
)
{
	gfloat percent;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return(0);

	if(max > 0l)
	    percent = (gfloat)cur / (gfloat)max;
	else
	    percent = 1.0f;

	EditorSetStatusProgress(editor, percent);

	return(0);
}

/*
 *      Editor file save progress callback.
 *
 *      Updates the progress bar.
 */
static gint EditorFileSaveProgressCB(
	gulong cur, gulong max, gpointer data
)
{ 
	gfloat percent;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return(0);

	if(max > 0l)
	    percent = (gfloat)cur / (gfloat)max;
	else
	    percent = 1.0f;

	EditorSetStatusProgress(editor, percent);

	return(0);
}


/*
 *	Gets the header values to the item.
 */
static void EditorFileOpenParseHeader(
	editor_struct *editor,
	editor_item_struct *item,
	MPHeader *head
)
{
	g_free(item->header_name);
	item->header_name = STRDUP(head->name);
	g_free(item->header_section_number);
	item->header_section_number = STRDUP(head->section_number);
	g_free(item->header_version);
	item->header_version = STRDUP(head->version);
	g_free(item->header_author);
	item->header_author = STRDUP(head->author);
	g_free(item->header_catagory);
	item->header_catagory = STRDUP(head->catagory);

	/* Comment lines have already been transfered */
}


/*
 *	Preprocessing for the manual page header structure, coppies values
 *	from the item data structure to the manual page header structure
 *	and formats it.
 *
 *	Last line on the manual page header structure will be appended
 *	as a title heading statement and all other lines will be added
 *	first as comments.
 */
static void EditorFileSaveFormatHeader(
	editor_struct *editor,
	editor_item_struct *item, 
	MPHeader *head
)
{
	gint i;

	/* Values */
	head->name = STRDUP(item->header_name);
	head->section_number = STRDUP(item->header_section_number);
	head->version = STRDUP(item->header_version);
	head->author = STRDUP(item->header_author);
	head->catagory = STRDUP(item->header_catagory);

	/* Comments */
	head->ncomments_lines = item->nlines;
	if(head->ncomments_lines > 0)
	{
	    head->comments_lines_list = (gchar **)g_malloc(
		head->ncomments_lines * sizeof(gchar *)
	    );
	    if(head->comments_lines_list == NULL)
	    {
		head->ncomments_lines = 0;
		return;
	    }
	    for(i = 0; i < item->nlines; i++)
		head->comments_lines_list[i] = STRDUP(item->lines_list[i]);
	}
}

/*
 *      Preprocessing for the manual page section structure, coppies
 *	values from the item data structure to the manual page section
 *	structure and formats it.
 *
 *	Section name will be coppied as well.
 */
static void EditorFileSaveFormatSection(
	editor_struct *editor,
	editor_item_struct *item,
	MPSection *sect
)
{
	/* Section name */
	sect->name = STRDUP(item->section_name);

	/* Lines */
	sect->nlines = item->nlines;
	if(sect->nlines > 0)
	{
	    gint i;

	    sect->lines_list = (gchar **)g_malloc(
		sect->nlines * sizeof(gchar *)
	    );
	    if(sect->lines_list == NULL)
	    {
		sect->nlines = 0;
		return;
	    }
	    for(i = 0; i < item->nlines; i++)
		sect->lines_list[i] = STRDUP(item->lines_list[i]);
	}
}


/*
 *	Opens a manual page.
 *
 *	The path specifies the manual page to open.
 *
 *	The node specifies the GtkCTreeNode to insert the opened
 *	manual page's toplevel GtkCTreeNode after.
 *
 *	If as_template is TRUE then the opened manual page will not
 *	have its full_path set and the Editor's last opened path will
 *	not be modified.
 *
 *	Returns 0 on success or non-zero on error.
 */
gint EditorFileOpen(
	editor_struct *editor,
	const gchar *path,
	GtkCTreeNode *insert_node,
	const gboolean as_template
)
{
	gint		status,
			nsections;
	GtkCTree *ctree;
	GtkCList *clist;
	GtkCTreeNode	*node,
			*new_toplevel_node = NULL;
	MPHeader *head;
	MPSection **sections_list, *sect;
	editor_item_struct *editor_item;
	me_core_struct *core;
	me_pixmaps_list_struct *pixmaps_list;

	if((editor == NULL) || (path == NULL))
	{
	    errno = EINVAL;
	    return(-2);
	}

	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Set the values from the widgets to the EditorItem so that
	 * we have the most up to date data
	 */
	EditorSetValues(
	    editor,
	    EditorCurrentSelectedNode(editor)
	);

	/* Get the toplevel node of the node to insert after */
	insert_node = EditorGetToplevelNode(insert_node);

	/* The insert_node is now set as the node to insert or NULL */

	editor->freeze_count++;
	EditorSetBusy(editor, TRUE);
	if(TRUE)
	{
	    gchar *msg = g_strdup_printf(
		"Opening manual page \"%s\"...",
		g_basename(path)
	    );
	    EditorSetStatusMessage(editor, msg);
	    g_free(msg);
	}
	EditorSetStatusProgress(editor, 0.0f);

	/* Open the Manual Page file */
	head = NULL;
	sections_list = NULL;
	nsections = 0;
	status = MPFileOpen(
	    path,
	    &head,
	    &sections_list, &nsections,
	    EditorFileOpenProgressCB, editor
	);
	if(status != 0)
	{
	    /* Open failed */
	    const gint error_code = (gint)errno;

	    MPHeaderDelete(head);
	    MPSectionsListClear(sections_list, nsections);

	    EditorSetStatusMessage(editor, "Open manual page failed");
	    EditorSetStatusProgress(editor, 0.0f);
	    EditorUpdateMenus(editor);

	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);   

	    errno = (int)error_code;
	    return(status);
	}
	else
	{
	    /* Opened successfully */
	    const gint ncolumns = MAX(clist->columns, 1);
	    gint i;
	    gchar	*name = STRDUP(g_basename(path)),
			*section_name,
			**strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	    for(i = 0; i < ncolumns; i++)
		strv[i] = "";

	    if(ncolumns > 0)
		strv[0] = name;

	    /* Insert the toplevel node to represent the opened
	     * manual page
	     */
	    new_toplevel_node = node = gtk_ctree_insert_node(
		ctree,
		NULL,				/* Toplevel */
		insert_node,
		strv,
		ME_LIST_ICON_TEXT_SPACING,
		pixmaps_list->manual_closed_20x20,
		pixmaps_list->manual_closed_20x20_mask,
		pixmaps_list->manual_opened_20x20,
		pixmaps_list->manual_opened_20x20_mask,
		FALSE,				/* Not leaf */
		TRUE				/* Expanded */
	    );

	    /* Create a new toplevel EditorItem */
	    editor_item = EditorItemNewWithValues(
		EDITOR_ITEM_FILE,
		editor,
		node
	    );
	    if(editor_item != NULL)
	    {
		/* Open as template? */
		if(as_template)
		{
		    /* Increment the core's untitled count */
		    core->untitled_count++;

		    /* Set the name as untitled */
		    g_free(editor_item->name);
		    editor_item->name = g_strdup_printf(
		        "Untitled %i",
		        core->untitled_count
		    );

		    /* Set the has changes flag on open */
		    editor_item->has_changes = TRUE;
		}
		else
		{
		    /* Open normally, set the name and path */
		    g_free(editor_item->name);
		    editor_item->name = STRDUP(name);

		    g_free(editor_item->path);
		    editor_item->path = STRDUP(path);
		}

		/* Set the EditorItem as the GtkCTreeNode's data */
		EditorBranchSetData(
		    ctree,
		    node,
		    editor_item, EditorItemDestroyCB
		);
	    }

	    /* Load the header */
	    if(head != NULL)
	    {
		/* Append a new GtkCTreeNode for this header */
		strv[0] = "Header";
		node = gtk_ctree_insert_node(
		    ctree,
		    new_toplevel_node,
		    NULL,
		    strv,
		    ME_LIST_ICON_TEXT_SPACING,
		    pixmaps_list->manpage_heading_20x20,
		    pixmaps_list->manpage_heading_20x20_mask,
		    pixmaps_list->manpage_heading_20x20,
		    pixmaps_list->manpage_heading_20x20_mask,
		    TRUE,			/* Is leaf */
		    FALSE			/* Not expanded */
		);

		/* Create a new EditorItem for this header */
		editor_item = EditorItemNewWithValues(
		    EDITOR_ITEM_HEADER,
		    editor,
		    node
		);
		if(editor_item != NULL)
		{
		    /* Set/transfer the header's comments lines to the
		     * new EditorItem
		     */
		    EditorItemSetLines(
			editor_item,
			head->comments_lines_list, head->ncomments_lines
		    );
		    head->comments_lines_list = NULL;
		    head->ncomments_lines = 0;
		}

		/* Set the EditorItem as the GtkCTreeNode's data */
		EditorBranchSetData(
		    ctree,
		    node,
		    editor_item, EditorItemDestroyCB
		);

		/* Get the header values */
		EditorFileOpenParseHeader(
		    editor,
		    editor_item,
		    head
		);
	    }

	    /* Load each section */
	    for(i = 0; i < nsections; i++)
	    {
		sect = sections_list[i];
		if(sect == NULL)
		    continue;

		if(sect->name != NULL)
		    section_name = g_strdup(sect->name);
		else
		    section_name = g_strdup_printf(
			"Section %i",
			i
		    );
		if(ncolumns > 0)
		    strv[0] = section_name;

		/* Append the GtkCTreeNode for this section */
		node = gtk_ctree_insert_node(
		    ctree,
		    new_toplevel_node,
		    NULL,
		    strv,
		    ME_LIST_ICON_TEXT_SPACING,
		    pixmaps_list->manpage_section_20x20,
		    pixmaps_list->manpage_section_20x20_mask,
		    pixmaps_list->manpage_section_20x20,
		    pixmaps_list->manpage_section_20x20_mask,
		    TRUE,			/* Is leaf */
		    FALSE			/* Not expanded */
		);

		/* Create a new EditorItem for this section */
		editor_item = EditorItemNewWithValues(
		    EDITOR_ITEM_SECTION,
		    editor,
		    node
		);
		if(editor_item != NULL)
		{
		    g_free(editor_item->section_name);
		    editor_item->section_name = STRDUP(section_name);

		    /* Set/transfer the section's lines to the new
		     * EditorItem
		     */
		    EditorItemSetLines(
			editor_item,
			sect->lines_list, sect->nlines
		    );
		    sect->lines_list = NULL;
		    sect->nlines = 0;
		}

		g_free(section_name);

		/* Set the EditorItem as the GtkCTreeNode's data */
		EditorBranchSetData(
		    ctree,
		    node,
		    editor_item, EditorItemDestroyCB
		);
	    }

	    g_free(strv);
	    g_free(name);
	}

	MPHeaderDelete(head);
	MPSectionsListClear(sections_list, nsections);

	/* If not opening as a template then update the Editor's last
	 * open path
	 */
	if(!as_template)
	{
	    g_free(editor->last_open_path);
	    editor->last_open_path = g_path_is_absolute(path) ?
		g_dirname(path) : STRDUP(g_get_current_dir());
	}

	/* Update the Layout GtkCTree's column width */
	gtk_clist_set_column_width(
	    clist,
	    0,					/* Column */
	    gtk_clist_optimal_column_width(clist, 0)
	);

	/* Select the opened manual page */
	if(new_toplevel_node != NULL)
	{
	    node = new_toplevel_node;
	    EditorSelect(
		editor,
		node
	    );
	    EditorGetValues(
		editor,
		node
	    );
	}

	if(TRUE)
	{
	    gchar *msg = g_strdup_printf(
		"Manual page \"%s\" opened",
		g_basename(path)
	    );
	    EditorSetStatusMessage(editor, msg);
	    g_free(msg);
	}
	EditorSetStatusProgress(editor, 0.0f);
	EditorUpdateMenus(editor);

	editor->freeze_count--;

	EditorSetBusy(editor, FALSE);

	return(0);
}


/*
 *	Save.
 *
 *	Does not modify the manual page's has changes flags.
 *
 *	The node specifies the manual page to save. The node may
 *	be either a child or toplevel node of the manual page.
 *
 *      Returns 0 on success or non-zero on error.
 */
gint EditorFileSave(
	editor_struct *editor,
	GtkCTreeNode *node
)
{
	gint		i,
			status,
			error_code,
			nsections;
	const gchar *path = NULL;
	GtkWidget *toplevel;
	GtkCTree *ctree;
	GtkCTreeNode *toplevel_node;
	MPHeader *head;
	MPSection	**sections_list,
			*sect;
	editor_item_struct *item;
	me_core_struct *core;

	if((editor == NULL) || (node == NULL))
	{
	    errno = EINVAL;
	    return(-2);
	}

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;

	/* Set the values from the widgets to the EditorItem so that
	 * we have the most up to date data
	 */
	EditorSetValues(
	    editor,
	    EditorCurrentSelectedNode(editor)
	);

	/* Get the toplevel node of the specified node */
	toplevel_node = EditorGetToplevelNode(node);
	if(toplevel_node != NULL)
	{
	    /* Get the toplevel EditorItem's path */
	    item = EditorBranchGetData(
		ctree,
		toplevel_node
	    );
	    if(item != NULL)
		path = item->path;
	}
	if(path == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	/* Check if the location exists and is not writable, a common
	 * case when users try to save to a globally installed manual
	 * page
	 */
	if(!access(path, F_OK))
	{
	    if(access(path, W_OK))
	    {
		errno = EPERM;
		return(-1);
	    }
	}


	/* Mark editor as processing */
	editor->freeze_count++;
	EditorSetBusy(editor, TRUE);

	if(TRUE)
	{
	    gchar *msg = g_strdup_printf(
		"Saving manual page \"%s\"...",
		g_basename(path)
	    );
	    EditorSetStatusMessage(editor, msg);
	    g_free(msg);
	}
	EditorSetStatusProgress(editor, 0.0f);


	/* Get the first child node of the toplevel node */
	node = NULL;
	if(toplevel_node != NULL)
	{
	    GtkCTreeRow *node_row = GTK_CTREE_ROW(toplevel_node);
	    if(node_row != NULL)
		node = node_row->children;
	}

	/* Iterate through each child node */
	head = NULL;
	sections_list = NULL;
	nsections = 0;
	while(node != NULL)
	{
	    GtkCTreeRow *node_row;

	    item = EditorBranchGetData(ctree, node);
	    if(item != NULL)
	    {
		/* Handle item data by type */
		switch(item->type)
		{
		  case EDITOR_ITEM_HEADER:
		    /* Destroy current header if exists */
		    MPHeaderDelete(head);

		    /* Allocate new header structure */		    
		    head = MPHeaderNew();
		    if(head != NULL)
		    {
			/* Copy values from the item data structure
			 * to the manual page header structure formatted
			 */
			EditorFileSaveFormatHeader(
			    editor,
			    item,
			    head
			);
		    }
		    break;

		  case EDITOR_ITEM_SECTION:
		    /* Increase total and allocate more pointers */
		    if(nsections < 0)
			nsections = 0;
		    i = nsections;
		    nsections = i + 1;
		    sections_list = (MPSection **)g_realloc(
			sections_list,
			nsections * sizeof(MPSection *)
		    );
		    if(sections_list == NULL)
		    {
			nsections = 0;   
			sect = NULL;
		    }
		    else
		    {
			sections_list[i] = sect = MPSectionNew();
		    }
		    if(sect != NULL)
		    {
			/* Copy values from the item data structure
			 * to the manual page section structure formatted.
			 */
			EditorFileSaveFormatSection(
			    editor,
			    item,
			    sect
			);
		    }
		    break;

		  case EDITOR_ITEM_FILE:
		    break;
		}
	    }

	    /* Get sibling of node */
	    node_row = GTK_CTREE_ROW(node);
	    if(node_row == NULL)
		node = NULL;
	    else
		node = node_row->sibling;
	}


	/* Save to file */
	status = MPFileSave(
	    path,
	    head,
	    sections_list, nsections,
	    EditorFileSaveProgressCB, editor
	);
	error_code = (gint)errno;

	MPHeaderDelete(head);
	MPSectionsListClear(sections_list, nsections);

	if(status != 0)
	{
	    EditorSetStatusMessage(editor, "Save manual page failed");
	}
	else
	{
	    gchar *msg = g_strdup_printf(
		"Manual page \"%s\" saved",
		g_basename(path)
	    );
	    EditorSetStatusMessage(editor, msg);
	    g_free(msg);
	}
	EditorSetStatusProgress(editor, 0.0f);
	EditorUpdateMenus(editor);

	editor->freeze_count--;

	EditorSetBusy(editor, FALSE);

	errno = (int)error_code;

	return(status);
}

/*
 *	Save as.
 *
 *	Saves the manual page and updates the manual page's file name
 *	and name to that of the specified path. The Editor's last
 *	save path will be updated.
 *
 *	Does not modify the manual page's has changes flags.
 *
 *	The path specifies the new file to save to.
 *
 *	The node specifies the manual page to save.
 *
 *	Returns 0 on success or non-zero on error.
 */
gint EditorFileSaveAs(
	editor_struct *editor,
	const gchar *path,
	GtkCTreeNode *node
)
{
	struct stat stat_buf;
	gint		status,
			error_code;
	gchar *s;
	GtkWidget *toplevel;
	GtkCTree *ctree;
	GtkCTreeNode *toplevel_node;
	editor_item_struct *item;
	me_core_struct *core;
	me_pixmaps_list_struct *pixmaps_list;

	if((editor == NULL) || (node == NULL) || (path == NULL))
	{
	    errno = EINVAL;
	    return(-2);
	}

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Get the toplevel node */
	toplevel_node = EditorGetToplevelNode(node);
	if(toplevel_node == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	/* Get the EditorItem of the toplevel node */
	item = EditorBranchGetData(
	    ctree,
	    toplevel_node
	);
	if(item == NULL)
	{
	    errno = EINVAL;
	    return(-2);
	}

	/* Check if the file exists */
	if(!stat((const char *)path, &stat_buf))
	{
	    gchar *msg = g_strdup_printf(
"Overwrite existing file:\n\
\n\
    %s",
		path
	    );
	    CDialogSetTransientFor(toplevel);
	    status = CDialogGetResponse(
"Confirm Overwrite",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		CDIALOG_BTNFLAG_NO
	    );
	    g_free(msg);
	    CDialogSetTransientFor(NULL);
	    switch(status)
	    {
	      case CDIALOG_RESPONSE_YES:
	      case CDIALOG_RESPONSE_YES_TO_ALL:
	      case CDIALOG_RESPONSE_OK:
		break;

	      case CDIALOG_RESPONSE_NO:
		errno = EINTR;
		return(-5);
		break;

	      default:
		errno = EINTR;
		return(-4);
		break;
	    }
	}


	/* Got all required data, now update file name on item data */

	/* Update the manual page's path and name */
	g_free(item->path);
	item->path = STRDUP(path);

	g_free(item->name);
	item->name = STRDUP(g_basename(path));

	/* Update the toplevel node's name */
	if(item->name != NULL)
	{
	    GtkCTreeRow *row_ptr = GTK_CTREE_ROW(toplevel_node);
	    if(row_ptr != NULL)
		gtk_ctree_set_node_info(
		    ctree,
		    toplevel_node,
		    item->name,
		    ME_LIST_ICON_TEXT_SPACING,
		    row_ptr->pixmap_closed,
		    row_ptr->mask_closed,
		    row_ptr->pixmap_opened,
		    row_ptr->mask_opened,
		    row_ptr->is_leaf,
		    row_ptr->expanded
		);
	}

	/* Save */
	status = EditorFileSave(
	    editor,
	    toplevel_node
	);
	error_code = (gint)errno;

	/* Update the last save as path */
	s = g_dirname(path);
	if(s != NULL)
	{
	    g_free(editor->last_save_as_path);
	    editor->last_save_as_path = s;
	}

	errno = (int)error_code;

	return(status);
}
