#include <stdlib.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <unistd.h>		/* For unlink() */

#include "cdialog.h"

#include "editor.h"
#include "editor_preview.h"
#include "editor_fio.h"

#include "viewer.h"
#include "viewer_fio.h"

#include "manedit.h"
#include "me_op.h"
#include "config.h"


void EditorPreview(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gint viewer_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Preview.
 *
 *	Converts and displays a manual page currently opened on the
 *	Editor to a Viewer.
 *
 *	The node specifies the manual page to preview.
 *
 *	The viewer_num specifies the Viewer to display the preview of
 *	the manual page.
 */
void EditorPreview(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gint viewer_num
)
{
	gint		i,
			status,
			fd,
			editor_num;
	gchar		*orig_path,
			*tmp_path,
			*prog_tmp_path;
	GtkWidget *toplevel;
	GtkCTree *ctree;
	GtkCTreeNode *toplevel_node;
	editor_item_struct *item;
	viewer_struct *v;
	me_core_struct *core;

	if(editor == NULL)
	    return;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;

	/* Get the Editor's index */
	for(i = 0; i < core->total_editors; i++)
	{
	    if(core->editor[i] == editor)
		break;
	}
	if(i < core->total_editors)
	    editor_num = i;
	else
	    editor_num = -1;

	/* Check if the specified Viewer is valid */
	v = ((viewer_num >= 0) && (viewer_num < core->total_viewers)) ?
	    core->viewer[viewer_num] : NULL;
	if(v == NULL)
	{
	    gchar *msg = g_strdup_printf(
"Viewer #%i is invalid.",
		viewer_num + 1
	    );
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Preview Failed",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(msg);
	    return;
	}

	/* Record the Editor's index on the Viewer */
	v->editor_num = editor_num;

	/* Get the toplevel node of the specified node */
	toplevel_node = EditorGetToplevelNode(node);
	if(toplevel_node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Preview Failed",
"The specified node has no toplevel node.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Get the EditorItem of the manual page to preview */
	item = EditorBranchGetData(
	    ctree,
	    toplevel_node
	);
	if(item == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Preview Failed",
"No EditorItem is available on the selected node.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Get/create the temporary directory */
	prog_tmp_path = MECreateTemporaryDirectory(core);
	if(prog_tmp_path == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Preview Failed",
"Unable to create the temporary directory.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Create a tempory file to convert the manual page to
	 * preview to
	 */
	tmp_path = g_strconcat(
	    prog_tmp_path,
	    G_DIR_SEPARATOR_S,
	    "preview",
	    "XXXXXX",
	    NULL
	);
	fd = (gint)mkstemp((char *)tmp_path);
	if(fd > -1)
	    (void)close((int)fd);

	/* To preview, first record the actual path of the manual
	 * page on the EditorItem and replace it with our temporary
	 * file path so that it gets saved to that, then after
	 * saving, we will set it back to the actual path
	 */
	orig_path = item->path;
	item->path = tmp_path;
	tmp_path = NULL;

	/* Save the manual page to the temporary file */
	status = EditorFileSave(
	    editor,
	    node
	);
	if(status != 0)
	{
	    const gint error_code = (gint)errno;
	    gchar *msg;
	    switch(status)
	    {
	      case -5:
	      case -4:
		break;

	      default:
		msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
		    g_strerror(error_code),
		    item->path
		);
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"Preview Failed",
		    msg,
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
		break;
	    }
	}
	else
	{
	    /* Clear the Viewer and open the manual page file */
	    ViewerViewTextRecordScrollPositions(v);
	    ViewerTextDelete(v, 0, -1);
	    ViewerTextInsertPosition(v, 0);
	    status = ViewerOpenFile(
		v,
		item->path,
		orig_path
	    );
	}

	/* Remove our temporary file that we saved the manual page
	 * to preview to
	 */
	if(item->path != NULL)
	{
	    (void)unlink(item->path);
	    g_free(item->path);
	}

	/* Set back the path on the EditorItem */
	item->path = orig_path;
	orig_path = NULL;

	/* Delete the temporary directory path */
	g_free(prog_tmp_path);
}
