#include <gtk/gtk.h>

#include "editor.h"
#include "editor_syntax_highlight.h"

#include "manedit.h"


void EditorSyntaxHighlight(
	editor_struct *editor,
	GtkWidget *w,
	gint cursor_pos,
	gint start_pos,
	gint end_pos
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Performs syntax highlighting on the GtkText widget w.
 *
 *	Any given position input values may be -1. If all are -1
 *	then entire buffer will have syntax highlighting performed on
 *	it.
 *
 *	Avoid making this call if the text widget is frozen.
 */
void EditorSyntaxHighlight(
	editor_struct *editor,
	GtkWidget *w,
	gint cursor_pos,
	gint start_pos,
	gint end_pos
)
{
	gboolean	in_tag,
			in_symbol,
			orig_has_selection;
	gint		c,
			orig_cur_pos, cur_pos,
			text_start, text_end,
			orig_sel_start = -1, orig_sel_end = -1;
	GtkEditable *editable = GTK_EDITABLE(w);
	GtkText *text = GTK_TEXT(w);
	me_core_struct *core;

	if((editor == NULL) || (w == NULL))
	    return;

	/* No syntax highlighting? */
	if(!editor->syntax_highlighting)
	    return;

	editor->freeze_count++;

	core = editor->core;

	/* Get the text bounds */
	text_start = 0;
	text_end = (gint)gtk_text_get_length(text);

	/* Clip the start and end positions */
	if(cursor_pos >= text_end)
	    cursor_pos = text_end - 1;
	if(start_pos >= text_end)
	    start_pos = text_end - 1;
	if(end_pos > text_end)
	    end_pos = text_end;

	/* Use the start position as cursor position if the cursor
	 * position is invalid and start position is valid
	 */
	if((cursor_pos < 0) && (start_pos > -1))
	    cursor_pos = start_pos;

	/* Record the original cursor position */
	orig_cur_pos = gtk_editable_get_position(editable);

	/* Record the original selection */
	orig_has_selection = editable->has_selection;
	if(orig_has_selection)
	{
	    orig_sel_start = editable->selection_start_pos;
	    orig_sel_end = editable->selection_end_pos;
	}

	/* Cursor position defined? */
	if(cursor_pos > -1)
	{
	    /* Set the starting cursor position */
	    cur_pos = cursor_pos;

	    /* Iterate backwards until a deliminator is encountered,
	     * if a deliminator is encountered then it will be colored
	     * and then the style will be updated for regular or tag
	     * text and in_tag will be set accordingly
	     *
	     * If no tags were encountered then the regular style will
	     * be set
	     */
	    in_tag = FALSE;
	    in_symbol = FALSE;
gtk_text_freeze(text);
	    while(cur_pos >= text_start)
	    {
		c = GTK_TEXT_INDEX(text, cur_pos);
		/* Is character a tag deliminator start? */
		if(c == '<')
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->tag_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			"<",
			1
		    );
		    in_tag = TRUE;
		    cur_pos++;	/* Seek past tag */
		    break;
		}
		/* Is character an tag deliminator end? */
		else if(c == '>')
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->tag_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			">",
			1
		    );
		    cur_pos++;	/* Seek past tag */
		    break;
		}
		/* Is character a start of a symbol representation? */
		else if(c == '&')
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->symbol_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			"&",
			1
		    );
		    in_symbol = TRUE;
		    cur_pos++;  /* Seek past symbol rep */
		    break;
		}
		/* Skip symbol representation end ';' when itterating
		 * backwards.
		 */
		else
		{
		    cur_pos--;
		}
	    }
	    /* Make sure current position stays at or above text_start */
	    if(cur_pos < text_start)
		cur_pos = text_start;

	    if(TRUE)
	    {
		/* Itterate to end of text or to a start or end
		 * deliminator after end_pos.
		 */
		while(cur_pos < text_end)
		{
		    c = GTK_TEXT_INDEX(text, cur_pos);
		    /* Is character a tag deliminator start? */
		    if((c == '<') && !in_tag)
		    {
			GdkFont *font = core->editable_font;
			GdkColor *fg_color = &editor->deliminator_text_color;

			gtk_text_set_point(text, cur_pos);
			gtk_text_forward_delete(text, 1);
			gtk_text_insert(
			    text,
			    font,
			    fg_color,
			    NULL,
			    "<",
			    1
			);         

			in_tag = TRUE;	/* Mark as in a tag */

			/* Encountered a deliminator after end_pos? */
			if(end_pos > -1)
			{
			    if(cur_pos > end_pos)
				break;
			}
		    }
		    /* Is character a tag deliminator end? */
		    else if((c == '>') && in_tag)
		    {
			GdkFont *font = core->editable_font;
			GdkColor *fg_color = &editor->deliminator_text_color;

			gtk_text_set_point(text, cur_pos);
			gtk_text_forward_delete(text, 1);
			gtk_text_insert(
			    text,
			    font,
			    fg_color,
			    NULL,
			    ">",
			    1  
			);

			in_tag = FALSE;		/* Mark no longer in a tag */

			/* Encountered a deliminator after end_pos? */
			if(end_pos > -1)
			{
			    if(cur_pos > end_pos)
				break;
			}
		    }
		    /* Is character a symbol representation start? */
		    else if((c == '&') && !in_symbol)
		    {
			GdkFont *font = core->editable_font;
			GdkColor *fg_color = &editor->symbol_text_color;

			gtk_text_set_point(text, cur_pos);
			gtk_text_forward_delete(text, 1);
			gtk_text_insert(
			    text,
			    font,
			    fg_color,
			    NULL,
			    "&",
			    1
			);

			in_symbol = TRUE;		/* Mark as in a sym */

			/* Encountered a deliminator after end_pos? */
			if(end_pos > -1)
			{
			    if(cur_pos > end_pos)
				break;
			}
		    }
		    /* Is character a symbol representation end? */
		    else if((c == ';') && in_symbol)
		    {
			GdkFont *font = core->editable_font;
			GdkColor *fg_color = &editor->symbol_text_color;

			gtk_text_set_point(text, cur_pos);
			gtk_text_forward_delete(text, 1);
			gtk_text_insert(
			    text,
			    font,
			    fg_color,
			    NULL,
			    ";",
			    1
			);

			in_symbol = FALSE;		/* Mark as no longer in symbol */

			/* Encountered a deliminator after end_pos? */
			if(end_pos > -1)
			{
			    if(cur_pos > end_pos)
				break;
			}
		    }
		    else
		    {
			gchar tmp_str[2];
			GdkFont *font = core->editable_font;
			GdkColor *fg_color;

			if(in_tag)
			    fg_color = &editor->tag_text_color;
			else if(in_symbol)
			    fg_color = &editor->symbol_text_color;
			else
			    fg_color = &editor->normal_text_color;

			tmp_str[0] = (char)c;
			tmp_str[1] = '\0';

			gtk_text_set_point(text, cur_pos);
			gtk_text_forward_delete(text, 1);
			gtk_text_insert(
			    text,
			    font,
			    fg_color,
			    NULL,
			    tmp_str,
			    1
			);
		    }

		    /* Increment to next position */
		    cur_pos++;
		}
	    }
gtk_text_thaw(text);
	}
	else
	{
	    /* Highlight entire buffer */

	    /* Set the cursor position to the start of the text */
	    cur_pos = text_start;

	    /* Reset the contexts */
	    in_tag = FALSE;
	    in_symbol = FALSE;

	    /* Iterate to the end of the text */
gtk_text_freeze(text);
	    while(cur_pos < text_end)
	    {
		c = GTK_TEXT_INDEX(text, cur_pos);
		/* Is character a tag deliminator start? */
		if((c == '<') && !in_tag)
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->deliminator_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			"<",
			1
		    );
		    in_tag = TRUE;		/* Mark as in a tag */
		}
		/* Is character a tag deliminator end? */
		else if((c == '>') && in_tag)
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->deliminator_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			">",
			1
		    );
		    in_tag = FALSE;		/* Mark no longer in a tag */
		}
		/* Is character a symbol representation start? */
		else if((c == '&') && !in_symbol)
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->symbol_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1); 
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			"&", 
			1   
		    );
		    in_symbol = TRUE;		/* Mark as in a symbol */
		}
		/* Is character a symbol representation end? */
		else if((c == ';') && in_symbol)
		{
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color = &editor->symbol_text_color;
		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(
			text,
			font,
			fg_color,
			NULL,
			";",
			1
		    );
		    in_symbol = FALSE;	/* Mark no longer in a sym */
		}
		else
		{
		    gchar tmp_str[2];
		    GdkFont *font = core->editable_font;
		    GdkColor *fg_color;

		    if(in_tag)
			fg_color = &editor->tag_text_color;
		    else if(in_symbol)
			fg_color = &editor->symbol_text_color;
		    else
			fg_color = &editor->normal_text_color;

		    tmp_str[0] = (char)c;
		    tmp_str[1] = '\0';

		    gtk_text_set_point(text, cur_pos);
		    gtk_text_forward_delete(text, 1);
		    gtk_text_insert(   
			text,
			font,
			fg_color,
			NULL,
			tmp_str,
			1
		    );
		}
		cur_pos++;
	    }
gtk_text_thaw(text);
	}

	/* Set back original cursor position */
	gtk_text_set_point(text, orig_cur_pos);
	gtk_editable_set_position(editable, orig_cur_pos);

	/* Set back originally had selection? */
	if(orig_has_selection)
	{
	    gtk_editable_select_region(
		editable,
		orig_sel_start, orig_sel_end
	    );
	}

	editor->freeze_count--;
}
