#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <locale.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <unistd.h>

#include "../include/disk.h"
#include "../include/string.h"

#include "guiutils.h"
#include "fprompt.h"
#include "cdialog.h"
#include "csd.h"
#include "fsd.h"
#include "fb.h"
#include "pdialog.h"
#include "progressdialog.h"

#include "me_types.h"
#include "manedit.h"
#include "me_op.h"
#include "config.h"


static void MEUnixSignalCB(int s);


static int		segfault_count;
static me_core_struct	*me_core;
gboolean		use_text_delete;


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Signal handler.
 */
void MEUnixSignalCB(int s)
{
	static gboolean doing_crash_save = FALSE;

	switch(s)
	{
	  case SIGINT: 
	  case SIGTERM:
	    if(me_core != NULL)
		me_core->need_close_all_windows = TRUE;
	    break;

	  case SIGSEGV:
	    signal(SIGSEGV, MEUnixSignalCB);	/* Watch for SIGSEGV again */
	    segfault_count++;			/* Increment SIGSEGV count */
	    /* Print warning of segmentation fault */
	    g_printerr(
		"%s triggered a segmentation fault (%i times)\n",
		PROG_NAME,
		segfault_count
	    );
	    /* Already doing crash save? */
	    if(doing_crash_save)
	    {

	    }
	    else
	    {
		/* Mark and perform emergency save procedure */
		doing_crash_save = TRUE;
		MEEmergencySave(me_core);
	    }
	    /* Got 3 or more segfaults? */
	    if(segfault_count >= 3)
	    {
		g_printerr(
		    "%s attempting immediate process exit().\n",
		    PROG_NAME
		);
		exit(1);
	    }
	    else
	    {
		if(me_core != NULL)
		    me_core->need_close_all_windows = TRUE;
	    }
	    break;

	  case SIGPIPE:
	    signal(SIGPIPE, MEUnixSignalCB);
	    break;
	}
}


int main(int argc, char **argv)
{
	gint i, status;
	const gchar *arg;
	me_core_struct *core;

	segfault_count = 0;
	me_core = core = NULL;

use_text_delete = TRUE;

	/* Connect the UNIX signal callbacks */
	signal(SIGINT, MEUnixSignalCB);
	signal(SIGTERM, MEUnixSignalCB);
	signal(SIGSEGV, MEUnixSignalCB);
	signal(SIGPIPE, MEUnixSignalCB);

	/* Parse the help and version arguments, other arguments will
	 * be handled later in MENew()
	 */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!g_strcasecmp(arg, "--help") ||
	       !g_strcasecmp(arg, "-help") ||
	       !g_strcasecmp(arg, "--h") ||
	       !g_strcasecmp(arg, "-h")
	    )
	    {
		g_print("%s\n", PROG_USAGE_MESG);
		return(0);
	    }
	    /* Version */
	    else if(!g_strcasecmp(arg, "--version") ||
		    !g_strcasecmp(arg, "-version")
	    )
	    {
		g_print(
		    "%s Version %s\n%s\n",
		    PROG_NAME_FULL, PROG_VERSION,
		    PROG_COPYRIGHT
		);
		return(0);
	    }
	}

	/* Set up the localization code */
	setlocale(LC_ALL, "");

	gtk_set_locale();

	/* Initialize GTK */
	gtk_init(&argc, &argv);

	/* Initialize the dialogs */
	FPromptInit();
	CDialogInit();
	FileBrowserInit();
	CSDInit();
	FSDInit();
	PDialogInit();
	ProgressDialogInit();

	/* Create the ManEdit core */
	core = MENew(
	    argc, argv,
	    &status
	);
	if(status != 0)
	{
	    MEDelete(core, FALSE);
	    return(status);
	}

	/* Enter the main GTK loop */
	gtk_main();

	/* Delete the ManEdit core */
	me_core = NULL;
	MEDelete(
	    core,
	    TRUE				/* Save the CfgList */
	);
	core = NULL;

	/* Shutdown the dialogs */
	ProgressDialogShutdown();
	PDialogShutdown();
	FSDShutdown();
	CSDShutdown();
	FileBrowserShutdown();
	CDialogShutdown();
	FPromptShutdown();

	/* clean up */
	{
		gchar *tmpdir = MECreateTemporaryDirectory(NULL);
		if (tmpdir) rmdir (tmpdir);
		free(tmpdir);
	}

	return(0);
}
