#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#ifdef HAVE_GZIP
#include <zlib.h>
#endif
#ifdef HAVE_BZIP2
#include <bzlib.h>
#endif

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/fio.h"
#include "../include/disk.h"

#include "manpage_fio.h"


/* MPHeader */
MPHeader *MPHeaderNew(void);
void MPHeaderDelete(MPHeader *head);

/* MPSection */
MPSection *MPSectionNew(void);
void MPSectionDelete(MPSection *sect);
void MPSectionsListClear(
        MPSection **sections_list, const int nsections
);

/* Low level file IO (for loading and saving as regular or compressed) */
static char *MPFileOpenBZip2(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
static char *MPFileOpenGZip(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
static char *MPFileOpenNormal(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
static int MPFileSaveBZip2(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
static int MPFileSaveGZip(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
static int MPFileSaveNormal(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);

/* Manual Page Data Parsers */
static const char *MPLoadStringNextCR(const char *buf);
static const char *MPLoadStringNextDoubleQuote(const char *buf);

/* Line Exploder */
static void MPLoadExplodeNewLines(
	char ***line, int *total_lines,
	const char *buf, const char *buf_end
);

/* Manual Page Data To XML converter */
static void MPConvertManualPageDataToXMLLines(
	const char *buf, const char *buf_end,
	char ***line, int *total_lines
);
/* XML To Manual Page Data converter */
static char *MPConvertXMLLineToManualPageData(char *line_ptr);

/* String Substitutions */
static char *MPSubStringIterate(
	char *buf, char **sub_start,
	const char *token, const char *replace,
	int strip_tail
);
static char *MPSubString(
	char *buf, const char *token, const char *replace,
	int strip_tail
);

/* Header Parser */
static void MPHeaderParseHeaderItem(
	MPHeader *mp_header,
	const char *line
);
static void MPHeaderParse(
	MPHeader *mp_header,
	char **lines_list, const int nlines
);

/* Open */
int MPOpenNexus(
	const char *path,
	const char *data,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
int MPFileOpen(
	const char *path,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
int MPDataOpen(
	const char *data,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);

/* Save */
int MPSaveNexus(
	const char *path,
	char **data_rtn,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
int MPFileSave(
	const char *path,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);
int MPDataSave(
	char **data_rtn,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define ISCR(c)		(((c) == '\n') || ((c) == '\r'))


/*
 *	Creates a new MPHeader.
 */
MPHeader *MPHeaderNew(void)
{
	return(MP_HEADER(calloc(1, sizeof(MPHeader))));
}

/*
 *	Deletes the MPHeader.
 */
void MPHeaderDelete(MPHeader *head)
{
	if(head == NULL)
	    return;

	free(head->name);
	free(head->section_number);
	free(head->version);
	free(head->author);
	free(head->catagory);
	strlistfree(head->comments_lines_list, head->ncomments_lines);

	free(head);
}


/*
 *	Creates a new MPSection.
 */
MPSection *MPSectionNew(void)
{
	return(MP_SECTION(calloc(1, sizeof(MPSection))));
}

/*
 *	Deletes the MPSection.
 */
void MPSectionDelete(MPSection *sect)
{
	if(sect == NULL)
	    return;

	free(sect->name);
	strlistfree(
	    sect->lines_list,
	    sect->nlines
	);

	free(sect);
}

/*
 *	Deletes the MPSections list.
 */
void MPSectionsListClear(
        MPSection **sections_list, const int nsections
)
{
	int i;

	for(i = 0; i < nsections; i++)
	    MPSectionDelete(sections_list[i]);

	free(sections_list);
}


/*
 *	Called by MPFileOpenNormal() to open the BZip2 file.
 */
static char *MPFileOpenBZip2(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
#ifdef HAVE_BZIP2
	int bzerror;
	long units_read;
	unsigned long	buf_cur_pos,
			buf_len,
			file_size,
			block_size;
	char *buf;
	FILE *fp;
	BZFILE *zfp;
	struct stat stat_buf;

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return(NULL);

	fstat(fileno(fp), &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)MAX(stat_buf.st_blksize, 1);

	/* Initialize BZip2 to read from the file */
	zfp = BZ2_bzReadOpen(
	    &bzerror,
	    fp,
	    0,					/* Do not conserve memory,
						 * do this fast */
	    0,					/* No need to be verbose */
	    NULL, 0				/* Nonsense buffer, always NULL and 0 */
	);
	if(bzerror != BZ_OK)
	{
	    BZ2_bzReadClose(&bzerror, zfp);
	    fclose(fp);
	    return(NULL);
	}

	/* Read the file */
	buf = NULL;
	buf_len = 0l;
	while(1)
	{
	    /* Report the progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (unsigned long)ftell(fp),
		    (unsigned long)(2 * file_size),
		    progress_data
		))
		{

		    BZ2_bzReadClose(&bzerror, zfp);
		    fclose(fp);
		    free(buf);
		    return(NULL);
		}
	    }

	    /* Record the current buffer position and increase the buffer
	     * allocation
	     */
	    buf_cur_pos = buf_len;
	    buf_len += block_size;
	    buf = (char *)realloc(
		buf,
		(buf_len + 1) * sizeof(char)
	    );
	    if(buf == NULL)
	    {
		buf_len = 0;
		break;
	    }

	    /* Read and decompress this block */
	    units_read = BZ2_bzRead(
		&bzerror,
		zfp,
		&buf[buf_cur_pos],
		block_size
	    );
	    if((units_read <= 0l) ||
	       ((bzerror != BZ_OK) && (bzerror != BZ_STREAM_END))
	    )
	    {
		buf[buf_cur_pos] = '\0';
		break;
	    }

	    if((buf_cur_pos + (unsigned long)units_read) <= buf_len)
		buf[buf_cur_pos + (unsigned long)units_read] = '\0';
	    else
		buf[buf_len - 1] = '\0';
	}

	/* Close the file */
	BZ2_bzReadClose(&bzerror, zfp);
	fclose(fp);

	*rtn_buf_len = (int)buf_len;

	return(buf);
#else
	return(NULL);
#endif
}

/*
 *	Called by MPFileOpenNormal() to open the GZip file.
 */
static char *MPFileOpenGZip(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
#ifdef HAVE_GZIP
	long units_read;
	unsigned long	buf_cur_pos,
			buf_len,
			file_size,
			block_size;
	char *buf;
	gzFile zfp;
	struct stat stat_buf;

	/* Open the file for reading */
	zfp = gzopen(path, "rb");
	if(zfp == NULL)
	    return(NULL);

	stat(path, &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)MAX(stat_buf.st_blksize, 1);

	/* Read the file */
	buf = NULL;
	buf_len = 0l;
	while(1)
	{
	    /* Report the progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (unsigned long)gztell(zfp),
		    (unsigned long)(2 * file_size),
		    progress_data
		))
		{
		    gzclose(zfp);
		    free(buf);
		    return(NULL);
		}
	    }

	    /* Record current buffer index and increase buffer
	     * allocation.
	     */
	    buf_cur_pos = buf_len;
	    buf_len += block_size;
	    buf = (char *)realloc(
		buf,
		(buf_len + 1) * sizeof(char)
	    );
	    if(buf == NULL)
	    {
		buf_len = 0;
		break;
	    }

	    /* Read and decompress this block */
	    units_read = gzread(
		zfp,
		&buf[buf_cur_pos],
		block_size
	    );
	    if(units_read <= 0l)
	    {
		buf[buf_cur_pos] = '\0';
		break;
	    }

	    if((buf_cur_pos + (unsigned long)units_read) <= buf_len)
		buf[buf_cur_pos + (unsigned long)units_read] = '\0';
	    else
		buf[buf_len - 1] = '\0';
	}

	/* Close the file */
	gzclose(zfp);

	*rtn_buf_len = (int)buf_len;

	return(buf);
#else
	return(NULL);
#endif
}

/*
 *	Called by MPFileOpenNexus() to open the file.
 */
static char *MPFileOpenNormal(
	const char *path,
	int *rtn_buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	unsigned long	file_size,
			block_size;
	const char	*s,
			*ext;
	char		*buf,
			*buf_ptr,
			*io_buf;
	FILE *fp;
	struct stat stat_buf;

	*rtn_buf_len = 0;

	/* Get the extension */
	s = strrchr(path, DIR_DELIMINATOR);
	if(s != NULL)
	    s++;
	else
	    s = path;
	ext = strrchr(s, '.');
	if(ext != NULL)
	{
	    /* Check if the file is compressed and open it accordingly */
	    if(!strcasecmp(ext, ".bz2"))
		return(MPFileOpenBZip2(
		    path,
		    rtn_buf_len,
		    progress_cb,
		    progress_data
		));
	    else if(!strcasecmp(ext, ".gz"))
		return(MPFileOpenGZip(
		    path,
		    rtn_buf_len,
		    progress_cb,
		    progress_data
		));
	}

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return(NULL);

	fstat(fileno(fp), &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)MAX(stat_buf.st_blksize, 1);

	/* Allocate the return buffer */
	if(file_size > 0l)
	    buf = (char *)malloc((file_size + 1) * sizeof(char));
	else
	    buf = NULL;
	if(buf == NULL)
	{
	    fclose(fp);
	    return(NULL);
	}

        /* Allocate the I/O buffer */
        io_buf = (char *)malloc(block_size * sizeof(char));

	/* Read the file */
	buf_ptr = buf;
	if(io_buf != NULL)
        {
            size_t units_read;

            while(!feof(fp))
            {
		/* Report the progress */
		if(progress_cb != NULL)
		{
		    if(progress_cb(
			(unsigned long)ftell(fp),
			(unsigned long)(2 * file_size),
			progress_data
		    ))
		    {
			fclose(fp);
			free(io_buf);
			free(buf);
			return(NULL);
		    }
	        }

                /* Read this block */
                units_read = fread(
                    io_buf,
                    sizeof(char),
                    (size_t)block_size,
                    fp
                );
                /* Read error? */
                if(ferror(fp))
		    break;
                else if(units_read <= 0l)
                    break;

		/* Copy this block to the return buffer */
		memcpy(
		    buf_ptr,
		    io_buf,
		    units_read * sizeof(char)
		);
		buf_ptr += units_read;
	    }

	    free(io_buf);
	}

	/* Close the file */
	fclose(fp);

	/* Null terminate the return buffer string */
	buf[file_size] = '\0';

	*rtn_buf_len = (int)file_size;

	return(buf);
}


/*
 *	Called by MPFileSaveNormal() to save to a BZip2 compressed file.
 */
static int MPFileSaveBZip2(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
#ifdef HAVE_BZIP2
	int		bzerror;
	unsigned long	buf_cur_pos,
			buf_cur_len,
			file_size,
			block_size;
	FILE *fp;
	BZFILE *zfp;
	struct stat stat_buf;

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	    return(-1);

	fstat(fileno(fp), &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)MAX(stat_buf.st_blksize, 1);

	/* Initialize BZip2 to write to the file */
	zfp = BZ2_bzWriteOpen(
	    &bzerror,
	    fp,
	    9,          /* Compression strength (0 to 9) */
	    0,          /* No need to be verbose */
	    30          /* Work factor (default is 30) */
	);
	if(bzerror != BZ_OK)
	{
	    BZ2_bzWriteClose(&bzerror, zfp, 0, NULL, NULL);
	    fclose(fp);
	    return(-1);
	}

	/* Write the file */
	buf_cur_pos = 0l;
	while(buf_cur_pos < buf_len)
	{
	    /* Calculate the current segment length, making it no
	     * longer than (buf_len - buf_cur_pos) or block_size,
	     * whichever is shorter
	     */
	    buf_cur_len = buf_len - buf_cur_pos;
	    if(buf_cur_len > block_size)
		buf_cur_len = block_size;
	    if(buf_cur_len <= 0l)
		break;

	    /* Report the progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (1 * buf_len) + buf_cur_pos,
		    2 * buf_len,
		    progress_data
		))
		{
		    BZ2_bzWriteClose(&bzerror, zfp, 0, NULL, NULL);
		    fclose(fp);
		    return(-4);
		}
	    }

	    /* Write the current calculated buffer segment */
	    BZ2_bzWrite(
		&bzerror,
		zfp,
		(void *)&buf[buf_cur_pos],
		buf_cur_len
	    );
	    if(bzerror != Z_OK)
		break;

	    buf_cur_pos += buf_cur_len;
	}

	/* Close the file */
	BZ2_bzWriteClose(&bzerror, zfp, 0, NULL, NULL);
	fclose(fp);

	return(0);
#else
	return(-2);
#endif
}

/*
 *	Called by MPFileSaveNormal() to save to a GZip file.
 */
static int MPFileSaveGZip(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
#ifdef HAVE_GZIP
	int status;
	gzFile zfp;
	unsigned long	file_size,
			block_size,
			buf_cur_pos,
			buf_cur_len;
	struct stat stat_buf;

	/* Open the file for writing */
	zfp = gzopen(path, "wb");
	if(zfp == NULL)
	    return(-1);

	stat(path, &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)MAX(stat_buf.st_blksize, 1);

	/* Write to the file */
	buf_cur_pos = 0l;
	while(buf_cur_pos < buf_len)
	{
	    /* Calculate current segment length, making it no longer
	     * than (buf_len - buf_cur_pos) or block_size
	     * whichever is shorter.
	     */
	    buf_cur_len = buf_len - buf_cur_pos;
	    if(buf_cur_len > block_size)
		buf_cur_len = block_size;
	    if(buf_cur_len <= 0)
		break;

	    /* Report the progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (1 * buf_len) + buf_cur_pos,
		    2 * buf_len,
		    progress_data
		))
		{
		    gzclose(zfp);
		    return(-4);
		}
	    }

	    /* Write this block */
	    status = gzwrite(
		zfp,
		(const voidp)&buf[buf_cur_pos],
		buf_cur_len
	    );
	    if(status <= 0)
		break;

	    buf_cur_pos += buf_cur_len;
	}

	/* Close the file */
	gzclose(zfp);

	return(0);
#else
	return(-2);
#endif
}

/*
 *	Called by MPFileSaveNexus() to save the file.
 */
static int MPFileSaveNormal(
	const char *path,
	const char *buf, const int buf_len,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	const char	*s,
			*ext;
	unsigned long	file_size,
			block_size,
			buf_cur_pos,
			buf_cur_len;
	size_t units_written;
	FILE *fp;
	struct stat stat_buf;

	/* Get the extension */
	s = strrchr(path, DIR_DELIMINATOR);
	if(s != NULL)
	    s++;
	else
	    s = path;
	ext = strrchr(s, '.');
	if(ext != NULL)
	{
	    /* Check if the file should be saved as compressed */
	    if(!strcasecmp(ext, ".bz2"))
		return(MPFileSaveBZip2(
		    path,
		    buf, buf_len,
		    progress_cb, progress_data
		));
	    else if(!strcasecmp(ext, ".gz"))
		return(MPFileSaveGZip(
		    path,
		    buf, buf_len,
		    progress_cb, progress_data
		));
	}

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	    return(-1);

	fstat(fileno(fp), &stat_buf);
	file_size = (unsigned long)stat_buf.st_size;
	block_size = (unsigned long)stat_buf.st_blksize;

	/* Write to the file */
	buf_cur_pos = 0l;
	while(buf_cur_pos < buf_len)
	{
	    /* Calculate current segment length, making it no longer
	     * than (buf_len - buf_cur_pos) or block_size
	     * whichever is shorter.
	     */
	    buf_cur_len = buf_len - buf_cur_pos;
	    if(buf_cur_len > block_size)
		buf_cur_len = block_size;
	    if(buf_cur_len <= 0)
		break;

	    /* Report the progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (1 * buf_len) + buf_cur_pos,
		    2 * buf_len,
		    progress_data
		))
		{
		    fclose(fp);
		    return(-4);
		}
	    }

	    /* Write this block */
	    units_written = fwrite(
		buf + buf_cur_pos,
		sizeof(char),
		buf_cur_len,
		fp
	    );
	    if(units_written <= 0l)
		break;

	    buf_cur_pos += (unsigned long)units_written;
	}

	/* Close the file */
	fclose(fp);

	return(0);
}


/*
 *	Returns the pointer in buf to the next occurance of a
 *	non-escaped newline character or NULL if there are no more.
 */
static const char *MPLoadStringNextCR(const char *buf)
{
	while((*buf) != '\0')
	{
	    if((*buf) == '\\')
	    {
		buf++;		/* Increment past current '\\' char */

		/* Next char is a null? */
		if((*buf) == '\0')
		    return(NULL);

		/* Skip whatever next char was */
		buf++;
		continue;
	    }

	    if(ISCR(*buf))
	       return(buf);

	    buf++;
	}

	return(NULL);
}

/*
 *      Returns pointer in buf to next occurance of non-escaped
 *      double quote character or NULL if there isn't any.
 */
static const char *MPLoadStringNextDoubleQuote(const char *buf)
{
	while((*buf) != '\0')
	{
	    if((*buf) == '\\')
	    {
		buf++;          /* Increment past current '\\' char */

		/* Next char is a null? */
		if((*buf) == '\0')
		    return(NULL);

		/* Skip whatever next char was */
		buf++;
		continue;
	    }

	    if((*buf) == '"')
	       return(buf);

	    buf++;
	}

	return(NULL);
}


/*
 *	Explodes the buf into lines from buf to buf_end (but not
 *	including buf_end) at all non-escaped deliminated newline
 *	characters.
 *
 *	Inputs are assumed valid.
 */
static void MPLoadExplodeNewLines(
	char ***line, int *total_lines,
	const char *buf, const char *buf_end
)
{
	const char *buf_limit_ptr;
	const char *buf_cur_ptr;
	int n, cur_seg_len;
	char *line_ptr;

	if((*total_lines) < 0)
	    (*total_lines) = 0;

	buf_cur_ptr = buf;

	do
	{
	    buf_limit_ptr = MPLoadStringNextCR(buf_cur_ptr);
	    if((buf_limit_ptr == NULL) || (buf_limit_ptr > buf_end))
	        buf_limit_ptr = buf_end;

	    cur_seg_len = MAX(
		(int)buf_limit_ptr - (int)buf_cur_ptr, 0
	    );

	    /* Allocate a new line */
	    n = (*total_lines);
	    (*total_lines) = n + 1;
	    (*line) = (char **)realloc(*line, (*total_lines) * sizeof(char *));
	    if((*line) == NULL)
	    {
		(*total_lines) = 0;
		break;
	    }
	    else
	    {
		line_ptr = (char *)malloc((cur_seg_len + 1) * sizeof(char));
		(*line)[n] = line_ptr;
		if(line_ptr != NULL)
		{
		    if(cur_seg_len > 0)
			strncpy(line_ptr, buf_cur_ptr, cur_seg_len);
		    line_ptr[cur_seg_len] = '\0';
		}
	    }

	    /* Seek past current buffer limit pointer */
	    buf_cur_ptr = buf_limit_ptr + 1;

	} while(buf_cur_ptr < buf_end);

}


/*
 *	Converts the data in manual page format into XML format.
 *
 *	The buf and buf_end specifies the manual page data.
 *
 *	The line and total_lines specifies the return values for the
 *	converted XML lines.
 */
static void MPConvertManualPageDataToXMLLines(
	const char *buf, const char *buf_end,
	char ***line, int *total_lines
)
{
	char *tmp_buf;
	int cur_seg_len;

	/* Copy the data to a temporary buffer */
	cur_seg_len = MAX((int)buf_end - (int)buf, 0);
	tmp_buf = (char *)malloc((cur_seg_len + 1) * sizeof(char));
	if(tmp_buf == NULL)
	    return;

	if(cur_seg_len > 0)
	    memcpy(
		tmp_buf,
		buf,
		cur_seg_len * sizeof(char)
	    );
	tmp_buf[cur_seg_len] = '\0';

	/* Make substitutions to tmp_buf, note that the order is
	 * important.
	 *
	 * When adding new substitutions, make sure the inverse operation
	 * is included in function MPConvertXMLLineToManualPageData() so that saving does
	 * the inverse substitution. Remember to make each inverse
	 * substitution in reverse order from substitutions here.
	 *
	 * Don't forget to add an inverse operation below in
	 * MPConvertXMLLineToManualPageData()!
	 */
	tmp_buf = MPSubString(tmp_buf, "\\-",	"-",		0);
	tmp_buf = MPSubString(tmp_buf, "&",     "&amp;",	0);
	tmp_buf = MPSubString(tmp_buf, "<",	"&lt;", 	0);
	tmp_buf = MPSubString(tmp_buf, ">",	"&gt;", 	0);
	/* Now we can replace to replace strings with < and > */
	tmp_buf = MPSubString(tmp_buf, ".BI",	"<BI>",		1);
	tmp_buf = MPSubString(tmp_buf, ".BR",   "<BR>", 	1);
	tmp_buf = MPSubString(tmp_buf, ".br",	"<br>", 	1);
	tmp_buf = MPSubString(tmp_buf, ".fi",	"<fi>", 	1);
	tmp_buf = MPSubString(tmp_buf, ".If",	"<If>",		1);
	tmp_buf = MPSubString(tmp_buf, ".Im",	"<Im>",		1);
	tmp_buf = MPSubString(tmp_buf, ".IP",   "<IP>", 	1);
	tmp_buf = MPSubString(tmp_buf, ".LP",   "<LP>", 	1);
	tmp_buf = MPSubString(tmp_buf, ".RP",   "<RP>",		1);
	tmp_buf = MPSubString(tmp_buf, ".TP",   "<TP>",         1);
	tmp_buf = MPSubString(tmp_buf, ".nf",   "<nf>",		1);
	tmp_buf = MPSubString(tmp_buf, ".PP",   "<PP>",		1);
	tmp_buf = MPSubString(tmp_buf, ".RB",   "<RB>",		1);
	tmp_buf = MPSubString(tmp_buf, ".RI",   "<RI>",		1);
	tmp_buf = MPSubString(tmp_buf, ".Sp",	"<Sp>",		1);
	tmp_buf = MPSubString(tmp_buf, ".ta",	"<ta>",		1);
	/* Single character id substitutions after two character ids */
	tmp_buf = MPSubString(tmp_buf, ".B",    "<B>",		1);
	tmp_buf = MPSubString(tmp_buf, ".I",    "<I>",		1);
/*
 * Need to handle:

.ad b
.hy 1
.if t

 */
	tmp_buf = MPSubString(tmp_buf, "\\1i",	"<li>",		1);
	tmp_buf = MPSubString(tmp_buf, "\\fB",  "<fB>",		0);
	tmp_buf = MPSubString(tmp_buf, "\\fI",	"<fI>",		0);
	tmp_buf = MPSubString(tmp_buf, "\\fR",  "<fR>",         0);
	tmp_buf = MPSubString(tmp_buf, "\\fP",	"<fP>",		0);


	/* Get new length, since it may have changed after the above
	 * substitutions.
	 */
	cur_seg_len = strlen(tmp_buf);

	/* Explode at all new line occurances */
	MPLoadExplodeNewLines(
	    line, total_lines,
	    tmp_buf, (const char *)(tmp_buf + cur_seg_len)
	);

	/* Delete the temporary buffer */
	free(tmp_buf);
}


/*
 *      Takes the given buffer and makes XML to manual page substitutions
 *	then returns a reallocated buffer. Basically doing the opposite
 *	of MPConvertManualPageDataToXMLLines() but does not concatinate
 *	any lines.
 *
 *	The given buffer should no longer be referenced after this call.
 *
 *      Inputs are assumed valid.
 */
static char *MPConvertXMLLineToManualPageData(char *line_ptr)
{
	char *tmp_buf = line_ptr;

	/* Make substitutions to tmp_buf, note that the order is
	 * important and should be in reverse inverse order from the
	 * function MPConvertManualPageDataToXMLLines().
	 */
	tmp_buf = MPSubString(tmp_buf, "<fP>", 	"\\fP",		0);
	tmp_buf = MPSubString(tmp_buf, "<fR>",  "\\fR",         0);
	tmp_buf = MPSubString(tmp_buf, "<fI>",	"\\fI",		0);
	tmp_buf = MPSubString(tmp_buf, "<fB>",  "\\fB",		0);
	tmp_buf = MPSubString(tmp_buf, "<li>",	"\\1i ",	0);

	/* Single character tags */
	tmp_buf = MPSubString(tmp_buf, "<I>",	".I ",		0);
	tmp_buf = MPSubString(tmp_buf, "<B>",	".B ",		0);

	/* Two character tags */
	tmp_buf = MPSubString(tmp_buf, "<ta>",	".ta ",		0);
	tmp_buf = MPSubString(tmp_buf, "<Sp>",	".Sp ",		0);
	tmp_buf = MPSubString(tmp_buf, "<RI>",	".RI ",		0);
	tmp_buf = MPSubString(tmp_buf, "<RB>",	".RB ",		0);
	tmp_buf = MPSubString(tmp_buf, "<PP>",	".PP ",		0);
	tmp_buf = MPSubString(tmp_buf, "<nf>",	".nf ",		0);
	tmp_buf = MPSubString(tmp_buf, "<TP>",  ".TP ",         0);
	tmp_buf = MPSubString(tmp_buf, "<RP>",	".RP ",		0);
	tmp_buf = MPSubString(tmp_buf, "<LP>",  ".LP ",		0);
	tmp_buf = MPSubString(tmp_buf, "<IP>",	".IP ",		0);
	tmp_buf = MPSubString(tmp_buf, "<Im>",	".Im ",		0);
	tmp_buf = MPSubString(tmp_buf, "<If>",	".If ",		0);
	tmp_buf = MPSubString(tmp_buf, "<fi>",	".fi ",		0);
	tmp_buf = MPSubString(tmp_buf, "<br>",	".br ",		0);
	tmp_buf = MPSubString(tmp_buf, "<BR>",	".BR ",		0);
	tmp_buf = MPSubString(tmp_buf, "<BI>",	".BI ",		0);

	/* Literal deliminator characters */
	tmp_buf = MPSubString(tmp_buf, "&gt;",	">",		0);
	tmp_buf = MPSubString(tmp_buf, "&lt;",	"<",		0);
	tmp_buf = MPSubString(tmp_buf, "&amp;",  "&",		0);
	tmp_buf = MPSubString(tmp_buf, "-",	"\\-",		0);

	return(tmp_buf);
}


/*
 *	Used by MPSubString(), no other functions should call this.
 *	Inputs assumed valid.
 *
 *	Returns a reallocated string from buf, the given buf should no
 *	longer be referenced. The sub_start will be a pointer to a pointer
 *	in buf in which the substitution should start at. If a
 *	substitution is made then sub_start will be set to the pointer
 *	in the new buffer after the substituted segment, otherwise if no
 *	substitution is made then sub_start will be set to point to NULL.
 *
 *	Inputs assumed valid.
 */
static char *MPSubStringIterate(
	char *buf, char **sub_start,
	const char *token, const char *replace, int strip_tail
)
{
	char *bp;
	int i, start_offset;
	int buf_len, token_len, replace_len;


	token_len = strlen(token);
	replace_len = strlen(replace);

	/* Search for token in buf at the given starting position
	 * assumed to be in buf.
	 */
	(*sub_start) = strstr(*sub_start, token);
	if((*sub_start) == NULL)
	    return(buf);

	/* Number of bytes into buf that matched the token */
	start_offset = (int)(*sub_start) - (int)buf;

	/* Now (*sub_start) is positioned at the start of the
	 * substitution. Begin removing token string from buf.
	 */
	bp = (*sub_start);
	while(1)
	{
	    (*bp) = (*(bp + token_len));
	    if((*bp) == '\0')
		break;
	    else
		bp++;
	}

	/* Reallocate buffer to accomidate the current length plus
	 * the added replace string.
	 */
	buf_len = strlen(buf) + replace_len;
	buf = (char *)realloc(buf, (buf_len + 1) * sizeof(char));
	if(buf == NULL)
	    return(NULL);

	/* Since reallocated buffer, need to calculate new start of
	 * substitution pointer.
	 */
	(*sub_start) = buf + start_offset;

	/* Replace string has length? */
	if(replace_len > 0)
	{
	    /* Shift to make room for incoming string */
	    bp = buf + buf_len;
	    while((bp - replace_len) >= (*sub_start))
	    {
		(*bp) = (char)(*(bp - replace_len));
		bp--;
	    }

	    /* Put in replace buffer */
	    memcpy(*sub_start, replace, replace_len * sizeof(char));
	}

	/* Remove tailing spaces after replaced string */
	if(strip_tail > 0)
	{
	    for(i = 0; i < strip_tail; i++)
	    {
		bp = (char *)((*sub_start) + replace_len);
		if(!ISBLANK(*bp))
		    break;

		while((*bp) != '\0')
		{
		    (*bp) = (char)(*(bp + 1));
		    bp++;
		}
	    }
	}


	/* Move (*sub_start) pointer past the replaced substitution */
	(*sub_start) = (char *)((*sub_start) + replace_len);

	return(buf);
}

/*
 *	Replaces all occurances of token in the given string buf with the
 *	value of replace. If replace is NULL then an empty string "" will
 *	be used as the replace string.
 *
 *	If strip_tail is positive then the number of spaces (if any or
 *	less) occuring after the replaced string will be stripped.
 *	If strip_tail is 2 then "hi there  baby" "there" "ya" becomes
 *	"hi yababy".
 *
 *	The returned buffer will be either buf, NULL or a reallocated
 *	buf, in any case you should always use the return value as
 *	the new buf (works like realloc()).
 */
static char *MPSubString(
	char *buf, const char *token, const char *replace,
	int strip_tail
)
{
	char *last_sub_ptr;


	if(buf == NULL)
	    return(NULL);

	if(token == NULL)
	    return(buf);
	else if((*token) == '\0')
	    return(buf);

	if(replace == NULL)
	    replace = "";

	/* Start last substitution pointer at buf */
	last_sub_ptr = buf;
	do
	{
	    buf = MPSubStringIterate(
		buf, &last_sub_ptr,
		token, replace, strip_tail
	    );
	}
	while(last_sub_ptr != NULL);

	return(buf);
}


/*
 *	Called by MPHeaderParse() to parse a single header item line.
 */
static void MPHeaderParseHeaderItem(
	MPHeader *mp_header,
	const char *line
)
{
	const char	*line_ptr = line,
			*line_next = line_ptr;

#define SEEK_NEXT {				\
 while(ISBLANK(*line_next))			\
  line_next++;					\
 line_ptr = line_next;				\
 if(*line_next == '"') {			\
  line_ptr++;					\
  line_next++;					\
  while((*line_next != '"') && (*line_next != '\0')) { \
   if(*line_next == '\\') {			\
    line_next++;				\
    if(*line_next != '\0')			\
     line_next++;				\
   } else {					\
    line_next++;				\
   }						\
  }						\
 } else {					\
  while(!ISBLANK(*line_next) && (*line_next != '\0')) \
   line_next++;					\
 }						\
}
#define GET_VALUE(_vs_) {			\
 /* Allocate memory for new argument */		\
 const int seg_len = line_next - line_ptr;	\
 char *s = (char *)malloc((seg_len + 1) * sizeof(char)); \
 if(s != NULL) {				\
  if(seg_len > 0)				\
   memcpy(					\
    s,						\
    line_ptr,					\
    seg_len					\
   );						\
  s[seg_len] = '\0';				\
 }						\
						\
 /* Set the value */				\
 free(_vs_);					\
 (_vs_) = s;					\
						\
 /* Seek past the current quote if any */	\
 if(*line_next == '"')				\
  line_next++;					\
}

	SEEK_NEXT
	GET_VALUE(mp_header->name);

	SEEK_NEXT
	GET_VALUE(mp_header->section_number);

	SEEK_NEXT
	GET_VALUE(mp_header->version);

	SEEK_NEXT
	GET_VALUE(mp_header->author);

	SEEK_NEXT
	GET_VALUE(mp_header->catagory);

#undef SEEK_NEXT
#undef GET_VALUE
}

/*
 *	Called by MPOpenNexus() to gets the header values from the
 *	header lines.
 */
static void MPHeaderParse(
	MPHeader *mp_header,
	char **lines_list, const int nlines
)
{
	int i;
	const char	*line,
			*line_ptr;

	for(i = 0; i < nlines; i++)
	{
	    line = lines_list[i];
	    if(line == NULL)
		continue;

	    line_ptr = line;
	    while(ISBLANK(*line_ptr))
		line_ptr++;

	    /* Title Heading */
	    if(strpfx(line_ptr, ".TH"))
	    {
		/* Seek to the first argument */
		while(!ISBLANK(*line_ptr) && ((*line_ptr) != '\0'))
		    line_ptr++;
		while(ISBLANK(*line_ptr))
		    line_ptr++;

		MPHeaderParseHeaderItem(
		    mp_header,
		    line_ptr
		);
	    }
	    /* Comment */
	    else if(strpfx(line_ptr, ".\\\""))
	    {
		/* Seek to the first and only argument */
		while(!ISBLANK(*line_ptr) && (*line_ptr != '\0'))
		    line_ptr++;

		while(ISBLANK(*line_ptr))
		    line_ptr++;

		mp_header->comments_lines_list = strlistappend(
		    mp_header->comments_lines_list,
		    &mp_header->ncomments_lines,
		    line_ptr
		);
	    }
	}
}



/*
 *	Loads the contents of the manual page file specified by path
 *	or data into the given pointers to substructures which this
 *	function will allocate.
 *
 *	If the progress_cb function is not NULL and it returns non-zero
 *	then the loading will be aborted. In which case return will be
 *	success but returned data will be short.
 *
 *	The loaded data needs to be freed with a call to MPHeaderDelete()
 *	and MPSectionsListClear().
 *
 *      Returns non-zero on error, where -1 is a general error and -2 is
 *      file not found/not allowed to read.
 */
int MPOpenNexus(
	const char *path,
	const char *data,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	int		n,
			data_len,
			buf_len,
			cur_seg_len,
			mp_total_sections;
	char		*buf,			/* Entire loaded buffer and length */
			*buf_end,
			*cur_buf_ptr,
			*limit_buf_ptr,
			*cur_section_name;
	MPHeader *mp_header;
	MPSection	**mp_section,
			*mp_section_ptr;

	/* Reset returns */
	if(mp_header_rtn != NULL)
	    (*mp_header_rtn) = NULL;
	if(mp_section_rtn != NULL)
	    (*mp_section_rtn) = NULL;
	if(mp_total_sections_rtn != NULL)
	    (*mp_total_sections_rtn) = 0;


	/* Load from file? */
	if(path != NULL)
	{
	    /* Report the initial progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(0l, 1l, progress_data))
		    return(0);
	    }

	    /* Load data from file, this will return a dynamically
	     * allocated buffer containing the literal data loaded from
	     * file.  The buffer length buf_len will not include the null
	     * terminating byte but buf will contain it if it is not NULL.
	     */
	    buf = MPFileOpenNormal(
		path,
		&buf_len,
		progress_cb,
		progress_data
	    );
	    if((buf == NULL) || (buf_len <= 0))
	    {
		free(buf);
		return(-1);
	    }

	    data_len = buf_len;
	}
	/* Load from data? */
	else if(data != NULL)
	{
	    data_len = strlen(data);

	    /* Report the initial progress */
	    if(progress_cb != NULL)
	    {
		if(progress_cb(0l, (unsigned long)data_len, progress_data))
		    return(-4);
	    }

	    /* How much buffer do we need? */
	    buf_len = data_len;
	    if(buf_len < 0)
		buf_len = 0;

	    /* Allocate buffer to load file into */
	    buf = (char *)malloc((buf_len + 1) * sizeof(char));
	    if(buf == NULL)
		return(-1);

	    /* Copy data to new buf */
	    memcpy(
		buf,
		data,
		buf_len * sizeof(char)
	    );
	    buf[buf_len] = '\0';	/* Yes its allocated */
	}
	else
	{
	    /* No given path or data to load from */
	    return(-1);
	}


	/* At this point the buffer buf is not NULL and should contain the
	 * given data or the data loaded from file.
	 */

	/* Begin parsing the buffer
	 *
	 * Get the buffer boundary pointers
	 */
	buf_end = buf + buf_len;
	cur_buf_ptr = buf;

	/* Parse the header */
	mp_header = MPHeaderNew();
	if(mp_header != NULL)
	{
	    char **header_lines = NULL;
	    int nheader_lines = 0;

	    /* Get pointer to start of first section heading (if any) */
	    limit_buf_ptr = strstr(cur_buf_ptr, ".SH");
	    if(limit_buf_ptr == NULL)
		limit_buf_ptr = buf_end;

	    cur_seg_len = MAX((int)limit_buf_ptr - (int)cur_buf_ptr, 0);

	    /* Explode the header lines */
	    MPLoadExplodeNewLines(
		&header_lines, &nheader_lines,
		cur_buf_ptr, limit_buf_ptr
	    );

	    /* Get the header values from the header lines */
	    MPHeaderParse(
		mp_header,
		header_lines, nheader_lines
	    );

	    /* Delete the header lines */
	    strlistfree(header_lines, nheader_lines);

	    /* Seek to the start of the first section for section
	     * parsing
	     */
	    cur_buf_ptr = limit_buf_ptr;
	}
	if(mp_header_rtn != NULL)
	    *mp_header_rtn = mp_header;
	else
	    MPHeaderDelete(mp_header);
	mp_header = NULL;

	/* Call progress callback */
	if(progress_cb != NULL)
	{
	    const int	i = cur_buf_ptr - buf,
			m = data_len;
	    if(progress_cb(
		(unsigned long)((1 * m) + i),
		(unsigned long)(2 * m),
		progress_data
	    ))
	    {
		free(buf);
		return(-4);
	    }
	}


	/* Begin loading sections */
	mp_section = NULL;
	mp_total_sections = 0;
	while(cur_buf_ptr < buf_end)
	{
	    /* Call progress callback? */
	    if(progress_cb != NULL)
	    {
		const int	i = cur_buf_ptr - buf,
				m = data_len;
		if(progress_cb(
		    (unsigned long)((1 * m) + i),
		    (unsigned long)(2 * m),
		    progress_data
		))
		{
		    if((mp_section_rtn == NULL) || (mp_total_sections_rtn == NULL))
		    {
			MPSectionsListClear(mp_section, mp_total_sections);
		    }
		    else 
		    {
			(*mp_section_rtn) = mp_section;
			(*mp_total_sections_rtn) = mp_total_sections;
		    }
		    free(buf);
		    return(-4);
		}
	    }


	    /* Increment current buffer pointer past the ".SH" marking
	     * plus any spaces.
	     */
	    cur_buf_ptr += strlen(".SH");
	    while(ISBLANK(*cur_buf_ptr))
		cur_buf_ptr++;

	    /* Now positioned at start of section name, check if
	     * section name is encapsulated withing double quotes.
	     */
	    cur_section_name = NULL;
	    if((*cur_buf_ptr) == '"')
	    {
		/* Need to seek to other end of '"' */
		cur_buf_ptr++;
		limit_buf_ptr = (char *)MPLoadStringNextDoubleQuote(cur_buf_ptr);
		if(limit_buf_ptr == NULL)
		    limit_buf_ptr = buf_end;

		cur_seg_len = MAX((int)limit_buf_ptr - (int)cur_buf_ptr, 0);
		cur_section_name = (char *)malloc((cur_seg_len + 1) * sizeof(char));
		if(cur_section_name != NULL)
		{
		    strncpy(cur_section_name, cur_buf_ptr, cur_seg_len);
		    cur_section_name[cur_seg_len] = '\0';
		}

		/* Seek current buffer to start of section body */
		cur_buf_ptr = limit_buf_ptr;
		while((cur_buf_ptr < buf_end) &&
		      (((*cur_buf_ptr) == '"') ||
		       ISBLANK(*cur_buf_ptr) ||
		       ISCR(*cur_buf_ptr)
		      )
		)
		    cur_buf_ptr++;
	    }
	    else
	    {
		/* Seek to next newline character */
		limit_buf_ptr = cur_buf_ptr;
		while((limit_buf_ptr < buf_end) &&
/* Some older man pages need to have entire .SH line loaded as section name
		      !ISBLANK(*limit_buf_ptr) && */
		      !ISCR(*limit_buf_ptr)
		)
		    limit_buf_ptr++;

		cur_seg_len = MAX((int)limit_buf_ptr - (int)cur_buf_ptr, 0);
		cur_section_name = (char *)malloc((cur_seg_len + 1) * sizeof(char));
		if(cur_section_name != NULL)
		{
		    strncpy(cur_section_name, cur_buf_ptr, cur_seg_len);
		    cur_section_name[cur_seg_len] = '\0';
		}

		/* Seek current buffer to start of section body */
		cur_buf_ptr = limit_buf_ptr;
		while((cur_buf_ptr < buf_end) &&
		      (ISBLANK(*cur_buf_ptr) || 
		       ISCR(*cur_buf_ptr)
		      )
		)
		    cur_buf_ptr++;
	    }

	    /* Current buffer position is now positioned at start of
	     * section body.
	     */

	    /* Allocate new section structure */
	    n = mp_total_sections;
	    mp_total_sections = n + 1;
	    mp_section = (MPSection **)realloc(
		mp_section, mp_total_sections * sizeof(MPSection *)
	    );
	    if(mp_section == NULL)
	    {
		mp_total_sections = 0;
		cur_buf_ptr = buf_end;	/* Set current buffer to end */
		break;
	    }
	    else
	    {
		mp_section[n] = mp_section_ptr = MPSectionNew();
		if(mp_section_ptr == NULL)
		{
		    cur_buf_ptr = buf_end;	/* Set current buffer to end */
		    break;
		}
	    }

	    /* Set the section's name */
	    mp_section_ptr->name = cur_section_name;
	    cur_section_name = NULL;

	    /* Fetch section's body */
	    limit_buf_ptr = strstr(cur_buf_ptr, ".SH");
	    if(limit_buf_ptr == NULL)
		limit_buf_ptr = buf_end;

	    /* Explode and parse the manual page format, storing the
	     * lines into the given lines array after parsing cur_buf_ptr
	     * to limit_buf_ptr.
	     */
	    MPConvertManualPageDataToXMLLines(
		cur_buf_ptr, limit_buf_ptr,
		&mp_section_ptr->lines_list, &mp_section_ptr->nlines
	    );

	    /* Move current position to start of next section's
	     * deliminator.
	     */
	    cur_buf_ptr = limit_buf_ptr;
	}
	if((mp_section_rtn == NULL) || (mp_total_sections_rtn == NULL))
	{
	    MPSectionsListClear(mp_section, mp_total_sections);
	}
	else
	{
	    (*mp_section_rtn) = mp_section;
	    (*mp_total_sections_rtn) = mp_total_sections;
	}
	mp_section = NULL;
	mp_total_sections = 0;

	/* Report the final progress */
	if(progress_cb != NULL)
	{
	    const int m = data_len;
	    if(progress_cb(
		(unsigned long)(2 * m),
		(unsigned long)(2 * m),
		progress_data
	    ))
	    {
		free(buf);
		return(-4);
	    }
	}   

	free(buf);

	return(0);
}

int MPFileOpen(
	const char *path,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	return(MPOpenNexus(
	    path,
	    NULL,
	    mp_header_rtn,
	    mp_section_rtn, mp_total_sections_rtn,
	    progress_cb, progress_data
	));
}

int MPDataOpen(
	const char *data,
	MPHeader **mp_header_rtn,
	MPSection ***mp_section_rtn, int *mp_total_sections_rtn,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	return(MPOpenNexus(
	    NULL,
	    data,
	    mp_header_rtn,
	    mp_section_rtn, mp_total_sections_rtn,
	    progress_cb, progress_data
	));
}


/*
 *	Saves the manual page data into either the specified path
 *	or data_rtn (whichever is not NULL).
 *
 *      If the progress_cb function is not NULL and it returns non-zero
 *      then the saving will be aborted. In which case return will be
 *      success but saved data will be short.
 *
 *      Returns non-zero on error, where -1 is a general error and -2 is
 *      cannot write to file.
 */
int MPSaveNexus(
	const char *path,
	char **data_rtn,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	int		i,
			nitems_to_save,
			status,
			section_num,
			buf_len;
	char *buf;
	const char *line_ptr;
	MPSection *section_ptr;

	/* Both path and data return may not be both NULL */
	if((path == NULL) && (data_rtn == NULL))
	    return(-1);

	buf = NULL;
	buf_len = 0;

	nitems_to_save = 1 + mp_total_sections;

/* Appends the string pointed to by out_ptr (which is assumed not
 * NULL) and appends it to the buf, reallocating buf as needed and
 * increasing the buf_len.
 */
#define WRITE_OUT(_s_)	{			\
 const int	cur_out_len = strlen(_s_),	\
		cur_out_pos = buf_len;		\
						\
 if(cur_out_len > 0) {				\
  buf_len += cur_out_len;			\
  buf = (char *)realloc(			\
   buf,						\
   (buf_len + 1) * sizeof(char)			\
  );						\
  if(buf == NULL) {				\
   buf_len = 0;					\
  } else {					\
   memcpy(					\
    &buf[cur_out_pos],				\
    (_s_),					\
    cur_out_len					\
   );						\
   buf[buf_len] = '\0';				\
  }						\
 }						\
}

	/* Report the initial progress */
	if(progress_cb != NULL)
	{
	    if(progress_cb(
		0l,
		(unsigned long)(2 * nitems_to_save),
		progress_data
	    ))
	    {
		free(buf);
		return(-4);
	    }
	}

	/* Header */
	if(mp_header != NULL)
	{
	    /* Comments */
	    for(i = 0; i < mp_header->ncomments_lines; i++)
	    {
		line_ptr = mp_header->comments_lines_list[i];
		if(line_ptr == NULL)
		    continue;

		WRITE_OUT(".\\\" ");
		WRITE_OUT(line_ptr);
		WRITE_OUT("\n");
	    }

	    WRITE_OUT(".TH");

	    WRITE_OUT(" \"");
	    if(mp_header->name != NULL)
		WRITE_OUT(mp_header->name);
	    WRITE_OUT("\"");

	    WRITE_OUT(" \"");
	    if(mp_header->section_number != NULL)
		WRITE_OUT(mp_header->section_number);
	    WRITE_OUT("\"");

	    WRITE_OUT(" \"");
	    if(mp_header->version != NULL)
		WRITE_OUT(mp_header->version);
	    WRITE_OUT("\"");

	    WRITE_OUT(" \"");
	    if(mp_header->author != NULL)
		WRITE_OUT(mp_header->author);
	    WRITE_OUT("\"");

	    WRITE_OUT(" \"");
	    if(mp_header->catagory != NULL)
		WRITE_OUT(mp_header->catagory);
	    WRITE_OUT("\"");

	    WRITE_OUT("\n");
	}

	/* Write each section */
	if(mp_section != NULL)
	{
	    /* Go through each section */
	    for(section_num = 0; section_num < mp_total_sections; section_num++)
	    {
		section_ptr = mp_section[section_num];
		if(section_ptr == NULL)
		    continue;

		/* Report the progress */
		if(progress_cb != NULL)
		{
		    if(progress_cb(
			(unsigned long)section_num,
			(unsigned long)(2 * nitems_to_save),
			progress_data
		    ))
		    {
			free(buf);
			return(-4);
		    }
		}

		/* Write section deliminator */
		WRITE_OUT(".SH ");

		/* Section name not available? */
		if(section_ptr->name == NULL)  
		{
		    WRITE_OUT("\"UNTITLED\"");
		}
		else
		{
		    int len = strlen(section_ptr->name) + 4;
		    char *s = (char *)malloc((len + 1) * sizeof(char));
		    if(s != NULL)
		    {
			sprintf(
			    s,
			    "\"%s\"",
			    section_ptr->name
			);
			WRITE_OUT(s);
			free(s);
		    }
		}

		/* Add new line at the end */
		WRITE_OUT("\n");

		/* Write out each line */
		for(i = 0; i < section_ptr->nlines; i++)
		{
		    line_ptr = section_ptr->lines_list[i];
		    if(line_ptr != NULL)
		    {
			char *s = strdup(line_ptr);
			if(s == NULL)
			    continue;

			/* Make substitutions and other formatting to
			 * the coppied line before writing it out
			 */
			s = MPConvertXMLLineToManualPageData(s);
			WRITE_OUT(s);
			free(s);

			WRITE_OUT("\n");
		    }
		}
	    }
	}

#undef WRITE_OUT

	/* At this point buf and buf_len specify the (null terminated)
	 * data to be written out.
	 */

	/* Write out to file? */
	status = -2;
	if(path != NULL)
	{
	    if((buf != NULL) && (buf_len > 0))
		status = MPFileSaveNormal(
		    path,
		    buf, buf_len,
		    progress_cb,
		    progress_data
		);
	}
	else if(data_rtn != NULL)
	{
/* Need to work on this */
	}

	free(buf);

	return(status);
}

int MPFileSave(
	const char *path,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	return(MPSaveNexus(
	    path,
	    NULL,
	    mp_header,
	    mp_section, mp_total_sections,
	    progress_cb, progress_data
	));
}

int MPDataSave(
	char **data_rtn,
	MPHeader *mp_header,
	MPSection **mp_section, int mp_total_sections,
        int (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data
)
{
	return(MPSaveNexus(
	    NULL,
	    data_rtn,
	    mp_header,
	    mp_section, mp_total_sections,
	    progress_cb, progress_data
	));
}
