#include <gtk/gtk.h>
#include <string.h>

#include "guiutils.h"

#include "ca_string.h"
#include "manpage_view.h"
#include "manpage_view_private.h"


/* MPVStrings */
MPVString *MPVStringNew(void);
void MPVStringDelete(MPVString *ms);

gboolean MPVStringIsEmpty(MPVString *ms);

void MPVStringClearLineEnds(MPVString *ms);

gchar *MPVStringToRegularString(
	MPVString *ms,
	const gint len
);

void MPVStringSegmentAddAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);
void MPVStringSegmentRemoveAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);

void MPVStringSelect(
	MPVString *ms,
	const gint selection_start_pos,
	const gint selection_end_pos,
	const gboolean clear_existing_selection
);
void MPVStringUnselectAll(MPVString *ms);

void MPVStringInsertCharacter(
	MPVString *ms,
	const gint i,
	const gchar c,
	const CAStringAttributes attr
);
void MPVStringInsertText(
	MPVString *ms,
	const gint i,
	const CAString *ca_string
);
void MPVStringRemoveCharacter(
	MPVString *ms,
	const gint i
);
void MPVStringRemoveText(
	MPVString *ms,
	const gint i,
	const gint len
);


/* MPVObjects */
MPVObject *MPVObjectNew(void);
void MPVObjectDelete(MPVObject *o);

MPVString *MPVObjectGetString(
	MPVObject *o,
	const gint string_num
);
void MPVObjectRemoveSelectedCharacters(MPVObject *o);


/* Translate Coordinates */
static CAString *MPVStringTranslateCoordinateToLine(
	MPView *v,
	CAString *cas,
	const gint y,
	gint *line_num_rtn
);
static gint MPVStringTranslateCoordinatesToIndex(
	MPView *v,
	CAString *cas,
	const gint x,
	const gint y
);
gboolean MPViewTranslateObjectCoordinatesToIndex(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint x, const gint y,
	gint *cursor_pos_rtn
);
gboolean MPViewTranslateCoordinatesToIndex(
	GtkWidget *w,
	const gint x, const gint y,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn
);
gboolean MPViewTranslateCoordinatesToLine(
	GtkWidget *w,
	const gint y,
	const gint start_obj_num,
	const gint start_string_num,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *line_num_rtn
);

/* MPView Get Font, Color, MPVObject */
static GdkFont *MPViewGetFontFromAttribute(
	MPView *v,
	const CAStringAttributes attr
);
static GdkColor *MPViewGetColorFromAttribute(
	MPView *v,
	const CAStringAttributes attr
);
MPVObject *MPViewGetObject(
	MPView *v,
	const gint obj_num
);

/* MPView Set/Get Cursor Position */
void MPViewSetCursorPosition(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint cursor_pos
);
void MPViewGetCursorPosition(
	GtkWidget *w,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn,
	MPGeometry **cursor_geo_rtn
);

/* MPView Select */
void MPViewSelect(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint selection_start_pos,
	const gint selection_end_pos
);
void MPViewSelectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
);
void MPViewSelectObjectsRange(
	GtkWidget *w,
	const gint start_obj_num,
	const gint start_string_num,
	const gint start_cursor_pos,
	const gint end_obj_num,
	const gint end_string_num,
	const gint end_cursor_pos
);
void MPViewSelectObjectAll(
	GtkWidget *w,
	const gint obj_num
);

/* MPView Unselect */
void MPViewUnselectAll(GtkWidget *w);
void MPViewUnselectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
);
void MPViewUnselectObjectAll(
	GtkWidget *w,
	const gint obj_num
);

/* MPView Remove Selected Characters */
void MPViewRemoveSelectedCharacters(GtkWidget *w);

/* MPView Visibility */
GdkVisibilityState MPViewGetObjectVisibility(
	GtkWidget *w,
	const gint obj_num
);
GdkVisibilityState MPViewGetCursorVisibilityX(GtkWidget *w);
GdkVisibilityState MPViewGetCursorVisibilityY(GtkWidget *w);
GdkVisibilityState MPViewGetCursorVisibility(GtkWidget *w);

/* MPView Move To */
void MPViewMoveToObject(
	GtkWidget *w,
	const gint obj_num,
	const gfloat x_coeff, const gfloat y_coeff
);
void MPViewMoveToCursorX(
	GtkWidget *w,
	const gfloat x_coeff
);
void MPViewMoveToCursorY(
	GtkWidget *w,
	const gfloat y_coeff
);
void MPViewMoveToCursor(
	GtkWidget *w,
	const gfloat x_coeff, const gfloat y_coeff
);

/* MPView Editing */
void MPViewCutString(GtkWidget *w);
void MPViewCopyString(GtkWidget *w);
void MPViewCutObjects(GtkWidget *w);
void MPViewCopyObjects(GtkWidget *w);
void MPViewPaste(GtkWidget *w);
void MPViewDelete(GtkWidget *w);

/* MPView */
MPView *MPViewGetWidgetData(
	GtkWidget *w,
	const gchar *func_name
);
GtkWidget *MPViewNew(void);
static CAString *MPViewResizeMPVStringCalculateLineWidth(
	MPView *v,
	MPVString *ms,
	CAString *ca_string_ptr,
	GdkFont **cur_font,
	const gint width_max,
	gint *line_width_rtn, gint *line_height_rtn
);
static void MPViewResizeMPVString(
	MPView *v,
	MPVString *ms,
	const gint width_max
);
static void MPViewResizeObject(
	MPView *v,
	MPVObject *o,
	const gint width_max
);
void MPViewResize(
	GtkWidget *w,
	const gint width, const gint height
);
void MPViewQueueResize(GtkWidget *w);
void MPViewUpdateCursorCoordinates(GtkWidget *w);
void MPViewDrawLineBar(GtkWidget *w);
static void MPVDrawObjectString(
	MPView *v,
	MPVObject *o,
	const gint string_num,
	MPVString *ms,
	GtkWidget *w,
	GdkDrawable *drawable,
	GdkGC *gc,
	GtkStyle *style,
	const gint start_x, const gint start_y
);
static void MPViewDrawObject(
	MPView *v,
	MPVObject *o,
	GtkWidget *w,
	GdkDrawable *drawable,
	GdkGC *gc,
	GtkStyle *style,
	const gint width, const gint height,
	const gint scroll_x, const gint scroll_y
);
void MPViewDraw(GtkWidget *w);
void MPViewQueueDraw(GtkWidget *w);


#define MP_VIEW_DEF_X_MARGIN			10
#define MP_VIEW_DEF_Y_MARGIN			10

#define MP_VIEW_DEF_X_SPACING			5
#define MP_VIEW_DEF_Y_SPACING			5

#define MP_VIEW_OBJECT_DEF_X_MARGIN		2
#define MP_VIEW_OBJECT_DEF_Y_MARGIN		2


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Creates a new MPVString.
 */
MPVString *MPVStringNew(void)
{
	return(MPV_STRING(g_malloc0(sizeof(MPVString))));
}

/*
 *	Deletes the MPVString.
 */
void MPVStringDelete(MPVString *ms)
{
	if(ms == NULL)
	    return;

	CAStringDelete(ms->ca_string);
	g_free(ms);
}


/*
 *	Checks if the MPVString is NULL or empty.
 */
gboolean MPVStringIsEmpty(MPVString *ms)
{
	if(ms == NULL)
	    return(TRUE);

	return(CAStringIsEmpty(ms->ca_string));
}


/*
 *	Removes all CA_STRING_CHARACTER_LINE_END characters and
 *	replaces them with spaces.
 */
void MPVStringClearLineEnds(MPVString *ms)
{
	if(ms == NULL)
	    return;

	CAStringClearLineEnds(ms->ca_string);
}


/*
 *	Copy/converts the MPVString to a regular single-byte
 *      null-terminated string.
 */
gchar *MPVStringToRegularString(
	MPVString *ms,
	const gint len
)
{
	if(ms == NULL)
	    return(NULL);

	return(CAStringToRegularString(
	    ms->ca_string,
	    len
	));
}


/*
 *	Adds the attribute to the MPVString region.
 */
void MPVStringSegmentAddAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
)
{
	if(ms == NULL)
	    return;

	CAStringSegmentAddAttribute(
	    ms->ca_string,
	    start_i, end_i,
	    attr
	);
}

/*
 *	Removes the attribute from the MPVString region.
 */
void MPVStringSegmentRemoveAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
)
{
	if(ms == NULL)
	    return;

	CAStringSegmentRemoveAttribute(
	    ms->ca_string,
	    start_i, end_i,
	    attr
	);
}


/*
 *	Adds the CA_STRING_ATTRIBUTE_SELECTED attribute to the
 *	MPVString region and, if clear_existing_selection is TRUE,
 *	removes the CA_STRING_ATTRIBUTE_SELECTED attribute from the
 *	rest.
 */
void MPVStringSelect(
	MPVString *ms,
	const gint selection_start_pos,
	const gint selection_end_pos,
	const gboolean clear_existing_selection
)
{
	CAString	*ca_string,
			*ca_string_ptr;

	if(ms == NULL)
	    return;

	CAStringSelect(
	    ms->ca_string,
	    selection_start_pos, selection_end_pos,
	    clear_existing_selection
	);

	/* Update the string's selection bounds to cover all the
	 * extents of the selected region(s)
	 *
	 * Seek to the start of the selection
	 */
	ca_string_ptr = ca_string = ms->ca_string;
	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    if(CA_STRING_ATTRIBUTE(ca_string_ptr) & CA_STRING_ATTRIBUTE_SELECTED)
		break;

	    ca_string_ptr++;
	}
	ms->selection_start_pos = CAStringLengthPointers(
	    ca_string,
	    ca_string_ptr
	);

	/* Seek to the end */
	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	    ca_string_ptr++;

	/* Seek to the end of the selection */
	while(ca_string_ptr > ca_string)
	{
	    if(CA_STRING_ATTRIBUTE(ca_string_ptr) & CA_STRING_ATTRIBUTE_SELECTED)
		break;

	    ca_string_ptr--;
	}
	ms->selection_end_pos = CAStringLengthPointers(
	    ca_string,
	    ca_string_ptr 
	) + 1;
}

/*
 *	Removes the CA_STRING_ATTRIBUTE_SELECTED attribute from the
 *      MPVString region.
 */
void MPVStringUnselectAll(MPVString *ms)
{
	if(ms == NULL)
	    return;

	ms->selection_start_pos = -1;
	ms->selection_end_pos = -1;

	CAStringUnselectAll(ms->ca_string);
}


/*
 *	Inserts a character.
 *
 *	The i specifies the insert position, if i is -1 or out of
 *	bounds then the new character will be appended.
 */
void MPVStringInsertCharacter(
	MPVString *ms,
	const gint i,
	const gchar c,
	const CAStringAttributes attr
)
{
	if(ms == NULL)
	    return;

	ms->ca_string = CAStringInsertCharacter(
	    ms->ca_string,
	    i,
	    c,
	    attr
	);
	ms->length = CAStringLength(ms->ca_string);
}

/*
 *	Inserts a section of text.
 *
 *	The i specifies the insert position, if i is -1 or out of
 *	bounds then the new text will be appended.
 */
void MPVStringInsertText(
	MPVString *ms,
	const gint i,
	const CAString *ca_string
)
{
	if(ms == NULL)
	    return;

	ms->ca_string = CAStringInsertText(
	    ms->ca_string,
	    i,
	    ca_string
	);
	ms->length = CAStringLength(ms->ca_string);
}

/*
 *	Removes a character.
 */
void MPVStringRemoveCharacter(
	MPVString *ms,
	const gint i
)
{
	if(ms == NULL)
	    return;

	ms->ca_string = CAStringRemoveCharacter(
	    ms->ca_string,
	    i
	);
	ms->length = CAStringLength(ms->ca_string);
}

/*
 *	Removes a section of text.
 */
void MPVStringRemoveText(
	MPVString *ms,
	const gint i,
	const gint len
)
{
	if(ms == NULL)
	    return;

	ms->ca_string = CAStringRemoveText(
	    ms->ca_string,
	    i,
	    len
	);
	ms->length = CAStringLength(ms->ca_string);
}


/*
 *	Creates a new MPVObject.
 */
MPVObject *MPVObjectNew(void)
{
	return(MPV_OBJECT(g_malloc0(sizeof(MPVObject))));
}

/*
 *	Deletes the MPVObject.
 */
void MPVObjectDelete(MPVObject *o)
{
	if(o == NULL)
	    return;

	if(o->strings_list != NULL)
	{
	    g_list_foreach(
		o->strings_list,
		(GFunc)MPVStringDelete,
		NULL
	    );
	    g_list_free(o->strings_list);
	}
	g_free(o);
}


/*
 *	Gets the MPVString on the MPVObject.
 */
MPVString *MPVObjectGetString(
	MPVObject *o,
	const gint string_num
)
{
	if((o == NULL) || (string_num < 0))
	    return(NULL);

	return(MPV_STRING(g_list_nth_data(
	    o->strings_list,
	    (guint)string_num
	)));
}

/*
 *	Removes all the selected characters from all the MOVObject's
 *	MPVStrings.
 */
void MPVObjectRemoveSelectedCharacters(MPVObject *o)
{
	GList *glist;
	CAString *ca_string_ptr;
	MPVString *ms;

	if(o == NULL)
	    return;

	/* Delete the selected characters on all the MPVStrings */
	for(glist = o->strings_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    ms = MPV_STRING(glist->data);
	    if(ms == NULL)
		continue;

	    ca_string_ptr = ms->ca_string;
	    if(ca_string_ptr == NULL)
		continue;

	    /* Delete this ACString's selected characters and update
	     * this MPVString's cursor position
	     */
	    while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	    {
		if(CA_STRING_ATTRIBUTE(ca_string_ptr) & CA_STRING_ATTRIBUTE_SELECTED)
		{
		    ms->cursor_pos = CAStringLengthPointers(
			ms->ca_string,
			ca_string_ptr
		    );
		    ms->ca_string = CAStringRemoveCharacter(
			ms->ca_string,
			ms->cursor_pos
		    );
		    ca_string_ptr = ms->ca_string;
		    if(ca_string_ptr != NULL)
			continue;
		    else
			break;
		}

		ca_string_ptr++;
	    }

	    /* Clear this MPVString's selection bounds */
	    ms->selection_start_pos = -1;
	    ms->selection_end_pos = -1;
	}

	/* Clear the MPVObject's has selection flag */
	o->flags &= ~MPV_OBJECT_HAS_SELECTION;
}


/*
 *	Seeks to the line that the y coordinate is on.
 */
static CAString *MPVStringTranslateCoordinateToLine(
	MPView *v,
	CAString *ca_string,
	const gint y,
	gint *line_num_rtn
)
{
	gint		line_num = 0,
			line_y = 0,
			line_height = 0;
	gchar characters_list[] = {
	    CA_STRING_CHARACTER_LINE_END,
	    CA_STRING_CHARACTER_LINE_BREAK
	};
	CAString	*ca_string_ptr = ca_string,
			*cas_line_start = ca_string_ptr,
			*cas_next,
			*cas_next_attr,
			*cas_next_line_end;
	GdkFont *font;

	if(y < line_y)
	{
	    *line_num_rtn = line_num;
	    return(ca_string_ptr);
	}

	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    /* Get the GdkFont for the curent attribute */
	    font = MPViewGetFontFromAttribute(v, CA_STRING_ATTRIBUTE(ca_string_ptr));
	    if(font != NULL)
	    {
		/* Update the current line height */
		if(line_height < (font->ascent + font->descent))
		    line_height = font->ascent + font->descent;
	    }

	    /* Seek to the next change in attribute and line end */
	    cas_next_attr = CAStringSeekToNextAttribute(ca_string_ptr);
	    cas_next_line_end = CAStringSeekToNextCharacters(
		ca_string_ptr,
		characters_list,
		sizeof(characters_list) / sizeof(gchar)
	    );

	    /* Does a line end occure at or before the attribute
	     * changes?
	     */
	    if(cas_next_line_end <= cas_next_attr)
	    {
		cas_next = cas_next_line_end;

		/* Does the y coordinate fall on this line? */
		if((y >= line_y) && (y < (line_y + line_height)))
		{
		    *line_num_rtn = line_num;
		    return(cas_line_start);
		}

		/* If the next segment is the end of the string then
		 * return the start of this line
		 */
		if(CA_STRING_CHARACTER(cas_next) == '\0')
		{
		    *line_num_rtn = line_num;
		    return(cas_line_start);
		}

		/* Update/reset the contexts */
		line_y += line_height;
		line_height = 0;
		line_num++;

		/* Go on to the next line */
		if(CA_STRING_CHARACTER(cas_next) != '\0')
		    cas_next++;
		ca_string_ptr = cas_line_start = cas_next;
	    }
	    else
	    {
		/* Attribute changes before any line end, so go on
		 * to the next portion with the different
		 * attribute
		 */
		cas_next = cas_next_attr;
		if(cas_next != ca_string_ptr)
		    ca_string_ptr = cas_next;
		else
		    ca_string_ptr++;
	    }
	}

	*line_num_rtn = line_num;

	return(cas_line_start);
}

/*
 *	Converts the coordinates to index position in the string.
 *
 *	The x and y specifies the coordinates which must be relative
 *	to the string's origin.
 */
static gint MPVStringTranslateCoordinatesToIndex(
	MPView *v,
	CAString *cas,
	const gint x,
	const gint y
)
{
	gchar c;
	gint		i,
			cur_x,
			line_num;
	CAString *cas_line;
	GdkFont *font;

	if(cas == NULL)
	    return(-1);

	if(x < 0)
	    return(0);

	/* Find the line that the y coordinate lies on */
	cas_line = MPVStringTranslateCoordinateToLine(
	    v,
	    cas,
	    y,
	    &line_num
	);
	if(cas_line == NULL)
	    return(-1);

	/* Search for the x position on this line */
	i = CAStringLengthPointers(cas, cas_line);
	cur_x = 0;
	cas = cas_line;
	while(CA_STRING_CHARACTER(cas) != '\0')
	{
	    font = MPViewGetFontFromAttribute(v, CA_STRING_ATTRIBUTE(cas));
	    if(font != NULL)
	    {
		const gint char_width = gdk_char_width(
		    font,
		    CA_STRING_CHARACTER(cas)
		);
		if((x >= cur_x) && (x < (cur_x + char_width)))
		    return(i);

		/* End of line? */
		c = CA_STRING_CHARACTER(cas);
		if((c == CA_STRING_CHARACTER_LINE_END) ||
		   (c == CA_STRING_CHARACTER_LINE_BREAK)
		)
		    return(i);

		cur_x += char_width;
	    }

	    i++;
	    cas++;
	}

	return(i);
}


/*
 *	Converts the coordinate position relative to the MPView's
 *	GtkDrawingArea to the specified MPVString's cursor position.
 *
 *	The obj_num specifies the MPVObject to calculate the cursor
 *	position relative to.
 *
 *	The string_num specifies the MPVString on the MPVObject to
 *	calculate the cursor position relative to.
 *
 *	The x and y specifies the coordinate position relative to the
 *	MPView's GtkDrawingArea.
 *
 *	The cursor_pos_rtn specifies the return value for the cursor
 *	position.
 *
 *	Returns TRUE if a valid cursor position was obtained.
 */
gboolean MPViewTranslateObjectCoordinatesToIndex(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint x, const gint y,
	gint *cursor_pos_rtn
)
{
	gint		dx, dy,
			o_dx, o_dy,
			cursor_pos;
	GtkAdjustment *adj;
	MPGeometry	*ms_geo,
			*o_geo;
	MPVString *ms;
	MPVObject *o;
	MPView *v;

	if(cursor_pos_rtn != NULL)
	    *cursor_pos_rtn = -1;

	v = MPViewGetWidgetData(
	    w,
	    "MPViewTranslateObjectCoordinatesToIndex"
	);
	if(v == NULL)
	    return(FALSE);

	o = MPViewGetObject(
	    v,
	    obj_num
	);
	if(o == NULL)
	    return(FALSE);

	o_geo = &o->geometry;

	ms = MPVObjectGetString(
	    o,
	    string_num
	);
	if(ms == NULL)
	    return(FALSE);

	ms_geo = &ms->geometry;

	adj = v->hadjustment;
	dx = x + (gint)adj->value;

	adj = v->vadjustment;
	dy = y + (gint)adj->value;

	o_dx = dx - o_geo->x;
	o_dy = dy - o_geo->y;

	cursor_pos = MPVStringTranslateCoordinatesToIndex(
	    v,
	    ms->ca_string,
	    o_dx - ms_geo->x,
	    o_dy - ms_geo->y
	);
	if(cursor_pos_rtn != NULL)
	    *cursor_pos_rtn = cursor_pos;

	return(TRUE);
}

/*
 *	Converts the coordinate position relative to the MPView's
 *	GtkDrawingArea to the MPVObject, MPVString, and MPVString's
 *	cursor position that it is over.
 *
 *	The x and y specifies the coordinate position relative to the
 *	MPView's GtkDrawingArea.
 *
 *	The obj_num_rtn, string_num_rtn, and cursor_pos_rtn specifies
 *	the return values for the MPVObject, MPVString, and cursor
 *	position.
 *
 *	Returns TRUE if the specified coordinates are over a
 *	MPVString and a valid cursor position was obtained.
 */
gboolean MPViewTranslateCoordinatesToIndex(
	GtkWidget *w,
	const gint x, const gint y,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn
)
{
	gint		dx, dy;
	GtkAdjustment *adj;
	MPView *v;

	if(obj_num_rtn != NULL)
	    *obj_num_rtn = -1;
	if(string_num_rtn != NULL)
	    *string_num_rtn = -1;
	if(cursor_pos_rtn != NULL)
	    *cursor_pos_rtn = -1;

	v = MPViewGetWidgetData(
	    w,
	    "MPViewTranslateCoordinatesToIndex"
	);
	if(v == NULL)
	    return(FALSE);

	adj = v->hadjustment;
	dx = x + (gint)adj->value;

	adj = v->vadjustment;
	dy = y + (gint)adj->value;

	if(v->objects_list != NULL)
	{
	    gint obj_num;
	    GList *glist;
	    MPGeometry *o_geo;
	    MPVObject *o;

	    for(glist = v->objects_list, obj_num = 0;
		glist != NULL;
		glist = g_list_next(glist), obj_num++
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		o_geo = &o->geometry;
		if((dx >= o_geo->x) && (dx < (o_geo->x + o_geo->width)) &&
		   (dy >= o_geo->y) && (dy < (o_geo->y + o_geo->height))
		)
		{
		    const gint	o_dx = dx - o_geo->x,
				o_dy = dy - o_geo->y;
		    gint string_num;
		    GList *glist;
		    MPGeometry *ms_geo;
		    MPVString *ms;

		    for(glist = o->strings_list, string_num = 0;
			glist != NULL;
			glist = g_list_next(glist), string_num++
		    )
		    {
			ms = MPV_STRING(glist->data);
			if(ms == NULL)
			    continue;

			ms_geo = &ms->geometry;
			if((o_dx >= ms_geo->x) && (o_dx < (ms_geo->x + ms_geo->width)) &&
			   (o_dy >= ms_geo->y) && (o_dy < (ms_geo->y + ms_geo->height))
			)
			{
			    const gint cursor_pos = MPVStringTranslateCoordinatesToIndex(
				v,
				ms->ca_string,
				o_dx - ms_geo->x,
				o_dy - ms_geo->y
			    );
			    if(obj_num_rtn != NULL)
				*obj_num_rtn = obj_num;
			    if(string_num_rtn != NULL)
				*string_num_rtn = string_num;
			    if(cursor_pos_rtn != NULL)
				*cursor_pos_rtn = cursor_pos;
			    return(TRUE);
			}
		    }
		}
	    }
	}

	return(FALSE);
}

/*
 *	Converts the coordinate position relative to the MPView's
 *	GtkDrawingArea to the MPVObject, MPVString, and MPVString's
 *	line number that it is over.
 *
 *	The y specifies the coordinate position relative to the
 *	MPView's Line Bar GtkDrawingArea.
 *
 *	The obj_num_rtn, string_num_rtn, and line_num_rtn specifies
 *	the return values for the MPVObject, MPVString, and line
 *	number.
 *
 *	Returns TRUE if the specified coordinate is over a
 *	MPVString and a valid line number was obtained.
 */
gboolean MPViewTranslateCoordinatesToLine(
	GtkWidget *w,
	const gint y,
	const gint start_obj_num,
	const gint start_string_num,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *line_num_rtn
)
{
	gint		_start_string_num = start_string_num,
			dy;
	GtkAdjustment *adj;
	MPView *v;

	if(obj_num_rtn != NULL)
	    *obj_num_rtn = -1;
	if(string_num_rtn != NULL)
	    *string_num_rtn = -1;
	if(line_num_rtn != NULL)
	    *line_num_rtn = -1;

	v = MPViewGetWidgetData(
	    w,
	    "MPViewTranslateCoordinatesToLine"
	);
	if(v == NULL)
	    return(FALSE);

	adj = v->vadjustment;
	dy = y + (gint)adj->value;

	if(v->objects_list != NULL)
	{
	    gint obj_num;
	    GList *glist;
	    MPGeometry *o_geo;
	    MPVObject *o;

	    for(glist = g_list_nth(v->objects_list, start_obj_num), obj_num = start_obj_num;
		glist != NULL;
		glist = g_list_next(glist), obj_num++
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		o_geo = &o->geometry;
		if((dy >= o_geo->y) && (dy < (o_geo->y + o_geo->height)))
		{
		    const gint o_dy = dy - o_geo->y;
		    gint string_num;
		    GList *glist;
		    MPGeometry *ms_geo;
		    MPVString *ms;

		    for(glist = g_list_nth(o->strings_list, _start_string_num), string_num = _start_string_num;
			glist != NULL;
			glist = g_list_next(glist), string_num++
		    )
		    {
			ms = MPV_STRING(glist->data);
			if(ms == NULL)
			    continue;

			ms_geo = &ms->geometry;
			if((o_dy >= ms_geo->y) && (o_dy < (ms_geo->y + ms_geo->height)))
			{
			    /* Find the line that the y coordinate lies on */
			    gint line_num;
			    (void)MPVStringTranslateCoordinateToLine(
				v,
				ms->ca_string,
				o_dy - ms_geo->y,
				&line_num
			    );
			    if(obj_num_rtn != NULL)
				*obj_num_rtn = obj_num;
			    if(string_num_rtn != NULL)
				*string_num_rtn = string_num;
			    if(line_num_rtn != NULL)
				*line_num_rtn = line_num;
			    return(TRUE);
			}
		    }
		}

		/* Start the search on the first string for
		 * subsequent MPVObjects
		 */
		_start_string_num = 0;
	    }
	}

	return(FALSE);
}


/*
 *	Gets the appropriate GdkFont based on the MPVCharacterAttribute.
 */
static GdkFont *MPViewGetFontFromAttribute(
	MPView *v,
	const CAStringAttributes attr
)
{
	GdkFont *font;

	if(attr & CA_STRING_ATTRIBUTE_BOLD)
	    font = v->bold_font;
	else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
	    font = v->underline_font;
	else if(attr & CA_STRING_ATTRIBUTE_HEADING)
	    font = v->heading_font;
	else
	    font = v->normal_font;

	if(font == NULL)
	    font = v->normal_font;

	return(font);
}

/*
 *	Gets the appropriate GdkColor based on the MPVCharacterAttribute.
 */
static GdkColor *MPViewGetColorFromAttribute(
	MPView *v,
	const CAStringAttributes attr
)
{
	GdkColor *c;

	if(attr & CA_STRING_ATTRIBUTE_SELECTED)
	{
	    GtkStyle *style = gtk_widget_get_style(v->da);
	    c = &style->text[GTK_STATE_SELECTED];
	}
	else if(attr & CA_STRING_ATTRIBUTE_BOLD)
	    c = v->bold_color;
	else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
	    c = v->underline_color;
	else if(attr & CA_STRING_ATTRIBUTE_HEADING)
	    c = v->heading_color;
	else
	    c = v->normal_color;

	if(c == NULL)
	{
	    GtkStyle *style = gtk_widget_get_style(v->da);
	    c = &style->text[GTK_STATE_NORMAL];
	}

	return(c);
}


/*
 *	Gets the MPVObject on the MPView.
 */
MPVObject *MPViewGetObject(
	MPView *v,
	const gint obj_num
)
{
	if((v == NULL) || (obj_num < 0))
	    return(NULL);

	return(MPV_OBJECT(g_list_nth_data(
	    v->objects_list,
	    (guint)obj_num
	)));
}


/*
 *	Sets the focus and cursor position.
 *
 *	The obj_num specifies the MPVObject to set into focus. If
 *	obj_num is -1 then the focus and cursor position will be
 *	cleared.
 *
 *	If string_num is -1 then the last MPVString on the MPVObject
 *	will be set into focus.
 *
 *	If cursor_pos is -1 then the cursor position will be set to the
 *	end of the MPVString's CAString.
 */
void MPViewSetCursorPosition(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint cursor_pos
)
{
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewSetCursorPosition"
	);
	if(v == NULL)
	    return;

	/* Reset the focus_string_num on all the MPVObjects and the
	 * cursor_pos on all their MPVStrings
	 */
	if(v->objects_list != NULL)
	{
	    GList *glist;
	    MPVObject *o;

	    for(glist = v->objects_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		if(o->strings_list != NULL)
		{
		    GList *glist;
		    MPVString *ms;

		    for(glist = o->strings_list;
			glist != NULL;
			glist = g_list_next(glist)
		    )
		    {
			ms = MPV_STRING(glist->data);
			if(ms == NULL)
			    continue;

			ms->cursor_pos = -1;
		    }
		}

		o->focus_string_num = -1;
	    }
	}

	/* Clear focus? */
	if(obj_num < 0)
	{
	    v->focus_obj_num = -1;
	    v->focus_string_num = -1;
	    memset(&v->last_cursor, 0x00, sizeof(MPGeometry));
	    MPViewQueueDraw(w);
	    return;
	}

	o = MPViewGetObject(
	    v,
	    (guint)obj_num
	);
	if(o == NULL)
	    return;

	v->focus_obj_num = obj_num;
	
	if(string_num > -1)
	{
	    ms = MPVObjectGetString(
		o,
		string_num
	    );
	    v->focus_string_num = string_num;
	}
	else
	{
	    GList *glist = g_list_last(o->strings_list);
	    ms = (glist != NULL) ? MPV_STRING(glist->data) : NULL;
	    v->focus_string_num = g_list_length(o->strings_list) - 1;
	}
	if(ms == NULL)
	    return;

	/* Clear the last_cursor MPGeometry, it will be set when
	 * MPViewDraw() is called
	 */
	memset(&v->last_cursor, 0x00, sizeof(MPGeometry));

	/* Set the focus string number on the MPVObject */
	o->focus_string_num = string_num;

	/* Set the MPVString's cursor position */
	ms->cursor_pos = (cursor_pos > -1) ? cursor_pos : ms->length;

	MPViewQueueDraw(w);
}

/*
 *	Gets the current cursor position.
 */
void MPViewGetCursorPosition(
	GtkWidget *w,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn,
	MPGeometry **cursor_geo_rtn
)
{
	MPVString *ms;
	MPVObject *o;
	MPView *v;

	if(obj_num_rtn != NULL)
	    *obj_num_rtn = -1;
	if(string_num_rtn != NULL)
	    *string_num_rtn = -1;
	if(cursor_pos_rtn != NULL)
	    *cursor_pos_rtn = -1;
	if(cursor_geo_rtn != NULL)
	    *cursor_geo_rtn = NULL;

	v = MPViewGetWidgetData(
	    w,
	    "MPViewGetCursorPosition"
	);
	if(v == NULL)
	    return;

	o = MPViewGetObject(
	    v,
	    v->focus_obj_num
	);
	if(o == NULL)
	    return;

	ms = MPVObjectGetString(
	    o,
	    v->focus_string_num
	);
	if(ms == NULL)
	    return;

	if(obj_num_rtn != NULL)
	    *obj_num_rtn = v->focus_obj_num;
	if(string_num_rtn != NULL)
	    *string_num_rtn = v->focus_string_num;
	if(cursor_pos_rtn != NULL)
	    *cursor_pos_rtn = ms->cursor_pos;
	if(cursor_geo_rtn != NULL)
	    *cursor_geo_rtn = &v->last_cursor;
}


/*
 *	Selects the region on the MPVString and sets the
 *	MPV_OBJECT_HAS_SELECTION flag on the MPVObject.
 *
 *	If selection_start_pos and selection_end_pos are both -1 then
 *	the MPVString is unselected and the MPV_OBJECT_HAS_SELECTION
 *	flag is removed from the MPVObject.
 */
void MPViewSelect(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint selection_start_pos,
	const gint selection_end_pos
)
{
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewSelect"
	);
	if(v == NULL)
	    return;

	/* Get the specified MPVObject */
	if(obj_num > -1)
	{
	    o = MPViewGetObject(
		v,
		obj_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(v->objects_list);
	    o = (glist != NULL) ? MPV_OBJECT(glist->data) : NULL;
	}
	if(o == NULL)
	    return;
	
	/* Get the specified MPVString on the MPVObject */
	if(string_num > -1)
	{
	    ms = MPVObjectGetString(
		o,
		string_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(o->strings_list);
	    ms = (glist != NULL) ? MPV_STRING(glist->data) : NULL;
	}
	if(ms == NULL)
	    return;

	/* Unselect? */
	if((selection_start_pos < 0) && (selection_end_pos < 0))
	{
	    /* Mark this MPVObject as not selected */
	    o->flags &= ~MPV_OBJECT_HAS_SELECTION;

	    /* Remove the selected attribute from the CAString */
	    MPVStringUnselectAll(ms);
	}
	else
	{
	    gint	_sel_start_pos,
			_sel_end_pos;

	    /* Mark this MPVObject as selected */
	    o->flags |= MPV_OBJECT_HAS_SELECTION;

	    if(selection_start_pos < 0)
		_sel_start_pos = 0;
	    else
		_sel_start_pos = selection_start_pos;
	    if(selection_end_pos < 0)
		_sel_end_pos = ms->length - 1;
	    else
		_sel_end_pos = selection_end_pos;

	    /* Add the selected attribute to the CAString region */
	    MPVStringSelect(
		ms,
		_sel_start_pos,
		_sel_end_pos,
		FALSE				/* Do not clear existing selection */
	    );
	}

	MPViewQueueDraw(w);
}

/*
 *	Selects the entire string on the object.
 */
void MPViewSelectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
)
{
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewSelectObjectStringAll"
	);
	if(v == NULL)
	    return;

	/* Get the specified MPVObject */
	if(obj_num > -1)
	{
	    o = MPViewGetObject(
		v,
		obj_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(v->objects_list);
	    o = (glist != NULL) ? MPV_OBJECT(glist->data) : NULL;
	}
	if(o == NULL)
	    return;

	/* Get the specified MPVString on the MPVObject */
	if(string_num > -1)
	{
	    ms = MPVObjectGetString(
		o,
		string_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(o->strings_list);
	    ms = (glist != NULL) ? MPV_STRING(glist->data) : NULL;
	}
	if(ms == NULL)
	    return;

	/* Mark this MPVObject as selected */
	o->flags |= MPV_OBJECT_HAS_SELECTION;

	/* Add the selected attribute to the CAString region */
	MPVStringSelect(
	    ms,
	    0,
	    ms->length,
	    TRUE				/* Clear existing selection */
	);

	MPViewQueueDraw(w);
}

/*
 *	Selects all the strings on all the objects from the starting
 *	position to the ending position.
 */
void MPViewSelectObjectsRange(
	GtkWidget *w,
	const gint start_obj_num,
	const gint start_string_num,
	const gint start_cursor_pos,
	const gint end_obj_num,
	const gint end_string_num,
	const gint end_cursor_pos
)
{
	const gint	obj_num = end_obj_num,
			string_num = end_string_num,
			cursor_pos = end_cursor_pos;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewSelectObjectsRange"
	);
	if(v == NULL)
	    return;


	if(TRUE)
	{
	    /* Over the same MPVObject and MPVString? */
	    if((obj_num == start_obj_num) &&
	       (string_num == start_string_num)
	    )
	    {
		/* Update the selection */
		if(cursor_pos >= start_cursor_pos)
		    MPViewSelect(
			v->toplevel,
			obj_num,
			string_num,
			start_cursor_pos,
			cursor_pos
		    );
		else
		    MPViewSelect(
			v->toplevel,
			obj_num,
			string_num,
			cursor_pos,
			start_cursor_pos
		    );
	    }
	    /* Same MPVObject but different MPVString? */
	    else if(obj_num == start_obj_num)
	    {
		MPVObject *o = MPViewGetObject(v, obj_num);
		MPVString *orig_ms = MPVObjectGetString(o, start_string_num);
		if((o != NULL) && (orig_ms != NULL))
		{
		    /* "Upper" MPVString? */
		    if(string_num < start_string_num)
		    {
			gint string_num_i;
			MPVString *ms;

			/* Complete the selection on the previous MPVString */
			MPViewSelect(
			    v->toplevel,
			    obj_num,
			    start_string_num,
			    0,
			    start_cursor_pos
			);

			/* Select all the MPVStrings in between */
			for(string_num_i = start_string_num - 1;
			    string_num_i > string_num;
			    string_num_i--
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				obj_num,
				string_num_i
			    );

			/* Start the selection on the new MPVString */
			ms = MPVObjectGetString(o, string_num);
			if(ms != NULL)
			{
			    MPViewSelect(
				v->toplevel,
				obj_num,
				string_num,
				cursor_pos,
				ms->length
			    );
			}
		    }
		    /* "Lower" MPVString */
		    else
		    {
			gint string_num_i;
			MPVString *ms;

			/* Complete the selection on the previous MPVString */
			MPViewSelect(
			    v->toplevel,
			    obj_num,
			    start_string_num,
			    start_cursor_pos,
			    orig_ms->length
			);

			/* Select all the MPVStrings in between */
			for(string_num_i = start_string_num + 1;
			    string_num_i < string_num;
			    string_num_i++
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				obj_num,
				string_num_i
			    );

			/* Start the selection on the new MPVString */
			ms = MPVObjectGetString(o, string_num);
			if(ms != NULL)
			{
			    MPViewSelect(
				v->toplevel,
				obj_num,
				string_num,
				0, cursor_pos
			    );
			}
		    }
		}
	    }
	    /* Different MPVObject and different MPVString */
	    else
	    {
		MPVObject	*orig_o = MPViewGetObject(v, start_obj_num),
				*o = MPViewGetObject(v, obj_num);
		MPVString	*orig_ms = MPVObjectGetString(orig_o, start_string_num),
				*ms = MPVObjectGetString(o, string_num);

		if((orig_o != NULL) && (o != NULL) &&
		   (orig_ms != NULL) && (ms != NULL)
		)
		{
		    /* "Upper" MPVObject? */
		    if(obj_num < start_obj_num)
		    {
			gint	nstrings,
				obj_num_i,
				string_num_i;

			/* Complete the selection on the original MPVString */
			MPViewSelect(
			    v->toplevel,
			    start_obj_num,
			    start_string_num,
			    0,
			    start_cursor_pos
			);

			for(string_num_i = start_string_num - 1;
			    string_num_i >= 0;
			    string_num_i--
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				start_obj_num,
				string_num_i
			    );

			/* Select the objects and strings in between */
			for(obj_num_i = start_obj_num - 1;
			    obj_num_i > obj_num;
			    obj_num_i--
			)
			    MPViewSelectObjectAll(
				v->toplevel,
				obj_num_i
			    );

			/* Start the selection on the new MPVString */
			nstrings = g_list_length(o->strings_list);
			for(string_num_i = nstrings - 1;
			    string_num_i > string_num;
			    string_num_i--
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				obj_num,
				string_num_i
			    );

			MPViewSelect(
			    v->toplevel,
			    obj_num,
			    string_num,
			    cursor_pos,
			    ms->length
			);
		    }
		    /* "Lower" MPVObject */
		    else
		    {
			gint	nstrings,
				obj_num_i,
				string_num_i;

			/* Complete the selection on the previous MPVString */
			MPViewSelect(
			    v->toplevel,
			    start_obj_num,
			    start_string_num,
			    start_cursor_pos,
			    orig_ms->length
			);

			nstrings = g_list_length(orig_o->strings_list);
			for(string_num_i = start_string_num + 1;
			    string_num_i < nstrings;
			    string_num_i++
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				start_obj_num,
				string_num_i
			    );

			/* Select the objects and strings in between */
			for(obj_num_i = start_obj_num + 1;
			    obj_num_i < obj_num;
			    obj_num_i++
			)
			    MPViewSelectObjectAll(
				v->toplevel,
				obj_num_i
			    );

			/* Start the selection on the new MPVString */
			for(string_num_i = 0;
			    string_num_i < string_num;
			    string_num_i++
			)
			    MPViewSelectObjectStringAll(
				v->toplevel,
				obj_num,
				string_num_i
			    );

			MPViewSelect(
			    v->toplevel,
			    obj_num,
			    string_num,
			    0,
			    cursor_pos
			);
		    }
		}
	    }
	}
}

/*
 *	Selects all the strings on the object.
 */
void MPViewSelectObjectAll(
	GtkWidget *w,
	const gint obj_num
)
{
	GList *glist;
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewSelectObjectAll"
	);
	if(v == NULL)
	    return;

	/* Get the specified MPVObject */
	if(obj_num > -1)
	{
	    o = MPViewGetObject(
		v,
		obj_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(v->objects_list);
	    o = (glist != NULL) ? MPV_OBJECT(glist->data) : NULL;
	}
	if(o == NULL)
	    return;

	/* Select all the MPVStrings on this MPVObject */
	for(glist = o->strings_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    ms = MPV_STRING(glist->data);
	    if(ms == NULL)
		continue;

	    /* Mark this MPVObject as selected */
	    o->flags |= MPV_OBJECT_HAS_SELECTION;

	    /* Add the selected attribute to the CAString region */
	    MPVStringSelect(
		ms,
		0,
		ms->length,
		TRUE				/* Clear existing selection */
	    );
	}

	MPViewQueueDraw(w);
}


/*
 *	Unselect all.
 */
void MPViewUnselectAll(GtkWidget *w)
{
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewUnselectAll"
	);
	if(v == NULL)
	    return;

	if(v->objects_list != NULL)
	{
	    GList *glist;
	    MPVObject *o;

	    for(glist = v->objects_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		if(o->strings_list != NULL)
		{
		    GList *glist;
		    MPVString *ms;

		    for(glist = o->strings_list;
			glist != NULL;
			glist = g_list_next(glist)
		    )
		    {
			ms = MPV_STRING(glist->data);
			if(ms == NULL)
			    continue;

			MPVStringUnselectAll(ms);
		    }
		}

		o->flags &= ~MPV_OBJECT_HAS_SELECTION;
	    }
	}

	MPViewQueueDraw(w);
}

/*
 *	Unselects the entire string on the object.
 */
void MPViewUnselectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
)
{
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewUnselectObjectStringAll"
	);
	if(v == NULL)
	    return;

	/* Get the specified MPVObject */
	if(obj_num > -1)
	{
	    o = MPViewGetObject(
		v,
		obj_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(v->objects_list);
	    o = (glist != NULL) ? MPV_OBJECT(glist->data) : NULL;
	}
	if(o == NULL)
	    return;

	/* Get the specified MPVString on the MPVObject */
	if(string_num > -1)
	{
	    ms = MPVObjectGetString(
		o,
		string_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(o->strings_list);
	    ms = (glist != NULL) ? MPV_STRING(glist->data) : NULL;
	}
	if(ms == NULL)
	    return;

	/* Mark this MPVObject as not selected */
	o->flags &= ~MPV_OBJECT_HAS_SELECTION;

	/* Remove the selected attribute from the entire CAString */
	MPVStringUnselectAll(ms);

	MPViewQueueDraw(w);
}

/*
 *	Unselects all the strings on the object.
 */
void MPViewUnselectObjectAll(
	GtkWidget *w,
	const gint obj_num
)
{
	GList *glist;
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewUnselectObjectAll"
	);
	if(v == NULL)
	    return;

	/* Get the specified MPVObject */
	if(obj_num > -1)
	{
	    o = MPViewGetObject(
		v,
		obj_num
	    );
	}
	else
	{
	    GList *glist = g_list_last(v->objects_list);
	    o = (glist != NULL) ? MPV_OBJECT(glist->data) : NULL;
	}
	if(o == NULL)
	    return;

	/* Select all the MPVStrings on this MPVObject */
	for(glist = o->strings_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    ms = MPV_STRING(glist->data);
	    if(ms == NULL)
		continue;

	    /* Mark this MPVObject as not selected */
	    o->flags &= ~MPV_OBJECT_HAS_SELECTION;

	    /* Remove the selected attribute from the entire CAString */
	    MPVStringUnselectAll(ms);
	}

	MPViewQueueDraw(w);
}


/*
 *	Removes all the selected characters.
 */
void MPViewRemoveSelectedCharacters(GtkWidget *w)
{
	GList *glist;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewRemoveSelectedCharacters"
	);
	if(v == NULL)
	    return;

	for(glist = v->objects_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    o = MPV_OBJECT(glist->data);
	    if(o == NULL)
		continue;

	    MPVObjectRemoveSelectedCharacters(o);
	}

	MPViewUnselectAll(w);
	MPViewQueueResize(w);
	MPViewQueueDraw(w);
}


/*
 *	Gets the MPVObject's visibility.
 */
GdkVisibilityState MPViewGetObjectVisibility(
	GtkWidget *w,
	const gint obj_num
)
{
	gint y;
	GtkAdjustment *adj;
	MPGeometry *o_geo;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewGetObjectVisibility"
	);
	if(w == NULL)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	o = MPViewGetObject(
	    v,
	    obj_num
	);
	if(o == NULL)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	adj = v->vadjustment;
	o_geo = &o->geometry;
	y = o_geo->y - (gint)adj->value;

	if(((y + o_geo->height) <= 0) || (y >= (gint)adj->page_size))
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	if((y >= 0) && ((y + o_geo->height) <= (gint)adj->page_size))
	    return(GDK_VISIBILITY_UNOBSCURED);

	return(GDK_VISIBILITY_PARTIAL);
}

/*
 *	Gets the current cursor's visibility.
 */
GdkVisibilityState MPViewGetCursorVisibilityX(GtkWidget *w)
{
	gint x;
	GtkAdjustment *adj;
	MPGeometry *cur_geo;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewGetCursorVisibilityX"
	);
	if(w == NULL)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	adj = v->hadjustment;
	cur_geo = &v->last_cursor;
	x = cur_geo->x;

	if(((x + cur_geo->width) <= 0) || (x >= (gint)adj->page_size))
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	if((x >= 0) && ((x + cur_geo->width) <= (gint)adj->page_size))
	    return(GDK_VISIBILITY_UNOBSCURED);

	return(GDK_VISIBILITY_PARTIAL);
}

GdkVisibilityState MPViewGetCursorVisibilityY(GtkWidget *w)
{
	gint y;
	GtkAdjustment *adj;
	MPGeometry *cur_geo;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewGetCursorVisibilityY"
	);
	if(w == NULL)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	adj = v->vadjustment;
	cur_geo = &v->last_cursor;
	y = cur_geo->y;

	if(((y + cur_geo->height) <= 0) || (y >= (gint)adj->page_size))
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	if((y >= 0) && ((y + cur_geo->height) <= (gint)adj->page_size))
	    return(GDK_VISIBILITY_UNOBSCURED);

	return(GDK_VISIBILITY_PARTIAL);
}

GdkVisibilityState MPViewGetCursorVisibility(GtkWidget *w)
{
	gint		x, y;
	GtkAdjustment	*hadj,
			*vadj;
	MPGeometry *cur_geo;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewGetCursorVisibility"
	);
	if(w == NULL)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	hadj = v->hadjustment;
	vadj = v->vadjustment;
	cur_geo = &v->last_cursor;
	x = cur_geo->x;
	y = cur_geo->y;

	if(((x + cur_geo->width) <= 0) || (x >= (gint)hadj->page_size) ||
	   ((y + cur_geo->height) <= 0) || (y >= (gint)vadj->page_size)
	)
	    return(GDK_VISIBILITY_FULLY_OBSCURED);

	if((x >= 0) && ((x + cur_geo->width) <= (gint)hadj->page_size) &&
	   (y >= 0) && ((y + cur_geo->height) <= (gint)vadj->page_size)
	)
	    return(GDK_VISIBILITY_UNOBSCURED);

	return(GDK_VISIBILITY_PARTIAL);
}


/*
 *	Scrolls to the MPVObject.
 */
void MPViewMoveToObject(
	GtkWidget *w,
	const gint obj_num,
	const gfloat x_coeff, const gfloat y_coeff
)
{
	GtkAdjustment *adj;
	MPGeometry *o_geo;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewMoveToObject"
	);
	if(w == NULL)
	    return;

	o = MPViewGetObject(
	    v,
	    obj_num
	);
	if(o == NULL)
	    return;

	o_geo = &o->geometry;

	adj = v->hadjustment;
	GTK_ADJUSTMENT_SET_VALUE(
	    adj,
	    (gfloat)(
		o_geo->x - (x_coeff * ((gint)adj->page_size - o_geo->width))
	    )
	);
	adj = v->vadjustment;
	GTK_ADJUSTMENT_SET_VALUE(
	    adj,
	    (gfloat)(
		o_geo->y - (y_coeff * ((gint)adj->page_size - o_geo->height))
	    )
	);
}

/*
 *	Scrolls to the current cursor's X coordinate.
 */
void MPViewMoveToCursorX(
	GtkWidget *w,
	const gfloat x_coeff
)
{
	gint x;
	GtkAdjustment *adj;
	MPGeometry *cur_geo;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewMoveToCursorX"
	);
	if(w == NULL)
	    return;

	adj = v->hadjustment;
	cur_geo = &v->last_cursor;
	x = cur_geo->x + (gint)adj->value;

	GTK_ADJUSTMENT_SET_VALUE(
	    adj,
	    (gfloat)(
		x - (x_coeff * ((gint)adj->page_size - cur_geo->width))
	    )
	);
}

/*
 *	Scrolls to the current cursor's Y coordinate.
 */
void MPViewMoveToCursorY(
	GtkWidget *w,
	const gfloat y_coeff
)
{
	gint y;
	GtkAdjustment *adj;
	MPGeometry *cur_geo;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewMoveToCursorY"
	);
	if(w == NULL)
	    return;

	adj = v->vadjustment;
	cur_geo = &v->last_cursor;
	y = cur_geo->y + (gint)adj->value;

	GTK_ADJUSTMENT_SET_VALUE(
	    adj,
	    (gfloat)(
		y - (y_coeff * ((gint)adj->page_size - cur_geo->height))
	    )
	);
}

void MPViewMoveToCursor(
	GtkWidget *w,
	const gfloat x_coeff, const gfloat y_coeff
)
{
	MPViewMoveToCursorX(w, x_coeff);
	MPViewMoveToCursorY(w, y_coeff);
}


/*
 *	Cut string.
 */
void MPViewCutString(GtkWidget *w)
{

}

/*
 *	Copy string.
 */
void MPViewCopyString(GtkWidget *w)
{
	GList *glist;
	CAString *coppied_ca_string;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewCopyString"
	);
	if(v == NULL)
	    return;

	coppied_ca_string = NULL;

	/* Concat all the selected CAStrings on all the MPVObjects to
	 * the coppied CAString
	 */
	for(glist = v->objects_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    o = MPV_OBJECT(glist->data);
	    if(o == NULL)
		continue;

	    if(o->strings_list != NULL)
	    {
		GList *glist;
		CAString *ca_string_ptr;
		MPVString *ms;

		for(glist = o->strings_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    ms = MPV_STRING(glist->data);
		    if(ms == NULL)
			continue;

		    ca_string_ptr = ms->ca_string;
		    if(ca_string_ptr == NULL)
			continue;

		    /* Append the selected regions of this CAString to
		     * the coppied CAString
		     */
		    while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
		    {
			/* Is this character selected? */
			if(CA_STRING_ATTRIBUTE(ca_string_ptr) & CA_STRING_ATTRIBUTE_SELECTED)
			{
			    gint	len,
					sel_len;

			    /* Seek to the end of the selected region */
			    CAString *ca_string_next = ca_string_ptr + 1;
			    while(CA_STRING_CHARACTER(ca_string_next) != '\0')
			    {
				if(!(CA_STRING_ATTRIBUTE(ca_string_next) & CA_STRING_ATTRIBUTE_SELECTED))
				    break;

				ca_string_next++;
			    }

			    /* Calculate the length of the selected
			     * region
			     */
			    sel_len = ca_string_next - ca_string_ptr;

			    /* Increase the allocation of the coppied
			     * CAString and append the selected region
			     * to it
			     */
			    len = CAStringLength(coppied_ca_string);
			    coppied_ca_string = (CAString *)g_realloc(
				coppied_ca_string,
				(len + sel_len + 1) * sizeof(CAString)
			    );
			    if(coppied_ca_string != NULL)
			    {
				memcpy(
				    coppied_ca_string + len,
				    ca_string_ptr,
				    sel_len * sizeof(CAString)
				);
				memset(
				    coppied_ca_string + len + sel_len,
				    0x00,
				    sizeof(CAString)
				);
			    }

			    ca_string_ptr = ca_string_next;
			}
			else
			{
			    ca_string_ptr++;
			}
		    }
		}
	    }
	}

	if(coppied_ca_string != NULL)
	{
	    /* Remove the selected attribute from the coppied CAString */
	    CAStringUnselectAll(coppied_ca_string);

	    /* Change all line end characters to spaces */
	    CAStringClearLineEnds(coppied_ca_string);

	    /* Transfer the coppied CAString to the selection */
	    GUIDDESetDirect(
		v->da,
		GDK_SELECTION_PRIMARY,
		GDK_CURRENT_TIME,
		v->ca_string_target_type,	/* As a CAString */
		(guint8 *)coppied_ca_string,
		(CAStringLength(coppied_ca_string) + 1) * sizeof(CAString)
	    );
/*	    coppied_ca_string = NULL; */
	}
}

void MPViewCutObjects(GtkWidget *w)
{

}

/*
 *	Copy objects.
 */
void MPViewCopyObjects(GtkWidget *w)
{
	gchar		*s,
			*coppied_xml_string;
	GList *glist;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewCopyObjects"
	);
	if(v == NULL)
	    return;

	coppied_xml_string = NULL;

	/* Concat all the selected MPVObjects to the coppied XML string */
	for(glist = v->objects_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    o = MPV_OBJECT(glist->data);
	    if(o == NULL)
		continue;

	    /* Does this object have any selections? */
	    if(!(o->flags & MPV_OBJECT_HAS_SELECTION))
		continue;

	    s = MPViewGetObjectXMLString(
		v->toplevel,
		o
	    );
	    if(s != NULL)
	    {
		if(coppied_xml_string != NULL)
		{
		    gchar *s2 = g_strconcat(
			coppied_xml_string,
			s,
			NULL
		    );
		    g_free(s);
		    if(s2 != NULL)
		    {
			g_free(coppied_xml_string);
			coppied_xml_string = s2;
		    }
		}
		else
		{
		    coppied_xml_string = s;
		}
	    }
	}

	if(coppied_xml_string != NULL)
	{
	    /* Transfer the coppied XML string to the selection */
	    GUIDDESetDirect(
		v->da,
		GDK_SELECTION_PRIMARY,
		GDK_CURRENT_TIME,
		v->mpv_object_target_type,	/* MPVObject */
		(guint8 *)coppied_xml_string,
		STRLEN(coppied_xml_string) + 1
	    );
/*	    coppied_xml_string = NULL; */
	}
}

/*
 *	Paste the selection.
 */
void MPViewPaste(GtkWidget *w)
{
	gint len;
	gchar *s;
	CAString *ca_string;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewPaste"
	);
	if(v == NULL)
	    return;

	/* Get the current primary selection based on type
	 *
	 * First try MPVObject
	 */
	s = (gchar *)GUIDDEGet(
	    v->da,
	    GDK_SELECTION_PRIMARY,
	    GDK_CURRENT_TIME,
	    v->mpv_object_target_type,		/* MPVObject */
	    &len
	);
	if(s != NULL)
	{
	    MPViewInsertObjectsFromXMLString(
		v->toplevel,
		v->focus_obj_num,
		s,
		NULL, NULL
	    );
	    g_free(s);
	    MPViewQueueResize(v->toplevel);
	    MPViewQueueDraw(v->toplevel);
	    return;
	}

	/* CAString */
	ca_string = (CAString *)GUIDDEGet(
	    v->da,
	    GDK_SELECTION_PRIMARY,
	    GDK_CURRENT_TIME,
	    v->ca_string_target_type,		/* As a CAString */
	    &len
	);
	if(ca_string != NULL)
	{
	    MPVObject *o = MPViewGetObject(
		v,
		v->focus_obj_num
	    );
	    if(o != NULL)
	    {
		MPVString *ms = MPVObjectGetString(
		    o,
		    o->focus_string_num
		);
		if(ms != NULL)
		{
		    MPVStringInsertText(
			ms,
			ms->cursor_pos,
			ca_string
		    );
		}
	    }
	    CAStringDelete(ca_string);
	    MPViewQueueResize(v->toplevel);
	    MPViewQueueDraw(v->toplevel);
	    return;
	}

	/* Regular string */
	s = GUIDDEGetString(
	    v->da,
	    GDK_SELECTION_PRIMARY,
	    GDK_CURRENT_TIME
	);
	if(s != NULL)
	{
	    MPVObject *o = MPViewGetObject(
		v,
		v->focus_obj_num
	    );
	    if(o != NULL)
	    {
		MPVString *ms = MPVObjectGetString(
		    o,
		    o->focus_string_num
		);
		if(ms != NULL)
		{
		    /* Convert to a CAString with no attributes */
		    CAString *ca_string = CAStringNewFromString(
			s,
			-1
		    );
		    MPVStringInsertText(
			ms,
			ms->cursor_pos,
			ca_string
		    );
		    CAStringDelete(ca_string);
		}
	    }
	    g_free(s);
	    MPViewQueueResize(v->toplevel);
	    MPViewQueueDraw(v->toplevel);
	    return;
	}

}

/*
 *	Deletes all the selected text regions, if nothing is selected
 *	then one character at the current cursor position is deleted.
 *
 *	If any object's string becomes empty then that object itself
 *	will be deleted.
 */
void MPViewDelete(GtkWidget *w)
{
	gboolean object_deleted;
	GList *glist;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewDelete"
	);
	if(v == NULL)
	    return;

	/* Check if nothing is selected */
	for(glist = v->objects_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    o = MPV_OBJECT(glist->data);
	    if(o == NULL)
		continue;

	    if(o->flags & MPV_OBJECT_HAS_SELECTION)
		break;
	}
	if(glist == NULL)
	{
	    /* Nothing selected so delete one character */
	    MPVObject *o = MPViewGetObject(
		v,
		v->focus_obj_num
	    );
	    if(o != NULL)
	    {
		GList *glist;
		MPVString *ms;

		/* Check if all the strings on this MPVObject are empty */
		for(glist = o->strings_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    if(!MPVStringIsEmpty(MPV_STRING(glist->data)))
			break;
		}
		if(glist == NULL)
		{
		    /* All the strings are empty, delete this MPVObject */
		    v->objects_list = g_list_remove(
			v->objects_list,
			o
		    );
		    MPVObjectDelete(o);

		    /* Set the next object into focus */
		    o = MPViewGetObject(
			v,
			v->focus_obj_num
		    );
		    if(o != NULL)
		    {
			v->focus_string_num = 0;
			o->focus_string_num = 0;
			ms = MPVObjectGetString(
			    o,
			    o->focus_string_num
			);
			if(ms != NULL)
			    ms->cursor_pos = 0;
		    }
		    else
		    {
			v->focus_obj_num = -1;
			v->focus_string_num = -1;
		    }

		    /* Report changed */
		    if(v->changed_cb != NULL)
			v->changed_cb(
			    v,
			    v->changed_data
			);
		}
		else
		{
		    /* Strings not empty, delete one character */
		    MPVString *ms = MPVObjectGetString(
			o,
			o->focus_string_num
		    );
		    if(ms != NULL)
		    {
			if(ms->cursor_pos > -1)
			{
			    /* Delete the next character */
			    MPVStringRemoveCharacter(
				ms,
				ms->cursor_pos
			    );

			    /* Report changed */
			    if(v->changed_cb != NULL)
				v->changed_cb(
				    v,
				    v->changed_data
				);
			}
		    }
		}
	    }
	    MPViewQueueResize(v->toplevel);
	    MPViewQueueDraw(v->toplevel);
	    return;
	}

	/* Delete all the selected string segments */
	object_deleted = FALSE;
	glist = v->objects_list;
	while(glist != NULL)
	{
	    o = MPV_OBJECT(glist->data);
	    if(o == NULL)
	    {
		glist = g_list_next(glist);
		object_deleted = FALSE;
		continue;
	    }

	    if(o->strings_list != NULL)
	    {
		GList *glist;
		CAString *ca_string_ptr;
		MPVString *ms;

		for(glist = o->strings_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    ms = MPV_STRING(glist->data);
		    if(ms == NULL)
			continue;

		    ca_string_ptr = ms->ca_string;
		    if(ca_string_ptr == NULL)
			continue;

		    /* Delete the selected regions of this CAString */
		    while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
		    {
			/* Is this character selected? */
			if(CA_STRING_ATTRIBUTE(ca_string_ptr) & CA_STRING_ATTRIBUTE_SELECTED)
			{
			    /* Seek to the end of the selected region */
			    CAString *ca_string_next = ca_string_ptr + 1;
			    while(CA_STRING_CHARACTER(ca_string_next) != '\0')
			    {
				if(!(CA_STRING_ATTRIBUTE(ca_string_next) & CA_STRING_ATTRIBUTE_SELECTED))
				    break;

				ca_string_next++;
			    }

			    /* Delete this segment */
			    MPVStringRemoveText(
				ms,
				ca_string_ptr - ms->ca_string,	/* Index */
				ca_string_next - ca_string_ptr	/* Length */
			    );

			    /* start from the beginning again */
			    ca_string_ptr = ms->ca_string;
			    if(ca_string_ptr == NULL)
				break;
			}
			else
			{
			    ca_string_ptr++;
			}
		    }
		}

		/* If all the MPVStrings are empty then delete this
		 * MPVObject
		 */
		for(glist = o->strings_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    ms = MPV_STRING(glist->data);
		    if(!MPVStringIsEmpty(ms))
			break;
		}
		if(glist == NULL)
		{
		    v->objects_list = g_list_remove(
			v->objects_list,
			o
		    );
		    MPVObjectDelete(o);
		    object_deleted = TRUE;
		}
	    }
	    /* If this object was removed from the list then start
	     * from the beginning again
	     */
	    if(object_deleted)
		glist = v->objects_list;
	    else
		glist = g_list_next(glist);
	    object_deleted = FALSE;
	}

	MPViewQueueResize(w);
	MPViewQueueDraw(w);

	/* Report changed */
	if(v->changed_cb != NULL)
	    v->changed_cb(
		v,
		v->changed_data
	    );
}


/*
 *      Gets the MPView data from the GtkWidget.
 */
MPView *MPViewGetWidgetData(
	GtkWidget *w,
	const gchar *func_name
)
{
	const gchar *key = MP_VIEW_KEY;
	MPView *v;

	if(w == NULL)
	    return(NULL);

	v = MP_VIEW(gtk_object_get_data(
	    GTK_OBJECT(w),
	    key
	));
	if(v == NULL)
	{
	    g_printerr(
"%s(): Warning: GtkWidget %p:\
 Unable to find the data that matches the key \"%s\".\n",
		func_name,
		w,
		key
	    );
	    return(NULL);
	}

	return(v);
}


/*
 *	Creates a new MPView.
 */
GtkWidget *MPViewNew(void)
{
	const gint border_minor = 3;
	GtkAdjustment *adj;
	GtkWidget	*w,
			*parent, *parent2,
			*toplevel;

	/* Create a new MPView */
	MPView *v = MP_VIEW(g_malloc0(sizeof(MPView)));
	if(v == NULL)
	    return(NULL);

	v->toplevel = toplevel = gtk_table_new(2, 3, FALSE);

/*
	v->freeze_count = 0;
	v->colormap = NULL;
	v->gc = NULL;
 */
	v->flags = MP_VIEW_SHOW_OBJECT_BOUNDS |
	    MP_VIEW_SHOW_STRING_MARKERS;

	v->ca_string_target_type = gdk_atom_intern(
	    CA_STRING_TARGET_TYPE,
	    FALSE
	);
	v->mpv_object_target_type = gdk_atom_intern(
	    MPV_OBJECT_TARGET_TYPE,
	    FALSE
	);
/*
	v->normal_font = NULL;
	v->bold_font = NULL;
	v->underline_font = NULL;
	v->heading_font = NULL;

	v->normal_color = NULL;
	v->bold_color = NULL;
	v->underline_color = NULL;
	v->heading_color = NULL;
	v->bg_color = NULL;

	v->line_break_bitmap = NULL;
	v->string_end_bitmap = NULL;
 */

	v->text_cursor = gdk_cursor_new(GDK_XTERM);
	v->line_bar_cursor = gdk_cursor_new(GDK_RIGHT_PTR);

	v->x_margin = MP_VIEW_DEF_X_MARGIN;
	v->y_margin = MP_VIEW_DEF_Y_MARGIN;
	v->x_spacing = MP_VIEW_DEF_X_SPACING;
	v->y_spacing = MP_VIEW_DEF_Y_SPACING;
	v->object_x_margin = MP_VIEW_DEF_X_SPACING;
	v->object_y_margin = MP_VIEW_DEF_Y_SPACING;

/*
	v->pixmap = NULL;
	v->line_bar_pixmap = NULL;

	v->objects_list = NULL;
 */

	v->focus_obj_num = -1;

/*
	v->last_motion_x = 0;
	v->last_motion_y = 0;
 */
	v->last_drag_obj_num = -1;
	v->last_drag_string_num = -1;
	v->last_drag_cursor_pos = -1;


	v->freeze_count++;

	/* Toplevel GtkTable */
	w = toplevel;
	gtk_object_set_data_full(
	    GTK_OBJECT(w), MP_VIEW_KEY,
	    v, MPViewDestroyCB
	);
	gtk_table_set_row_spacing(GTK_TABLE(w), 0, 0);
	gtk_table_set_row_spacing(GTK_TABLE(w), 1, (guint)border_minor);
/*	gtk_table_set_row_spacings(GTK_TABLE(w), (guint)border_minor); */
	gtk_table_set_col_spacings(GTK_TABLE(w), (guint)border_minor);
	parent = w;

	/* Line Bar GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    0, 1,
	    0, 1,
	    0,
	    GTK_SHRINK | GTK_FILL | GTK_EXPAND,
	    0, 0
	);
	gtk_widget_show(w);
	parent2 = w;

	/* Line Bar GtkDrawingArea */
	v->line_bar_da = w = gtk_drawing_area_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_CAN_FOCUS);
	gtk_widget_set_usize(w, 30, -1);
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
	    GDK_FOCUS_CHANGE_MASK |
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "realize",
	    GTK_SIGNAL_FUNC(MPViewLineBarRealizeCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarConfigureEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarExposeEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_in_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarFocusEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_out_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarFocusEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarKeyEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarKeyEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarButtonEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarButtonEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(MPViewLineBarMotionEventCB), v
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);


	/* View GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    1, 2,
	    0, 1,
	    GTK_SHRINK | GTK_FILL | GTK_EXPAND,
	    GTK_SHRINK | GTK_FILL | GTK_EXPAND,
	    0, 0
	);
	gtk_widget_show(w);
	parent2 = w;

	/* View GtkDrawingArea */
	v->da = w = gtk_drawing_area_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_CAN_FOCUS);
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
	    GDK_FOCUS_CHANGE_MASK |
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "realize",
	    GTK_SIGNAL_FUNC(MPViewRealizeCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(MPViewConfigureEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(MPViewExposeEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_in_event",
	    GTK_SIGNAL_FUNC(MPViewFocusEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_out_event",
	    GTK_SIGNAL_FUNC(MPViewFocusEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(MPViewKeyEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(MPViewKeyEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(MPViewButtonEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(MPViewButtonEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(MPViewMotionEventCB), v
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);

	/* Vertical GtkAdjustment and GtkScrollBar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f, 10.0f,
	    10.0f, 10.0f, 10.0f
	);
	v->vadjustment = adj = (GtkAdjustment *)GTK_OBJECT_REF(GTK_OBJECT(adj));
	v->vscroll_bar = w = gtk_vscrollbar_new(adj);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    2, 3,
	    0, 1,
	    GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(MPViewAdjustmentValueChangedCB), v
	);

	/* Horizontal GtkAdjustment and GtkScrollBar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f, 10.0f,
	    10.0f, 10.0f, 10.0f
	);
	v->hadjustment = adj = (GtkAdjustment *)GTK_OBJECT_REF(GTK_OBJECT(adj));
	v->hscroll_bar = w = gtk_hscrollbar_new(adj);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    1, 2,
	    1, 2,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    GTK_FILL,
	    0, 0
	);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(MPViewAdjustmentValueChangedCB), v
	);

	v->freeze_count--;

	return(v->toplevel);
}

/*
 *	Called by MPViewResizeMPVString() to calculate the size of
 *	one line.
 */
static CAString *MPViewResizeMPVStringCalculateLineWidth(
	MPView *v,
	MPVString *ms,
	CAString *ca_string_ptr,
	GdkFont **cur_font,
	const gint width_max,
	gint *line_width_rtn, gint *line_height_rtn
)
{
	gint		len,
			width, height;
	gchar characters_list[] = {
	    ' ',
	    CA_STRING_CHARACTER_LINE_BREAK
	};
	GdkFont *font = *cur_font;
	CAString	*cas_next,
			*cas_next_attr,
			*cas_next_space,
			*cas_last_space;

	/* Set the starting line size */
	*line_width_rtn = 1;
	*line_height_rtn = (font != NULL) ? (font->ascent + font->descent) : 0;

	cas_last_space = NULL;

	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    /* Get the GdkFont for the curent attribute */
	    font = MPViewGetFontFromAttribute(v, CA_STRING_ATTRIBUTE(ca_string_ptr));

	    /* Seek to the next change in attribute and space */
	    cas_next_attr = CAStringSeekToNextAttribute(ca_string_ptr);
	    cas_next_space = CAStringSeekToNextCharacters(
		ca_string_ptr,
		characters_list,
		sizeof(characters_list) / sizeof(gchar)
	    );

	    /* Does a space occure before the attribute changes? */
	    if(cas_next_space < cas_next_attr)
	    {
		gchar *s;

		/* Set the next string segment pointer at this space */
		cas_next = cas_next_space;

		/* Calculate the number of characters of this word
		 * including the space
		 */
		len = (gint)(cas_next - ca_string_ptr) + 1;

		/* Copy/convert this word and space into a regular
		 * string
		 */
		s = CAStringToRegularString(ca_string_ptr, len);
		if((s != NULL) && (font != NULL))
		{
		    /* Calculate the width and height of this word
		     * and space
		     */
		    width = gdk_text_width(font, s, len);
		    height = font->ascent + font->descent;

		    /* Exceeded the maximum width and a prior space
		     * was encountered?
		     */
		    if((((*line_width_rtn) + width) > width_max) &&
		       (cas_last_space != NULL)
		    )
		    {
			/* Replace the last space character with a
			 * line end character
			 */
			CA_STRING_SET_CHARACTER(
			    cas_last_space,
			    CA_STRING_CHARACTER_LINE_END
			);

			/* Seek past the last space and break */
			ca_string_ptr = cas_last_space + 1;
			g_free(s);
			break;
		    }
		    /* Is this space a line break? */
		    else if(CA_STRING_CHARACTER(cas_next) == CA_STRING_CHARACTER_LINE_BREAK)
		    {
			*line_width_rtn = (*line_width_rtn) + width;

			/* Seek past the line break and break */
			ca_string_ptr = cas_next + 1;
			g_free(s);
			break;
		    }
		    else
		    {
			*line_width_rtn = (*line_width_rtn) + width;

			/* Record the last space position */
			cas_last_space = cas_next;
		    }
		    if(height > *line_height_rtn)
			*line_height_rtn = height;
		}
		g_free(s);

		/* Seek past the space to the next word */
		ca_string_ptr = cas_next + 1;
	    }
	    else
	    {
		/* Attribute changes before any space is encountered */
		gchar *s;

		/* Set the next string segment pointer at the start of
		 * the next attribute changed segment
		 */
		cas_next = cas_next_attr;

		/* Calculate the number of characters of the same
		 * attribute
		 */
		len = (gint)(cas_next - ca_string_ptr);

		/* Copy/convert these characters into a regular string */
		s = CAStringToRegularString(ca_string_ptr, len);
		if((s != NULL) && (font != NULL))
		{
		    /* Calculate the width and height of this segment */
		    width = gdk_text_width(font, s, len);
		    height = font->ascent + font->descent;

		    /* Exceeded the maximum width and a prior space
		     * was encountered?
		     */
		    if((((*line_width_rtn) + width) > width_max) &&
		       (cas_last_space != NULL)
		    )
		    {
			/* Replace the last space character with a
			 * line end character
			 */
			CA_STRING_SET_CHARACTER(
			    cas_last_space,
			    CA_STRING_CHARACTER_LINE_END
			);

			/* Seek past the last space and break */
			ca_string_ptr = cas_last_space + 1;
			g_free(s);
			break;
		    }
		    else
		    {
			*line_width_rtn = (*line_width_rtn) + width;
		    }
		    if(height > *line_height_rtn)
			*line_height_rtn = height;
		}
		g_free(s);

		if(cas_next != ca_string_ptr)
		    ca_string_ptr = cas_next;
		else
		    ca_string_ptr++;
	    }
	}

	*cur_font = font;

	return(ca_string_ptr);
}

/*
 *	Called by MPViewResizeObject() to resize the MPVString.
 *
 *	Updates the size of the MPVString and adds newline characters
 *	to its MPVAttributeString.
 */
static void MPViewResizeMPVString(
	MPView *v,
	MPVString *ms,
	const gint width_max
)
{
	gint		line_width,
			line_height,
			min_width,
			min_height;
	GdkFont *font = v->normal_font;
	MPGeometry *ms_geo = &ms->geometry;
	CAString *ca_string_ptr;

	/* Replace all the newline characters with spaces */
	MPVStringClearLineEnds(ms);

	/* Reset the MPVString's size */
	ms_geo->width = 0;
	ms_geo->height = 0;

	ca_string_ptr = ms->ca_string;
	if(ca_string_ptr == NULL)
	    return;

	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    /* Get the size of this line */
	    ca_string_ptr = MPViewResizeMPVStringCalculateLineWidth(
		v,
		ms,
		ca_string_ptr,
		&font,
		width_max,
		&line_width, &line_height
	    );

	    /* Update the size of the MPVString */
	    if(line_width > ms_geo->width)
		ms_geo->width = line_width;
	    ms_geo->height += line_height;
	}

	/* Maintain minimum size */
	font = v->normal_font;
	min_width = 3;
	min_height = (font != NULL) ? (font->ascent + font->descent) : 0;
	if(ms_geo->width < min_width)
	    ms_geo->width = min_width;
	if(ms_geo->height < min_height)
	    ms_geo->height = min_height;
}

/*
 *	Called by MPViewResize() to resize the MPVObject.
 *
 *	Updates the size of the MPVObject and all of its MPVStrings
 *	and adds newline characters to its MPAttrStrings.
 *
 *	The width and height specifies the size of the MPView.
 *
 *	The width_max specifies the prefered maximum width of the
 *	MPVObject.
 */
static void MPViewResizeObject(
	MPView *v,
	MPVObject *o,
	const gint width_max
)
{
	gint		string_num,
			new_width, new_height;
	GList *glist;
	MPGeometry	*ms_geo,
			*o_geo = &o->geometry;
	MPVString *ms;

	/* Reset the MPVObject's size */
	o_geo->width = 0;
	o_geo->height = 0;

	/* Resize each MPVString */
	for(glist = o->strings_list, string_num = 0;
	    glist != NULL;
	    glist = g_list_next(glist), string_num++
	)
	{
	    ms = MPV_STRING(glist->data);
	    if(ms == NULL)
		continue;

	    /* Set the MPVString's position based on the MPVObject's
	     * type
	     */
	    ms_geo = &ms->geometry;
	    switch(o->type)
	    {
	      case MPV_OBJECT_LEFT_PARAGRAPH:
		ms_geo->x = v->object_x_margin;
		ms_geo->y = v->object_y_margin;
		break;

	      case MPV_OBJECT_ITEM:
		if(string_num == 0)
		{
		    /* Position the item at the origin */
		    ms_geo->x = v->object_x_margin;
		    ms_geo->y = v->object_y_margin;
		}
		else
		{
		    /* Position subsequent strings relative to the
		     * previous string which should be the item since
		     * only two strings should be on this object
		     */
		    MPVString *ms_item = MPVObjectGetString(
			o,
			string_num - 1
		    );
		    if(ms_item != NULL)
		    {
			/* If the item's width exceeds half the
			 * maximum width then position it below the
			 * item, otherwise position it to the right
			 * of the item
			 */
			MPGeometry *item_geo = &ms_item->geometry;
			if((item_geo->width + item_geo->x) > (width_max / 2))
			{
			    /* Position this string below the item
			     * string with X indent
			     */
			    ms_geo->x = item_geo->x + 50;
			    ms_geo->y = item_geo->height + item_geo->y;
			}
			else
			{
			    /* Position this string to the right of the
			     * item
			     */
			    ms_geo->x = item_geo->width + item_geo->x + 10;
			    ms_geo->y = item_geo->y;
			}
		    }
		    else
		    {
			ms_geo->x = v->object_x_margin;
			ms_geo->y = v->object_y_margin;
		    }
		}
		break;

	      case MPV_OBJECT_HEADING:
		ms_geo->x = v->object_x_margin;
		ms_geo->y = v->object_y_margin;
		break;
	    }

	    /* Update this MPVString's size */
	    MPViewResizeMPVString(
		v,
		ms,
		width_max - ms_geo->x - v->object_x_margin
	    );

	    /* Update the MPVObject's size after resizing this MPVString */
	    new_width = ms_geo->width + ms_geo->x + v->object_x_margin;
	    if(new_width > o_geo->width)
		o_geo->width = new_width;
	    new_height = ms_geo->height + ms_geo->y + v->object_y_margin;
	    if(new_height > o_geo->height)
		o_geo->height = new_height;
	}
}

/*
 *	Resizes the MPView.
 */
void MPViewResize(
	GtkWidget *w,
	const gint width, const gint height
)
{
	gboolean need_container_check_resize;
	gint		_width = width,
			_height = height;
	gfloat		prev_value;
	GtkAdjustment *adj;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewResize"
	);
	if(v == NULL)
	    return;

	/* If there is no change in size then get the current size */
	if((_width <= 0) || (_height <= 0))
	{
	    w = v->da;
	    if(w->window == NULL)
		return;

	    gdk_window_get_size(w->window, &_width, &_height);
	}
	if((_width <= 0) || (_height <= 0))
	    return;

	need_container_check_resize = FALSE;

	/* Resize the MPVObjects */
	if(v->objects_list != NULL)
	{
	    gint	x = v->x_margin,
			y = v->y_margin,
			greatest_x = 0,
			width_max;
	    GList *glist;
	    MPGeometry *o_geo;
	    MPVObject *o;

	    for(glist = v->objects_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		o_geo = &o->geometry;

		/* Set the MPVObject's coordinates based on its type */
		switch(o->type)
		{
		  case MPV_OBJECT_LEFT_PARAGRAPH:
		  case MPV_OBJECT_ITEM:
		    o_geo->x = x;
		    o_geo->y = y;
		    break;

		  case MPV_OBJECT_HEADING:
		    /* Double the y spacing for subsequent headings */
		    if(glist != v->objects_list)
			y += v->y_spacing;
		    o_geo->x = x;
		    o_geo->y = y;
		    break;
		}

		/* Calculate the prefered maximum width of the
		 * MPVObject, the MPVObject still may exceed it (such
		 * as when it has a single long word in its MPVString)
		 */
		width_max = _width - o_geo->x - v->x_margin;

		/* Update this MPVObject's size */
		MPViewResizeObject(
		    v,
		    o,
		    width_max
		);

		/* Updated the furthest X coordinate value after
		 * resizing this MPVObject
		 */
		if((o_geo->x + o_geo->width) > greatest_x)
		    greatest_x = o_geo->x + o_geo->width;

		/* Increment the Y coordinate for the next MPVObject */
		y += o_geo->height;
		if(g_list_next(glist) != NULL)
		    y += v->y_spacing;
	    }

	    greatest_x += v->x_margin;		/* Add right edge margin */
	    y += v->y_margin;			/* Add bottom edge margin */

	    /* Update the virtual ranges of the GtkAdjustments */
	    adj = v->hadjustment;
	    adj->lower = 0.0f;
	    adj->upper = (gfloat)greatest_x;

	    adj = v->vadjustment;
	    adj->lower = 0.0f;
	    adj->upper = (gfloat)y;
	}


	/* Update the Horizontal GtkAdjustment */
	adj = v->hadjustment;
	prev_value = adj->value;
	adj->page_size = (gfloat)_width;
	adj->step_increment = adj->page_size * 0.1f;
	adj->page_increment = adj->page_size * 1.0f;
	gtk_adjustment_changed(adj);

	/* Map/unmap the Horizontal GtkScrollBar */
	if((adj->upper - adj->lower) > adj->page_size)
	{
	    GtkWidget *w = v->hscroll_bar;
	    if(!GTK_WIDGET_MAPPED(w))
	    {
		need_container_check_resize = TRUE;
		gtk_widget_show(w);
	    }
	}
	else
	{
	    GtkWidget *w = v->hscroll_bar;
	    if(GTK_WIDGET_MAPPED(w))
	    {
		need_container_check_resize = TRUE;
		gtk_widget_hide(v->hscroll_bar);
	    }
	}

	/* Clip the GtkAdjustment value */
	if(adj->value > (adj->upper - adj->page_size))
	    adj->value = adj->upper - adj->page_size;
	if(adj->value < adj->lower)
	    adj->value = adj->lower;
	if(prev_value != adj->value)
	    gtk_adjustment_value_changed(adj);


	/* Update the Vertical GtkAdjustment */
	adj = v->vadjustment;
	prev_value = adj->value;
	adj->page_size = (gfloat)_height;
	adj->step_increment = adj->page_size * 0.1f;
	adj->page_increment = adj->page_size * 1.0f;
	gtk_adjustment_changed(adj);

	/* Map/unmap the Vertical GtkScrollBar */
	if((adj->upper - adj->lower) > adj->page_size)
	{
	    GtkWidget *w = v->vscroll_bar;
	    if(!GTK_WIDGET_MAPPED(w))
	    {
		need_container_check_resize = TRUE;
		gtk_widget_show(w);
	    }
	}
	else
	{
	    GtkWidget *w = v->vscroll_bar;
	    if(GTK_WIDGET_MAPPED(w))
	    {
		need_container_check_resize = TRUE;
		gtk_widget_hide(w);
	    }
	}

	/* Clip the GtkAdjustment value */
	if(adj->value > (adj->upper - adj->page_size))
	    adj->value = adj->upper - adj->page_size;
	if(adj->value < adj->lower)
	    adj->value = adj->lower;
	if(prev_value != adj->value)
	    gtk_adjustment_value_changed(adj);

	/* Recreate the back buffer GdkPixmap if there is a change
	 * in size
	 */
	w = v->da;
	if((w->window != NULL) && (width > 0) && (height > 0))
	{
	    GDK_PIXMAP_UNREF(v->pixmap);
	    v->pixmap = gdk_pixmap_new(
		w->window,
		width, height,
		-1
	    );
	}

	if(need_container_check_resize)
	    gtk_container_check_resize(GTK_CONTAINER(v->toplevel));
}

/*
 *	Queue the MPView to resize.
 */
void MPViewQueueResize(GtkWidget *w)
{
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewQueueResize"
	);
	if(v == NULL)
	    return;

	if(v->resize_idle_id != 0)
	    return;

	v->resize_idle_id = gtk_idle_add_priority(
	    GTK_PRIORITY_RESIZE,
	    MPViewResizeIdleCB, v
	);
}


/*
 *	Updates the MPView's cursor geometry.
 *
 *	Calculates the cursor geometry identically to the MPViewDraw*()
 *	functions and updates the last_cursor MPGeomtry on the MPView
 *	but does not perform any actual drawing.
 */
void MPViewUpdateCursorCoordinates(GtkWidget *w)
{
	gint		x, y,
			start_x, start_y;
	gchar characters_list[] = {
	    CA_STRING_CHARACTER_LINE_END,
	    CA_STRING_CHARACTER_LINE_BREAK
	};
	GtkAdjustment *adj;
	CAString	*ca_string_ptr,
			*cas_next,
			*cas_next_attr,
			*cas_next_line_end,
			*cas_cursor_ptr;
	GdkFont *font;
	CAStringAttributes cur_attr;
	MPGeometry	*ms_geo,
			*o_geo;
	MPVString *ms;
	MPVObject *o;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewUpdateCursorCoordinates"
	);
	if(v == NULL)
	    return;

	o = MPViewGetObject(
	    v,
	    v->focus_obj_num
	);
	if(o == NULL)
	    return;

	o_geo = &o->geometry;

	ms = MPVObjectGetString(
	    o,
	    o->focus_string_num
	);
	if(ms == NULL)
	    return;

	ca_string_ptr = ms->ca_string;
	if(ca_string_ptr == NULL)
	    return;

	ms_geo = &ms->geometry;

	adj = v->hadjustment;
	start_x = x = o_geo->x - (gint)adj->value + ms_geo->x;
	adj = v->vadjustment;
	start_y = y = o_geo->y - (gint)adj->value + ms_geo->y;

	/* If the string is empty then set the cursor's geometry
	 * to the origin of th eobject
	 */
	if(CA_STRING_CHARACTER(ca_string_ptr) == '\0')
	{
	    MPGeometry *cur_geo = &v->last_cursor;
	    font = v->normal_font;
	    cur_geo->x = x;
	    cur_geo->y = y;
	    cur_geo->width = 1;
	    cur_geo->height = (font != NULL) ? (font->ascent + font->descent) : 0;
	    return;
	}

	/* Get the MPVString's cursor position pointer */
	cas_cursor_ptr = CAStringSeekToIndex(
	    ca_string_ptr,
	    ms->cursor_pos
	);

	/* Iterate through the MPVString's CAString and calculate
	 * the cursor's coordinates
	 */
	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    /* Get the GdkFont for the curent attribute */
	    cur_attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
	    font = MPViewGetFontFromAttribute(v, cur_attr);

	    /* Seek to the next change in attribute and line end */
	    cas_next_attr = CAStringSeekToNextAttribute(ca_string_ptr);
	    cas_next_line_end = CAStringSeekToNextCharacters(
		ca_string_ptr,
		characters_list,
		sizeof(characters_list) / sizeof(gchar)
	    );

	    /* Does a line end occure at or before the attribute
	     * changes?
	     */
	    if(cas_next_line_end <= cas_next_attr)
	    {
		/* One complete line */
		gint len;
		gchar *s;

		cas_next = cas_next_line_end;

		/* Calculate the number of characters of this line */
		len = CAStringLengthPointers(ca_string_ptr, cas_next);

		/* Copy/convert this line into a regular string */
		s = CAStringToRegularString(ca_string_ptr, len);

		/* Does the current cursor position lie in this
		 * segment?
		 */
		if((cas_cursor_ptr >= ca_string_ptr) &&
		   (cas_cursor_ptr <= cas_next)
		)
		{
		    /* Calculate the cursor's position */
		    MPGeometry *cur_geo = &v->last_cursor;
		    const gint cur_i = CAStringLengthPointers(
			ca_string_ptr,
			cas_cursor_ptr
		    );
		    cur_geo->x = x;
		    cur_geo->y = y;
		    if((cur_i > 0) && (s != NULL) && (font != NULL))
			cur_geo->x += gdk_text_width(
			    font,
			    s,
			    MIN(cur_i, len)
			);
		    cur_geo->width = 1;
		    cur_geo->height = (font != NULL) ?
			(font->ascent + font->descent) : 1;

		    g_free(s);
		    return;
		}

		g_free(s);

		/* Go to the next line */
		x = start_x;
		if(font != NULL)
		    y += font->ascent + font->descent;

		/* Seek past the newline to the next line */
		ca_string_ptr = cas_next;
		if(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
		    ca_string_ptr++;
	    }
	    else
	    {
		/* Attribute changes before any newline is encountered */
		gint len;
		gchar *s;

		cas_next = cas_next_attr;

		/* Calculate the number of characters of the same
		 * attribute
		 */
		len = CAStringLengthPointers(ca_string_ptr, cas_next);

		/* Copy/convert these characters into a regular string */
		s = CAStringToRegularString(ca_string_ptr, len);

		/* Does the current cursor position lie in this
		 * segment (excluding cas_next)?
		 */
		if((cas_cursor_ptr >= ca_string_ptr) &&
		   (cas_cursor_ptr < cas_next)
		)
		{
		    /* Calculate the cursor's position */
		    MPGeometry *cur_geo = &v->last_cursor;
		    const gint cur_i = CAStringLengthPointers(
			ca_string_ptr,
			cas_cursor_ptr
		    );
		    cur_geo->x = x;
		    cur_geo->y = y;
		    if((cur_i > 0) && (s != NULL) && (font != NULL))
			cur_geo->x += gdk_text_width(
			    font,
			    s,
			    MIN(cur_i, len)
			);
		    cur_geo->width = 1;
		    cur_geo->height = (font != NULL) ?
			(font->ascent + font->descent) : 1;

		    g_free(s);
		    return;
		}

		if((s != NULL) && (font != NULL))
		    x += gdk_text_width(font, s, len);

		g_free(s);

		if(cas_next != ca_string_ptr)
		    ca_string_ptr = cas_next;
		else
		    ca_string_ptr++;
	    }
	}
}

/*
 *	Draws the Line Bar.
 */
void MPViewDrawLineBar(GtkWidget *w)
{
	const gint border_minor = 2;
	gint		scroll_x, scroll_y,
			width, height;
	GdkDrawable *drawable;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkGC *gc;
	GtkAdjustment *adj;
	GtkStateType state;
	GtkStyle *style;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewDrawLineBar"
	);
	if(v == NULL)
	    return;

	w = v->line_bar_da;
	if(!GTK_WIDGET_MAPPED(w))
	    return;

	window = w->window;
	gc = v->gc;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if((window == NULL) || (gc == NULL) || (style == NULL))
	    return;

	pixmap = v->line_bar_pixmap;
	drawable = (pixmap != NULL) ?
	    (GdkDrawable *)pixmap : (GdkDrawable *)window;

	gdk_window_get_size((GdkWindow *)drawable, &width, &height);
	if((width <= 0) || (height <= 0))
	    return;

	adj = v->hadjustment;
	scroll_x = (gint)adj->value;
	adj = v->vadjustment;
	scroll_y = (gint)adj->value;

	gdk_gc_set_fill(gc, GDK_SOLID);
	gdk_gc_set_foreground(
	    gc,
	    (v->normal_color != NULL) ?
		v->normal_color : &style->text[state]
	);

	/* Draw the background (insensitive)*/
	gdk_draw_rectangle(
	    drawable,
	    style->base_gc[GTK_STATE_INSENSITIVE],
	    TRUE,				/* Fill */
	    0, 0,
	    width, height
	);

	/* Draw each object */
	if(v->objects_list != NULL)
	{
	    gint x, y;
	    GList *glist;
	    MPGeometry *o_geo;
	    MPVObject *o;

	    for(glist = v->objects_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		o = MPV_OBJECT(glist->data);
		if(o == NULL)
		    continue;

		o_geo = &o->geometry;
		x = 0;
		y = o_geo->y - scroll_y;

		if(((y + o_geo->height) < 0) || (y >= height))
		    continue;

		/* Draw the background and border */
		if(TRUE)
		{
		    gint8 dashes_list[] = {1, 1};

		    /* Draw the sensitive portion to denote the
		     * object's area
		     */
		    gdk_gc_set_foreground(
			gc,
			(v->bg_color != NULL) ?
			    v->bg_color : &style->base[state]
		    );
		    gdk_draw_rectangle(
			drawable,
			gc,
			TRUE,				/* Fill */
			0, y,
			width, o_geo->height + 1
		    );

		    gdk_gc_set_function(gc, GDK_INVERT);
		    gdk_gc_set_foreground(
			gc,
			&style->white
		    );
		    gdk_gc_set_line_attributes(
			gc,
			1,
			GDK_LINE_ON_OFF_DASH,
			GDK_CAP_NOT_LAST,
			GDK_JOIN_BEVEL
		    );
		    gdk_gc_set_dashes(
			gc,
			1,
			dashes_list, sizeof(dashes_list)
		    );
		    gdk_draw_line(
			drawable,
			gc,
			0, y,
			width, y
		    );
		    gdk_draw_line(
			drawable,
			gc,
			0, y + o_geo->height,
			width, y + o_geo->height
		    );
		    gdk_gc_set_function(gc, GDK_COPY);
		    gdk_gc_set_line_attributes(
			gc,
			1,
			GDK_LINE_SOLID,
			GDK_CAP_NOT_LAST,
			GDK_JOIN_MITER
		    );
		}

		/* Draw the type text */
		if(TRUE)
		{
		    const gchar *text = NULL;
		    switch(o->type)
		    {
		      case MPV_OBJECT_LEFT_PARAGRAPH:
			text = "LP";
			break;

		      case MPV_OBJECT_ITEM:
			text = "TP";
			break;

		      case MPV_OBJECT_HEADING:
			text = "SH";
			break;
		    }
		    if((text != NULL) && (style->font != NULL))
		    {
			GdkFont *font = style->font;
			GdkTextBounds b;
			gdk_string_bounds(font, text, &b);
			gdk_draw_string(
			    drawable,
			    font,
			    style->text_gc[state],
			    width - border_minor - b.width - b.lbearing,
			    y + font->ascent,
			    text
			);
		    }
		}
	    }
	}

	/* Draw the cursor */
	if((GTK_WIDGET_HAS_FOCUS(w) || GTK_WIDGET_HAS_FOCUS(v->da)) &&
	   (v->focus_obj_num > -1)
	)
	{
	    MPGeometry *cur_geo = &v->last_cursor;
	    const gint	arrow_height = MAX(cur_geo->height, 5),
			arrow_width = MIN(arrow_height, (width - 4));
#if 1
	    gtk_draw_arrow(
		style,
		(GdkWindow *)drawable,
		state,
		GTK_SHADOW_OUT,
		GTK_ARROW_RIGHT,
		TRUE,				/* Fill */
		(width - arrow_width) - border_minor,
		cur_geo->y,
		arrow_width,
		arrow_height
	    );
#else
	    gdk_gc_set_function(gc, GDK_INVERT);
	    gdk_gc_set_foreground(
		gc,
		&style->white
	    );
	    gdk_draw_rectangle(
		drawable,
		gc,
		TRUE,				/* Fill */
		0, cur_geo->y,
		width, cur_geo->height
	    );
	    gdk_gc_set_function(gc, GDK_COPY);
#endif
	}

	/* Send the drawable to window if drawable is not the window */
	if(drawable != (GdkDrawable *)window)
	    gdk_draw_pixmap(
		window,
		gc,
		drawable,
		0, 0,
		0, 0,
		width, height
	    );
}

/*
 *	Called by MPVDrawObject() to draw the MPVString and cursor.
 *
 *	The cursor's geometry will be updated.
 */
static void MPVDrawObjectString(
	MPView *v,
	MPVObject *o,
	const gint string_num,
	MPVString *ms,
	GtkWidget *w,
	GdkDrawable *drawable,
	GdkGC *gc,
	GtkStyle *style,
	const gint start_x, const gint start_y
)
{
	gint		x, y;
	gchar characters_list[] = {
	    CA_STRING_CHARACTER_LINE_END,
	    CA_STRING_CHARACTER_LINE_BREAK
	};
	const GtkStateType state = GTK_WIDGET_STATE(w);
	CAString	*ca_string_ptr = ms->ca_string,
			*cas_next,
			*cas_next_attr,
			*cas_next_line_end,
			*cas_cursor_ptr;
	GdkFont *font;
	GdkColor *color;
	CAStringAttributes cur_attr;

	if(ca_string_ptr == NULL)
	    return;

	x = start_x;
	y = start_y;

	/* If this MPVObject has focus then get the pointer to the
	 * cursor position in its MPVString's CAString
	 */
	cas_cursor_ptr = (o->focus_string_num == string_num) ?
	    CAStringSeekToIndex(ca_string_ptr, ms->cursor_pos) : NULL;

	/* If the cursor is set on this MPVObject and the string is
	 * empty then draw just the cursor
	 */
	if((cas_cursor_ptr != NULL) &&
	   CAStringIsEmpty(ca_string_ptr)
	)
	{
	    MPGeometry *cur_geo = &v->last_cursor;
	    font = v->normal_font;
	    cur_geo->x = x;
	    cur_geo->y = y;
	    cur_geo->width = 1;
	    cur_geo->height = (font != NULL) ? (font->ascent + font->descent) : 0;

	    /* Draw the cursor */
	    if(GTK_WIDGET_HAS_FOCUS(w) || GTK_WIDGET_HAS_FOCUS(v->line_bar_da))
	    {
		gdk_gc_set_function(gc, GDK_INVERT);
		gdk_gc_set_foreground(
		    gc,
		    &style->white
		);
		gdk_draw_line(
		    drawable,
		    gc,
		    cur_geo->x, cur_geo->y,
		    cur_geo->x, cur_geo->y + cur_geo->height
		);
		gdk_gc_set_function(gc, GDK_COPY);
	    }
	    return;
	}

#define DRAW_STRING_END_ICON(_x_,_y_)	{	\
 if(v->string_end_bitmap != NULL) {		\
  gint	x2, y2,					\
	width2, height2;			\
  GdkBitmap *mask = v->string_end_bitmap;	\
  GdkColor *color2 = v->normal_color;		\
						\
  gdk_window_get_size(				\
   (GdkWindow *)mask,				\
   &width2, &height2				\
  );						\
  x2 = (_x_);					\
  y2 = (_y_) + ((font != NULL) ?		\
   (font->ascent - height2) : 0);		\
  gdk_gc_set_clip_origin(gc, x2, y2);		\
  gdk_gc_set_clip_mask(gc, mask);		\
  gdk_gc_set_foreground(			\
   gc,						\
   (color2 != NULL) ?				\
   color2 : &style->fg[state]			\
  );						\
  gdk_draw_rectangle(				\
   drawable,					\
   gc,						\
   TRUE,			/* Fill */	\
   x2, y2,					\
   width2, height2				\
  );						\
  gdk_gc_set_clip_mask(gc, NULL);		\
 }						\
}


	/* Draw each line */
	while(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
	{
	    /* Get the GdkFont and GdkColor for the curent attribute */
	    cur_attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
	    font = MPViewGetFontFromAttribute(v, cur_attr);
	    color = MPViewGetColorFromAttribute(v, cur_attr);

	    /* Seek to the next change in attribute and line end */
	    cas_next_attr = CAStringSeekToNextAttribute(ca_string_ptr);
	    cas_next_line_end = CAStringSeekToNextCharacters(
		ca_string_ptr,
		characters_list,
		sizeof(characters_list) / sizeof(gchar)
	    );

	    /* Does a line end occure at or before the attribute
	     * changes?
	     */
	    if(cas_next_line_end <= cas_next_attr)
	    {
		/* Drawing one complete line */
		gint	len,
			x_end = x;
		gchar *s;

		cas_next = cas_next_line_end;

		/* Calculate the number of characters on this line */
		len = (gint)(cas_next - ca_string_ptr);

		/* Copy/convert this line into a regular string */
		s = CAStringToRegularString(ca_string_ptr, len);
		if((s != NULL) && (font != NULL))
		{
		    GdkTextBounds b;
		    gdk_text_bounds(font, s, len, &b);
		    if(cur_attr & CA_STRING_ATTRIBUTE_SELECTED)
		    {
			const gint	x2 = x,
					y2 = y,
					width2 = b.width,
					height2 = font->ascent + font->descent;
			gdk_gc_set_foreground(
			    gc,
			    &style->bg[GTK_STATE_SELECTED]
			);
			gdk_draw_rectangle(
			    drawable,
			    gc,
			    TRUE,		/* Fill */
			    x2, y2,
			    width2, height2
			);
		    }
		    gdk_gc_set_font(gc, font);
		    gdk_gc_set_foreground(
			gc,
			(color != NULL) ?
			    color : &style->text[state]
		    );
		    gdk_draw_text(
			drawable,
			font,
			gc,
			x /* - b.lbearing */,
			y + font->ascent,
			s, len
		    );
		    /* Draw the underline? */
		    if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
		    {
			const gint	x2 = x + b.width,
					y2 = y + font->ascent;
			gdk_draw_line(
			    drawable,
			    gc,
			    x, y2,
			    x2, y2
			);
		    }
		    x_end += b.width;
		}

		/* Draw line break icon */
		if((v->flags & MP_VIEW_SHOW_STRING_MARKERS) &&
		   (CA_STRING_CHARACTER(cas_next) == CA_STRING_CHARACTER_LINE_BREAK) &&
		   (v->line_break_bitmap != NULL)
		)
		{
		    gint	x2, y2,
				width2, height2;
		    GdkBitmap *mask = v->line_break_bitmap;
		    GdkColor *color2 = v->normal_color;

		    gdk_window_get_size((GdkWindow *)mask, &width2, &height2);
		    x2 = x_end;
		    y2 = y + ((font != NULL) ? (font->ascent - height2) : 0);
		    gdk_gc_set_clip_origin(gc, x2, y2);
		    gdk_gc_set_clip_mask(gc, mask);
		    gdk_gc_set_foreground(
			gc,
			(color2 != NULL) ?
			    color2 : &style->fg[state]
		    );
		    gdk_draw_rectangle(
			drawable,
			gc,
			TRUE,			/* Fill */
			x2, y2,
			width2, height2
		    );
		    gdk_gc_set_clip_mask(gc, NULL);
		}

		/* Does the current cursor position lie in this
		 * segment?
		 */
		if((cas_cursor_ptr >= ca_string_ptr) &&
		   (cas_cursor_ptr <= cas_next)
		)
		{
		    /* Calculate the cursor's position */
		    MPGeometry *cur_geo = &v->last_cursor;
		    const gint cur_i = CAStringLengthPointers(
			ca_string_ptr,
			cas_cursor_ptr
		    );
		    cur_geo->x = x;
		    cur_geo->y = y;
		    if((cur_i > 0) && (s != NULL) && (font != NULL))
			cur_geo->x += gdk_text_width(
			    font,
			    s,
			    MIN(cur_i, len)
			);
		    cur_geo->width = 1;
		    cur_geo->height = (font != NULL) ?
			(font->ascent + font->descent) : 1;

		    /* Draw the cursor */
		    if(GTK_WIDGET_HAS_FOCUS(w) ||
		       GTK_WIDGET_HAS_FOCUS(v->line_bar_da)
		    )
		    {
			gdk_gc_set_function(gc, GDK_INVERT);
			gdk_gc_set_foreground(
			    gc,
			    &style->white
			);
			gdk_draw_line(
			    drawable,
			    gc,
			    cur_geo->x, cur_geo->y,
			    cur_geo->x, cur_geo->y + cur_geo->height
			);
			gdk_gc_set_function(gc, GDK_COPY);
		    }
		}

		g_free(s);

		/* Seek past the line end character to the next line */
		ca_string_ptr = cas_next;
		if(CA_STRING_CHARACTER(ca_string_ptr) != '\0')
		    ca_string_ptr++;

		/* Draw the string end icon */
		if((v->flags & MP_VIEW_SHOW_STRING_MARKERS) &&
		   (CA_STRING_CHARACTER(ca_string_ptr) == '\0')
		)
		{
		    DRAW_STRING_END_ICON(x_end, y);
		}

		/* Increment the coordinates to the next line */
		x = start_x;
		if(font != NULL)
		    y += font->ascent + font->descent;
	    }
	    else
	    {
		/* Attribute changes before any line end is encountered */
		gint	len,
			text_width = 0;
		gchar *s;

		cas_next = cas_next_attr;

		/* Calculate the number of characters of the same
		 * attribute
		 */
		len = (gint)(cas_next - ca_string_ptr);

		/* Copy/convert these characters into a regular string */
		s = CAStringToRegularString(ca_string_ptr, len);
		if((s != NULL) && (font != NULL))
		{
		    GdkTextBounds b;
		    gdk_text_bounds(font, s, len, &b);
		    text_width = b.width;
		    if(cur_attr & CA_STRING_ATTRIBUTE_SELECTED)
		    {
			const gint	x2 = x,
					y2 = y,
					width2 = text_width,
					height2 = font->ascent + font->descent;
			gdk_gc_set_foreground(
			    gc,
			    &style->bg[GTK_STATE_SELECTED]
			);
			gdk_draw_rectangle(
			    drawable,
			    gc,
			    TRUE,		/* Fill */
			    x2, y2,
			    width2, height2
			);
		    }
		    gdk_gc_set_font(gc, font);
		    gdk_gc_set_foreground(
			gc,
			(color != NULL) ?
			    color : &style->text[state]
		    );
		    gdk_draw_text(
			drawable,
			font,
			gc,
			x /* - b.lbearing */,
			y + font->ascent,
			s, len
		    );
		    /* Draw the underline? */
		    if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
		    {
			const gint	x2 = x + text_width,
					y2 = y + font->ascent;
			gdk_draw_line(
			    drawable,
			    gc,
			    x, y2,
			    x2, y2
			);
		    }
		}

		/* Does the current cursor position lie in this
		 * segment (excluding cas_next)?
		 */
		if((cas_cursor_ptr >= ca_string_ptr) &&
		   (cas_cursor_ptr < cas_next)
		)
		{
		    /* Calculate the cursor's position */
		    MPGeometry *cur_geo = &v->last_cursor;
		    const gint cur_i = CAStringLengthPointers(
			ca_string_ptr,
			cas_cursor_ptr
		    );
		    cur_geo->x = x;
		    cur_geo->y = y;
		    if((cur_i > 0) && (s != NULL) && (font != NULL))
			cur_geo->x += gdk_text_width(
			    font,
			    s,
			    MIN(cur_i, len)
			);
		    cur_geo->width = 1;
		    cur_geo->height = (font != NULL) ?
			(font->ascent + font->descent) : 1;

		    /* Draw the cursor */
		    if(GTK_WIDGET_HAS_FOCUS(w) ||
		       GTK_WIDGET_HAS_FOCUS(v->line_bar_da)
		    )
		    {
			gdk_gc_set_function(gc, GDK_INVERT);
			gdk_gc_set_foreground(
			    gc,
			    &style->white
			);
			gdk_draw_line(
			    drawable,
			    gc,
			    cur_geo->x, cur_geo->y,
			    cur_geo->x, cur_geo->y + cur_geo->height
			);
			gdk_gc_set_function(gc, GDK_COPY);
		    }
		}

		g_free(s);

		if(cas_next != ca_string_ptr)
		    ca_string_ptr = cas_next;
		else
		    ca_string_ptr++;

		/* Draw the string end icon */
		if((v->flags & MP_VIEW_SHOW_STRING_MARKERS) &&
		   (CA_STRING_CHARACTER(ca_string_ptr) == '\0')
		)
		{
		    DRAW_STRING_END_ICON(x, y);
		}

		/* Increment the X coordinate past this segment */
		x += text_width;
	    }
	}
#undef DRAW_STRING_END_ICON
}

/*
 *	Called by MPViewDraw() to draw the MPVObject.
 */
static void MPViewDrawObject(
	MPView *v,
	MPVObject *o,
	GtkWidget *w,
	GdkDrawable *drawable,
	GdkGC *gc,
	GtkStyle *style,
	const gint width, const gint height,
	const gint scroll_x, const gint scroll_y
)
{
	gint x, y;
	MPGeometry *o_geo;

	if(o == NULL)
	    return;

	/* Calculate the object's coordinates on the GtkDrawingArea
	 * with the scroll GtkAdjustments applied
	 */
	o_geo = &o->geometry;
	x = o_geo->x - scroll_x;
	y = o_geo->y - scroll_y;

	if((y >= height) || ((y + o_geo->height) < 0))
	    return;

	if(o->strings_list != NULL)
	{
	    gint string_num;
	    GList *glist;
	    MPGeometry *ms_geo;
	    MPVString *ms;

	    for(glist = o->strings_list, string_num = 0;
		glist != NULL;
		glist = g_list_next(glist), string_num++
	    )
	    {
		ms = MPV_STRING(glist->data);
		if(ms == NULL)
		    continue;

		ms_geo = &ms->geometry;

		MPVDrawObjectString(
		    v,
		    o,
		    string_num,
		    MPV_STRING(glist->data),
		    w,
		    drawable,
		    gc,
		    style,
		    x + ms_geo->x,
		    y + ms_geo->y
		);
	    }
	}

	/* Draw the focus/object bounds */
	if((GTK_WIDGET_HAS_FOCUS(w) || GTK_WIDGET_HAS_FOCUS(v->line_bar_da)) &&
	   (o->focus_string_num > -1) &&
	   (v->flags & MP_VIEW_SHOW_OBJECT_BOUNDS)
	)
	{
	    gint8 dashes_list[] = {1, 1};
	    gdk_gc_set_function(gc, GDK_INVERT);
	    gdk_gc_set_foreground(
		gc,
		&style->white
	    );
	    gdk_gc_set_line_attributes(
		gc,
		1,
		GDK_LINE_ON_OFF_DASH,
		GDK_CAP_NOT_LAST,
		GDK_JOIN_BEVEL
	    );
	    gdk_gc_set_dashes(
		gc,
		1,
		dashes_list, sizeof(dashes_list)
	    );
	    gdk_draw_rectangle(
		drawable,
		gc,
		FALSE,				/* Outline */
		x, y,
		o_geo->width, o_geo->height
	    );
	    gdk_gc_set_function(gc, GDK_COPY);
	    gdk_gc_set_line_attributes(
		gc,
		1,
		GDK_LINE_SOLID,
		GDK_CAP_NOT_LAST,
		GDK_JOIN_MITER
	    );
	}
}

/*
 *	Draws the MPView.
 */
void MPViewDraw(GtkWidget *w)
{
	gint		scroll_x, scroll_y,
			width, height;
	GdkDrawable *drawable;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkGC *gc;
	GtkStateType state;
	GtkAdjustment *adj;
	GtkStyle *style;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewDraw"
	);
	if(v == NULL)
	    return;

	/* Reset the cursor position off the visible area, if the
	 * cursor should be visible then it will be drawn and its
	 * geometry updated later in this call
	 */
	if(TRUE)
	{
	    MPGeometry *cur_geo = &v->last_cursor;
	    cur_geo->x = -10;
	    cur_geo->y = -10;
	    cur_geo->width = 1;
	    cur_geo->height = 1;
	}

	w = v->da;
	if(!GTK_WIDGET_MAPPED(w))
	    return;

	window = w->window;
	gc = v->gc;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if((window == NULL) || (gc == NULL) || (style == NULL))
	    return;

	pixmap = v->pixmap;
	drawable = (pixmap != NULL) ?
	    (GdkDrawable *)pixmap : (GdkDrawable *)window;

	gdk_window_get_size((GdkWindow *)drawable, &width, &height);
	if((width <= 0) || (height <= 0))
	    return;

	adj = v->hadjustment;
	scroll_x = (gint)adj->value;
	adj = v->vadjustment;
	scroll_y = (gint)adj->value;

	gdk_gc_set_fill(gc, GDK_SOLID);
	gdk_gc_set_foreground(
	    gc,
	    (v->bg_color != NULL) ?
		v->bg_color : &style->base[state]
	);

	/* Draw the background */
	gdk_draw_rectangle(
	    drawable,
	    gc,
	    TRUE,				/* Fill */
	    0, 0,
	    width, height
	);

	/* Draw each object */
	if(v->objects_list != NULL)
	{
	    GList *glist;

	    for(glist = v->objects_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		MPViewDrawObject(
		    v,
		    MPV_OBJECT(glist->data),
		    w,
		    drawable,
		    gc,
		    style,
		    width, height,
		    scroll_x, scroll_y
		);
	    }
	}

	/* Send drawable to window if drawable is not the window */
	if(drawable != (GdkDrawable *)window)
	    gdk_draw_pixmap(
		window,
		gc,
		drawable,
		0, 0,
		0, 0,
		width, height
	    );

	/* Draw the Line Bar afterwards, since the cursor's geometry
	 * is not set until all the MPVObjects have been drawn
	 */
	MPViewDrawLineBar(v->toplevel);
}

/*
 *	Queues the MPView to draw.
 */
void MPViewQueueDraw(GtkWidget *w)
{
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewQueueDraw"
	);
	if(v == NULL)
	    return;

	gtk_widget_queue_draw(v->da);
}
