#include <string.h>
#include <gtk/gtk.h>

#include "ca_string.h"
#include "manpage_view.h"
#include "manpage_view_private.h"


static gboolean MPViewXMLIsTagObject(const gchar *tag);
static const gchar *MPViewXMLGetSymbol(
	const gchar *s,
	char *c_rtn
);
static const gchar *MPViewXMLGetTag(
	const gchar *s,
	gchar **tag_rtn
);
static const gchar *MPViewXMLInsertObjectLP(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
);
static const gchar *MPViewXMLInsertObjectTP(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
);
static const gchar *MPViewXMLInsertObjectSH(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
);

/* Insert From XMLString */
gint MPViewInsertObjectsFromXMLString(
	GtkWidget *w,
	const gint i,
	const gchar *s,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
);

gchar *MPViewGetObjectXMLString(
        GtkWidget *w,
        MPVObject *o
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Checks if the tag represents a new object.
 */
static gboolean MPViewXMLIsTagObject(const gchar *tag)
{
	if(tag == NULL)
	    return(FALSE);

	if(!strcmp((const char *)tag, "LP") ||
	   !strcmp((const char *)tag, "TP") ||
	   !strcmp((const char *)tag, "RP") ||
	   !strcmp((const char *)tag, "SH")
	)
	    return(TRUE);
	else
	    return(FALSE);
}

/*
 *	Gets the character represented by the symbol and seeks
 *	past the symbol (past the ';' character) or end of string.
 */
static const gchar *MPViewXMLGetSymbol(
	const gchar *s,
	char *c_rtn
)
{
	const gchar *s_next;

	/* Seek to the start of the symbol */
	while((*s != '\0') && (*s != '&'))
	    s++;

	/* Seek past the start of the symbol */
	if(*s == '&')
	    s++;

	/* Seek past the symbol */
	s_next = s;
	while((*s_next != '\0') && (*s_next != ';'))
	    s_next++;
	if(*s_next == ';')
	    s_next++;

	if(!strncmp((const char *)s, "lt;", 3))
	    *c_rtn = '<';
	else if(!strncmp((const char *)s, "gt;", 3))
	    *c_rtn = '>';
	else if(!strncmp((const char *)s, "amp;", 4))
	    *c_rtn = '&';
	else
	    *c_rtn = '&';

	return(s_next);
}

/*
 *	Gets the string in the tag (without the '<' and '>'
 *	characters) and seeks past the tag (past the '>' character)
 *	or end of string.
 */
static const gchar *MPViewXMLGetTag(
	const gchar *s,
	gchar **tag_rtn
)
{
	gint len;
	gchar *tag;
	const gchar *s_next;

	/* Seek to the start of the tag */
	while((*s != '\0') && (*s != '<'))
	    s++;

	/* Seek past the start of the tag */
	if(*s == '<')
	    s++;

	/* Seek to the end of the tag */
	s_next = s;
	while((*s_next != '\0') && (*s_next != '>'))
	    s_next++;

	/* Get the length of this tag */
	len = (gint)(s_next - s);

	/* Seek past the end of the tag */
	if(*s_next == '>')
	    s_next++;

	tag = (gchar *)g_malloc((len + 1) * sizeof(gchar));
	if(tag == NULL)
	{
	    *tag_rtn = NULL;
	    return(s_next);
	}

	if(len > 0)
	    memcpy(
		tag,
		s,
		len * sizeof(gchar)
	    );
	tag[len] = '\0';

	*tag_rtn = tag;

	return(s_next);
}

/*
 *	Inserts a left paragraph object.
 *
 *	The s specifies the start of the data for the new object (past
 *	the tag).
 *
 *	Returns the string position at the start of the tag for the next
 *	object or end of string.
 */
static const gchar *MPViewXMLInsertObjectLP(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
)
{
	gchar c;
	gint obj_num;
	CAString *ca_string;
	CAStringAttributes attr;
	MPVString *ms;
	MPVObject *o;

	while((*s == ' ') || (*s == '\t') || (*s == '\r') || (*s == '\n'))
	    s++;

	o = MPVObjectNew();
	if(o == NULL)
	    return(s);

	if(insert_pos > -1)
	{
	    obj_num = insert_pos;
	    v->objects_list = g_list_insert(
		v->objects_list,
		o,
		insert_pos
	    );
	}
	else
	{
	    obj_num = g_list_length(v->objects_list);
	    v->objects_list = g_list_append(
		v->objects_list,
		o
	    );
	}

	o->type = MPV_OBJECT_LEFT_PARAGRAPH;
	o->focus_string_num = -1;

	ms = MPVStringNew();
	if(ms == NULL)
	    return(s);

	o->strings_list = g_list_append(
	    o->strings_list,
	    ms
	);

	ca_string = NULL;
	attr = 0;
	while(*s != '\0')
	{
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (gulong)(s - s_start),
		    progress_m,
		    progress_data
		))
		    return(s_start + progress_m);
	    }

	    /* Start of tag? */
	    if(*s == '<')
	    {
		const gchar *s_prev = s;
		gchar *tag;

		/* Get the tag and seek past the tag */
		s = MPViewXMLGetTag(
		    s,
		    &tag
		);
		if(tag == NULL)
		    break;

		/* Bold */
		if(!strcmp((const char *)tag, "fB"))
		{
		    attr |= CA_STRING_ATTRIBUTE_BOLD;
		    g_free(tag);
		    continue;
		}
		/* Italic */
		else if(!strcmp((const char *)tag, "fI"))
		{
		    attr |= CA_STRING_ATTRIBUTE_UNDERLINE;
		    g_free(tag);
		    continue;
		}
		/* Normal */
		else if(!strcmp((const char *)tag, "fR"))
		{
		    attr &= ~CA_STRING_ATTRIBUTE_BOLD;
		    attr &= ~CA_STRING_ATTRIBUTE_UNDERLINE;
		    attr &= ~CA_STRING_ATTRIBUTE_SELECTED;
		    g_free(tag);
		    continue;
		}
		/* Line Break? */
		else if(!strcmp((const char *)tag, "br"))
		{
		    c = CA_STRING_CHARACTER_LINE_BREAK;
		}
		/* Is this tag the start of a new object? */
		else if(MPViewXMLIsTagObject(tag))
		{
		    /* Seek back to the start of the tag and stop
		     * adding to this string
		     */
		    s = s_prev;
		    g_free(tag);
		    break;
		}
		else
		{
		    /* All else add the tag to the string */
		    c = *s_prev;
		    s = s_prev + 1;
		}

		g_free(tag);
	    }
	    /* Symbol? */
	    else if(*s == '&')
	    {
		s = MPViewXMLGetSymbol(
		    s,
		    &c
		);
	    }
	    else
	    {
		/* Regular character */
		c = *s;
		s++;

		/* Turn tabs and newlines into spaces */
		if((c == '\t') || (c == '\n') || (c == '\r'))
		    c = ' ';
	    }

	    /* Append this character */
	    ca_string = CAStringAppendCharacter(
		ca_string,
		c,
		attr
	    );
	}

	if(ca_string == NULL)
	    ca_string = CAStringNewFromString("", -1);

	/* Strip any leading or tailing spaces */
	ca_string = CAStringStrip(ca_string);

	ms->ca_string = ca_string;
	ms->length = CAStringLength(ms->ca_string);
	ms->cursor_pos = -1;
	ms->selection_start_pos = ms->selection_end_pos = -1;

	return(s);
}

/*
 *	Inserts an item object.
 *
 *	The s specifies the start of the data for the new object (past
 *	the tag).
 *
 *	Returns the string position at the start of the tag for the next
 *	object or end of string.
 */
static const gchar *MPViewXMLInsertObjectTP(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
)
{
	gchar c;
	gint obj_num;
	CAString *ca_string;
	CAStringAttributes attr;
	MPVString *ms;
	MPVObject *o;

	while((*s == ' ') || (*s == '\t') || (*s == '\r') || (*s == '\n'))
	    s++;

	o = MPVObjectNew();
	if(o == NULL)
	    return(s);

	if(insert_pos > -1)
	{
	    obj_num = insert_pos;
	    v->objects_list = g_list_insert(
		v->objects_list,
		o,
		insert_pos
	    );
	}
	else
	{
	    obj_num = g_list_length(v->objects_list);
	    v->objects_list = g_list_append(
		v->objects_list,
		o
	    );
	}

	o->type = MPV_OBJECT_ITEM;
	o->focus_string_num = -1;

	/* Add the first MPVString */
	ms = MPVStringNew();
	if(ms == NULL)
	    return(s);

	o->strings_list = g_list_append(
	    o->strings_list,
	    ms
	);

	ca_string = NULL;
	attr = 0;
	while(*s != '\0')
	{
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (gulong)(s - s_start),
		    progress_m,
		    progress_data
		))
		    return(s_start + progress_m);
	    }

	    /* Start of tag? */
	    if(*s == '<')
	    {
		const gchar *s_prev = s;
		gchar *tag;

		/* Get the tag and seek past the tag */
		s = MPViewXMLGetTag(
		    s,
		    &tag
		);
		if(tag == NULL)
		    break;

		/* Bold */
		if(!strcmp((const char *)tag, "fB"))
		{
		    attr |= CA_STRING_ATTRIBUTE_BOLD;
		    g_free(tag);
		    continue;
		}
		/* Italic */
		else if(!strcmp((const char *)tag, "fI"))
		{
		    attr |= CA_STRING_ATTRIBUTE_UNDERLINE;
		    g_free(tag);
		    continue;
		}
		/* Normal */
		else if(!strcmp((const char *)tag, "fR"))
		{
		    attr &= ~CA_STRING_ATTRIBUTE_BOLD;
		    attr &= ~CA_STRING_ATTRIBUTE_UNDERLINE;
		    attr &= ~CA_STRING_ATTRIBUTE_SELECTED;
		    g_free(tag);
		    continue;
		}
		/* Line Break? */
		else if(!strcmp((const char *)tag, "br"))
		{
		    c = CA_STRING_CHARACTER_LINE_BREAK;
		}
		/* Is this tag the start of a new object? */
		else if(MPViewXMLIsTagObject(tag))
		{
		    /* Seek back to the start of the tag and stop
		     * adding to this string
		     */
		    s = s_prev;
		    g_free(tag);
		    break;
		}
		else
		{
		    /* All else add the tag to the string */
		    c = *s_prev;
		    s = s_prev + 1;
		}

		g_free(tag);
	    }
	    /* Symbol? */
	    else if(*s == '&')
	    {
		s = MPViewXMLGetSymbol(
		    s,
		    &c
		);
	    }
	    /* End of item? */
	    else if(*s == '\n')
	    {
		s++;
		break;
	    }
	    else
	    {
		/* Regular character */
		c = *s;
		s++;

		/* Turn tabs and newlines into spaces */
		if((c == '\t') || (c == '\n') || (c == '\r'))
		    c = ' ';
	    }

	    /* Append this character */
	    ca_string = CAStringAppendCharacter(
		ca_string,
		c,
		attr
	    );
	}
	if(ca_string == NULL)
	    ca_string = CAStringNewFromString("", -1);

	/* Strip any leading or tailing spaces */
	ca_string = CAStringStrip(ca_string);

	ms->ca_string = ca_string;
	ms->length = CAStringLength(ms->ca_string);
	ms->cursor_pos = -1;
	ms->selection_start_pos = ms->selection_end_pos = -1;


	/* Add the second MPVString */
	ms = MPVStringNew();
	if(ms == NULL)
	    return(s);

	o->strings_list = g_list_append(
	    o->strings_list,
	    ms
	);

	ca_string = NULL;
	attr = 0;
	while(*s != '\0')
	{
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (gulong)(s - s_start),
		    progress_m,
		    progress_data
		))
		    return(s_start + progress_m);
	    }

	    /* Start of tag? */
	    if(*s == '<')
	    {
		const gchar *s_prev = s;
		gchar *tag;

		/* Get the tag and seek past the tag */
		s = MPViewXMLGetTag(
		    s,
		    &tag
		);
		if(tag == NULL)
		    break;

		/* Bold */
		if(!strcmp((const char *)tag, "fB"))
		{
		    attr |= CA_STRING_ATTRIBUTE_BOLD;
		    g_free(tag);
		    continue;
		}
		/* Italic */
		else if(!strcmp((const char *)tag, "fI"))
		{
		    attr |= CA_STRING_ATTRIBUTE_UNDERLINE;
		    g_free(tag);
		    continue;
		}
		/* Normal */
		else if(!strcmp((const char *)tag, "fR"))
		{
		    attr &= ~CA_STRING_ATTRIBUTE_BOLD;
		    attr &= ~CA_STRING_ATTRIBUTE_UNDERLINE;
		    attr &= ~CA_STRING_ATTRIBUTE_SELECTED;
		    g_free(tag);
		    continue;
		}
		/* Line Break? */
		else if(!strcmp((const char *)tag, "br"))
		{
		    c = CA_STRING_CHARACTER_LINE_BREAK;
		}
		/* Is this tag the start of a new object? */
		else if(MPViewXMLIsTagObject(tag))
		{
		    /* Seek back to the start of the tag and stop
		     * adding to this string
		     */
		    s = s_prev;
		    g_free(tag);
		    break;
		}
		else
		{
		    /* All else add the tag to the string */
		    c = *s_prev;
		    s = s_prev + 1;
		}

		g_free(tag);
	    }
	    /* Symbol? */
	    else if(*s == '&')
	    {
		s = MPViewXMLGetSymbol(
		    s,
		    &c
		);
	    }
	    else
	    {
		/* Regular character */
		c = *s;
		s++;

		/* Turn tabs and newlines into spaces */
		if((c == '\t') || (c == '\n') || (c == '\r'))
		    c = ' ';
	    }

	    /* Append this character */
	    ca_string = CAStringAppendCharacter(
		ca_string,
		c,
		attr
	    );
	}
	if(ca_string == NULL)
	    ca_string = CAStringNewFromString("", -1);

	/* Strip any leading or tailing spaces */
	ca_string = CAStringStrip(ca_string);

	ms->ca_string = ca_string;
	ms->length = CAStringLength(ms->ca_string);
	ms->cursor_pos = -1;
	ms->selection_start_pos = ms->selection_end_pos = -1;

	return(s);
}

/*
 *	Inserts an heading object.
 *
 *	The s specifies the start of the data for the new object (past
 *	the tag).
 *
 *	Returns the string position at the start of the tag for the next
 *	object or end of string.
 */
static const gchar *MPViewXMLInsertObjectSH(
	MPView *v,
	const gchar *s,
	const gint insert_pos,
	const gchar *s_start,
	gulong progress_m,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
)
{
	gchar c;
	gint obj_num;
	CAString *ca_string;
	CAStringAttributes attr;
	MPVString *ms;
	MPVObject *o;

	while((*s == ' ') || (*s == '\t') || (*s == '\r') || (*s == '\n'))
	    s++;

	o = MPVObjectNew();
	if(o == NULL)
	    return(s);

	if(insert_pos > -1)
	{
	    obj_num = insert_pos;
	    v->objects_list = g_list_insert(
		v->objects_list,
		o,
		insert_pos
	    );
	}
	else
	{
	    obj_num = g_list_length(v->objects_list);
	    v->objects_list = g_list_append(
		v->objects_list,
		o
	    );
	}

	o->type = MPV_OBJECT_HEADING;
	o->focus_string_num = -1;

	/* Add the first MPVString */
	ms = MPVStringNew();
	if(ms == NULL)
	    return(s);

	o->strings_list = g_list_append(
	    o->strings_list,
	    ms
	);

	ca_string = NULL;
	attr = CA_STRING_ATTRIBUTE_HEADING;
	while(*s != '\0')
	{
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (gulong)(s - s_start),
		    progress_m,
		    progress_data
		))
		    return(s_start + progress_m);
	    }

	    /* Start of tag? */
	    if(*s == '<')
	    {
		const gchar *s_prev = s;
		gchar *tag;

		/* Get the tag and seek past the tag */
		s = MPViewXMLGetTag(
		    s,
		    &tag
		);
		if(tag == NULL)
		    break;

		/* Bold */
		if(!strcmp((const char *)tag, "fB"))
		{
		    /* Ignore */
		    g_free(tag);
		    continue;
		}
		/* Italic */
		else if(!strcmp((const char *)tag, "fI"))
		{
		    /* Ignore */
		    g_free(tag);
		    continue;
		}
		/* Normal */
		else if(!strcmp((const char *)tag, "fR"))
		{
		    /* Ignore */
		    g_free(tag);
		    continue;
		}
		/* Line Break? */
		else if(!strcmp((const char *)tag, "br"))
		{
		    c = CA_STRING_CHARACTER_LINE_BREAK;
		}
		/* Is this tag the start of a new object? */
		else if(MPViewXMLIsTagObject(tag))
		{
		    /* Seek back to the start of the tag and stop
		     * adding to this string
		     */
		    s = s_prev;
		    g_free(tag);
		    break;
		}
		else
		{
		    /* All else add the tag to the string */
		    c = *s_prev;
		    s = s_prev + 1;
		}

		g_free(tag);
	    }
	    /* Symbol? */
	    else if(*s == '&')
	    {
		s = MPViewXMLGetSymbol(
		    s,
		    &c
		);
	    }
	    else
	    {
		/* Regular character */
		c = *s;
		s++;

		/* Turn tabs and newlines into spaces */
		if((c == '\t') || (c == '\n') || (c == '\r'))
		    c = ' ';
	    }

	    /* Append this character */
	    ca_string = CAStringAppendCharacter(
		ca_string,
		c,
		attr
	    );
	}
	if(ca_string == NULL)
	    ca_string = CAStringNewFromString("", -1);

	/* Strip any leading or tailing spaces */
	ca_string = CAStringStrip(ca_string);

	ms->ca_string = ca_string;
	ms->length = CAStringLength(ms->ca_string);
	ms->cursor_pos = -1;
	ms->selection_start_pos = ms->selection_end_pos = -1;

	return(s);
}

/*
 *	Inserts new MPVObjects into the MPView from the XML string.
 *
 *	The i specifies the MPVObject to insert at or -1 to append.
 *
 *	The s specifies a string describing the manual page with XML
 *	tags.
 *
 *	The progress_cb and progress_data specifies the progress
 *	callbacks.
 *
 *	Returns the number of MPVObjects added or negative on error.
 */
gint MPViewInsertObjectsFromXMLString(
	GtkWidget *w,
	const gint i,
	const gchar *s,
	gint progress_cb(const gulong, const gulong, gpointer),
	gpointer progress_data
)
{
	gint		nobjs_added,
			insert_pos = i;
	gulong progress_m;
	const gchar	*s_start,
			*s_end;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewInsertObjectFromXMLString"
	);
	if((v == NULL) || (s == NULL))
	    return(-2);

	nobjs_added = 0;
	progress_m = STRLEN(s);
	s_start = s;
	s_end = s_start + progress_m;

	while(*s != '\0')
	{
	    if(progress_cb != NULL)
	    {
		if(progress_cb(
		    (gulong)(s - s_start),
		    progress_m,
		    progress_data
		))
		    break;
	    }

	    /* Start of tag? */
	    if(*s == '<')
	    {
		const gchar *prev_s = s;
		gchar *tag;

		s = MPViewXMLGetTag(
		    s,
		    &tag
		);
		if(tag == NULL)
		    break;

		/* Does this tag describe a new object? */
		if(MPViewXMLIsTagObject(tag))
		{
		    if(!strcmp((const char *)tag, "LP"))
		    {
			s = MPViewXMLInsertObjectLP(
			    v,
			    s,
			    insert_pos,
			    s_start, progress_m,
			    progress_cb, progress_data
			);
		    }
		    else if(!strcmp((const char *)tag, "TP"))
		    {
			s = MPViewXMLInsertObjectTP(
			    v,
			    s,
			    insert_pos,
			    s_start, progress_m,
			    progress_cb, progress_data
			);
		    }
		    else if(!strcmp((const char *)tag, "SH"))
		    {
			s = MPViewXMLInsertObjectSH(
			    v,
			    s,
			    insert_pos,
			    s_start, progress_m,
			    progress_cb, progress_data
			);
		    }
		    else
		    {
			s = MPViewXMLInsertObjectLP(
			    v,
			    s,
			    insert_pos,
			    s_start, progress_m,
			    progress_cb, progress_data
			);
		    }
		}
		else
		{
		    /* Tag does not describe an object but no object
		     * is in context so added a new object
		     */
		    s = MPViewXMLInsertObjectLP(
			v,
			prev_s,
			insert_pos,
			s_start, progress_m,
			progress_cb, progress_data
		    );
		}
		if(insert_pos > -1)
		    insert_pos++;

		nobjs_added++;

		g_free(tag);
	    }
	    else
	    {
		/* Default to adding a object as a LP string */
		s = MPViewXMLInsertObjectLP(
		    v,
		    s,
		    insert_pos,
		    s_start, progress_m,
		    progress_cb, progress_data
		);
		if(insert_pos > -1)
		    insert_pos++;

		nobjs_added++;
	    }
	}

	MPViewQueueResize(v->toplevel);
	MPViewQueueDraw(v->toplevel);

	return(nobjs_added);
}


/*
 *	Gets an XML string describing the the MPVObject.
 *
 *	The o specifies the MPVObject.
 *
 *	Returns a dynamically allocated string describing the MPVObject
 *	with XML tags.
 */
gchar *MPViewGetObjectXMLString(
        GtkWidget *w,
        MPVObject *o
)
{
	gchar *xml_string;
	MPView *v = MPViewGetWidgetData(
	    w,
	    "MPViewGetObjectXMLString"
	);
	if((v == NULL) || (o == NULL))
	    return(NULL);

	xml_string = NULL;

#define APPEND_CHARACTER(_c_)	{			\
 if(xml_string != NULL) {				\
  gchar *s2 = g_strdup_printf(				\
   "%s%c",						\
   xml_string,						\
   (_c_)						\
  );							\
  g_free(xml_string);					\
  xml_string = s2;					\
 } else {						\
  xml_string = g_strdup_printf("%c", (_c_));		\
 }							\
}
#define APPEND_STRING(_s_)	{			\
 if((_s_) != NULL) {					\
  if(xml_string != NULL) {				\
   gchar *s2 = g_strconcat(xml_string, (_s_), NULL);	\
   g_free(xml_string);					\
   xml_string = s2;					\
  } else {						\
   xml_string = g_strdup(_s_);				\
  }							\
 }							\
}
	switch(o->type)
	{
	  case MPV_OBJECT_LEFT_PARAGRAPH:
	    if(o->strings_list != NULL)
	    {
		MPVString *ms;

		APPEND_STRING("<LP>\n");

		ms = MPV_STRING(g_list_nth_data(
		    o->strings_list,
		    0
		));
		if(ms != NULL)
		{
		    if(ms->ca_string != NULL)
		    {
			CAStringAttributes	attr,
						cur_attr = 0;
			CAString *ca_string_ptr = ms->ca_string;
			gchar c = CA_STRING_CHARACTER(ca_string_ptr);
			while(c != '\0')
			{
			    /* Change in attribute? */
			    attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
			    if(attr != cur_attr)
			    {
				/* Append XML tag to change attribute */
				if(attr & CA_STRING_ATTRIBUTE_BOLD)
				{
				    APPEND_STRING("<fB>");
				    cur_attr = attr;
				}
				else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				{
				    APPEND_STRING("<fI>");
				    cur_attr = attr;
				}
				else
				{
				    if(cur_attr & CA_STRING_ATTRIBUTE_BOLD)
				    {
					APPEND_STRING("<fR>");
				    }
				    else if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				    {
					APPEND_STRING("<fR>");
				    }
				    cur_attr = attr;
				}
			    }

			    if(c == CA_STRING_CHARACTER_LINE_BREAK)
			    {
				APPEND_STRING("<br>");
			    }
			    else if(c == CA_STRING_CHARACTER_LINE_END)
			    {
				APPEND_STRING(" ");
			    }
			    else if(c == '&')
			    {
				APPEND_STRING("&amp;");
			    }
			    else if(c == '<')
			    {
				APPEND_STRING("&lt;");
			    }
			    else if(c == '>')
			    {
				APPEND_STRING("&gt;");
			    }
			    else
			    {
				APPEND_CHARACTER(c);
			    }

			    ca_string_ptr++;
			    c = CA_STRING_CHARACTER(ca_string_ptr);
			}
		    }
		    APPEND_STRING("\n"); 
		}
	    }
	    break;

	  case MPV_OBJECT_ITEM:
	    if(o->strings_list != NULL)
	    {
		MPVString *ms;

		APPEND_STRING("<TP>\n");

		/* Item */
		ms = MPV_STRING(g_list_nth_data(
		    o->strings_list,
		    0
		));
		if(ms != NULL)
		{
		    if(ms->ca_string != NULL)
		    {
			CAStringAttributes	attr,
						cur_attr = 0;
			CAString *ca_string_ptr = ms->ca_string;
			gchar c = CA_STRING_CHARACTER(ca_string_ptr);
			while(c != '\0')
			{
			    /* Change in attribute? */
			    attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
			    if(attr != cur_attr)
			    {
				/* Append XML tag to change attribute */
				if(attr & CA_STRING_ATTRIBUTE_BOLD)
				{
				    APPEND_STRING("<fB>");
				    cur_attr = attr;
				}
				else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				{
				    APPEND_STRING("<fI>");
				    cur_attr = attr;
				}
				else
				{
				    if(cur_attr & CA_STRING_ATTRIBUTE_BOLD)
				    {
					APPEND_STRING("<fR>");
				    }
				    else if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				    {
					APPEND_STRING("<fR>");
				    }
				    cur_attr = attr;
				}
			    }

			    if(c == CA_STRING_CHARACTER_LINE_BREAK)
			    {
				APPEND_STRING("<br>");
			    }
			    else if(c == CA_STRING_CHARACTER_LINE_END)
			    {
				APPEND_STRING(" ");
			    }
			    else if(c == '&')
			    {
				APPEND_STRING("&amp;");
			    }
			    else if(c == '<')
			    {
				APPEND_STRING("&lt;");
			    }
			    else if(c == '>')
			    {
				APPEND_STRING("&gt;");
			    }
			    else
			    {
				APPEND_CHARACTER(c);
			    }

			    ca_string_ptr++;
			    c = CA_STRING_CHARACTER(ca_string_ptr);
			}
		    }
		    APPEND_STRING("\n"); 
		}

		/* Value */
		ms = MPV_STRING(g_list_nth_data(
		    o->strings_list,
		    1
		));
		if(ms != NULL)
		{
		    if(ms->ca_string != NULL)
		    {
			CAStringAttributes	attr,
						cur_attr = 0;
			CAString *ca_string_ptr = ms->ca_string;
			gchar c = CA_STRING_CHARACTER(ca_string_ptr);
			while(c != '\0')
			{
			    /* Change in attribute? */
			    attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
			    if(attr != cur_attr)
			    {
				/* Append XML tag to change attribute */
				if(attr & CA_STRING_ATTRIBUTE_BOLD)
				{
				    APPEND_STRING("<fB>");
				    cur_attr = attr;
				}
				else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				{
				    APPEND_STRING("<fI>");
				    cur_attr = attr;
				}
				else
				{
				    if(cur_attr & CA_STRING_ATTRIBUTE_BOLD)
				    {
					APPEND_STRING("<fR>");
				    }
				    else if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				    {
					APPEND_STRING("<fR>");
				    }
				    cur_attr = attr;
				}
			    }

			    if(c == CA_STRING_CHARACTER_LINE_BREAK)
			    {
				APPEND_STRING("<br>");
			    }
			    else if(c == CA_STRING_CHARACTER_LINE_END)
			    {
				APPEND_STRING(" ");
			    }
			    else if(c == '&')
			    {
				APPEND_STRING("&amp;");
			    }
			    else if(c == '<')
			    {
				APPEND_STRING("&lt;");
			    }
			    else if(c == '>')
			    {
				APPEND_STRING("&gt;");
			    }
			    else
			    {
				APPEND_CHARACTER(c);
			    }

			    ca_string_ptr++;
			    c = CA_STRING_CHARACTER(ca_string_ptr);
			}
		    }
		    APPEND_STRING("\n"); 
		}
	    }
	    break;

	  case MPV_OBJECT_HEADING:
	    if(o->strings_list != NULL)
	    {
		MPVString *ms;

		APPEND_STRING("<SH>\n");

		ms = MPV_STRING(g_list_nth_data(
		    o->strings_list,
		    0
		));
		if(ms != NULL)
		{
		    if(ms->ca_string != NULL)
		    {
			CAStringAttributes	attr,
						cur_attr = 0;
			CAString *ca_string_ptr = ms->ca_string;
			gchar c = CA_STRING_CHARACTER(ca_string_ptr);
			while(c != '\0')
			{
			    /* Change in attribute? */
			    attr = CA_STRING_ATTRIBUTE(ca_string_ptr);
			    if(attr != cur_attr)
			    {
				/* Append XML tag to change attribute */
				if(attr & CA_STRING_ATTRIBUTE_BOLD)
				{
				    APPEND_STRING("<fB>");
				    cur_attr = attr;
				}
				else if(attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				{
				    APPEND_STRING("<fI>");
				    cur_attr = attr;
				}
				else
				{
				    if(cur_attr & CA_STRING_ATTRIBUTE_BOLD)
				    {
					APPEND_STRING("<fR>");
				    }
				    else if(cur_attr & CA_STRING_ATTRIBUTE_UNDERLINE)
				    {
					APPEND_STRING("<fR>");
				    }
				    cur_attr = attr;
				}
			    }

			    if(c == CA_STRING_CHARACTER_LINE_BREAK)
			    {
				APPEND_STRING("<br>");
			    }
			    else if(c == CA_STRING_CHARACTER_LINE_END)
			    {
				APPEND_STRING(" ");
			    }
			    else if(c == '&')
			    {
				APPEND_STRING("&amp;");
			    }
			    else if(c == '<')
			    {
				APPEND_STRING("&lt;");
			    }
			    else if(c == '>')
			    {
				APPEND_STRING("&gt;");
			    }
			    else
			    {
				APPEND_CHARACTER(c);
			    }

			    ca_string_ptr++;
			    c = CA_STRING_CHARACTER(ca_string_ptr);
			}
		    }
		    APPEND_STRING("\n"); 
		}
	    }
	    break;

	}

#undef APPEND_CHARACTER
#undef APPEND_STRING

	return(xml_string);
}
