#include <gtk/gtk.h>

#include "guiutils.h"
#include "pulist.h"
#include "tool_bar.h"
#include "cdialog.h"
#include "fb.h"
#include "fsd.h"
#include "csd.h"
#include "cfg_win.h"

#include "me_types.h"

#include "options_win.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_op.h"

#include "me_cfg_list.h"
#include "config.h"

#include "images/icon_tool_bar_20x20.xpm"
#include "images/icon_fonts_20x20.xpm"
#include "images/icon_colors_20x20.xpm"

#include "images/icon_add_20x20.xpm"
#include "images/icon_edit_20x20.xpm"
#include "images/icon_remove_20x20.xpm"

#include "images/icon_options_48x48.xpm"


/* Callbacks */
static void MEOptionsWinBrowsePathCB(GtkWidget *widget, gpointer data);

static void MEOptionsWinSyntaxHighlightingCB(GtkWidget *widget, gpointer data);

static void MEOptionsWinManualPagePathsAddCB(GtkWidget *widget, gpointer data);
static void MEOptionsWinManualPagePathsEditCB(GtkWidget *widget, gpointer data);
static void MEOptionsWinManualPagePathsRemoveCB(GtkWidget *widget, gpointer data);

static void MEOptionsWinManualPagePathsCListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

/* Build Pages */
static void MEOptionsWinBuildGeneralPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
);
static void MEOptionsWinBuildEditorPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
);
static void MEOptionsWinBuildViewerPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
);
static void MEOptionsWinBuildLocationsPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
);
static void MEOptionsWinBuildCommandsPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
);
static void MEOptionsWinBuildPages(
	me_core_struct *core,
	GtkWidget *cfg_win
);

GtkWidget *MEOptionsWinNew(me_core_struct *core);
void MEOptionsWinSelectPage(
	GtkWidget *cfg_win,
	const gchar *page_name
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Browse path callback.
 */
static void MEOptionsWinBrowsePathCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint		npaths = 0,
			nftypes = 0;
	gchar **paths_list;
	GtkWidget *cfg_win;
	fb_type_struct	**ftypes_list = NULL,
			*ftype_rtn = NULL;
	GtkEntry *entry = (GtkEntry *)data;
	if(entry == NULL)
	    return;

	cfg_win = gtk_widget_get_toplevel(widget);

	CfgWinSetBusy(cfg_win, TRUE);

	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*", "All files"
	);

	FileBrowserSetTransientFor(cfg_win);
	response = FileBrowserGetResponse(
	    "Select Path",
	    "Select", "Cancel",
	    gtk_entry_get_text(entry),
	    ftypes_list, nftypes,
	    &paths_list, &npaths, &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gchar *path = STRDUP(paths_list[npaths - 1]);
	    if(path != NULL)
	    {
		gtk_entry_set_text(entry, path);
		g_free(path);
	    }
	}

	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	CfgWinSetBusy(cfg_win, FALSE);
}


/*
 *	Syntax Highlighting "toggled" signal callback.
 */
static void MEOptionsWinSyntaxHighlightingCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	me_core_struct *core;
	GtkWidget *cfg_win = (GtkWidget *)data;
	if((widget == NULL) || (cfg_win == NULL))
	    return;

	if(!GTK_WIDGET_MAPPED(widget))
	    return;

	toplevel = cfg_win;

	core = ME_CORE(gtk_object_get_data(
	    GTK_OBJECT(cfg_win),
	    ME_CORE_KEY
	));
	if(core == NULL)
	    return;

	/* If turning on syntax highlighting then warn the user about
	 * problems with the GtkText widget
	 */
	if(GTK_TOGGLE_BUTTON(widget)->active)
	{
	    gchar       *prog_tmp_path = MECreateTemporaryDirectory(core),
			*msg = g_strdup_printf(
"This program makes use of the GTK+ GtkText widget which\n\
has known bugs that may cause applications to crash\n\
under certain circumstances (such as syntax highlighting).\n\
\n\
You are advised not to use syntax highlighting, however\n\
if you do and a crash does occure then this program will\n\
try its best to save any opened documents into the\n\
program's tempory directory:\n\
\n\
    %s",
		prog_tmp_path
	    );
	    g_free(prog_tmp_path);

	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Syntax Highlighting Warning",
"Enabling syntax highlighting may cause crashes due to\n\
bugs in the GTK+ GtkText widget.\n\
\n\
Click on `help' for explaination.",
		msg,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    g_free(msg);
	    CDialogSetTransientFor(NULL);
	}
}


/*
 *	Add Manual Page Path callback.
 */
static void MEOptionsWinManualPagePathsAddCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint		npaths = 0,
			nftypes = 0;
	gchar **paths_list;
	const gchar *path = NULL;
	GtkCList *clist;
	CfgWinLink *cfg_link;
	fb_type_struct	**ftypes_list = NULL,
			*ftype_rtn = NULL;
	GtkWidget *cfg_win = (GtkWidget *)data;
	if(cfg_win == NULL)
	    return;

	cfg_link = CfgWinGetCfgWinLinkByParameter(
	    cfg_win,
	    ME_CFG_PARM_PATHS_MANPAGES
	);
	if(cfg_link == NULL)
	    return;

	clist = (GtkCList *)CfgWinLinkGetWidget(cfg_link);
	if(clist == NULL)
	    return;

	CfgWinSetBusy(cfg_win, TRUE);

	if(clist->selection_end != NULL)
	{
	    GList *glist = clist->selection_end;
	    path = (const gchar *)gtk_clist_get_row_data(
		clist,
		(gint)glist->data
	    );
	}

	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*", "All files"
	);

	FileBrowserSetTransientFor(cfg_win);
	response = FileBrowserGetResponse(
	    "Add New Manual Page Path",
	    "Add", "Cancel",
	    path,
	    ftypes_list, nftypes,
	    &paths_list, &npaths, &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gchar *path = STRDUP(paths_list[npaths - 1]);
	    if(path != NULL)
	    {
		const gint ncolumns = MAX(clist->columns, 1);
		gint	i,
			row;
		gchar **strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));

		for(i = 0; i < ncolumns; i++)
		    strv[i] = "";

		strv[0] = path;

		gtk_clist_freeze(clist);
		if(clist->selection_end != NULL)
		{
		    GList *glist = clist->selection_end;
		    row = gtk_clist_insert(
			clist,
			(gint)glist->data,
			strv
		    );
		}
		else
		{
		    row = gtk_clist_append(
			clist,
			strv
		    );
		}
		if(row > -1)
		{
		    gtk_clist_set_row_data_full(
			clist,
			row,
			g_strdup(path), (GtkDestroyNotify)g_free
		    );
		}
		gtk_clist_thaw(clist);

		g_free(strv);

		CfgWinSetHasChanges(cfg_win, TRUE);
		CfgWinUpdateMenus(cfg_win);

		g_free(path);
	    }
	}
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	CfgWinSetBusy(cfg_win, FALSE);
}

/*
 *	Edit Manual Page Path callback.
 */
static void MEOptionsWinManualPagePathsEditCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint row = -1, nftypes = 0, npaths;
	gchar **paths_list;
	const gchar *path = NULL;
	GtkCList *clist;
	CfgWinLink *cfg_link;
	fb_type_struct **ftypes_list = NULL, *ftype_rtn = NULL;
	GtkWidget *cfg_win = (GtkWidget *)data;
	if(cfg_win == NULL)
	    return;

	cfg_link = CfgWinGetCfgWinLinkByParameter(
	    cfg_win,
	    ME_CFG_PARM_PATHS_MANPAGES
	);
	if(cfg_link == NULL)
	    return;

	clist = (GtkCList *)CfgWinLinkGetWidget(cfg_link);
	if(clist == NULL)
	    return;

	CfgWinSetBusy(cfg_win, TRUE);

	if(clist->selection_end != NULL)
	{
	    GList *glist = clist->selection_end;
	    row = (gint)glist->data;
	    path = (const gchar *)gtk_clist_get_row_data(
		clist,
		row
	    );
	}
	if(path == NULL)
	{
	    CfgWinSetBusy(cfg_win, FALSE);
	    return;
	}

	FileBrowserTypeListNew(
	    &ftypes_list, &nftypes,
	    "*.*", "All files"
	);

	FileBrowserSetTransientFor(cfg_win);
	response = FileBrowserGetResponse(
	    "Edit Manual Page Path",
	    "Set", "Cancel",
	    path,
	    ftypes_list, nftypes,
	    &paths_list, &npaths, &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	if(response && (npaths > 0))
	{
	    gchar *path = STRDUP(paths_list[npaths - 1]);
	    if(path != NULL)
	    {
		gtk_clist_freeze(clist);
		gtk_clist_set_text(
		    clist,
		    row, 0,
		    path
		);
		gtk_clist_set_row_data_full(
		    clist,
		    row,
		    g_strdup(path), (GtkDestroyNotify)g_free
		);
		gtk_clist_thaw(clist);

		CfgWinSetHasChanges(cfg_win, TRUE);
		CfgWinUpdateMenus(cfg_win);

		g_free(path);
	    }
	}
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	CfgWinSetBusy(cfg_win, FALSE);
}

/*
 *	Remove Manual Page Path callback.
 */
static void MEOptionsWinManualPagePathsRemoveCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	CfgWinLink *cfg_link;
	GtkWidget *cfg_win = (GtkWidget *)data;
	if(cfg_win == NULL)
	    return;

	cfg_link = CfgWinGetCfgWinLinkByParameter(
	    cfg_win,
	    ME_CFG_PARM_PATHS_MANPAGES
	);
	if(cfg_link == NULL)
	    return;

	clist = (GtkCList *)CfgWinLinkGetWidget(cfg_link);
	if(clist == NULL)
	    return;

	CfgWinSetBusy(cfg_win, TRUE);

	gtk_clist_freeze(clist);
	while(clist->selection != NULL)
	{
	    glist = clist->selection;
	    row = (gint)glist->data;
	    gtk_clist_remove(clist, row);
	}
	gtk_clist_thaw(clist);

	CfgWinSetBusy(cfg_win, FALSE);
}

/*
 *	Manual Path Paths GtkCList "select_row" signal callbakc.
 */
static void MEOptionsWinManualPagePathsCListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	GtkWidget *cfg_win = (GtkWidget *)data;
	if((clist == NULL) || (cfg_win == NULL))
	    return;

	if(event != NULL)
	{
	    if(event->type == GDK_2BUTTON_PRESS)
	    {
		MEOptionsWinManualPagePathsEditCB(
		    GTK_WIDGET(clist),
		    cfg_win
		);
	    }
	}
}


/*
 *	Builds the General Page.
 */
static void MEOptionsWinBuildGeneralPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
)
{
	const gint border_major = 5;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent, *parent2,
			*cfg_page;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* General Page */
	pixmap = pixmaps_list->manedit_20x20;
	mask = pixmaps_list->manedit_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "General",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Startup With GtkFrame */
	w = gtk_frame_new("Startup With");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Editor */
	w = gtk_check_button_new_with_label("Editor");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
	    w,
	    ME_CFG_PARM_STARTUP_EDITOR
	);
	gtk_widget_show(w);

	/* Viewer */
	w = gtk_check_button_new_with_label("Viewer");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
	    w,
	    ME_CFG_PARM_STARTUP_VIEWER
	);
	gtk_widget_show(w);

	/* Window Placement GtkFrame */
	w = gtk_frame_new("Window Placement");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Record Last Position & Size */
	w = gtk_check_button_new_with_label("Record Last Position & Size");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
	    w,
	    ME_CFG_PARM_RECORD_WIN_POS
	);
	gtk_widget_show(w);


	/* Tooltips */
	w = gtk_frame_new("Tooltips");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Show Tooltips */
	w = gtk_check_button_new_with_label("Show Tooltips");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
	    w,
	    ME_CFG_PARM_SHOW_TOOLTIPS
	);
	gtk_widget_show(w);
}

/*
 *	Builds the Editor Page.
 */
static void MEOptionsWinBuildEditorPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
)
{
	const gint	border_major = 5,
			label_width = 100,
			entry_width = 250;
	GSList *group;
	GdkWindow *window = cfg_win->window;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent, *parent2,
			*cfg_page;
	CfgWinLink *cfg_link;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* Editor Page */
	pixmap = pixmaps_list->editor_20x20;
	mask = pixmaps_list->editor_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Editor",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    TRUE				/* Is a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Display */
	w = gtk_frame_new("Display");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Syntax Highlighting */
	w = gtk_check_button_new_with_label("Syntax Highlighting");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(MEOptionsWinSyntaxHighlightingCB), cfg_win
	);
	CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING
	);
	gtk_widget_show(w);




	/* Build the subpages */
	cfg_parent_page = cfg_page;

	/* Tool Bar Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_tool_bar_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Tool Bar",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Button Decals */
	w = gtk_frame_new("Button Decals");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Button Decals GtkRadioButtons */
	group = NULL;
	w = gtk_radio_button_new_with_label(group, "Text");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_TEXT
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Pictures");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Pictures & Text");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES_AND_TEXT
	);
	gtk_widget_show(w);


	/* Relief */
	w = gtk_frame_new("Relief");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Relief GtkRadioButtons */
	group = NULL;
	w = gtk_radio_button_new_with_label(group, "Normal");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_NORMAL
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Half");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_HALF
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "None");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_NONE
	);
	gtk_widget_show(w);


	/* Fonts Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_fonts_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Fonts",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Editable */
	w = gtk_frame_new("Editable");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = FSDPromptNewSimple(
	    "Normal:",
	    label_width,
	    entry_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    FSDPromptGetEntry(w),
	    ME_CFG_PARM_FONT_EDIT_STD
	);
	gtk_widget_show(w);


	/* Colors Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_colors_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Colors",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Editable */
	w = gtk_frame_new("Editable");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Normal */
	w = CSDColorButtonNewSimple(
	    "Normal:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_EDITOR_FG
	);
	gtk_widget_show(w);

	/* Deliminators */
	w = CSDColorButtonNewSimple(
	    "Deliminators:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_EDITOR_DELIMINATOR
	);
	gtk_widget_show(w);

	/* Tags */
	w = CSDColorButtonNewSimple(
	    "Tags:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_EDITOR_TAG
	);
	gtk_widget_show(w);

	/* Symbols */
	w = CSDColorButtonNewSimple(
	    "Symbols:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_EDITOR_SYMBOL
	);
	gtk_widget_show(w);

	/* Background */
	w = CSDColorButtonNewSimple(
	    "Background:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_EDITOR_BG
	);
	gtk_widget_show(w);
}

/*
 *	Builds the Viewer Page.
 */
static void MEOptionsWinBuildViewerPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
)
{
	const gint	border_major = 5,
			border_minor = 2,
			label_width = 100,
			entry_width = 250;
	GSList *group;
	GdkWindow *window = cfg_win->window;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent, *parent2, *parent3,
			*cfg_page,
			*pulist;
	CfgWinLink *cfg_link;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* Viewer Page */
	pixmap = pixmaps_list->viewer_20x20;
	mask = pixmaps_list->viewer_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Viewer",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    TRUE				/* Is a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Display */
	w = gtk_frame_new("Display");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_label_new("Startup Page:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Startup Page Popup List Box */
	w = PUListBoxNew(
	    150, -1
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	pulist = PUListBoxGetPUList(w);
	gtk_widget_show(w);
	if(pulist != NULL)
	{
	    gint i;
#define PULIST_APPEND(_s_,_d_)	{		\
 i = PUListAddItem(pulist, (_s_));		\
 PUListSetItemData(				\
  pulist, i,					\
  (gpointer)(_d_)				\
 );						\
}
	    PULIST_APPEND("Manual Page", VIEWER_PAGE_MANPAGE);
	    PULIST_APPEND("Index", VIEWER_PAGE_INDEX);
#undef PULIST_APPEND
	    PUListBoxSetLinesVisible(
		w,
		MIN((i + 1), 10)
	    );
	}
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_POPUP_LIST_BOX,
	    w,
	    ME_CFG_PARM_VIEWER_STARTUP_PAGE
	);




	cfg_parent_page = cfg_page;

	/* Tool Bar Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_tool_bar_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Tool Bar",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Button Decals */
	w = gtk_frame_new("Button Decals");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Button Decals GtkRadioButtons */
	group = NULL;
	w = gtk_radio_button_new_with_label(group, "Text");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_TEXT
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Pictures");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Pictures & Text");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES_AND_TEXT
	);
	gtk_widget_show(w);

	/* Relief */
	w = gtk_frame_new("Relief");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Relief GtkRadioButtons */
	group = NULL;
	w = gtk_radio_button_new_with_label(group, "Normal");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_NORMAL
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "Half");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_HALF
	);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(group, "None");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
	    w,
	    ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
	    cfg_link,
	    (gpointer)GTK_RELIEF_NONE
	);
	gtk_widget_show(w);


	/* Fonts Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_fonts_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Fonts",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Manual Page */
	w = gtk_frame_new("Manual Page");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = FSDPromptNewSimple(
	    "Normal:",
	    label_width,
	    entry_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    FSDPromptGetEntry(w),
	    ME_CFG_PARM_FONT_MANPAGE_STD
	);
	gtk_widget_show(w);

	w = FSDPromptNewSimple(
	    "Bold:",
	    label_width,
	    entry_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    FSDPromptGetEntry(w),
	    ME_CFG_PARM_FONT_MANPAGE_BOLD
	);
	gtk_widget_show(w);

	w = FSDPromptNewSimple(
	    "Underline:",
	    label_width,
	    entry_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    FSDPromptGetEntry(w),
	    ME_CFG_PARM_FONT_MANPAGE_UNDERLINE
	);
	gtk_widget_show(w);


	/* Colors Page */
	pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    NULL,
	    (gchar **)icon_colors_20x20_xpm
	);
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Colors",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Manual Page */
	w = gtk_frame_new("Manual Page");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Normal */
	w = CSDColorButtonNewSimple(
	    "Normal:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_MANPAGE_FG
	);
	gtk_widget_show(w);

	/* Bold */
	w = CSDColorButtonNewSimple(
	    "Bold:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_MANPAGE_BOLD
	);
	gtk_widget_show(w);

	/* Underline */
	w = CSDColorButtonNewSimple(
	    "Underline:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_MANPAGE_UNDERLINE
	);
	gtk_widget_show(w);

	/* Background */
	w = CSDColorButtonNewSimple(
	    "Background:",
	    label_width
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_COLOR_BUTTON,
	    w,
	    ME_CFG_PARM_COLOR_MANPAGE_BG
	);
	gtk_widget_show(w);
}

/*
 *	Builds the Locations Page.
 */
static void MEOptionsWinBuildLocationsPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
)
{
	const gint	border_minor = 2,
			border_major = 5;
	gpointer	entry_rtn,
			browse_rtn;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*cfg_page;
	GtkCList *clist;
	CfgWinLink *cfg_link;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* Locations Page */
	pixmap = pixmaps_list->options_20x20;
	mask = pixmaps_list->options_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Locations",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    TRUE				/* Is a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Directories */
	w = gtk_frame_new("Directories");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* ManEdit Global */
	w = GUIPromptBarWithBrowse(
	    NULL,
	    "ManEdit Global:",
	    NULL,
	    &entry_rtn,
	    &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    entry_rtn,
	    ME_CFG_PARM_PATH_GLOBAL_DATA
	);
	gtk_widget_show(w);

	w = (GtkWidget *)browse_rtn;
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinBrowsePathCB), entry_rtn
	);


	/* Temporary */
	w = GUIPromptBarWithBrowse(
	    NULL,
	    "Temporary:",
	    NULL,
	    &entry_rtn,
	    &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    entry_rtn,
	    ME_CFG_PARM_PATH_TMP
	);
	gtk_widget_show(w);

	w = (GtkWidget *)browse_rtn;
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinBrowsePathCB), entry_rtn
	);



	cfg_parent_page = cfg_page;

	/* Manual Page Paths Page */
	pixmap = pixmaps_list->options_20x20;
	mask = pixmaps_list->options_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Manual Page Paths",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Description GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Description GtkLabel */
	w = gtk_label_new(
"These paths refer to the directories on your computer that contain\n\
manual pages. They are used by the Viewer's Index Page to list all the\n\
available manual pages."
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* GtkHBox for the GtkCList and Buttons */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkScrolledWindow for the GtkCList */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkCList */
	w = gtk_clist_new(1);
	clist = GTK_CLIST(w);
	gtk_signal_connect(
	    GTK_OBJECT(w), "select_row",
	    GTK_SIGNAL_FUNC(MEOptionsWinManualPagePathsCListSelectRowCB), cfg_win
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	gtk_clist_set_row_height(clist, 20);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_clist_set_column_auto_resize(clist, 0, TRUE);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_CLIST,
	    w,
	    ME_CFG_PARM_PATHS_MANPAGES
	);
	gtk_widget_show(w);


	/* GtkVBox for the Buttons */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkVBox for the Add, Edit, and Remove Buttons */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Add GtkButton */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_add_20x20_xpm,
	    "Add",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinManualPagePathsAddCB), cfg_win
	);
	GUISetWidgetTip(
	    w,
	    "Add a new manual page path"
	);
	gtk_widget_show(w);

	/* Edit GtkButton */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_edit_20x20_xpm,
	    "Edit",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinManualPagePathsEditCB), cfg_win
	);
	GUISetWidgetTip(
	    w,
	    "Edit the selected manual page path"
	);
	gtk_widget_show(w);

	/* Remove GtkButton */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_remove_20x20_xpm,
	    "Remove",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinManualPagePathsRemoveCB), cfg_win
	);
	GUISetWidgetTip(
	    w,
	    "Remove the selected manual page path(s)"
	);
	gtk_widget_show(w);
}

/*
 *	Builds the Commands Page.
 */
static void MEOptionsWinBuildCommandsPage(
	me_core_struct *core,
	GtkWidget *cfg_win,
	GtkWidget *cfg_parent_page
)
{
	const gint border_major = 5;
	gpointer	entry_rtn,
			browse_rtn;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent, *parent2,
			*cfg_page;
	CfgWinLink *cfg_link;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* Commands Page */
	pixmap = pixmaps_list->options_20x20;
	mask = pixmaps_list->options_20x20_mask;
	cfg_page = CfgWinAddPageTree(
	    cfg_win,
	    cfg_parent_page,
	    NULL,				/* Append */
	    "Commands",
	    pixmap, mask,
	    pixmap, mask,
	    TRUE,				/* Create heading */
	    FALSE				/* Not a parent */
	);
	parent = cfg_page;
	gtk_box_set_spacing(GTK_BOX(parent), border_major);

	/* Manual Page Interpreter */
	w = gtk_frame_new("Manual Page Interpreter");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = GUIPromptBarWithBrowse(
	    NULL,
	    NULL,
	    NULL,
	    &entry_rtn,
	    &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    (GtkWidget *)entry_rtn,
"Enter the command to run the Manual Page Interpreter, the\
 program that displays Manual Pages to a terminal, available\
 substitutions are; %p specifies the section number and %f\
 specifies the Manual Page file."
	);
	cfg_link = CfgWinAddLink(
	    cfg_win,
	    cfg_page,
	    CFG_WIN_LINK_WIDGET_EDITABLE,
	    entry_rtn,
	    ME_CFG_PARM_CMD_MANPAGE_INTERPRETER
	);
	gtk_widget_show(w);

	w = (GtkWidget *)browse_rtn;
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MEOptionsWinBrowsePathCB), entry_rtn
	);
}

/*
 *	Builds the Options Window pages.
 */
static void MEOptionsWinBuildPages(
	me_core_struct *core,
	GtkWidget *cfg_win
)
{
	if(!GTK_WIDGET_REALIZED(cfg_win))
	    gtk_widget_realize(cfg_win);

	MEOptionsWinBuildGeneralPage(
	    core,
	    cfg_win,
	    NULL
	);
	MEOptionsWinBuildEditorPage(
	    core,
	    cfg_win,
	    NULL
	);
	MEOptionsWinBuildViewerPage(
	    core,
	    cfg_win,
	    NULL
	);
	MEOptionsWinBuildLocationsPage(
	    core,
	    cfg_win,
	    NULL
	);
	MEOptionsWinBuildCommandsPage(
	    core,
	    cfg_win,
	    NULL
	);
}


/*
 *	Creates a new Options Window.
 */
GtkWidget *MEOptionsWinNew(me_core_struct *core)
{
	GtkWidget *cfg_win;

	if(core == NULL)
	    return(NULL);

	cfg_win = CfgWinNew(
	    CFG_WIN_STYLE_TREE,
	    PROG_NAME ": Options",
	    (guint8 **)icon_options_48x48_xpm,
	    CFG_WIN_BUTTON_OK | CFG_WIN_BUTTON_APPLY |
		CFG_WIN_BUTTON_CANCEL | CFG_WIN_BUTTON_CLOSE,
	    ME_OPTIONS_WIN_DEF_WIDTH, ME_OPTIONS_WIN_DEF_HEIGHT,
	    core->cfg_list,
	    core->cfg_file,
	    0,
	    NULL
	);
	if(cfg_win == NULL)
	    return(NULL);

	gtk_object_set_data(
	    GTK_OBJECT(cfg_win),
	    ME_CORE_KEY,
	    core
	);

	CfgWinSetApplyCB(
	    cfg_win,
	    MEOptionsWinApplyCB,
	    core
	);

	MEOptionsWinBuildPages(core, cfg_win);

	CfgWinUpdateMenus(cfg_win);

	return(cfg_win);
}

/*
 *	Selects the page on the Options Window.
 */
void MEOptionsWinSelectPage(
	GtkWidget *cfg_win,
	const gchar *page_name
)
{
	GtkWidget *cfg_win_page;

	if(cfg_win == NULL)
	    return;

	/* Find the page who's name matches the specified page name */
	if(page_name != NULL)
	    cfg_win_page = CfgWinGetPageByLabel(
		cfg_win,
		NULL,				/* Search from toplevel */
		page_name,
		TRUE				/* Search recursively */
	    );
	else
	    cfg_win_page = CfgWinGetDefaultPage(cfg_win);
	if(cfg_win_page == NULL)
	    return;

	CfgWinSetPage(
	    cfg_win,
	    cfg_win_page
	);
}
