#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/disk.h"
#include "../include/string.h"

#include "cfg.h"

#include "guiutils.h"
#include "cfg_gtk.h"
#include "pulist.h"
#include "tool_bar.h"
#include "fb.h"

#include "viewer.h"
#include "viewer_cb.h"
#include "viewer_dnd.h"
#include "editor_dnd.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_op.h"
#include "messages.h"

#include "me_cfg_list.h"
#include "config.h"


#include "images/icon_manedit_editor_16x16.xpm"
#include "images/icon_manedit_viewer_16x16.xpm"
#include "images/icon_manedit_editor_20x20.xpm"
#include "images/icon_manedit_viewer_48x48.xpm"

#include "images/icon_manual_opened_20x20.xpm"
#include "images/icon_search_20x20.xpm"
#include "images/icon_open_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_clear_20x20.xpm"
#include "images/icon_copy_20x20.xpm"
#include "images/icon_stop_20x20.xpm"
#include "images/icon_exit_20x20.xpm"
#include "images/icon_folder_parent_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_options_20x20.xpm"


/* Index GtkCTree Item */
ViewerIndexItem *ViewerIndexItemNew(void);
void ViewerIndexItemDelete(ViewerIndexItem *item);

/* Section */
void ViewerSetSection(
	viewer_struct *v,
	const gint section_num
);
gint ViewerCurrentSection(viewer_struct *v);

/* View Text */
void ViewerTextInsertPosition(
	viewer_struct *v,
	const gint start_pos
);
void ViewerTextInsert(
	viewer_struct *v,
	const gchar *buf, const gint buf_len,
	gint (*progress_cb)(const gulong, const gulong, gpointer),
	gpointer progress_data
);
void ViewerTextDelete(
	viewer_struct *v,
	const gint start_pos, const gint end_pos
);

/* Index GtkCTree */
static gint ViewerIndexLoadCountIterate(const gchar *path);
static GtkCTreeNode *ViewerIndexLoadIterate(
	viewer_struct *v,
	GtkCTree *ctree,
	GtkCTreeNode *parent,
	GtkCTreeNode *sibling,
	const gchar *path,
	gint *nobjs_loaded,
	const gint nobjs
);
void ViewerIndexLoad(
	viewer_struct *v,
	GList *paths_list
);
void ViewerIndexRefresh(viewer_struct *v);

/* Find */
gboolean ViewerDoFind(
	viewer_struct *v, GtkText *text,
	gchar *haystack, gchar *needle,
	gint haystack_len, gint start_pos,
	gboolean case_sensitive,
	gboolean move_to,
	gboolean *search_wrapped
);

/* Record Scroll Position */
void ViewerViewTextRecordScrollPositions(viewer_struct *v);

/* Viewer */
static void ViewerCreateMenuBar(
	viewer_struct *v,
	GtkWidget *parent
);
static void ViewerCreateToolBar(
	viewer_struct *v,
	GtkWidget *parent
);
static void ViewerCreateStatusBar(
	viewer_struct *v,
	GtkWidget *parent
);
viewer_struct *ViewerNew(
	me_core_struct *core,
	const gint editor_num
);
void ViewerRecordConfiguration(viewer_struct *v);
void ViewerUpdateTitle(viewer_struct *v);
void ViewerUpdateMenus(viewer_struct *v);
void ViewerSetPage(
	viewer_struct *v,
	viewer_page page
);
void ViewerSetStatusPosition(
	viewer_struct *v,
	const gint row, const gint column
);
void ViewerSetStatusMessage(
	viewer_struct *v,
	const gchar *msg
);
void ViewerSetStatusProgress(
	viewer_struct *v,
	const gfloat val
);
void ViewerSetBusy(
	viewer_struct *v,
	const gboolean busy
);
gboolean ViewerIsMapped(viewer_struct *v);
void ViewerMap(viewer_struct *v);
void ViewerUnmap(viewer_struct *v);
void ViewerDelete(viewer_struct *v);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Creates a new Viewer Index Item.
 */
extern ViewerIndexItem *ViewerIndexItemNew(void)
{
	return(VIEWER_INDEX_ITEM(g_malloc0(sizeof(ViewerIndexItem))));
}

/*
 *	Deletes the Viewer Index Item;
 */
void ViewerIndexItemDelete(ViewerIndexItem *item)
{
	if(item == NULL)
	    return;

	g_free(item->path);
	g_free(item);
}


/*
 *	Sets the section number.
 *
 *	The section_num specifies the section number starting from 1.
 */
void ViewerSetSection(
	viewer_struct *v,
	const gint section_num
)
{
	gint i, n;
	const gchar	*s,
			*section_str;
	GtkWidget	*pulist,
			*pulistbox;

	if(v == NULL)
	    return;

	pulistbox = v->section_pulistbox;
	pulist = PUListBoxGetPUList(pulistbox);
	switch(section_num)
	{
	  case 1:
	    section_str = ME_MANPAGE_SECTION_NAME_1;
	    break;
	  case 2:
	    section_str = ME_MANPAGE_SECTION_NAME_2;
	    break;
	  case 3:
	    section_str = ME_MANPAGE_SECTION_NAME_3;
	    break;
	  case 4:
	    section_str = ME_MANPAGE_SECTION_NAME_4;
	    break;
	  case 5:
	    section_str = ME_MANPAGE_SECTION_NAME_5;
	    break;
	  case 6:
	    section_str = ME_MANPAGE_SECTION_NAME_6;
	    break;
	  case 7:
	    section_str = ME_MANPAGE_SECTION_NAME_7;
	    break;
	  case 8:
	    section_str = ME_MANPAGE_SECTION_NAME_8;
	    break;
	  case 9:
	    section_str = ME_MANPAGE_SECTION_NAME_9;
	    break;

	  case -1:
	    section_str = ME_MANPAGE_SECTION_NAME_EXACT;
	    break;

	  default:
	    section_str = ME_MANPAGE_SECTION_NAME_ANY;
	    break;
	}

	n = PUListGetTotalItems(pulist);
	for(i = 0; i < n; i++)
	{
	    s = (const gchar *)PUListGetItemData(pulist, i);
	    if(s == NULL)
		continue;

	    if(!g_strcasecmp(s, section_str))
	    {
		PUListBoxSelect(pulistbox, i);
		break;
	    }
	}
}

/*
 *	Gets the current section number.
 *
 *	Returns the current section starting from 1, -1 for "exact"
 *	(filename), -2 for "any" section, or 0 on error..
 */
gint ViewerCurrentSection(viewer_struct *v)
{
	gint section_num;
	const gchar *section;

	if(v == NULL)
	    return(0);

	section = (const gchar *)PUListBoxGetSelectedData(
	    v->section_pulistbox
	);
	if(section == NULL)
	    return(0);

	if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_1))
	{
	    section_num = 1;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_2))
	{
	    section_num = 2;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_3))
	{
	    section_num = 3;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_4))
	{
	    section_num = 4;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_5))
	{
	    section_num = 5;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_6))
	{
	    section_num = 6;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_7))
	{
	    section_num = 7;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_8))
	{
	    section_num = 8;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_9))
	{
	    section_num = 9;
	}
	else if(!g_strcasecmp(section, ME_MANPAGE_SECTION_NAME_EXACT))
	{
	    section_num = -1;
	}
	else
	{
	    /* All else assume any (-2) */
	    section_num = -2;
	}

	return(section_num);
}


/*
 *	Sets the view GtkText insert position.
 *
 *	The start_pos specifies the insert position, if start_pos is
 *	-1 then it will be set to the end.
 */
void ViewerTextInsertPosition(
	viewer_struct *v,
	const gint start_pos
)
{
	gint		_start_pos = start_pos,
			end_pos;
	GtkWidget *w;
	GtkText *text;

	if(v == NULL)
	    return;

	w = v->view_text;
	text = GTK_TEXT(w);

	/* Get length of text */
	end_pos = gtk_text_get_length(text);

	if(_start_pos < 0)
	    _start_pos = end_pos;
	if(_start_pos > end_pos)
	    _start_pos = end_pos;

	gtk_text_set_point(text, _start_pos);
}

/*
 *	Inserts the given buffer to the viewer's view text parsed.
 *
 *	If buf_len is -1 then the entire given buf will be inserted,
 *	and in which case the given buf needs to be null terminated.
 *
 *	Calling function is responsible for setting the starting
 *	text insert position.
 *
 *	If the callback function is not NULL and returns non-zero then
 *	the insert will be aborted.
 */
void ViewerTextInsert(
	viewer_struct *v,
	const gchar *buf, const gint buf_len,
	gint (*progress_cb)(const gulong, const gulong, gpointer),
	gpointer progress_data
)
{
	gint _buf_len = buf_len;
	gchar last_char;
	const gchar	*buf_ptr,
			*buf_next,
			*buf_end;
	GdkFont *font;
	GdkColor *c;
	GtkWidget *w;
	GtkText *text;
	me_core_struct *core;

	if((v == NULL) || (buf == NULL))
	    return;

	w = v->view_text;
	text = GTK_TEXT(w);
	core = v->core;
	font = core->manpage_normal_font;
	c = &v->view_normal_text_color;

	if(_buf_len < 0)
	    _buf_len = STRLEN(buf);

	if(_buf_len == 0)
	    return;


#define IS_CTL_CHAR(c)	(((c) == 0x08))

	gtk_text_freeze(text);

	/* Iterate through the buffer */
	last_char = '\0';	/* Reset last character to be printed */
	buf_ptr = buf;
	buf_end = (const gchar *)(buf + _buf_len);
	while(buf_ptr < buf_end)
	{
	    if(progress_cb != NULL)
	    {
#if 0
/* Do not call the callback while GtkText widget is frozen */
		status = progress_cb(
		    (gulong)MIN(buf_ptr - buf, _buf_len), (gulong)_buf_len,
		    progress_data
		);
		if(status != 0)
		    break;
#endif
	    }

	    /* Seek buf_next to the next control character or end of
	     * buffer
	     */
	    buf_next = buf_ptr;
	    while(buf_next < buf_end)
	    {
		if(IS_CTL_CHAR(*buf_next))
		    break;
		else
		    buf_next++;
	    }
	    /* Is the current character a control character? */
	    if(IS_CTL_CHAR(*buf_ptr))
	    {
		buf_ptr++;			/* Seek past the control char */

		/* Set the font and color
		 *
		 * Last print character an underline?
		 */
		if(last_char == '_')
		{
		    font = core->manpage_underline_font;
		    c = &v->view_underline_text_color;
		}
		/* Last print character same as this one? */
		else if((buf_ptr < buf_end) ?
		    (last_char == (*buf_ptr)) : FALSE
		)
		{
		    font = core->manpage_bold_font;
		    c = &v->view_bold_text_color;
		}
		else
		{
		    font = core->manpage_normal_font;
		    c = &v->view_normal_text_color;
		}
		if(font == NULL)
		    font = core->manpage_normal_font;

		/* Delete one character backwards */
		gtk_text_backward_delete(text, 1);

		/* Increment and insert the next character */
		if(buf_ptr < buf_end)
		{
		    gtk_text_insert(
			text,
			font,
			c,
			NULL,
			buf_ptr,
			1
		    );
		}

		/* Increment to next character */
		buf_ptr++;
	    }
	    else
	    {
		/* Store to next control character or end of buffer */

		/* Calculate this segment length */
		gint seg_len = (gint)(buf_next - buf_ptr);

		/* Use ManPage standard text style */
		c = &v->view_normal_text_color;
		font = core->manpage_normal_font;

		/* Any characters to store? */
		if(seg_len > 0)
		{
		    gtk_text_insert(
			text,
			font,
			c,
			NULL,
			buf_ptr,
			seg_len
		    );
		}

		/* Seek past this segment */
		buf_ptr = (const gchar *)(buf_ptr + MAX(seg_len, 1));

		/* Record last character to be printed */
		if((buf_ptr <= buf_end) && (buf_ptr > buf))
		    last_char = (*(buf_ptr - 1));
	    }
	}

	/* Clear the selection */
	gtk_editable_select_region(GTK_EDITABLE(text), -1, -1);

#undef IS_CTL_CHAR

	gtk_text_thaw(text);
}

/*
 *	Deletes the segment of text.
 *
 *	If end_pos is -1 then all text will be deleted from start_pos.
 */
void ViewerTextDelete(
	viewer_struct *v,
	const gint start_pos, const gint end_pos
)
{
	GtkWidget *w;

	if(v == NULL)
	    return;

	w = v->view_text;

	gtk_text_freeze(GTK_TEXT(w));
	gtk_editable_delete_text(
	    GTK_EDITABLE(w),
	    start_pos, end_pos
	);
	gtk_text_thaw(GTK_TEXT(w));
}


/*
 *	Called by ViewerIndexLoad() to recursively count the number
 *	of directories and manual pages to load.
 */
static gint ViewerIndexLoadCountIterate(const gchar *path)
{
	gint		i,
			nobjs = 0;
	const gchar *name;
	gchar **names_list;

	/* Get the list of child objects in this directory */
	names_list = GetDirEntNames(path);
	/*
	names_list may return NULL, for example when user has no
	right to access path
	*/
	for(i = 0; names_list && names_list[i]; i++)
	{
	    name = names_list[i];

	    /* Count only names of manual pages or child directories */
	    if(strcmp(name, ".") &&
	       strcmp(name, "..") &&
	       strcmp(name, "whatis")
	    )
	    {
		gchar *child_path = STRDUP(PrefixPaths(
		    path,
		    name
		));
		if(child_path != NULL)
		{
		    /* Check if this object is a directory */
		    struct stat stat_buf;
		    if(!stat((const char *)child_path, &stat_buf))
		    {
			if(S_ISDIR(stat_buf.st_mode))
			{
			    /* Count the number of manual pages and
			     * subdirectories in this directory
			     */
			    nobjs += ViewerIndexLoadCountIterate(child_path);
			}
		    }

		    nobjs++;			/* Count this object */

		    g_free(child_path);
		}
	    }

	    g_free(names_list[i]);
	}

	g_free(names_list);

	return(nobjs);
}

/*
 *	Called by ViewerIndexLoad() to recursively load the Index
 *	GtkCTree.
 */
static GtkCTreeNode *ViewerIndexLoadIterate(
	viewer_struct *v,
	GtkCTree *ctree,
	GtkCTreeNode *parent,
	GtkCTreeNode *sibling,
	const gchar *path,
	gint *nobjs_loaded,
	const gint nobjs
)
{
	struct stat stat_buf;
	gint i;
	const gchar *name;
	gchar **strv;
	GtkCList *clist = GTK_CLIST(ctree);
	const gint ncolumns = MAX(clist->columns, 1);
	GtkCTreeNode *new_branch = NULL;
	ViewerIndexItem *item;
	me_core_struct *core = v->core;
	me_pixmaps_list_struct *pixmaps_list = &core->pixmaps_list;

	/* Check for user abort */
	if(v->stop_count > 0)
	    return(new_branch);

	if(path == NULL)
	    return(new_branch);

	name = g_basename(path);
	if(name == NULL)
	    name = path;

	if(stat((const char *)path, &stat_buf))
	    return(new_branch);

	/* Allocate the row cell values */
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	if(strv == NULL)
	    return(new_branch);

	for(i = 0; i < ncolumns; i++)
	    strv[i] = NULL;

	if(S_ISDIR(stat_buf.st_mode))
	{
	    /* Use the full path for toplevel node s*/
	    if(parent == NULL)
	    {
		strv[0] = g_strdup(path);
	    }
	    /* Substitute the directory name with verbose catagory
	     * name if possible
	     */
	    else if(!strcmp((const char *)name, "man1"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_1);
	    }
	    else if(!strcmp((const char *)name, "man2"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_2);
	    }
	    else if(!strcmp((const char *)name, "man3"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_3);
	    }
	    else if(!strcmp((const char *)name, "man4"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_4);
	    }
	    else if(!strcmp((const char *)name, "man5"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_5);
	    }
	    else if(!strcmp((const char *)name, "man6"))
	    {       
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_6);
	    }
	    else if(!strcmp((const char *)name, "man7"))
	    {
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_7);
	    }
	    else if(!strcmp((const char *)name, "man8"))
	    { 
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_8);
	    }
	    else if(!strcmp((const char *)name, "man9"))
	    { 
		strv[0] = g_strdup(ME_MANPAGE_SECTION_NAME_9);
	    }
	    else
	    {
		strv[0] = g_strdup(name);
	    }

	    /* Description value */
	    if(ncolumns > 1)
		strv[1] = g_strdup("*Directory*");

	    /* Create a new branch */
	    new_branch = gtk_ctree_insert_node(
		ctree,
		parent,
		sibling,
		strv,
		ME_LIST_ICON_TEXT_SPACING,
		pixmaps_list->folder_closed_20x20,
		pixmaps_list->folder_closed_20x20_mask,
		pixmaps_list->folder_opened_20x20,
		pixmaps_list->folder_opened_20x20_mask,
		FALSE,				/* Not leaf */
		FALSE				/* Not expanded */
	    );

	    /* Create a new ViewerIndexItem */
	    item = ViewerIndexItemNew();
	    if(item != NULL)
	    {
		item->type = VIEWER_INDEX_ITEM_TYPE_DIRECTORY;
		item->path = g_strdup(path);
		gtk_ctree_node_set_row_data_full(
		    ctree,
		    new_branch,
		    item, (GtkDestroyNotify)ViewerIndexItemDelete
		);
	    }

	    /* Delete the row cell values */
	    strlistfree(strv, ncolumns);
	}
	else
	{
	    strv[0] = g_strdup(name);
	    /* Strip the extension from the name */
	    if(strv[0] != NULL)
	    {
		gchar *s = strrchr(strv[0], '.');
		if(s != NULL)
		    *s = '\0';
	    }

	    /* Description */
	    if(ncolumns > 1)
		strv[1] = g_strdup("Manual Page File");

	    /* Create a new branch */
	    new_branch = gtk_ctree_insert_node(
		ctree,
		parent,
		sibling,
		strv,
		ME_LIST_ICON_TEXT_SPACING,
		pixmaps_list->manual_closed_20x20,
		pixmaps_list->manual_closed_20x20_mask,
		pixmaps_list->manual_opened_20x20,
		pixmaps_list->manual_opened_20x20_mask,
		TRUE,				/* Is leaf */
		FALSE				/* Not expanded */
	    );

	    /* Create a new ViewerIndexItem */
	    item = ViewerIndexItemNew();
	    if(item != NULL)
	    {
		item->type = VIEWER_INDEX_ITEM_TYPE_MANPAGE;
		item->path = g_strdup(path);
		gtk_ctree_node_set_row_data_full(
		    ctree,
		    new_branch,
		    item, (GtkDestroyNotify)ViewerIndexItemDelete
		);
	    }

	    /* Delete the row cell values */
	    strlistfree(strv, ncolumns);
	}

	/* Count this object loaded */
	*nobjs_loaded = (*nobjs_loaded) + 1;

	/* If the specified path is a directory then recurse into it */
	if(S_ISDIR(stat_buf.st_mode))
	{
	    gint i, npaths;
	    const gchar *name;
	    gchar **names_list;

	    /* Print status message */
	    gchar *msg = g_strdup_printf(
		"Scanning %s...",
		path
	    );
	    ViewerSetStatusMessage(v, msg);
	    g_free(msg);

	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    /* Check for user abort */
	    if(v->stop_count > 0)
		return(new_branch);

	    /* Get the list of child objects in this directory */
	    names_list = GetDirEntNames2(path, &npaths);
	    names_list = StringQSort(names_list, npaths);

	    /* Iterate through each child */
	    for(i = 0; names_list && names_list[i]; i++)
	    {
		name = names_list[i];

		/* Delete the rest of the names if user aborted */
		if(v->stop_count > 0)
		{
		    g_free(names_list[i]);
		    continue;
		}

		ViewerSetStatusProgress(
		    v,
		    (gfloat)(*nobjs_loaded) / (gfloat)nobjs
		);

		/* Handle only names of manual pages or child
		 * directories
		 */
		if(strcmp(name, ".") &&
		   strcmp(name, "..") &&
		   strcmp(name, "whatis")
		)
		{
		    gchar *child_path = STRDUP(PrefixPaths(
			path,
			name
		    ));
		    if(child_path != NULL)
		    {
			/* Load the manual pages and subdirectories in
			 * this directory
			 */
			ViewerIndexLoadIterate(
			    v,
			    ctree,
			    new_branch,		/* Parent */
			    NULL,		/* Append */
			    child_path,
			    nobjs_loaded,
			    nobjs
			);

			/* Thaw and refreeze the Index GtkCTree after
			 * every 100 objects loaded to optimize speed
			 * and interactiveness
			 */
			if(((*nobjs_loaded) % 100) == 0)
			{
			    gtk_clist_thaw(clist);
			    gtk_clist_freeze(clist);
			}

			g_free(child_path);
		    }
		}

		g_free(names_list[i]);
	    }

	    g_free(names_list);
	}

	return(new_branch);
}

/*
 *	Loads the manual page paths to the Index GtkCTree.
 *
 *	The paths_list specifies a GList of gchar * paths.
 */
void ViewerIndexLoad(
	viewer_struct *v,
	GList *paths_list
)
{
	gint nobjs, nobjs_loaded;
	const gchar *path;
	gchar *msg;
	GList *glist;
	GtkCList *clist;
	GtkCTree *ctree;

	if((v == NULL) || (paths_list == NULL))
	    return;

	ctree = GTK_CTREE(v->index_ctree);
	clist = GTK_CLIST(ctree);

	v->processing = TRUE;
	v->stop_count = 0;

	ViewerSetBusy(v, TRUE);
	ViewerUpdateMenus(v);

	ViewerSetStatusMessage(v, "Loading index...");

	/* Count the number of manual pages and directories to load */
	nobjs = 0;
	for(glist = paths_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    path = (const gchar *)glist->data;
	    if(path == NULL)
		continue;

	    nobjs++;
	    nobjs += ViewerIndexLoadCountIterate(path);
	}

	gtk_clist_freeze(clist);

	/* Add the paths to the Index GtkCTree */
	nobjs_loaded = 0;
	for(glist = paths_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    path = (const gchar *)glist->data;
	    if(path == NULL)
		continue;

	    /* Manage GTK events */
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    /* Check for user abort */
	    if(v->stop_count > 0)
		break;

	    msg = g_strdup_printf(
		"Scanning %s...",
		path
	    );
	    ViewerSetStatusMessage(v, msg);
	    g_free(msg);

	    ViewerSetStatusProgress(
		v,
		(gfloat)(nobjs_loaded) / (gfloat)nobjs
	    );

	    /* Load this path and any subdirs */
	    ViewerIndexLoadIterate(
		v,
		ctree,
		NULL,				/* Toplevel */
		NULL,				/* Append */
		path,
		&nobjs_loaded,
		nobjs
	    );

	    ViewerSetStatusProgress(
		v,
		(gfloat)(nobjs_loaded) / (gfloat)nobjs
	    );

	    /* Thaw and refreeze the Index GtkCTree after every 100
	     * objects loaded to optimize speed and interactiveness
	     */
	    if((nobjs_loaded % 100) == 0)
	    {
		gtk_clist_thaw(clist);
		gtk_clist_freeze(clist);
	    }
	}

	gtk_clist_thaw(clist);

	if(v->stop_count > 0)
	    ViewerSetStatusMessage(v, "Index load interrupted");
	else
	    ViewerSetStatusMessage(v, "Index loaded");
	ViewerSetStatusProgress(v, 0.0f);

	v->processing = FALSE;
	ViewerSetBusy(v, FALSE);
	ViewerUpdateMenus(v);
}

/*
 *	Reloads the Index GtkCTree.
 */
void ViewerIndexRefresh(viewer_struct *v)
{
	const gchar *path;
	GList		*glist,
			*manpage_paths_list;
	GtkCList *clist;
	GtkCTree *ctree;
	CfgList *cfg_list;
	CfgStringList *string_list;
	me_core_struct *core;

	if(v == NULL)
	    return;

	ctree = GTK_CTREE(v->index_ctree);
	clist = GTK_CLIST(ctree);
	core = v->core;
	cfg_list = core->cfg_list;

	/* Get the manpage paths */
	string_list = ME_GET_STRING_LIST(ME_CFG_PARM_PATHS_MANPAGES);

	/* Copy the manpage paths */
	manpage_paths_list = NULL;
	for(glist = string_list->list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    path = (const gchar *)glist->data;
	    if(path == NULL)
		continue;

	    manpage_paths_list = g_list_append(
		manpage_paths_list,
		g_strdup(path)
	    );
	}

	/* Clear the Index GtkCTree */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);

	/* Load the index */
	if(manpage_paths_list != NULL)
	{
	    ViewerIndexLoad(
		v,
		manpage_paths_list
	    );
	    g_list_foreach(
		manpage_paths_list,
		(GFunc)g_free,
		NULL
	    );
	    g_list_free(manpage_paths_list);
	}
}


/*
 *      Procedure to find needle from haystack with respect to
 *      the given text widget.
 *
 *      Both needle and haystack may be modified by this function,
 *      but never deleted.
 *
 *      Given GtkText widget will be updated and instructed to select
 *      the matched text if a match was made. If case_sensitive is TRUE
 *      then the search will be made case sensitive, and if move_to is
 *      set to TRUE then gtk_editable_set_position() will be called to
 *      scroll to the new matched position.
 *
 *      If haystack_len is -1 then strlen() will be called on haystack
 *      to get the actual length. Given start_pos will be sanitized,
 *      if start_pos is >= haystack_len then start_pos will be set to 0.
 *
 *      Returns TRUE if a match is made.
 *
 *      Pointer to search_wrapped will be set to TRUE if a match
 *      was made after a search wrapped through the entire buffer.
 */
gboolean ViewerDoFind(
	viewer_struct *v, GtkText *text,
	gchar *haystack, gchar *needle,
	gint haystack_len, gint start_pos,
	gboolean case_sensitive,
	gboolean move_to,
	gboolean *search_wrapped
)
{
	gchar *bp, *match_ptr;


	/* Reset search wrapped */
	if(search_wrapped != NULL)
	    (*search_wrapped) = FALSE;

	/* Inputs valid? */
	if((v == NULL) || (text == NULL) ||
	   (haystack == NULL) || (needle == NULL)
	)
	    return(FALSE);

	/* Need to get haystack length ourself? */
	if(haystack_len < 0)
	    haystack_len = strlen(haystack);

	/* Empty haystack? */
	if(haystack_len < 1)
	    return(FALSE);

	/* Empty needle? */
	if((*needle) == '\0')
	    return(FALSE);

	ViewerSetStatusMessage(v, "Searching...");
	
	/* Sanitize starting position */
	if(start_pos >= haystack_len)
	    start_pos = 0;
	else if(start_pos < 0)
	    start_pos = 0;
 
	
	/* Case insensitive search? */
	if(!case_sensitive)
	{
	    /* Need to modify buffers to upper case */
	    
	    /* Modify ending portion of haystack */
	    bp = (gchar *)(haystack + start_pos);
	    while((*bp) != '\0')
	    {
		(*bp) = toupper(*bp);
		bp++;
	    }
	
	    /* Modify needle */
	    bp = needle;
	    while((*bp) != '\0')
	    {
		(*bp) = toupper(*bp);
		bp++;
	    }
	}
	
	    
	/* Begin search from starting position */
	match_ptr = strstr(
	    (gchar *)(haystack + start_pos),
	    needle
	);
	if(match_ptr == NULL)
	{
	    /* Did not get match, so start from beginning of haystack
	     * and search again. First check if this is case insensitive.
	     */
	    if(!case_sensitive)
	    {
		/* Toupper the beginning portion of the buffer */
		bp = haystack;
		while(bp < (gchar *)(haystack + start_pos))
		{
		    (*bp) = toupper(*bp);
		    bp++;
		}
	    }
	    /* Now try match again for the second time starting form the
	     * beginning.
	     */
	    match_ptr = strstr( 
		(gchar *)(haystack + 0),
		needle
	    );
	    if(match_ptr != NULL)
	    {
		/* Got match the second time, now mark search wrapped
		 * as true.
		 */
		if(search_wrapped != NULL)
		    (*search_wrapped) = TRUE;
	    }
	}
	    
	ViewerSetStatusMessage(v, "Search done");
	
	/* Did we get a match? */
	if(match_ptr == NULL)
	{
	    /* Sorry, no match made */
	    return(FALSE);
	}
	else
	{
	    /* Got match */
	    gint match_start_pos = (gint)(match_ptr - haystack);
	    gint needle_len = strlen(needle);
		    
	    if(match_start_pos < 0)
		match_start_pos = 0;
	    
	    /* Now modify the text widget, set new cursor position and
	     * select the matched string.
	     */
	    if(1)
	    {
		GtkEditable *editable = (GtkEditable *)text;
	    
		/* Scroll to new matched position? */
		if(move_to)
		{
		    gtk_text_set_point(text, match_start_pos);
		    gtk_editable_set_position(
			editable, match_start_pos
		    );
		}

		if(needle_len > 0)
		{
		    gtk_editable_select_region(
			editable,
			match_start_pos,
			match_start_pos + needle_len
		    );
		}
	    }

	    return(TRUE);
	}
}


/*
 *	Records the viewer's view text scroll bar position values.
 */
void ViewerViewTextRecordScrollPositions(viewer_struct *v)
{
	GtkText *text;
	GtkAdjustment *adj;


	if(v == NULL)
	    return;

	text = (GtkText *)v->view_text;
	if(text == NULL)
	    return;

	adj = text->hadj;
	if(adj != NULL)
	    v->last_scroll_hpos = adj->value;

	adj = text->vadj;
	if(adj != NULL)
	    v->last_scroll_vpos = adj->value;

	return;
}


/* 
 *      Creates the menu bar for the viewer, this function should be
 *      called by ViewerNew().
 *
 *      Given parent is assumed to be a vbox.
 */
static void ViewerCreateMenuBar(viewer_struct *v, GtkWidget *parent)
{
	guint		accel_key,
			accel_mods;
	gchar *label = NULL;
	guint8 **icon;
	gpointer old_data, data = v;
	GtkAccelGroup *accelgrp = v->accelgrp;
	GtkWidget	*w,
			*parent2,
			*menu_bar,
			*menu;
	void (*func_cb)(GtkWidget *w, gpointer);

	/* Menu Bar */
	v->menu_bar = menu_bar = GUIMenuBarCreate(NULL);
	gtk_widget_show(menu_bar);

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_LABEL,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SUBMENU,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
 if(w != NULL)					\
  GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_CHECK,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL,						\
  NULL,						\
  0, 0,						\
  NULL, NULL					\
 );						\
}

	/* Manual */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_open_20x20_xpm;
	    label = "Open...";
	    accel_key = GDK_o;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerOpenCB;
	    ADD_MENU_ITEM_LABEL
	    v->open_mi = w;

	    icon = (guint8 **)icon_clear_20x20_xpm;
	    label = "Clear";
	    accel_key = GDK_k;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerClearCB;
	    ADD_MENU_ITEM_LABEL
	    v->clear_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_close_20x20_xpm;
	    label = "Close";
	    accel_key = GDK_w;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerCloseCB;
	    ADD_MENU_ITEM_LABEL
	    v->close_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_exit_20x20_xpm;
	    label = "Exit";
	    accel_key = GDK_x;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerExitCB;
	    ADD_MENU_ITEM_LABEL 
	    v->exit_mi = w;
	}
	v->manual_mh = GUIMenuAddToMenuBar(
	    menu_bar, menu,
	    "Manual",
	    GTK_JUSTIFY_LEFT
	);

	/* Edit */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_copy_20x20_xpm;
	    label = "Copy";
	    accel_key = GDK_c;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerCopyCB;
	    ADD_MENU_ITEM_LABEL 
	    v->edit_copy_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Select All";
	    accel_key = GDK_a;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerSelectAllCB;
	    ADD_MENU_ITEM_LABEL
	    v->edit_select_all_mi = w;

	    icon = NULL;
	    label = "Unselect All";
	    accel_key = GDK_u;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerUnselectAllCB;
	    ADD_MENU_ITEM_LABEL 
	    v->edit_unselect_all_mi = w;
	}
	v->edit_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Edit",
	    GTK_JUSTIFY_LEFT
	);

	/* View */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_stop_20x20_xpm;
	    label = "Stop";
	    accel_key = GDK_Escape;
	    accel_mods = 0;
	    func_cb = ViewerStopCB;
	    ADD_MENU_ITEM_LABEL
	    v->view_stop_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Manual Page";
	    accel_key = GDK_1;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerPageToggleCB;
	    ADD_MENU_ITEM_CHECK
	    v->view_view_mi = w;

	    icon = NULL;
	    label = "Index";
	    accel_key = GDK_2;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerPageToggleCB;
	    ADD_MENU_ITEM_CHECK
	    v->view_index_mi = w;
	}
	v->view_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "View",
	    GTK_JUSTIFY_LEFT
	);

	/* Settings */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    icon = (guint8 **)icon_options_20x20_xpm;
	    label = "Options...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerOptionsCB;
	    ADD_MENU_ITEM_LABEL
	    v->settings_options_mi = w;
	}
	v->settings_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Settings",
	    GTK_JUSTIFY_LEFT
	);

	/* Windows */
	menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    old_data = data;
	    data = v->core;

	    icon = (guint8 **)icon_manedit_editor_20x20_xpm;
	    label = "New Editor";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = MENewEditorCB;
	    ADD_MENU_ITEM_LABEL
	    v->windows_new_editor = w;

	    icon = (guint8 **)icon_manual_opened_20x20_xpm;
	    label = "New Viewer";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = MENewViewerCB;
	    ADD_MENU_ITEM_LABEL
	    v->windows_new_viewer = w;

	    data = old_data;
	}
	v->windows_mh = GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Windows",
	    GTK_JUSTIFY_LEFT
	);

	/* Help */
	menu = MEBuildHelpMenu(
	    v->core,
	    accelgrp
	);
	GUIMenuAddToMenuBar(
	    menu_bar,
	    menu,
	    "Help",   
	    GTK_JUSTIFY_RIGHT
	);

#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK 
#undef ADD_MENU_ITEM_SEPARATOR


	/* Menu Bar GtkHandleBox */
	w = gtk_handle_box_new();
	v->menu_bar_dock = w;
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	gtk_container_add(GTK_CONTAINER(parent2), menu_bar);
}

/*
 *      Creates the tool bar for the viewer, this function should be
 *      called by ViewerNew(). 
 *
 *      Given parent is assumed to be a vbox.
 */
static void ViewerCreateToolBar(viewer_struct *v, GtkWidget *parent)
{
	const gint border_minor = 2;
	GList *items_list;
	GtkWidget	*w,
			*parent2;
	me_core_struct *core = v->core;
	CfgList *cfg_list = core->cfg_list;

	/* ToolBarItems */
	items_list = NULL;
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Open",
		(guint8 **)icon_open_20x20_xpm,
		VIEWER_TT_OPEN,
		VIEWER_OPID_OPEN,
		ViewerToolBarItemActivateCB, v,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Clear",
		(guint8 **)icon_clear_20x20_xpm,
		VIEWER_TT_CLEAR,
		VIEWER_OPID_CLEAR,
		ViewerToolBarItemActivateCB, v,
		NULL, NULL,
		NULL, NULL
	    )
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNewSeparator()
	);
	items_list = g_list_append(
	    items_list,
	    ToolBarItemNew(
		TOOL_BAR_ITEM_BUTTON,
		"Stop",
		(guint8 **)icon_stop_20x20_xpm,
		VIEWER_TT_STOP,
		VIEWER_OPID_STOP,
		ViewerToolBarItemActivateCB, v,
		NULL, NULL,
		NULL, NULL
	    )
	);

	/* GtkHandleBox */
	v->tool_bar_dock = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* ToolBar */
	v->tool_bar = w = ToolBarNew(
	    GTK_ORIENTATION_HORIZONTAL,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS),
	    (GtkReliefStyle)ME_GET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF),
	    GTK_POLICY_AUTOMATIC,
	    items_list
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	gtk_widget_show(w);

	if(items_list != NULL)
	{
	    g_list_foreach(
		items_list,
		(GFunc)ToolBarItemDelete,
		NULL
	    );
	    g_list_free(items_list);
	}
}

/*
 *	Creates the status bar for the viewer, this function should be
 *	called by ViewerNew().
 *
 *	Given parent is assumed to be a vbox.
 */
static void ViewerCreateStatusBar(
	viewer_struct *v,
	GtkWidget *parent
)
{
	const gint border_minor = 2;
	GtkWidget *w, *parent2, *parent3;
	GtkAdjustment *adj;

	/* Outer GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_widget_set_usize(
	    w,
	    -1,
	    ME_STATUS_BAR_HEIGHT
	);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Main GtkTable */
	w = gtk_table_new(
	    1, 4,				/* Rows columns */
	    FALSE				/* Not homogenious */
	);
	gtk_table_set_col_spacings(GTK_TABLE(w), 1);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkProgressBar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1.0f, 100.0f,
	    0.0f, 0.0f, 0.0f
	);
	v->status_bar_progress = w = gtk_progress_bar_new_with_adjustment(adj);
	gtk_widget_set_usize(
	    w,
	    100,
	    ME_STATUS_BAR_HEIGHT - (2 * 3)
	);
	gtk_progress_bar_set_orientation(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_LEFT_TO_RIGHT
	);
	gtk_progress_bar_set_bar_style(
	    GTK_PROGRESS_BAR(w),
	    GTK_PROGRESS_CONTINUOUS
	);
	gtk_progress_set_activity_mode(
	    GTK_PROGRESS(w),
	    FALSE
	);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    0, 1,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);


	/* Message GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    1, 2,
	    0, 1,
	    GTK_SHRINK | GTK_EXPAND | GTK_FILL,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Message GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Message Gtklabel */
	v->status_bar_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);


	/* Cursor Position GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    2, 3,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Cursor position GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Cursor position GtkLabel */
	v->status_bar_cursor_label = w = gtk_label_new("");
	gtk_widget_set_usize(w, 100, -1);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);


	/* Window buttons GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_table_attach(
	    GTK_TABLE(parent2),
	    w,
	    3, 4,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Editor GtkButton */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_manedit_editor_16x16_xpm
	);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, 25, -1);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MENewEditorCB), v->core
	);
	GUISetWidgetTip(w, MEDIT_SB_TT_NEW_EDITOR);
	gtk_widget_show(w);

	/* Viewer GtkButton */
	w = (GtkWidget *)GUIButtonPixmap(
	    (guint8 **)icon_manedit_viewer_16x16_xpm
	);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_set_usize(w, 25, -1);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(MENewViewerCB), v->core
	);
	GUISetWidgetTip(w, MEDIT_SB_TT_NEW_VIEWER);
	gtk_widget_show(w);
}

/*
 *	Creates a new viewer structure with its values reset and widgets
 *	built or NULL on error.
 */
viewer_struct *ViewerNew(
	me_core_struct *core,
	const gint editor_num
)
{
	const gint	border_minor = 2,
			border_major = 5;
	guint		accel_key,
			accel_mods;
	gchar *title[4];
	const gchar *label;
	guint8 **icon;
	gpointer	combo_rtn,
			data;
	const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",				0,	MEDIT_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",			0,	MEDIT_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",				0,	MEDIT_DND_TYPE_INFO_STRING},
{MEDIT_DND_TARGET_EDITOR_BRANCH,	GTK_TARGET_SAME_APP,	MEDIT_DND_TYPE_INFO_EDITOR_BRANCH_CMD}
	};
	const GtkTargetEntry dnd_src_types[] = {
{"text/plain",				0,	MEDIT_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",			0,	MEDIT_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",				0,	MEDIT_DND_TYPE_INFO_STRING}
	};
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*parent5,
			*toplevel,
			*scroll_parent,
			*pulist,
			*menu;
	GtkCList *clist;
	void (*func_cb)(GtkWidget *w, gpointer);
	CfgList *cfg_list;
	viewer_struct *v;

	if(core == NULL)
	    return(NULL);

	cfg_list = core->cfg_list;

	/* Create a new Viewer */
	v = VIEWER(g_malloc0(sizeof(viewer_struct)));
	if(v == NULL)
	    return(NULL);

	v->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	v->accelgrp = accelgrp = gtk_accel_group_new();
/*
	v->freeze_count = 0;
	v->busy_count = 0;
	v->stop_count = 0;
	v->processing = FALSE;
 */
	v->core = core;
	v->editor_num = editor_num;
	v->current_page = VIEWER_PAGE_MANPAGE;
/*
	v->last_selected_index_node = NULL;
	v->cur_manpage_path = NULL;
	v->cur_manpage_name = NULL;
	v->last_open_path = NULL;
 */

#define ADD_MENU_ITEM_LABEL {			\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_LABEL,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU {			\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SUBMENU,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
 GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_ITEM_CHECK {			\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_CHECK,			\
  accelgrp,					\
  icon,						\
  label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SEPARATOR {		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL,						\
  NULL,						\
  0, 0,						\
  NULL, NULL					\
 );						\
}

	v->freeze_count++;

	/* Toplevel */
	w = toplevel;
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, FALSE);
	gtk_window_set_title(GTK_WINDOW(w), ME_VIEWER_DEF_TITLE);
	if(ME_GET_B(ME_CFG_PARM_RECORD_WIN_POS))
	{
	    CfgGtkSetGtkWindow(
		cfg_list,
		ME_CFG_PARM_VIEWER_X,
		ME_CFG_PARM_VIEWER_Y,
		ME_CFG_PARM_VIEWER_WIDTH,
		ME_CFG_PARM_VIEWER_HEIGHT,
		w
	    );
	}
	else
	{
	    gtk_widget_set_usize(
		w,
	        ME_VIEWER_DEF_WIDTH,
		ME_VIEWER_DEF_HEIGHT
	    );
	}
	gtk_widget_realize(w);
	GUISetWMIcon(
	    w->window,
	    (guint8 **)icon_manedit_viewer_48x48_xpm
	);
	gtk_widget_add_events(
	    w,
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(ViewerDeleteEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(ViewerKeyEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(ViewerKeyEventCB), v
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;

	/* Main GtkVBox */
	v->main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Create the Menu Bar */
	ViewerCreateMenuBar(v, v->main_vbox);

	/* Create the Tool Bar */
	ViewerCreateToolBar(v, v->main_vbox);


	/* Main GtkNotebook */
	v->main_notebook = w = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(w), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 2);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(w), TRUE);
	gtk_signal_connect(
	    GTK_OBJECT(w), "switch_page",
	    GTK_SIGNAL_FUNC(ViewerSwitchPageCB), v
	);
	gtk_widget_show(w);
	parent2 = w;


	/* Begin creating page 0 on main_notebook */

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_set_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page( 
	    GTK_NOTEBOOK(parent2),
	    w,
	    gtk_label_new("Manual Page")
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Section and Manpage GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Section GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = gtk_label_new("Section:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Section PopUp List Box */
	v->section_pulistbox = w = PUListBoxNew(
	    180, -1
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	pulist = PUListBoxGetPUList(w);
	gtk_widget_show(w);
	if(pulist != NULL)
	{
	    gint i;
#define PULIST_APPEND(_s_)	{		\
 i = PUListAddItem(pulist, (_s_));		\
 PUListSetItemDataFull(				\
  pulist, i,					\
  STRDUP(_s_), (GtkDestroyNotify)g_free		\
 );						\
}
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_ANY);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_EXACT);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_1);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_2);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_3);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_4);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_5);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_6);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_7);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_8);
	    PULIST_APPEND(ME_MANPAGE_SECTION_NAME_9);
#undef PULIST_APPEND
	    PUListBoxSetLinesVisible(
		w,
		MIN((i + 1), 10)
	    );
	}

	/* Manpage GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = GUICreateMenuItemIcon(
	    (guint8 **)icon_manual_opened_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = GUIComboCreate(
	    "ManPage:",
	    "",			/* Initial value */
	    NULL,		/* Initial glist of items */
	    20,			/* Max items */
	    &combo_rtn,
	    (gpointer)v,
	    ViewerManPageActivateCB,
	    NULL
	);
	v->manpage_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	GUISetWidgetTip(GTK_COMBO(combo_rtn)->entry, VIEWER_TT_SEARCH_MANPAGE);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Find GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = GUICreateMenuItemIcon(
	    (guint8 **)icon_search_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = GUIComboCreate(
	    "Find:",
	    "",                 /* Initial value */
	    NULL,               /* Initial glist of items */
	    20,                 /* Max items */
	    &combo_rtn,
	    (gpointer)v,
	    ViewerFindActivateCB,
	    NULL
	);
	v->find_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	gtk_widget_show(w);



	/* View text table parent to hold view text and scroll bars */
	w = gtk_table_new(2, 2, FALSE);
	gtk_table_set_row_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_table_set_col_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* View GtkText */
	v->view_text = w = gtk_text_new(NULL, NULL);
	gtk_widget_modify_style(
	    w,
	    core->manpage_rcstyle
	);
	gtk_widget_set_events(
	    w,
	    GDK_BUTTON_PRESS_MASK | GDK_KEY_PRESS_MASK |
	    GDK_KEY_RELEASE_MASK
	);
	/* Need to connect key event signals to change editable
	 * state. Reason explained in ViewerViewKeyEventCB().
	 */
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(ViewerViewKeyEventCB), v
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(ViewerViewKeyEventCB), v
	);
	/* Need to connect button_press_event signal_after just
	 * after text widget is created, this is so that the right
	 * click menu can be mapped properly.
	 */
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(ViewerViewButtonEventCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(ViewerViewButtonEventCB), v
	);
	gtk_editable_set_editable(GTK_EDITABLE(w), FALSE);
	gtk_text_set_word_wrap(GTK_TEXT(w), TRUE);
	gtk_table_attach(
	    GTK_TABLE(parent4), w,
	    0, 1, 0, 1,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_widget_realize(w);
	if(w->window != NULL)
	{
	    GdkColormap *colormap = GDK_COLORMAP_REF(gtk_widget_get_colormap(w));
	    GdkColor *ct;
	    CfgColor *cs;

	    ct = &v->view_normal_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_FG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &v->view_bold_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_BOLD);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &v->view_underline_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_UNDERLINE);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    v->colormap = colormap;

	    gdk_window_set_cursor(
		GTK_TEXT(w)->text_area,
		core->text_cursor
	    );
	}
	GUIDNDSetTar(
	    w,
	    dnd_tar_types,
	    sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
	    GDK_ACTION_COPY | GDK_ACTION_MOVE,  /* Actions */
	    GDK_ACTION_MOVE,                    /* Default action if same */
	    GDK_ACTION_COPY,                    /* Default action */
	    ViewerViewTextDNDDataRecievedCB,
	    v,
	    TRUE
	);
	gtk_widget_show(w);

	/* Vertical scroll bar for text widget */
	scroll_parent = gtk_vscrollbar_new(GTK_TEXT(w)->vadj);
	gtk_table_attach(
	    GTK_TABLE(parent4), scroll_parent,
	    1, 2, 0, 1,
	    GTK_FILL,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
	    0, 0
	);
	gtk_widget_show(scroll_parent);


	/* View text right click GtkMenu */
	v->view_menu = menu = GUIMenuCreate();
	data = v;

	if(menu != NULL)
	{
	    icon = (guint8 **)icon_copy_20x20_xpm;
	    label = "Copy";
	    accel_key = GDK_c;
	    accel_mods = GDK_CONTROL_MASK;
	    func_cb = ViewerCopyCB;
	    ADD_MENU_ITEM_LABEL
	    v->view_copy_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = NULL;
	    label = "Select All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerSelectAllCB;
	    ADD_MENU_ITEM_LABEL
	    v->view_select_all_mi = w;

	    icon = NULL;
	    label = "Unselect All";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerUnselectAllCB;
	    ADD_MENU_ITEM_LABEL
	    v->view_unselect_all_mi = w;
	}


	/* Begin creating page 1 on main_notebook */

	/* Index ctree vbox parent to hold index ctree widgets */
	w = gtk_vbox_new(FALSE, 5);
	gtk_container_border_width(GTK_CONTAINER(w), 5);
	gtk_notebook_append_page(
	    GTK_NOTEBOOK(parent2),
	    w,
	    gtk_label_new("Index")
	);
	gtk_widget_show(w);
	parent3 = w;


	/* Hbox for find in pages combo */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = GUICreateMenuItemIcon(
	    (guint8 **)icon_search_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = GUIComboCreate(
	    "Find Content:",
	    "",                 /* Initial value */
	    NULL,               /* Initial glist of items */
	    20,                 /* Max items */
	    &combo_rtn,
	    v,
	    ViewerFindInPagesActivateCB,
	    NULL
	);
	v->index_find_in_pages_combo = (GtkWidget *)combo_rtn;
	gtk_combo_set_case_sensitive(
	    (GtkCombo *)combo_rtn,
	    TRUE
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Scrolled window for index ctree */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	scroll_parent = w;

	/* Index GtkCTree */
	title[0] = "Name";
	title[1] = "Description";
	v->index_ctree = w = gtk_ctree_new_with_titles(
	    2,
	    0,
	    title
	);
	clist = GTK_CLIST(w);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(ViewerMenuMapCB), v
	);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_column_titles_show(clist);
	gtk_clist_set_row_height(clist, ME_LIST_ROW_SPACING);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_clist_set_column_width(clist, 0, 300);
	gtk_clist_set_column_resizeable(clist, 0, TRUE);
	gtk_clist_set_column_width(clist, 1, 200);
	gtk_clist_set_column_resizeable(clist, 1, TRUE);
	gtk_container_add(GTK_CONTAINER(scroll_parent), w);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_select_row",
	    GTK_SIGNAL_FUNC(ViewerIndexCTreeSelectCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_unselect_row",
	    GTK_SIGNAL_FUNC(ViewerIndexCTreeUnselectCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_expand",
	    GTK_SIGNAL_FUNC(ViewerIndexCTreeExpandCB), v
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_collapse",
	    GTK_SIGNAL_FUNC(ViewerIndexCTreeExpandCB), v
	);
	GUIDNDSetSrc(
	    w,
	    dnd_src_types,
	    sizeof(dnd_src_types) / sizeof(GtkTargetEntry),
	    GDK_ACTION_COPY,			/* Actions */
	    GDK_BUTTON1_MASK,			/* Buttons */
	    NULL,
	    ViewerIndexCTreeDNDDataRequestCB,
	    ViewerDNDDataDeleteCB,
	    NULL,
	    v
	);
	gtk_widget_show(w);

	/* Index GtkCTree Right Click GtkMenu */
	v->index_menu = menu = GUIMenuCreate();
	data = v;

	if(menu != NULL)
	{
	    icon = NULL;
	    label = "Expand/Collapse";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerIndexBranchExpandCB;
	    ADD_MENU_ITEM_LABEL
	    v->index_expand_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_folder_parent_20x20_xpm;
	    label = "Parent";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerIndexGotoParentCB;
	    ADD_MENU_ITEM_LABEL
	    v->index_parent_mi = w;

	    icon = (guint8 **)icon_manual_opened_20x20_xpm;
	    label = "View";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerIndexViewCB;
	    ADD_MENU_ITEM_LABEL
	    v->index_view_mi = w;

	    ADD_MENU_ITEM_SEPARATOR

	    icon = (guint8 **)icon_reload_20x20_xpm;
	    label = "Refresh";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerIndexRefreshCB;
	    ADD_MENU_ITEM_LABEL
	    v->index_refresh_mi = w;

	    icon = (guint8 **)icon_stop_20x20_xpm;
	    label = "Stop";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ViewerStopCB; 
	    ADD_MENU_ITEM_LABEL
	    v->index_stop_mi = w;
	}

	/* Separator */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Hbox for index ctree buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 5);
	gtk_widget_show(w);
	parent4 = w;

	/* Parent GtkButton */
	v->index_parent_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_folder_parent_20x20_xpm, "Parent", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(ViewerIndexGotoParentCB), v
	);
	gtk_widget_show(w);

	/* View GtkButton */
	v->index_view_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_manual_opened_20x20_xpm,
	    "View",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(ViewerIndexViewCB), v
	);
	gtk_widget_show(w);

	/* Refresh GtkButton */
	v->index_refresh_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_reload_20x20_xpm, "Refresh", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(ViewerIndexRefreshCB), v
	);
	gtk_widget_show(w);



	/* Status Bar */
	ViewerCreateStatusBar(v, v->main_vbox);

#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_SUBMENU
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR

	ViewerUpdateMenus(v);

	v->freeze_count--;

	return(v);
}

/*
 *	Records the Viewer's values to the CfgList.
 */
void ViewerRecordConfiguration(viewer_struct *v)
{
	CfgList *cfg_list;
	me_core_struct *core;

	if(v == NULL)
	    return;

	core = v->core;
	cfg_list = core->cfg_list;

	/* Toplevel GtkWindow position and size */
	CfgGtkGetGtkWindow(
	    cfg_list,
	    ME_CFG_PARM_VIEWER_X,
	    ME_CFG_PARM_VIEWER_Y,
	    ME_CFG_PARM_VIEWER_WIDTH,
	    ME_CFG_PARM_VIEWER_HEIGHT,
	    v->toplevel
	);

	ME_SET_S(
	    ME_CFG_PARM_PATH_LAST_OPEN,
	    v->last_open_path
	);
}

/*
 *	Updates the title.
 */
void ViewerUpdateTitle(viewer_struct *v)
{
	gchar *title;

	if(v == NULL)
	    return;

	title = NULL;
	switch(v->current_page)
	{
	  case VIEWER_PAGE_MANPAGE:
	    if(STRISEMPTY(v->cur_manpage_name))
		title = STRDUP(ME_VIEWER_DEF_TITLE);
	    else
		title = g_strconcat(
		    ME_VIEWER_DEF_TITLE,
		    ": ",
		    v->cur_manpage_name,
		    NULL
		);
	    break;
	  case VIEWER_PAGE_INDEX:
	    title = g_strconcat(
		ME_VIEWER_DEF_TITLE,
		": Index",
		NULL
	    );
	    break;
	}
	if(title != NULL)
	{
	    gtk_window_set_title(GTK_WINDOW(v->toplevel), title);
	    g_free(title);
	}
}

/*
 *	Updates all menus on viewer.
 */
void ViewerUpdateMenus(viewer_struct *v)
{
	gboolean	sensitive,
			state;
	GtkWidget *w, *menu;
	GtkCTreeRow *branch_row;
	CfgList *cfg_list;
	ViewerIndexItem *item;
	me_core_struct *core;

	if(v == NULL)
	    return;

	core = v->core;
	cfg_list = core->cfg_list;

	v->freeze_count++;

#define SET_CHECK_MENU_ITEM_STATE(_w_,_state_)	{	\
 GUIMenuItemSetCheck((_w_), (_state_), TRUE);		\
}

	/* Manual */
	menu = v->manual_mh;
	if(menu != NULL)
	{
	    sensitive = TRUE;
	    GTK_WIDGET_SET_SENSITIVE(v->open_mi, sensitive);
	}

	/* Edit */
	menu = v->edit_mh;
	if(menu != NULL)
	{
	    sensitive = TRUE;
	    GTK_WIDGET_SET_SENSITIVE(v->edit_copy_mi, sensitive);
	    GTK_WIDGET_SET_SENSITIVE(v->edit_select_all_mi, sensitive);
	    GTK_WIDGET_SET_SENSITIVE(v->edit_unselect_all_mi, sensitive);
	}

	/* View */
	menu = v->view_mh;
	if(menu != NULL)
	{
	    sensitive = v->processing;
	    GTK_WIDGET_SET_SENSITIVE(v->view_stop_mi, sensitive);

	    w = v->view_view_mi;
	    state = (v->current_page == VIEWER_PAGE_MANPAGE) ? TRUE : FALSE;
	    SET_CHECK_MENU_ITEM_STATE(w, state);

	    w = v->view_index_mi;
	    state = (v->current_page == VIEWER_PAGE_INDEX) ? TRUE : FALSE;
	    SET_CHECK_MENU_ITEM_STATE(w, state);
	}

	/* Settings */
	menu = v->settings_mh;
	if(menu != NULL)
	{
	    sensitive = TRUE;
	    GTK_WIDGET_SET_SENSITIVE(v->settings_options_mi, sensitive);
	}


	/* View right click GtkMenu */
	menu = v->view_menu;
	if(menu != NULL)
	{
	    sensitive = TRUE;
	    GTK_WIDGET_SET_SENSITIVE(v->view_copy_mi, sensitive);
	    GTK_WIDGET_SET_SENSITIVE(v->view_select_all_mi, sensitive);
	    GTK_WIDGET_SET_SENSITIVE(v->view_unselect_all_mi, sensitive);
	}

	/* Index right click GtkMenu */
	menu = v->index_menu;
	if(menu != NULL)
	{
	    if(v->last_selected_index_node == NULL)
		branch_row = NULL;
	    else
		branch_row = GTK_CTREE_ROW(v->last_selected_index_node);
	    sensitive = (branch_row != NULL) ?
		(branch_row->parent != NULL) : FALSE;
	    GTK_WIDGET_SET_SENSITIVE(v->index_parent_mi, sensitive);

            /* sensitive may be 0 */
	    if((v->index_ctree != NULL) &&
	       (v->last_selected_index_node != NULL) && sensitive
	    )
		item = VIEWER_INDEX_ITEM(gtk_ctree_node_get_row_data(
		    GTK_CTREE(v->index_ctree), v->last_selected_index_node
		));
	    else
		item = NULL;
	    if(item == NULL)
		sensitive = FALSE;
	    else if(item->type == VIEWER_INDEX_ITEM_TYPE_DIRECTORY)
		sensitive = TRUE;
	    else
		sensitive = FALSE;
	    GTK_WIDGET_SET_SENSITIVE(v->index_expand_mi, sensitive);

	    if(item == NULL)
		sensitive = FALSE;
	    else if((item->type == VIEWER_INDEX_ITEM_TYPE_MANPAGE) ||
		    (item->type == VIEWER_INDEX_ITEM_TYPE_OTHER_FILE)
	    )
		sensitive = TRUE;
	    else
		sensitive = FALSE;
	    GTK_WIDGET_SET_SENSITIVE(v->index_view_mi, sensitive);

	    sensitive = TRUE;
	    GTK_WIDGET_SET_SENSITIVE(v->index_refresh_mi, TRUE);

	    sensitive = v->processing;
	    GTK_WIDGET_SET_SENSITIVE(v->index_stop_mi, sensitive);
	}

	/* Tool Bar */
	w = v->tool_bar;
	ToolBarSetButtonDecals(
	    w,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS)
	);
	ToolBarSetItemSensitive(
	    w,
	    VIEWER_OPID_OPEN,
	    !FileBrowserIsQuery()
	);
	ToolBarSetItemSensitive(
	    w,
	    VIEWER_OPID_CLEAR,
	    TRUE
	);
	ToolBarSetItemSensitive(
	    w,
	    VIEWER_OPID_STOP,
	    v->processing
	);


	/* Buttons on index page */
	if(v->last_selected_index_node == NULL)
	    branch_row = NULL;
	else
	    branch_row = GTK_CTREE_ROW(v->last_selected_index_node);
	sensitive = (branch_row != NULL) ?
	    (branch_row->parent != NULL) : FALSE;
	GTK_WIDGET_SET_SENSITIVE(v->index_parent_btn, sensitive);

        /* sensitive may be 0 */
	if((v->index_ctree != NULL) &&
	   (v->last_selected_index_node != NULL && sensitive)
	)
	    item = VIEWER_INDEX_ITEM(gtk_ctree_node_get_row_data(
		GTK_CTREE(v->index_ctree),
		v->last_selected_index_node
	    ));
	else
	    item = NULL;
	if(item == NULL)
	    sensitive = FALSE;
	else if((item->type == VIEWER_INDEX_ITEM_TYPE_MANPAGE) ||
		(item->type == VIEWER_INDEX_ITEM_TYPE_OTHER_FILE)
	)
	    sensitive = TRUE;
	else
	    sensitive = FALSE;
	GTK_WIDGET_SET_SENSITIVE(v->index_view_btn, sensitive);

	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(v->index_refresh_btn, sensitive);


	v->freeze_count--;

#undef SET_CHECK_MENU_ITEM_STATE
#undef SET_BUTTON_LAYOUT
}

/*
 *	Switches the Viewer's page.
 */
void ViewerSetPage(
	viewer_struct *v,
	viewer_page page
)
{
	if(v == NULL)
	    return;

	gtk_notebook_set_page(
	    GTK_NOTEBOOK(v->main_notebook),
	    (gint)page
	);
}

/*
 *	Updates the cursor position value on the status bar.
 */
void ViewerSetStatusPosition(
	viewer_struct *v,
	const gint row, const gint column
)
{
	gchar *s;
	GtkWidget *w;

	if(v == NULL)
	    return;

	w = v->status_bar_cursor_label;
	if(w == NULL)
	    return;

	s = g_strdup_printf(
	    "Row: %i / Col: %i",
	    row, column
	);
	gtk_label_set_text(GTK_LABEL(w), s);
	g_free(s);

	while(gtk_events_pending() > 0)
	    gtk_main_iteration();
}

/*
 *	Sets the status message for the viewer.
 *
 *	Passing NULL clears the message.
 *
 *      Note that gtk_main_iteration() will be called when this function
 *      is called to update the label properly.
 */
void ViewerSetStatusMessage(
	viewer_struct *v,
	const gchar *msg
)
{
	GtkWidget *w;

	if(v == NULL)
	    return;

	w = v->status_bar_label;
	if(w == NULL)
	    return;

	gtk_label_set_text(
	    GTK_LABEL(w),
	    (msg != NULL) ? msg : ""
	);

	while(gtk_events_pending() > 0)
	    gtk_main_iteration();
}

/*
 *	Sets the status bar's progress.
 *
 *	The val specifies the value from 0.0 to 1.0 inclusive or -1.0 for
 *	"unknown limit" which will cause the progress bar to be displayed
 *	in activity mode.
 *
 *	Note that gtk_main_iteration() will be called when this function
 *	is called to update the progress bar properly.
 */
void ViewerSetStatusProgress(
	viewer_struct *v,
	const gfloat val
)
{
	const gboolean allow_gtk_iteration = TRUE;
	gfloat value = val;
	GtkWidget *w;
	GtkProgress *progress;
	GtkProgressBar *progress_bar;

	if(v == NULL)
	    return;

	w = v->status_bar_progress;
	progress = GTK_PROGRESS(w);
	progress_bar = GTK_PROGRESS_BAR(w);

	/* Do activity? */
	if(value < 0.0f)
	{
	    GtkAdjustment *adj = progress->adjustment;

	    /* Get new value based on unknown position */
	    value = gtk_progress_get_value(progress) + 1.0f;
	    if(value > adj->upper)
		value = adj->lower;

	    /* Update the progress bar for `activity mode' (unknown limit)
	     * and set the new value
	     */
	    gtk_progress_set_activity_mode(
		progress,
		TRUE
	    );
	    gtk_progress_set_show_text(
		progress,
		FALSE
	    );
	    gtk_progress_set_value(
		progress,
		value
	    );

	    /* Reset the last progress position to -1.0, indicating
	     * that there is no known last position since we are
	     * doing activity mode
	     */
	    v->status_bar_progress_pos_last = -1.0f;
	}
	else
	{

	    /* Clip value to 0.0, 1.0 */
	    if(value > 1.0f)
		value = 1.0f;

	    /* Reset the last progress position if it is greater than
	     * the specified value, implying that the progress has
	     * warped back to the beginning
	     */
	    if(v->status_bar_progress_pos_last > value)
		v->status_bar_progress_pos_last = -1.0f;

	    /* Has the value not sufficiently changed? (less than
	     * 0.009 change in value)
	     */
	    if((v->status_bar_progress_pos_last > 0.0f) &&
	       ((value - v->status_bar_progress_pos_last) < 0.009f)
	    )
		return;

	    /* Update progress */
	    gtk_progress_set_activity_mode(
		progress,
		FALSE
	    );
	    gtk_progress_set_format_string(
		progress,
		"%p%%"
	    );
	    /* Display text only if value is positive */
	    gtk_progress_set_show_text(
		progress,
		(value > 0.0f) ? TRUE : FALSE
	    );
	    gtk_progress_bar_update(
		progress_bar,
		value 
	    );

	    /* Record the last progress value */
	    v->status_bar_progress_pos_last = value;
	}

	/* Manage any GTK events so that the value changes to the
	 * GtkWidgets take affect immediately
	 */
	if(allow_gtk_iteration)
	{
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();
	}
}

/*
 *	Sets the Viewer busy or ready.
 */
void ViewerSetBusy(
	viewer_struct *v,
	const gboolean busy
)
{ 
	GdkCursor	*cur,
			*text_cur;
	GdkWindow *window;
	GtkWidget *w;
	me_core_struct *core;

	if(v == NULL)
	    return;

	w = v->toplevel;
	core = v->core;

	if(busy)
	{
	    v->busy_count++;
	    if(v->busy_count > 1)
		return;
	    text_cur = cur = core->busy_cursor;
	}
	else
	{
	    v->busy_count--;
	    if(v->busy_count < 0)
		v->busy_count = 0;
	    if(v->busy_count > 0)
		return;
	    cur = NULL;				/* Default cursor */
	    text_cur = core->text_cursor;
	}

	window = w->window;
	if(window != NULL)
	    gdk_window_set_cursor(window, cur);

	window = GTK_TEXT(v->view_text)->text_area;
	if(window != NULL)
	    gdk_window_set_cursor(window, text_cur);

	gdk_flush();
}

/*
 *	Checks if the Viewer is mapped.
 */
gboolean ViewerIsMapped(viewer_struct *viewer)
{
	if(viewer == NULL)
	    return(FALSE);

	return(GTK_WIDGET_MAPPED(viewer->toplevel));
}

/*
 *	Maps the Viewer.
 */
void ViewerMap(viewer_struct *v)
{
	if(v == NULL)
	    return;

	gtk_widget_show(v->toplevel);
}

/*
 *	Unmaps the Viewer.
 */
void ViewerUnmap(viewer_struct *v)
{       
	if(v == NULL)
	    return;

	gtk_widget_hide(v->toplevel);
}


/*
 *	Deletes the Viewer.
 */
void ViewerDelete(viewer_struct *v)
{
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *branch, *prev_branch;
	GtkCTreeRow *branch_row;

	if(v == NULL)
	    return;

	v->freeze_count++;

	gtk_widget_hide(v->toplevel);

	/* Clear the search combo */
	GUIComboClearAll(v->manpage_combo);

	/* Remove all branches on index ctree */
	ctree = GTK_CTREE(v->index_ctree);
	clist = GTK_CLIST(ctree);
	gtk_clist_freeze(clist);
	branch = gtk_ctree_node_nth(ctree, 0);
	while(branch != NULL)  
	{
	    prev_branch = branch;	/* Record the current branch */

	    /* Get the next sibling branch */
	    branch_row = GTK_CTREE_ROW(branch);
	    branch = (branch_row != NULL) ?
		branch_row->sibling : NULL;

	    /* Remove the current branch */
	    gtk_ctree_remove_node(ctree, prev_branch);
	}
	gtk_clist_thaw(clist);

	/* Clear the text */
	ViewerTextDelete(v, 0, -1);

	/* Clear the sections list */
	PUListClear(PUListBoxGetPUList(v->section_pulistbox));

	GTK_WIDGET_DESTROY(v->toplevel);
	GTK_ACCEL_GROUP_UNREF(v->accelgrp);

	if(v->colormap != NULL)
	{
	    GdkColormap *colormap = v->colormap;
	    GDK_COLORMAP_FREE_COLOR(colormap, &v->view_normal_text_color);
	    GDK_COLORMAP_FREE_COLOR(colormap, &v->view_bold_text_color);
	    GDK_COLORMAP_FREE_COLOR(colormap, &v->view_underline_text_color);
	    GDK_COLORMAP_UNREF(colormap);
	}

	g_free(v->cur_manpage_path);
	g_free(v->cur_manpage_name);
	g_free(v->last_open_path);

	v->freeze_count--;

	g_free(v);
}
