#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/prochandle.h"

#include "cfg.h"

#include "guiutils.h"
#include "pulist.h"
#include "cdialog.h"

#include "viewer.h"
#include "viewer_fio.h"

#include "manedit.h"
#include "me_op.h"

#include "me_cfg_list.h"
#include "config.h"


static gint ViewerOpenProgressCB(
	const gulong i, const gulong n,
	gpointer data
);
gint ViewerOpenFile(
	viewer_struct *v,
	const gchar *path,
	const gchar *name
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Viewer load progress callback.
 */
static gint ViewerOpenProgressCB(
	const gulong i, const gulong n,
	gpointer data
)
{
	gfloat percent;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return(0);

	if(n > 0l)
	    percent = (gfloat)i / (gfloat)n;
	else
	    percent = 1.0f;

	ViewerSetStatusProgress(v, percent);

	return(0);
}


/*
 *	Opens the manual page.
 *
 *	The path specifies the manual page.
 *
 *	The name specifies the name of the manual page.
 */
gint ViewerOpenFile(
	viewer_struct *v,
	const gchar *path,
	const gchar *name
)
{
	FILE *fp;
	struct stat stat_buf;
	gint		p,
			fd,
			buf_len;
	gchar		*s,
			*cmd,
			*buf,
			*prog_tmp_dir,
			*stdout_path,
			*stderr_path;
	const gchar *converter_cmd;
	GtkWidget *toplevel;
	CfgList *cfg_list;
	me_core_struct *core;

	if((v == NULL) || (path == NULL))
	    return(-2);

	toplevel = v->toplevel;
	core = v->core;
	cfg_list = core->cfg_list;

	/* If the specified path is an absolute path, then check if
	 * it does not exist
	 */
	if(g_path_is_absolute(path))
	{
	    if(stat((const char *)path, &stat_buf))
		return(-1);
	}

	/* Get the manual page interpreter command */
	converter_cmd = ME_GET_S(ME_CFG_PARM_CMD_MANPAGE_INTERPRETER);
	if(STRISEMPTY(converter_cmd))
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Open Manual Page Failed",
"Manual page interpreter command was not set.\n\
\n\
To set the manual page interpreter command, go to\n\
Settings->Options->Commands",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return(-1);
	}

	/* Create the temporary directory */
	prog_tmp_dir = MECreateTemporaryDirectory(core);
	if(prog_tmp_dir == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Open Manual Page Failed",
		"Unable to create temporary directory.",
		NULL,
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return(-1);
	}

	/* Get the standard output file path to output conversion to,
	 * note that stdout_path needs to be deleted later
	 */
	stdout_path = STRDUP(PrefixPaths(
	    prog_tmp_dir,
	    "preview_stdoutXXXXXX"
	));
	fd = (gint)mkstemp((char *)stdout_path);
	if(fd > -1)
	    (void)close((int)fd);

	stderr_path = STRDUP(PrefixPaths(
	    prog_tmp_dir,
	    "preview_stderrXXXXXX"
	));
	fd = (gint)mkstemp((char *)stderr_path);
	if(fd > -1)
	    (void)close((int)fd);

	/* Delete the temprary directory path */
	g_free(prog_tmp_dir);
	prog_tmp_dir = NULL;

	/* No output file path available? */
	if(stdout_path == NULL)
	{
	    g_free(stdout_path);
	    g_free(stderr_path);
	    return(-1);
	}

	/* Begin converting */

	v->freeze_count++;
	ViewerSetBusy(v, TRUE);
	if(TRUE)
	{
	    gchar *msg = g_strdup_printf(
"Opening manual page \"%s\"...",
		name
	    );
	    ViewerSetStatusMessage(v, msg);
	    g_free(msg);
	}
	ViewerSetStatusProgress(v, 0.0f);

/* Define values in enviroment variables? */

	/* Format the convert command */
	cmd = (gchar *)strsub(
	    (const char *)converter_cmd,
	    "%f",
	    path
	);
	s = (gchar *)strsub(
	    (const char *)cmd,
	    "%p",
	    ""
	);
	g_free(cmd);
	cmd = s;

	/* Execute the convert command */
	p = (gint)ExecOE(
	    (const char *)cmd,
	    (const char *)stdout_path,
	    (const char *)stderr_path
	);
	g_free(cmd);
	if(p <= 0)
	{
	    (void)unlink(stdout_path);
	    g_free(stdout_path);
	    (void)unlink(stderr_path);
	    g_free(stderr_path);

	    ViewerSetStatusMessage(v, "Open failed");
	    ViewerSetStatusProgress(v, 0.0f);
	    ViewerSetBusy(v, FALSE);

	    v->freeze_count--;
	    return(-1);
	}

	/* Wait for the convert command */
	while(ExecProcessExists(p))
	{
	    ViewerSetStatusProgress(v, -1.0f);
	    usleep(8000);
	}

	ViewerSetStatusProgress(v, 0.0f);

	/* Report any errors during the conversion */
	MEDialogFromFile(
	    core,
	    "Open Manual Page Warning",
	    stderr_path,
	    toplevel
	);

	/* No output? */
	if(stat((const char *)stdout_path, &stat_buf))
	{
	    (void)unlink(stdout_path);
	    g_free(stdout_path);
	    (void)unlink(stderr_path);
	    g_free(stderr_path);

	    ViewerSetStatusMessage(v, "Open failed");
	    ViewerSetStatusProgress(v, 0.0f);
	    ViewerSetBusy(v, FALSE);

	    v->freeze_count--;
	    return(-2);
	}
	if(stat_buf.st_size == 0l)
	{
	    (void)unlink(stdout_path);
	    g_free(stdout_path);
	    (void)unlink(stderr_path);
	    g_free(stderr_path);

	    ViewerSetStatusMessage(v, "Open failed");
	    ViewerSetStatusProgress(v, 0.0f);
	    ViewerSetBusy(v, FALSE);

	    v->freeze_count--;
	    return(-2);
	}

	/* Allocate the buffer to open the file's contents */
	buf_len = (gint)stat_buf.st_size;
	if(buf_len < 0)
	    buf_len = 0;
	buf = (gchar *)g_malloc((buf_len + 1) * sizeof(gchar));
	if(buf == NULL)
	{
	    (void)unlink(stdout_path);
	    g_free(stdout_path);
	    (void)unlink(stderr_path);
	    g_free(stderr_path);

	    ViewerSetStatusMessage(v, "Open failed");
	    ViewerSetStatusProgress(v, 0.0f);
	    ViewerSetBusy(v, FALSE);

	    v->freeze_count--;
	    return(-3);
	}

	/* Open the converted file for reading */
	fp = fopen((const char *)stdout_path, "rb");
	if(fp == NULL)
	{
	    (void)unlink(stdout_path);
	    g_free(stdout_path);
	    (void)unlink(stderr_path);
	    g_free(stderr_path);

	    g_free(buf);

	    ViewerSetStatusMessage(v, "Open failed");
	    ViewerSetStatusProgress(v, 0.0f);
	    ViewerSetBusy(v, FALSE);

	    v->freeze_count--;
	    return(-2);
	}

	/* Begin opening the converted file */
	if(stat_buf.st_blksize > 0l)
	{
	    const gulong read_buf_len = (gulong)stat_buf.st_blksize;
	    glong units_read;
	    gchar	*buf_ptr = buf,
			*buf_end = buf_ptr + buf_len;

	    while(buf_ptr < buf_end)
	    {
		units_read = (glong)fread(
		    buf_ptr,
		    sizeof(gchar),
		    (size_t)MIN((buf_end - buf_ptr), read_buf_len),
		    fp
		);
		if(units_read <= 0l)
		{
		    *buf_ptr = '\0';
		    break;
		}

		buf_ptr += units_read;

		if(buf_len > 0)
		    ViewerSetStatusProgress(
			v,
			(gfloat)(buf_ptr - buf) / (gfloat)buf_len
		    );
	    }
	    buf[buf_len] = '\0';
	}
	else
	{
	    const glong units_read = (glong)fread(
		buf,
		sizeof(gchar),
		(size_t)buf_len,
		fp
	    );
	    if(units_read < buf_len)
	    {
		if(units_read > 0)
		    buf[units_read] = '\0';
		else
		    *buf = '\0';
	    }
	    else
	    {
		buf[buf_len] = '\0';
	    }

	    ViewerSetStatusProgress(v, 1.0f);
	}

	/* Close the file */
	fclose(fp);
	fp = NULL;

	/* Remove the tempory output files */
	(void)unlink(stdout_path);
	g_free(stdout_path);
	stdout_path = NULL;

	(void)unlink(stderr_path);
	g_free(stderr_path);
	stderr_path = NULL;

	/* Set the buffer to the viewer */
	ViewerTextInsert(
	    v,
	    buf, buf_len,
	    ViewerOpenProgressCB, v
	);

	ViewerSetStatusProgress(v, 1.0f);

	/* Delete the buffer */
	g_free(buf);
	buf = NULL;
	buf_len = 0;

	v->freeze_count--;


	/* Check if the last manual page path matches the new one */
	if((v->cur_manpage_name != NULL) && (name != NULL))
	{
	    /* Name same? */
	    if((v->cur_manpage_name == name) ?
	        TRUE : !strcmp(v->cur_manpage_name, name)
	    )
	    {
		/* Names match, so scroll to last view_text position */
		GtkText *text = GTK_TEXT(v->view_text);
		GTK_ADJUSTMENT_SET_VALUE(
		    text->vadj,
		    v->last_scroll_vpos
		);
	    }
	}


	/* Record the current opened manual page file on viewer */
	if((path != v->cur_manpage_path) && (path != NULL))
	{
	    g_free(v->cur_manpage_path);
	    v->cur_manpage_path = STRDUP(path);
	}

	/* Record the name */
	if((name != v->cur_manpage_name) && (name != NULL))
	{
	    g_free(v->cur_manpage_name);
	    v->cur_manpage_name = STRDUP(name);
	}


	/* Update the last opened path on the viewer */
	g_free(v->last_open_path);
	v->last_open_path = STRDUP(path);
	if(v->last_open_path != NULL)
	{
	    gchar *s = (gchar *)strrchr(
		(char *)v->last_open_path,
		G_DIR_SEPARATOR
	    );
	    if(s != NULL)
		*s = '\0';
	}

	/* Switch to the view page */
	gtk_notebook_set_page(
	    GTK_NOTEBOOK(v->main_notebook),
	    VIEWER_PAGE_MANPAGE
	);

	/* Update the text ManPage GtkEntry */
	if(v->manpage_combo != NULL)
	{
	    GtkWidget *w = v->manpage_combo;
	    GtkCombo *combo = GTK_COMBO(w);
	    const gchar *s = (const gchar *)strrchr(
		(const char *)path,
		' '
	    );
	    if(s != NULL)
	    {
		s++;
	        GUIComboAddItem(w, s);
	        gtk_entry_set_text(GTK_ENTRY(combo->entry), s);
	    }
	    else
	    {
		GUIComboAddItem(w, path);
		gtk_entry_set_text(GTK_ENTRY(combo->entry), path);
	    }
	    gtk_editable_set_position(GTK_EDITABLE(combo->entry), -1);

	    /* If absolute path then set the section to exact */
	    s = gtk_entry_get_text(GTK_ENTRY(combo->entry));
	    if((s != NULL) ? (*s == G_DIR_SEPARATOR) : FALSE)
		ViewerSetSection(
		    v,
		    -1				/* Exact */
		);
	}

	if(TRUE)
	{
	    gchar *msg = g_strdup_printf(
"Manual page \"%s\" opened",
		name
	    );
	    ViewerSetStatusMessage(v, msg);
	    g_free(msg);
	}
	ViewerSetStatusProgress(v, 0.0f);

	ViewerUpdateTitle(v);
	ViewerUpdateMenus(v);
	ViewerSetBusy(v, FALSE);

	return(0);
}
